
#include <dksf.h>
#include <dkstream.h>
#include <dkstr.h>

#include <stdio.h>
#if DK_HAVE_STDLIB_H
#include <stdlib.h>
#endif
#if DK_HAVE_UNISTD_H
#include <unistd.h>
#endif

#ifndef MY_BUFFERSIZE
#define MY_BUFFERSIZE 8192
#endif

static
int
copy_streams DK_P2(dk_stream_t *, dst, dk_stream_t *, src)
{
  int back;
  char buffer[MY_BUFFERSIZE];
  int  can_continue;
  size_t bufused, bufwritten;
  can_continue = back = 1;
  while(can_continue) {
    bufused = dkstream_read(src,buffer,sizeof(buffer));
    if(bufused) {
      bufwritten = dkstream_write(dst,buffer,bufused);
      if(bufwritten != bufused) {
	can_continue = 0;
	back = 0;
      }
    } else {
      can_continue = 0;
    }
  }
  return back;
}

static
int
get_type DK_P1(char *, suffix)
{
  int back = 0; char *ptr;
  if(suffix) {
    ptr = suffix; ptr++;
    if(dkstr_casecmp(ptr, "gz") == 0) {
      back = 1;
    } else {
      if(dkstr_casecmp(ptr, "bz2") == 0) {
	back = 2;
      }
    }
  }
  return back;
}

int main(int argc, char *argv[])
{
  int exval = 0;
  int inputtype, outputtype;
  dk_stream_t *instr, *outstr;
  char *p1, *p2;
  if(argc == 3) {
    inputtype = outputtype = 0;
    p1 = dksf_get_file_type_dot(argv[1]);
    p2 = dksf_get_file_type_dot(argv[2]);
    inputtype = get_type(p1);
    outputtype = get_type(p2);
    instr = outstr = NULL;
    switch(inputtype) {
      case 1: instr = dkstream_opengz(argv[1], "rb"); break;
      case 2: instr = dkstream_openbz2(argv[1], "rb"); break;
      default : instr = dkstream_openfile(argv[1], "rb"); break;
    }
    switch(outputtype) {
      case 1:  outstr = dkstream_opengz(argv[2], "wb"); break;
      case 2:  outstr = dkstream_openbz2(argv[2], "wb"); break;
      default: outstr = dkstream_openfile(argv[2], "wb"); break;
    }
    if(instr && outstr) {
      exval = copy_streams(outstr, instr);
    }
    if(outstr) { dkstream_close(outstr); outstr = NULL; }
    if(instr)  { dkstream_close(instr);  instr  = NULL; }
  }
  exval = (exval ? 0 : 1);
  exit(exval); return exval;
}

