

#include <dksto.h>
#include <dkmem.h>
#include <dkstr.h>

#include <stdio.h>
#if DK_HAVE_STDLIB_H
#include <stdlib.h>
#endif
#if DK_HAVE_UNISTD_H
#include <unistd.h>
#endif

/*
	g <account>
	s <account> <value>
	v <account>
	p a
	p v
*/
typedef struct {
  unsigned long customer_id;
  double        how_much;
} bank_account;

static
bank_account *
new_bank_account(unsigned long cid)
{
  bank_account *back = NULL;
  back = dk_new(bank_account,1);
  if(back) {
    back->customer_id = cid;
    back->how_much = 0.0;
  }
  return back;
}

static
void
delete_bank_account(bank_account *baptr)
{
  if(baptr) {
    dk_delete(baptr);
  }
}

#define COMPARE_BY_UID	    0
#define COMPARE_BY_MONEY    1
#define COMPARE_AGAINST_UID 2

static
int
compare_bank_accounts(void *bl, void *br, int what)
{
  int back = 0;
  bank_account *bal, *bar;
  unsigned long *uidptr;
  bal = (bank_account *)bl; bar = (bank_account *)br;
  switch(what) {
    case COMPARE_BY_MONEY: {
      if(bal->how_much > bar->how_much) {
	back = 1;
      } else {
	if(bal->how_much < bar->how_much) {
	  back = -1;
	}
      }
    } break;
    case COMPARE_AGAINST_UID : {
      uidptr = (unsigned long *)br;
      if(bal->customer_id > (*uidptr)) {
	back = 1;
      } else {
	if(bal->customer_id < (*uidptr)) {
	  back = -1;
	}
      }
    } break;
    default : {
      if(bal->customer_id > bar->customer_id) {
	back = 1;
      } else {
	if(bal->customer_id < bar->customer_id) {
	  back = -1;
	}
      }
    } break;
  }
  return back;
}

int main(int argc, char *argv[])
{
  dk_storage_t *st_cid, *st_val;
  dk_storage_iterator_t *it_cid, *it_val;
  int can_continue;
  char inputline[256], *cptr1, *cptr2;
  bank_account ba, *baptr1, *baptr2, *baptr3;
  unsigned long accountnumber;
  double value;

  st_cid = dksto_open(0); st_val = dksto_open(0);
  if(st_cid) { it_cid = dksto_it_open(st_cid); }
  if(st_val) { it_val = dksto_it_open(st_val); }
  if(st_cid && st_val && it_cid && it_val) {
    dksto_set_comp(st_cid, compare_bank_accounts, COMPARE_BY_UID);
    dksto_set_comp(st_val, compare_bank_accounts, COMPARE_BY_MONEY);
    /* ... + banking takes place here */
    can_continue = 1;
    while(can_continue) {
      if(fgets(inputline,sizeof(inputline),stdin)) {
	cptr1 = dkstr_chr(inputline, '#');
	if(cptr1) { *cptr1 = '\0'; }
	dkstr_chomp(inputline,NULL);
	cptr1 = dkstr_start(inputline,NULL);
	if(cptr1) {
	  printf("%s\n", cptr1);
	  switch(*cptr1) {
	    case 'q' : {
	      can_continue = 0;
	    } break;
	    case 's' : {
	      /* set value for account, add new account or
		 change exisiting account */
	      cptr1++;
	      cptr1 = dkstr_start(cptr1,NULL);
	      if(cptr1) {
		cptr2 = dkstr_next(cptr1,NULL);
		if(cptr2) {
		  if(sscanf(cptr1, "%lu", &accountnumber) == 1) {
		    if(sscanf(cptr2, "%lg", &value) == 1) {
		      baptr1 = dksto_it_find_like(
			it_cid,
			(void *)(&accountnumber),
			COMPARE_AGAINST_UID
		      );
		      if(baptr1) {
			/* ... + update existing account */
                        dksto_remove(st_val,baptr1);
			baptr1->how_much = value;
			if(!dksto_add(st_val,baptr1)) {
			  dksto_remove(st_cid,baptr1);
			  delete_bank_account(baptr1);
			}
			/* ... - update existing account */
		      } else {
			/* ... + create new account */
			baptr1 = dk_new(bank_account,1);
			if(baptr1) {
			  baptr1->customer_id = accountnumber;
			  baptr1->how_much = value;
			  if(dksto_add(st_cid,baptr1)) {
			    if(!dksto_add(st_val,baptr1)) {
			      dksto_remove(st_cid,baptr1);
			      delete_bank_account(baptr1);
			    }
			  } else {
                            delete_bank_account(baptr1);
			  }
			}
			/* ... - create new account */
		      }
		    }
		  }
		}
	      }
	    } break;
	    case 'g' : {
	      /* get value for account */
	      value = 0.0;
	      cptr1++;
	      cptr1 = dkstr_start(cptr1,NULL);
	      if(cptr1) {
		if(sscanf(cptr1, "%lu", &accountnumber) == 1) {
                  baptr1 = dksto_it_find_like(
		    it_cid,
		    &accountnumber,
		    COMPARE_AGAINST_UID
		  );
		  if(baptr1) {
		    value = baptr1->how_much;
		  }
		  printf("%10lu\t%10lg\n", accountnumber, value);
		}
	      }
	    } break;
	    case 'p' : {
	      cptr1++;
	      cptr1 = dkstr_start(cptr1,NULL);
	      if(cptr1) {
		switch(*cptr1) {
		  case 'a' : {
		    /* ... + print, sorted by account */
		    dksto_it_reset(it_cid);
		    while((baptr1 = dksto_it_next(it_cid)) != NULL) {
		      printf("%10lu\t%10lg\n", baptr1->customer_id,
			baptr1->how_much
		      );
		    }
		    /* ... - print, sorted by account */
		  } break;
		  case 'v' : {
		    /* ... + print, sorted by value */
		    dksto_it_reset(it_val);
		    while((baptr1 = dksto_it_next(it_val)) != NULL) {
		      printf("%10lu\t%10lg\n", baptr1->customer_id,
			baptr1->how_much
		      );
		    }
		    /* ... - print, sorted by value */
		  } break;
		}
	      }
	    } break;
	  }
	}
      } else {
	can_continue = 0;
      }
    }
    /* ... - banking takes place here */
    /* ... + remove all bank account data */
    dksto_it_reset(it_cid);
    while((baptr1 = dksto_it_next(it_cid)) != NULL) {
      delete_bank_account(baptr1);
    }
    /* ... - remove all bank account data */
  }
  if(it_val) { dksto_it_close(it_val); it_val = NULL; }
  if(it_cid) { dksto_it_close(it_cid); it_cid = NULL; }
  if(st_val) { dksto_close(st_val); st_val = NULL; }
  if(st_cid) { dksto_close(st_cid); st_cid = NULL; }
  exit(0); return 0;
}

