/* 
 * The dk libraries (libdkport + libdkc)
 * Copyright (C) 2000, 2001 - Dirk Krause
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 * In this package the copy of the GNU Library General Public License
 * is placed in file COPYING.
 */


#line 23 "dkapp.ctr"


#include <stdio.h>

#include <dk.h>
#include <dktypes.h>
#include <dkmem.h>
#include <dksf.h>
#include <dkstr.h>
#include <dksto.h>
#include <dklogc.h>
#include <dkstt.h>
#include <dkstream.h>
#include <dkcp.h>

#if DK_TIME_WITH_SYS_TIME
#include <sys/time.h>
#include <time.h>
#else
#if DK_HAVE_SYS_TIME_H
#include <sys/time.h>
#else
#if DK_HAVE_TIME_H
#include <time.h>
#endif
#endif
#endif

#if DK_HAVE_STDLIB_H
#include <stdlib.h>
#endif
#if DK_HAVE_UNISTD_H
#include <unistd.h>
#endif

#define DK_APP_C 1
#include <dkapp.h>


#if DK_HAVE_STRING_H
#include <string.h>
#endif

#if DK_HAVE_WINREG_H
#include <dkwin.h>
#endif

#if DK_HAVE_SYSLOG_H
#include <syslog.h>
#endif

#if DK_HAVE_CTYPE_H
#include <ctype.h>
#endif

#if DK_HAVE_ZLIB_H
#include <zlib.h>
#endif

static char *unix_sysconfdir;
extern char *kpse_var_value(char *);

typedef char *CHARPTR;

typedef struct {
  char *name;
  dk_stt_t *st;
} stt_entry;

static int set_silent = 0;

static char key_log_off[]= { "/log/off" };

int
dkapp_silence_check DK_P2(int,argc,char **,argv)
{
  int back = 0;
  int i;
  char *cptr, *vptr;
  for(i = 1; i < argc; i++) {
    cptr = argv[i]; vptr = NULL;
    if(cptr[0] == '-') {
    if(cptr[1] == '-') {
      vptr = dkstr_chr(cptr, '=');
      if(vptr) { *vptr = (char)0; }
      if(strcmp(&(cptr[2]), key_log_off) == 0) {
        if(vptr) {
	  if(dkstr_is_on(&(vptr[1]))) {
	    back = 1;
	  }
	} else {
	  back = 1;
	}
      }
      if(vptr) { *vptr = '='; }
    }
    }
  }
  return back;
}

static void stt_entry_free DK_P1(stt_entry *,s)
{
  char *cptr;
  cptr = s->name;
  if(cptr) { dk_delete(cptr); }
  s->name = NULL;
  if(s->st) { dkstt_close(s->st); }
  s->st = NULL;
  dk_delete(s) ;
}

static int
stt_entry_comp DK_P3(void *,p1,void *,p2,int,crit)
{
  int back = 0;
  stt_entry *s1, *s2;
  
  if(p1 && p2) {
    switch(crit) {
      case 1: {
	s1 = (stt_entry *)p1;
	if(s1->name) {
	  back = strcmp((s1->name),((char *)p2));
	} else {
	  back = -1;
	}
      } break;
      default: {
	s1 = (stt_entry *)p1; s2 = (stt_entry *)p2;
	if(s1->name) {
	  if(s2->name) {
	    back = strcmp((s1->name),(s2->name));
	  } else {
	    back = 1;
	  }
	} else {
	  if(s2->name) {
	    back = -1;
	  }
	}
      } break;
    }
  } 
  return back;
}

static char all_files[] = { "all" };
static char scope_all[] = { "*" };
static char dot[] = { "." };
static char app[] = { "app" };
static char suffix_log[] = { "log" };
static char dot_tmp[] = { ".TMP" };

#if DK_HAVE_WINREG_H
static char hklm[] = { "Software\\DkApp" };
static char hkcu[] = { "SOFTWARE\\DkApp" };
static char backslash[] = { "\\" };
#endif

static char fn_sep[] = {
#if DK_HAVE_FEATURE_BACKSLASH
  "\\"
#else
  "/"
#endif
};

static char bin_dir[] = { "bin" };
static char lib_dir[] = { "lib" };

static char defaults_sub[] = {
#if DK_HAVE_DOTFILENAMES
  ".defaults"
#else
  "defaults"
#endif
};

static char appdef[] = { "appdef" };



/* ********************************************************************* */
/* *                                                                   * */
/* *   Comparison of preference entries.                               * */
/* *   cr = 0: pref entry / pref entry                                 * */
/* *   cr = 1: pref entry / entry name as string                       * */
/* *                                                                   * */
/* ********************************************************************* */

#line 212 "dkapp.ctr"



void
dkpref_delete DK_P1(dk_preference_t *, p)
{
  char *x;
  
  if(p) {
    
    x = p->v;
    if(x) dk_delete(x);
    x = p->n;
    if(x) dk_delete(x);
    p->p = 0;
    p->v = NULL;
    p->n = NULL;
    dk_delete(p);
  }
  
}

dk_preference_t *
dkpref_new DK_P3(char *, name, char *, value, int, prio)
{
  dk_preference_t *back = NULL;
  
  if(name && value) {
    back = dk_new(dk_preference_t,1);
    if(back) {
      back->n = dkstr_dup(name);
      back->v = dkstr_dup(value);
      back->p = prio;
      if(!((back->v) && (back->n))) {
	dkpref_delete(back);
	back = NULL;
      }
    }
  } 
  return back;
}

int
dkapp_pref_compare DK_P3(void *, p1, void *, p2, int, cr)
{
  int back = 0;
  
  if(p1 && p2) {
    switch(cr) {
      case 1: {  /* search for specific entry */
	
	back = strcmp((((dk_preference_t *)p1)->n),(char *)p2);
      } break;
      default: { /* sort entries in container */
	
	back = strcmp((((dk_preference_t *)p1)->n),(((dk_preference_t *)p2)->n));
      } break;
    }
  } 
  return back;
}



static void
add_pref_to_list DK_P5(dk_storage_t *,s, dk_storage_iterator_t *,si,char *,n, char *,v, int,p)
{
  dk_preference_t *pptr;
  char *cptr;
  
  if(s && si && n && v) {
    pptr =
    (dk_preference_t *)dksto_it_find_like(si,n,1);
    if(pptr) { 
      if(p >= pptr->p) {
	cptr = pptr->v;
	if(cptr) dk_delete(cptr);
	pptr->v = dkstr_dup(v);
	pptr->p = p;
      }
    } else { 
      pptr = dkpref_new(n,v,p);
      if(pptr) {
	if(!dksto_add(s,pptr)) {
	  dkpref_delete(pptr);
	}
      }
    }
  } 
}


static void
read_cfg DK_P7(dk_storage_t *, s, dk_storage_iterator_t *, si, char *, fn, char *, un, char *, an, char *, hn, int, prio)
{
  FILE *inputfile;
  char inputline[256], *cp1, *cp2, *cp3;
  int can_continue;
  int p_to_save;
  int use_entries;
  
  if(fn && s && si) {
    inputfile = dksf_fopen(fn, "r");
    if(inputfile) { 
      can_continue = 1; use_entries = 1; p_to_save = 0;
      while(can_continue) {
	if(fgets(inputline,sizeof(inputline),inputfile)) {
	  cp2 = cp3 = NULL;
	  cp1 = dkstr_start(inputline, NULL);
	  if(cp1) {
	    cp2 = dkstr_chr(cp1, '#');
	    if(cp2) {
	      *cp2 = '\0';
	    }
	    cp1 = dkstr_start(cp1, NULL);
	    if(cp1) {
	      
	      if(*cp1 == '[') {
		use_entries = 0; p_to_save = 0;
		cp2 = dkstr_rchr(cp1, ']');
		if(cp2) {
		  *cp2 = '\0';
		  cp1++;
		  cp1 = dkstr_start(cp1, NULL);
		  if(cp1) {
		    dkstr_chomp(cp1, NULL);
		    cp2 = dkstr_chr(cp1, '/');
		    if(cp2) {
		      *(cp2++) = '\0';
		      cp3 = dkstr_chr(cp2, '/');
		      if(cp3) {
			*(cp3++) = '\0';
		      }
		    }
                    switch(prio) {
		      case 0: {
			if(cp1) {
			  if(strcmp(cp1, scope_all) == 0) {
			    use_entries = 1;
			  } else {
			    if(un) {
			      if(strcmp(cp1, un) == 0) {
				use_entries = 1; p_to_save = 4;
			      }
			    }
			  }
			  if(use_entries) {
			    use_entries = 0;
			    if(cp2) {
			      if(strcmp(cp2, scope_all) == 0) {
				use_entries = 1;
			      } else {
				if(an) {
				  if(strcmp(cp2, an) == 0) {
				    use_entries = 1; p_to_save |= 2;
				  }
				}
			      }
			      if(use_entries) {
				use_entries = 0;
				if(cp3) {
				  if(strcmp(cp3, scope_all) == 0) {
				    use_entries = 1;
				  } else {
				    if(hn) {
				      if(strcmp(cp3, hn) == 0) {
					use_entries = 1; p_to_save |= 1;
				      }
				    }
				  }
				} else {
				  use_entries = 1;
				}
			      }
			    } else {
			      use_entries = 1;
			    }
			  }
			} else {
			  use_entries = 1;
			}
		      } break;
		      case 2: {
			if(cp1) {
			  if(strcmp(cp1, scope_all) == 0) {
			    use_entries = 1;
			  } else {
			    if(un) {
			      if(strcmp(cp1, un) == 0) {
				use_entries = 1; p_to_save |= 4;
			      }
			    }
			  }
			  if(use_entries) {
			    use_entries = 0;
			    if(cp2) {
			     if(strcmp(cp2, scope_all) == 0) {
			       use_entries = 1;
			     } else {
			       if(hn) {
				 if(strcmp(cp2, hn) == 0) {
				   use_entries = 1; p_to_save |= 1;
				 }
			       }
			     }
			    } else {
			      use_entries = 1;
			    }
			  }
			} else {
			  use_entries = 1;
			}
		      } break;
		      case 4: {
			if(cp1) {
			  if(strcmp(cp1, scope_all) == 0) {
			    use_entries = 1;
			  } else {
			    if(an) {
			      if(strcmp(cp1, an) == 0) {
				use_entries = 1; p_to_save |= 2;
			      }
			    }
			  }
			  if(use_entries) {
			    use_entries = 0;
			    if(cp2) {
			      if(strcmp(cp2, scope_all) == 0) {
				use_entries = 1;
			      } else {
				if(hn) {
				  if(strcmp(cp2, hn) == 0) {
				    use_entries = 1; p_to_save |= 1;
				  }
				}
			      }
			    } else {
			      use_entries = 1;
			    }
			  }
			} else {
			  use_entries = 1;
			}
		      } break;
		      case 6: {
			if(cp1) {
			  if(strcmp(cp1, scope_all) == 0) {
			    use_entries = 1;
			  } else {
			    if(hn) {
			      if(strcmp(cp1, hn) == 0) {
				use_entries = 1; p_to_save |= 1;
			      }
			    }
			  }
			} else {
			  use_entries = 1;
			}
		      } break;
		    }
		  }
		}
	      } else { /* if(*cp1 == '[') */
	        if(use_entries) {
		  cp2 = dkstr_chr(cp1, '=');
		  if(cp2) {
		    *(cp2++) = '\0';
		    cp2 = dkstr_start(cp2, NULL);
		    if(cp2) {
		      cp1 = dkstr_start(cp1, NULL);
		      if(cp1) {
		        dkstr_chomp(cp1, NULL);
		        dkstr_chomp(cp2, NULL);
		        add_pref_to_list(s,si,cp1,cp2,(p_to_save | prio));
		      }
		    }
		  }
	        }
	      }
	    }
	  }
	} else {
	  can_continue = 0;
	}
      }
      fclose(inputfile);
    }
  }
  
}


static void
app_init_empty DK_P1(dk_app_t *, a)
{
  
  if(a) {
    (a->td).l = 0UL;
    (a->a).o.argc = 0; (a->a).o.argv = NULL;
    (a->a).a.argc = 0; (a->a).a.argv = NULL;
    (a->n).u = NULL;
    (a->n).a = NULL;
    (a->n).h = NULL;
    (a->n).g = NULL;
    (a->d).h = NULL;
    (a->d).t = NULL;
    (a->d).a = NULL;
    (a->d).s = NULL;
    (a->d).pt = NULL;
    (a->d).etc = NULL;
    (a->p).c = NULL;
    (a->p).ci = NULL;
    (a->p).a = NULL;
    (a->p).ai = NULL;
#if DK_HAVE_WINREG_H
    (a->p).what = 0;
#else
    (a->p).u = NULL;
    (a->p).ui = NULL;
    (a->p).s = NULL;
    (a->p).si = NULL;
#endif
    (a->p).unc  = 0;
    (a->x).f = NULL;
    (a->x).d = NULL;
    (a->l).max = DK_LOG_LEVEL_IGNORE;
    (a->l).nostdio = 0;
    (a->l).o.m = DK_LOG_LEVEL_NONE;
    (a->l).o.f = 0;
    (a->l).o.ide_type = 0;
    (a->l).e.m = DK_LOG_LEVEL_PROGRESS;
    (a->l).e.f = 0;
    (a->l).e.ide_type = 0;
    (a->l).f.m = DK_LOG_LEVEL_WARNING;
    (a->l).f.k = DK_LOG_LEVEL_ERROR;
    (a->l).f.f = 3;
    (a->l).f.t = NULL;
    (a->l).f.n = NULL;
    (a->l).f.ide_type = 0;
#if DK_HAVE_SYSLOG
    (a->l).s.m = DK_LOG_LEVEL_NONE;
    (a->l).s.i = NULL;
    (a->l).s.f = 0;
    (a->l).s.o = 0;
#endif
    (a->l).o.c = NULL;
    (a->l).e.c = NULL;
    (a->l).f.c = NULL;
    (a->l).ef.n = NULL;
    (a->l).ef.lineno = 0UL;
    (a->loc).l = NULL;
    (a->loc).r = NULL;
    (a->loc).e = NULL;
    (a->loc).s = NULL;
    (a->loc).si = NULL;
    a->relaxed_fopen_check = 0;
  }
}

#if DK_HAVE_WINREG_H
static char colon[] = { ":" };
static char slash[] = { "/" };
#endif

int dkapp_set_pref DK_P3(dk_app_t *,a,char *,key,char *,value)
{
  int back = 0;
#if DK_HAVE_WINREG_H
  char keybuffer[256];
#endif
  
  if(a && key && value) {
    if(((a->p).a) && ((a->p).ai)) {
      add_pref_to_list(((a->p).a), ((a->p).ai), key, value, 6);
      back = 1;
    }
#if DK_HAVE_WINREG_H
    if(((a->n).h) && (((a->p).what) & 8)) {
      if((strlen((a->n).h) + strlen(key) + 1) < sizeof(keybuffer)) {
	strcpy(keybuffer, (a->n).h);
	strcat(keybuffer, colon);
	strcat(keybuffer, key);
	RegSetValueExA(
	  (a->p).hkcu_app,
	  keybuffer,
	  0,
	  REG_SZ,
	  value,
	  (strlen(value) + 1)
	);
      }
    }
#endif
  } 
  return back;
}

int dkapp_get_pref DK_P5(dk_app_t *,a,char *,key,char *,b,size_t,lgt,int,excl)
{
  int back = 0;
  int found;
  int where;
#if DK_HAVE_WINREG_H
  char keybuffer[256], *uptr, *hptr;
  LONG retval;
  DWORD dwType, sz;
#endif
  dk_preference_t *pptr;

  
  found = 0;
  where = 31;
  if(a && key && b && lgt) { 
    while((where >= 0) && (found == 0)) { 
      switch(where & 24) {
	case 24: { /* programs own settings */
	  
	  if(!(excl & DK_APP_PREF_EXCL_PROG)) {
	    
	    if(((a->p).a) && ((a->p).ai)) {
	      
	      pptr =
	      (dk_preference_t *)dksto_it_find_like(((a->p).ai),key,1);
	      if(pptr) {
		
		found = 1;
		if(pptr->v) {
		  if(lgt > strlen(pptr->v)) {
		    
		    strcpy(b, pptr->v); back = 1;
		  }
		}
	      }
	    }
	  }
	  where = 24;
	} break;
	case 16: { /* command line settings */
	  
	  if(!(excl & DK_APP_PREF_EXCL_CMD)) {
	    
	    if(((a->p).c) && ((a->p).ci)) {
	      
	      pptr =
	      (dk_preference_t *)dksto_it_find_like(((a->p).ci),key,1);
	      if(pptr) {
		  
		  found = 1;
		  if(pptr->v) {
		    if(lgt > strlen(pptr->v)) {
		      
		      strcpy(b, pptr->v); back = 1;
		    }
		  }
	      }
	    }
	  }
	  where = 16;
	} break;
	case  8: { /* user settings */
	  
	  if(!(excl & DK_APP_PREF_EXCL_USER)) {
	    
#if DK_HAVE_WINREG_H
	    
	    switch(where & 7) {
	      case 7: 
	      case 6:
	      case 5:
	      case 4: {
		
		hptr = ((where & 1) ? ((a->n).h) : (all_files));
		if(hptr) {
		  if((strlen(key)+strlen(hptr)+1) < sizeof(keybuffer)) {
		    
		    strcpy(keybuffer, hptr);
		    strcat(keybuffer, colon);
		    strcat(keybuffer, key);
		    
		  }
		  retval = ERROR_SUCCESS - 1L;
		  sz = lgt;
		  if(where & 2) {
		    
		    if(((a->p).what) & 8) {
		      
		      retval = RegQueryValueExA(
			(a->p).hkcu_app,
			keybuffer,
			NULL,
			&dwType,
			b,
			&sz
		      );
		    }
		  } else {
		    
		    if(((a->p).what) & 4) {
		      
		      retval = RegQueryValueExA(
			(a->p).hkcu_all,
			keybuffer,
			NULL,
			&dwType,
			b,
			&sz
		      );
		    }
		  }
		  if(retval == ERROR_SUCCESS) {
		    
		    found = 1;
		    if(dwType == REG_SZ) {
		      if(sz < lgt) {
			
			b[sz] = '\0';
			back = 1;
			found = 1;
		      }
		    }
		  }
		}
	      } break;
	    }
#else
	    
	    if(((a->p).u) && ((a->p).ui)) {
	      
	      pptr =
	      (dk_preference_t *)dksto_it_find_like(((a->p).ui),key,1);
	      if(pptr) {
		
		found = 1;
		if(pptr->v) {
		  if(lgt > strlen(pptr->v)) {
		    
		    strcpy(b, pptr->v); back = 1;
		  }
		}
	      }
	    }
	    where =  8;
#endif
	  }
	} break;
	case  0: { /* system settings */
	  
	  if(!(excl & DK_APP_PREF_EXCL_SYSTEM)) {
	    
#if DK_HAVE_WINREG_H
	    
	    uptr = ((where & 4) ? ((a->n).u) : (all_files));
	    hptr = ((where & 1) ? ((a->n).h) : (all_files));
	    if(uptr && hptr) {
if((strlen(key) + strlen(uptr) + strlen(hptr) + 2) < sizeof(keybuffer)) {
	      
	      strcpy(keybuffer, uptr);
	      strcat(keybuffer, slash);
	      strcat(keybuffer, hptr);
	      strcat(keybuffer, colon);
	      strcat(keybuffer, key);
	      
	      sz = lgt;
	      retval = ERROR_SUCCESS - 1L;
	      if(where & 2) {
		
		if(((a->p).what) & 2) {
		      
		      retval = RegQueryValueExA(
			(a->p).hklm_app,
			keybuffer,
			NULL,
			&dwType,
			b,
			&sz
		      );
		}
	      } else {
		
		if(((a->p).what) & 1) {
		      
		      retval = RegQueryValueExA(
			(a->p).hklm_all,
			keybuffer,
			NULL,
			&dwType,
			b,
			&sz
		      );
		}
	      }
	      if(retval == ERROR_SUCCESS) {
		
		found = 1;
		if(dwType == REG_SZ) {
		  if(sz < lgt) {
		    
		    b[sz] = '\0'; back = 1;
		  }
		}
	      }
}
	    }
#else
	    
	    if(((a->p).s) && ((a->p).si)) {
	      
	      pptr =
	      (dk_preference_t *)dksto_it_find_like(((a->p).si),key,1);
	      if(pptr) {
		
	        found = 1;
	        if(pptr->v) {
		  if(lgt > strlen(pptr->v)) {
		    
		    strcpy(b, pptr->v); back = 1;
		  }
	        }
	      }
	    }
	    where = 0;
#endif
	  }
	} break;
      }
      where--;
    }
    
  } 
  return back;
}

static char *log_level_strings[] = {
  (char *)"n$one",
  (char *)"pa$nic",
  (char *)"f$atal",
  (char *)"e$rror",
  (char *)"warning",
  (char *)"i$nfo",
  (char *)"pr$ogress",
  (char *)"d$ebug",
  NULL
};

static int
get_log_level DK_P1(char *,str)
{
  int back = -1;
  if(str) {
    back = dkstr_array_abbr(log_level_strings, str, '$', 0);
  }
  return back;
}

static char log_file_ide[] = { "/log/file/ide" };
static char log_stdout_ide[] = { "/log/stdout/ide" };
static char log_stderr_ide[] = { "/log/stderr/ide" };
static char log_file_name[] = { "/log/file/name" };
static char log_file_level[] = { "/log/file/level" };
static char log_file_keep[] = { "/log/file/keep" };
static char log_file_time[] = { "/log/file/time" };
static char log_file_split[] = { "/log/file/split" };
static char log_stdout_level[] = { "/log/stdout/level" };
static char log_stdout_time[] = { "/log/stdout/time" };
static char log_stdout_split[] = { "/log/stdout/split" };
static char log_stderr_level[] = { "/log/stderr/level" };
static char log_stderr_time[] = { "/log/stderr/time" };
static char log_stderr_split[] = { "/log/stderr/split" };
static char log_syslog_level[] = { "/log/syslog/level" };
#if DK_HAVE_CODEPAGES
static char log_file_codepage[] = { "/log/file/codepage" };
static char log_stdout_codepage[] = { "/log/stdout/codepage" };
static char log_stderr_codepage[] = { "/log/stderr/codepage" };
#endif
static char app_dir[] = { "/dir/app" };
static char shared_dir[] = { "/dir/shared" };
static char tmp_dir[] = { "/dir/tmp" };
static char ui_lang[] = { "/ui/lang" };
static char ui_lang_env[] = { "/ui/lang/env" };
static char storage_trees[] = { "/storage/trees" };
static char minus_sign[] = { "-" };

static void
save_lang DK_P2(dk_app_t *, a, char *, str)
{
  char *cptr;
  cptr = dkstr_chr(str, '.');
  if(cptr) {
    *(cptr++) = '\0';
    if(strlen(cptr) > 0) {
      (a->loc).e = dkstr_dup(cptr);
    }
  }
  cptr = dkstr_chr(str, '_');
  if(cptr) {
    *(cptr++) = '\0';
    if(strlen(cptr) > 0) {
      (a->loc).r = dkstr_dup(cptr);
    }
  }
  if(strlen(str) > 0) {
    (a->loc).l = dkstr_dup(str);
  }
}

static char rb[] = { "rb" };

static int
file_check DK_P1(char *,fn)
{
  int back = 0;
  dk_stat_t st;
  if(dkstat_get(&st,fn)) {
    if(((st.filetype) & (~(DK_FT_SYMLINK))) == DK_FT_REG) {
      back = 1;
    }
  } 
  return back;
}

static int
file_search DK_P5(dk_app_t *,a,char *,name,char *,buffer,size_t,sz,char *,d)
{
  int back = 0;
  size_t lgt;
  if((a->loc).l) {
    /* l r e */
    /* l r */
    if((a->loc).r) {
      if((a->loc).e) {
	lgt = strlen(d) + strlen(name) + strlen((a->loc).l);
	lgt += strlen((a->loc).r) + strlen((a->loc).e) + 4;
	if(lgt < sz) {
	  strcpy(buffer,d);
	  strcat(buffer, fn_sep);
	  strcat(buffer, (a->loc).l);
	  strcat(buffer, fn_sep);
	  strcat(buffer, (a->loc).r);
	  strcat(buffer, fn_sep);
	  strcat(buffer, (a->loc).e);
	  strcat(buffer, fn_sep);
	  strcat(buffer, name);
	  back = file_check(buffer);
	}
      }
      if(!back) {
	lgt = strlen(d) + strlen(name) + strlen((a->loc).l);
	lgt += strlen((a->loc).r) + 3;
	if(lgt < sz) {
	  strcpy(buffer,d);
	  strcat(buffer, fn_sep);
	  strcat(buffer, (a->loc).l);
	  strcat(buffer, fn_sep);
	  strcat(buffer, (a->loc).r);
	  strcat(buffer, fn_sep);
	  strcat(buffer, name);
	  back = file_check(buffer);
	}
      }
    }
    /* l e */
    if(!back) {
      if((a->loc).e) {
	lgt = strlen(d) + strlen(name) + strlen((a->loc).l);
	lgt += strlen((a->loc).e) + 3;
	if(lgt < sz) {
	  strcpy(buffer,d);
	  strcat(buffer, fn_sep);
	  strcat(buffer, (a->loc).l);
	  strcat(buffer, fn_sep);
	  strcat(buffer, (a->loc).e);
	  strcat(buffer, fn_sep);
	  strcat(buffer, name);
	  back = file_check(buffer);
	}
      }
    }
    if(!back) {
      lgt = strlen(d) + strlen(name) + strlen((a->loc).l) + 2;
      if(lgt < sz) {
	strcpy(buffer,d);
	strcat(buffer, fn_sep);
	strcat(buffer, (a->loc).l);
	strcat(buffer, fn_sep);
	strcat(buffer, name);
	back = file_check(buffer);
      }
    }
  }
  if(!back) {
    lgt = strlen(d) + strlen(name) + 1;
    if(lgt < sz) {
      strcpy(buffer,d);
      strcat(buffer, fn_sep);
      strcat(buffer, name);
      back = file_check(buffer);
    }
  }
  return back;
}

static char *debug_strings[21];

static dk_string_finder_t debug_string_finder[] = {
  { (char *)"/d/00", &(debug_strings[0]), (char *)"Login name:      " },
  { (char *)"/d/01", &(debug_strings[1]), (char *)"Application:     " },
  { (char *)"/d/02", &(debug_strings[2]), (char *)"Host name:       " },
  { (char *)"/d/03", &(debug_strings[3]), (char *)"HOME dir:        " },
  { (char *)"/d/04", &(debug_strings[4]), (char *)"TEMP dir:        " },
  { (char *)"/d/05", &(debug_strings[5]), (char *)"Application dir: " },
  { (char *)"/d/06", &(debug_strings[6]), (char *)"Shared dir:      " },
  { (char *)"/d/07", &(debug_strings[7]), (char *)"Executable file: " },
  { (char *)"/d/08", &(debug_strings[8]), (char *)"Executable dir:  " },
  { (char *)"/d/09", &(debug_strings[9]), (char *)"Log file:        " },
  { (char *)"/d/10", &(debug_strings[10]), (char *)"Language:        " },
  { (char *)"/d/11", &(debug_strings[11]), (char *)"Region:          " },
  { (char *)"/d/12", &(debug_strings[12]), (char *)"Encoding:        " },
  { (char *)"/d/13", &(debug_strings[13]), (char *)"unknown" },
  { (char *)"/d/14", &(debug_strings[14]), (char *)"Proc temp dir:   " },
  { (char *)"/d/15", &(debug_strings[15]), (char *)"Checking file \"" },
  { (char *)"/d/16", &(debug_strings[16]), (char *)"\"." },
  { (char *)"/d/17", &(debug_strings[17]), (char *)"File \"" },
  { (char *)"/d/18", &(debug_strings[18]), (char *)"\" found." },
  { (char *)"/d/19", &(debug_strings[19]), (char *)"Program group:   " },
  { (char *)"/d/20", &(debug_strings[20]), (char *)"System conf dir: " },
  {NULL, NULL, NULL}
};

static char *open_and_close_strings[] = {
  NULL, NULL, NULL, NULL, NULL, NULL,
  NULL, NULL, NULL, NULL, NULL, NULL,
  NULL, NULL, NULL, NULL, NULL, NULL,
  NULL, NULL, NULL, NULL, NULL, NULL,
  NULL, NULL, NULL, NULL, NULL, NULL,
  NULL, NULL, NULL
};

static dk_string_finder_t open_and_close_finder[] = {
  { (char *)"/msg/01", &(open_and_close_strings[0]), (char *)"Application \"" },
  { (char *)"/msg/02", &(open_and_close_strings[1]), (char *)"\" started." },
  { (char *)"/msg/03", &(open_and_close_strings[2]), (char *)"Application \"" },
  { (char *)"/msg/04", &(open_and_close_strings[3]), (char *)"\" finished." },
  { (char *)"/msg/05", &(open_and_close_strings[4]), (char *)"File \"" },
  { (char *)"/msg/06", &(open_and_close_strings[5]), (char *)"\" found." },
  { (char *)"/msg/07", &(open_and_close_strings[6]), (char *)"File \"" },
  { (char *)"/msg/08", &(open_and_close_strings[7]), (char *)"\" not found!." },
  { (char *)"/msg/09", &(open_and_close_strings[8]), (char *)"File \"" },
  { (char *)"/msg/10", &(open_and_close_strings[9]), (char *)"\" opened." },
  { (char *)"/msg/11", &(open_and_close_strings[10]), (char *)"Failed to read \"" },
  { (char *)"/msg/12", &(open_and_close_strings[11]), (char *)"\"!" },
  { (char *)"/msg/13", &(open_and_close_strings[12]), (char *)"String table \"" },
  { (char *)"/msg/14", &(open_and_close_strings[13]), (char *)"\" not found!." },
  { (char *)"/msg/15", &(open_and_close_strings[14]), (char *)"String table \"" },
  { (char *)"/msg/16", &(open_and_close_strings[15]), (char *)"\" found." },
  { (char *)"/msg/17", &(open_and_close_strings[16]), (char *)"Different owners for symlink \""},
  { (char *)"/msg/18", &(open_and_close_strings[17]), (char *)"\" and target file!" },
  { (char *)"/msg/19", &(open_and_close_strings[18]), (char *)"Symlink \"" },
  { (char *)"/msg/20", &(open_and_close_strings[19]), (char *)"\" resides in a group writable directory!" },
  { (char *)"/msg/21", &(open_and_close_strings[20]), (char *)"Symlink \"" },
  { (char *)"/msg/22", &(open_and_close_strings[21]), (char *)"\" resides in a world writable directory!" },
  { (char *)"/msg/23", &(open_and_close_strings[22]), (char *)"No file name like \"" },
  { (char *)"/msg/24", &(open_and_close_strings[23]), (char *)"\"!" },
  { (char *)"/msg/25", &(open_and_close_strings[24]), (char *)"Too many file names for \"" },
  { (char *)"/msg/26", &(open_and_close_strings[25]), (char *)"\"!" },
  { (char *)"/msg/27", &(open_and_close_strings[26]), (char *)"\"" },
  { (char *)"/msg/28", &(open_and_close_strings[27]), (char *)"\" is a directory!" },

  { NULL, NULL, NULL }
};

static int app_err_conf = 0;

static char *app_err_str[21];

static dk_string_finder_t app_err_find[] = {
  { (char *)"/e/00", &(app_err_str[0]), (char *)"Cannot traverse directory \"" }, 
  { (char *)"/e/01", &(app_err_str[1]), (char *)"\"!" },
  { (char *)"/e/02", &(app_err_str[2]), (char *)"No information about \"" },
  { (char *)"/e/03", &(app_err_str[3]), (char *)"\"!" },
  { (char *)"/e/04", &(app_err_str[4]), (char *)"Failed to estimate current working directory!" },
  { (char *)"/e/05", &(app_err_str[5]), (char *)"Not enough memory! Failed to allocate " },
  { (char *)"/e/06", &(app_err_str[6]), (char *)" elements, " },
  { (char *)"/e/07", &(app_err_str[7]), (char *)" bytes per element." },
  { (char *)"/e/08", &(app_err_str[8]), (char *)"No file matching \"" },
  { (char *)"/e/09", &(app_err_str[9]), (char *)"\"!" },
  { (char *)"/e/10", &(app_err_str[10]), (char *)"No directory matching \"" },
  { (char *)"/e/11", &(app_err_str[11]), (char *)"\"!" },
  { (char *)"/e/12", &(app_err_str[12]), (char *)"Failed to open file \"" },
  { (char *)"/e/13", &(app_err_str[13]), (char *)"\" for read access!" },
  { (char *)"/e/14", &(app_err_str[14]), (char *)"Failed to open file \"" },
  { (char *)"/e/15", &(app_err_str[15]), (char *)"\" for write access!" },
  { (char *)"/e/16", &(app_err_str[16]), (char *)"Error while writing to file \"" },
  { (char *)"/e/17", &(app_err_str[17]), (char *)"\"!" },
  { (char *)"/e/18", &(app_err_str[18]), (char *)"Error while reading from file \"" },
  { (char *)"/e/19", &(app_err_str[19]), (char *)"\"!" },
  { (char *)"/e/20", &(app_err_str[20]), (char *)"TCP/IP not available!" },
  { NULL, NULL, NULL }
};

void dkapp_err_memory DK_P3(dk_app_t *, app, size_t, elsize, size_t, nelem)
{
  char *logmsg[5];
  char buffer1[32], buffer2[32];
  if(app) {
    sprintf(buffer1, "%lu", (unsigned long)nelem);
    sprintf(buffer2, "%lu", (unsigned long)elsize);
    logmsg[0] = app_err_str[5];
    logmsg[1] = buffer1;
    logmsg[2] = app_err_str[6];
    logmsg[3] = buffer2;
    logmsg[4] = app_err_str[7];
    dkapp_log_msg(app, DK_LOG_LEVEL_ERROR, logmsg, 5);
  }
}

void dkapp_err_nowrite DK_P3(dk_app_t *,app,char *,filename,int,reason)
{
  char *logmsg[3];
  if(app && filename) {
    switch(reason) {
      case DK_SF_SEC_WG : {
        logmsg[0] = open_and_close_strings[18];
	logmsg[1] = filename;
	logmsg[2] = open_and_close_strings[19];
	dkapp_log_msg(app, DK_LOG_LEVEL_ERROR, logmsg, 3);
      } break;
      case DK_SF_SEC_WO : {
        logmsg[0] = open_and_close_strings[20];
	logmsg[1] = filename;
	logmsg[2] = open_and_close_strings[21];
	dkapp_log_msg(app, DK_LOG_LEVEL_ERROR, logmsg, 3);
      } break;
      case DK_SF_SEC_OWNER : {
        logmsg[0] = open_and_close_strings[16];
	logmsg[1] = filename;
	logmsg[2] = open_and_close_strings[17];
	dkapp_log_msg(app, DK_LOG_LEVEL_ERROR, logmsg, 3);
      } break;
      case DK_SF_SEC_DIR: {
        logmsg[0] = open_and_close_strings[26];
	logmsg[1] = filename;
	logmsg[2] = open_and_close_strings[27];
	dkapp_log_msg(app, DK_LOG_LEVEL_ERROR, logmsg, 3);
      } break;
    }
  }
}

typedef struct {
  size_t mpl;		/* maximum path length */
  char   *d1;		/* buffer for directory to search for */
  char   *d2;		/* buffer for GetWindowsDirectory */
  char   *d3;		/* buffer for entire directory */
  int    logenab;	/* logging enabled */
  dk_app_t *a;		/* the application */
  char   *buffer;	/* destination buffer */
  size_t sz;		/* destination buffer size */
  char   *name;		/* name of file to search for */
  int    index_found;	/* -1: not found, 0: plain file, >0: use openstream */
  dk_stream_suffix_t *sl;	/* suffix list */
  int    useunsecdir;	/* use or not current directory */
} file_finder_t;

static void
init_file_finder DK_P1(file_finder_t *,ff)
{
  ff->mpl = 0;
  ff->d1 = ff->d2 = ff->d3 = NULL;
  ff->logenab = 0;
  ff->a = NULL;
  ff->buffer = NULL;
  ff->sz = 0;
  ff->index_found = -1;
  ff->sl = NULL;
  ff->name = NULL;
  ff->useunsecdir = 1;
  
}

static file_finder_t *
new_file_finder DK_P0()
{
  file_finder_t *back = NULL;
  char *d1, *d2, *d3;
  long l; size_t s;
  
  back = dk_new(file_finder_t,1);
  if(back) {
    init_file_finder(back);
    l = dksf_get_maxpathlen();
    if(l <= 0L) l = 1024L;
    s = (size_t)l;
    back ->mpl = s;
    d1 = dk_new(char,s); d2 = dk_new(char,s); d3 = dk_new(char,s);
    if(d1 && d2 && d3) {
      back->d1 = d1; back->d2 = d2; back->d3 = d3;
    } else {
      if(d1) { dk_delete(d1); }
      if(d2) { dk_delete(d2); }
      if(d3) { dk_delete(d3); }
      dk_delete(back);
      back = NULL;
    }
  } 
  return back;
}

static void
delete_file_finder DK_P1(file_finder_t *,ff)
{
  char *x;
  
  if(ff) {
    x = ff->d1;
    if(x) { dk_delete(x); }
    x = ff->d2;
    if(x) { dk_delete(x); }
    x = ff->d3;
    if(x) { dk_delete(x); }
    ff->mpl = 0; ff->d1 = ff->d2 = NULL;
    ff->logenab = 0; ff->a = NULL;
    dk_delete(ff);
  }
  
}

static int
logged_file_check DK_P3(dk_app_t *,a,char *,fn,int,logenab)
{
  int back = 0;
  char *logmsg[3];
  if(fn) {
    if(a && logenab) {
      logmsg[0] = debug_strings[15];
      logmsg[1] = fn;
      logmsg[2] = debug_strings[16];
      dkapp_log_msg(a,DK_LOG_LEVEL_DEBUG,logmsg,3);
    }
    back = file_check(fn);
    if(a && logenab && back) {
      logmsg[0] = debug_strings[17];
      logmsg[1] = fn;
      logmsg[2] = debug_strings[18];
      dkapp_log_msg(a,DK_LOG_LEVEL_DEBUG,logmsg,3);
    }
  }
  return back;
}

static void
ff_dir DK_P2(file_finder_t *,ff,int,usedir)
{
  size_t lgt; int lfd;
  dk_stream_suffix_t *sptr;

  
  lgt = strlen(ff->name);
  if(usedir) {
    lgt += (strlen(ff->d3) + strlen(fn_sep));
  }
  if(lgt < ff->sz) {
    if(usedir) {
      strcpy(ff->buffer, ff->d3);
      strcat(ff->buffer, fn_sep);
      strcat(ff->buffer, ff->name);
    } else {
      strcpy(ff->buffer, ff->name);
    }
    dksf_correct_fnsep(ff->buffer);
    if(logged_file_check(ff->a,ff->buffer,ff->logenab)) {
      ff->index_found = 0;
    }
  }
  if((ff->index_found) == -1) {
  if(ff->sl) {
    sptr = ff->sl; lfd = 1;
    while((sptr->suffix) && ((ff->index_found) == -1)) {
      lgt = strlen(ff->name) + strlen(sptr->suffix);
      if(usedir) {
        lgt += (strlen(ff->d3) + strlen(fn_sep));
      }
      if(lgt < ff->sz) {
        if(usedir) {
	  strcpy(ff->buffer, ff->d3);
	  strcat(ff->buffer, fn_sep);
	  strcat(ff->buffer, ff->name);
        } else {
          strcpy(ff->buffer, ff->name);
        }
	strcat(ff->buffer, sptr->suffix);
	dksf_correct_fnsep(ff->buffer);
	if(logged_file_check(ff->a,ff->buffer,ff->logenab)) {
	  ff->index_found = lfd;
	}
      }
      sptr++; lfd++;
    }
  }
  }
  
}

static void
ff_subs DK_P1(file_finder_t *,ff)
{
  int i1, do_the_check;
  size_t lgt;
  
  i1 = 8;
  while((i1 > 0) && ((ff->index_found) == -1)) {
    i1--;
    do_the_check = 1;
    lgt = strlen(ff->d1);
    if(i1 & 4) {
      if(((ff->a)->loc).l) {
        lgt++;
	lgt += strlen(((ff->a)->loc).l);
      } else {
        do_the_check = 0;
      }
    }
    if(i1 & 2) {
      if(((ff->a)->loc).r) {
        lgt++;
	lgt += strlen(((ff->a)->loc).r);
      } else {
        do_the_check = 0;
      }
    }
    if(i1 & 1) {
      if(((ff->a)->loc).e) {
        lgt++;
	lgt += strlen(((ff->a)->loc).e);
      } else {
        do_the_check = 0;
      }
    }
    if(do_the_check) {
      if(lgt < ff->mpl) {
        strcpy(ff->d3, ff->d1);
	if(i1 & 4) {
	  strcat(ff->d3, fn_sep);
	  strcat(ff->d3, ((ff->a)->loc).l);
	}
	if(i1 & 2) {
	  strcat(ff->d3, fn_sep);
	  strcat(ff->d3, ((ff->a)->loc).r);
	}
	if(i1 & 1) {
	  strcat(ff->d3, fn_sep);
	  strcat(ff->d3, ((ff->a)->loc).e);
	}
	ff_dir(ff, 1);
      }
    }
  }
  
}

char w32_sysconfdir[128];

static void
ff_run DK_P1(file_finder_t *,ff)
{
  int   i1; char *my_sysconfdir;

  my_sysconfdir = kpse_var_value("TEXMFMAIN");
  if(my_sysconfdir) {
    strcpy(w32_sysconfdir, my_sysconfdir);
    strcat(w32_sysconfdir, "/scripts/dktool");
    unix_sysconfdir = w32_sysconfdir;
    free(my_sysconfdir);
  }
  else {
    fprintf(stderr, "I cannot obtain $TEXMFMAIN\n");
    unix_sysconfdir = NULL;
  }
  /* nothing found yet */
  my_sysconfdir = unix_sysconfdir;
  if(ff->a) {
    if(((ff->a)->d).etc) {
      my_sysconfdir = ((ff->a)->d).etc;
    }
  }
  ff->index_found = -1;
  i1 = 10;
  while((i1 > 0) && ((ff->index_found) == -1)) {
    switch(i1) {
      case 10: {
        if(ff->useunsecdir) {
          ff_dir(ff, 0);
	}
      } break;
      case  9: {
        if(ff->useunsecdir) {
          if(((ff->a)->d).a) {
	    if(strlen(((ff->a)->d).a) < ff->mpl) {
	      strcpy(ff->d1, ((ff->a)->d).a);
	      ff_subs(ff);
	    }
	  }
	}
      } break;
      case  8: {
        if(ff->useunsecdir) {
          if((((ff->a)->d).s) && (((ff->a)->n).g)) {
	    if((1+strlen(((ff->a)->d).s) + strlen(((ff->a)->n).g)) < ff->mpl) {
	      strcpy(ff->d1, ((ff->a)->d).s);
	      strcat(ff->d1, fn_sep);
	      strcat(ff->d1, ((ff->a)->n).g);
	      ff_subs(ff);
	    }
	  }
	}
      } break;
      case  7: {
        if(ff->useunsecdir) {
          if(((ff->a)->d).s) {
	    if(strlen(((ff->a)->d).s) < ff->mpl) {
	      strcpy(ff->d1, ((ff->a)->d).s);
	      ff_subs(ff);
	    }
	  }
	}
      } break;
      case  6: {
#if DK_HAVE_GETWINDOWSDIRECTORY
        UINT uSize;
	
	uSize = GetWindowsDirectoryA((LPTSTR)(ff->d2), (UINT)(ff->mpl));
        if((uSize > 0) && (uSize < ((UINT)(ff->mpl)))) {
	  
	  if(((ff->a)->n).a) { 
	  if((1 + strlen(ff->d2) + strlen(((ff->a)->n).a)) < ff->mpl) {
	    
	    strcpy(ff->d1, ff->d2);
	    strcat(ff->d1, fn_sep);
	    strcat(ff->d1, ((ff->a)->n).a);
	    ff_subs(ff);
	  } else {
	    
	  }
	  } else { 
	  }
	} else {
	  i1 = 4;
	  if(uSize == 0) {
	    
	  }
	  
	}
#else
        i1 = 4;
	
#endif
      } break;
      case  5: {
#if DK_HAVE_GETWINDOWSDIRECTORY
        if(((ff->a)->n).g) {
	if((1 + strlen(ff->d2) + strlen(((ff->a)->n).g)) < ff->mpl) {
	  strcpy(ff->d1, ff->d2);
	  strcat(ff->d1, fn_sep);
	  strcat(ff->d1, ((ff->a)->n).g);
	  ff_subs(ff);
	}
	}
#else
        i1 = 4;
#endif
      } break;
      case  4: {
#if DK_HAVE_GETWINDOWSDIRECTORY
        if(strlen(ff->d2) < ff->mpl) {
	  strcpy(ff->d1, ff->d2);
	  ff_subs(ff);
	}
#endif
      } break;
      case  3: {
        if(((ff->a)->n).a) {
	if((1 + strlen(my_sysconfdir) + strlen(((ff->a)->n).a)) < ff->mpl) {
	  strcpy(ff->d1, my_sysconfdir);
	  strcat(ff->d1, fn_sep);
	  strcat(ff->d1, ((ff->a)->n).a);
	  ff_subs(ff);
	}
	}
      } break;
      case  2: {
        if(((ff->a)->n).g) {
	if((1 + strlen(my_sysconfdir) + strlen(((ff->a)->n).g)) < ff->mpl) {
	  strcpy(ff->d1, my_sysconfdir);
	  strcat(ff->d1, fn_sep);
	  strcat(ff->d1, ((ff->a)->n).g);
	  ff_subs(ff);
	}
	}
      } break;
      case  1: {
        if(strlen(my_sysconfdir) < ff->mpl) {
	  strcpy(ff->d1, my_sysconfdir);
	  ff_subs(ff);
	}
      } break;
    }
    i1--;
  }
  if(ff->logenab) {
    char *errmsgs[3];
    if((ff->index_found) == -1) {
      errmsgs[0] = open_and_close_strings[6];
      errmsgs[1] = ff->name;
      errmsgs[2] = open_and_close_strings[7];
      dkapp_log_msg(ff->a,DK_LOG_LEVEL_ERROR,errmsgs,3);
    } else {
      errmsgs[0] = open_and_close_strings[4];
      errmsgs[1] = ff->buffer;
      errmsgs[2] = open_and_close_strings[5];
      dkapp_log_msg(ff->a,DK_LOG_LEVEL_PROGRESS,errmsgs,3);
    }
  }
  
}

static dk_stream_t *
my_read_file DK_P4(dk_app_t *,a,char *,name,int,logenab,int,useunsecdir)
{
  dk_stream_t *back = NULL;
  file_finder_t *ff;
  dk_stream_suffix_t *sl;
  dk_stream_open_fct_t *fct;
  char *fnb;

  
  if(a && name) {
    ff = new_file_finder();
    if(ff) {
      fnb = dk_new(char,(ff->mpl));
      if(fnb) {
        ff->logenab = logenab;
        ff->a = a;
        ff->buffer = fnb;
        ff->sz = ff->mpl;
	ff->name = name;
	ff->useunsecdir = useunsecdir;
	sl = ff->sl = dkstream_get_read_suffixes();
	ff_run(ff);
	if((ff->index_found) > -1) {
	  if((ff->index_found) > 0) {
	    fct = sl[((ff->index_found) - 1)].fct;
	    if(fct) {
	      a->relaxed_fopen_reason = 0;
	      back = ((*fct)(fnb,rb,a->relaxed_fopen_check, &(a->relaxed_fopen_reason)));
	      if(logenab) {
	        if(back) {
		  char *errmsgs[3];
		  errmsgs[0] = open_and_close_strings[8];
		  errmsgs[1] = ff->buffer;
		  errmsgs[2] = open_and_close_strings[9];
		  dkapp_log_msg(a,DK_LOG_LEVEL_PROGRESS,errmsgs,3);
		} else {
	          if(a->relaxed_fopen_reason) {
	            dkapp_err_nowrite(a, fnb, a->relaxed_fopen_reason);
		  } else {
		    char *errmsgs[3];
		    errmsgs[0] = open_and_close_strings[10];
		    errmsgs[1] = ff->buffer;
		    errmsgs[2] = open_and_close_strings[11];
		    dkapp_log_msg(a,DK_LOG_LEVEL_ERROR,errmsgs,3);
		  }
	        }
	      }
	    }
	  } else {
	    back = dkapp_stream_openfile(a, fnb, rb);
	    if(logenab) {
	      if(back) {
		char *errmsgs[3];
		errmsgs[0] = open_and_close_strings[8];
		errmsgs[1] = ff->buffer;
		errmsgs[2] = open_and_close_strings[9];
		dkapp_log_msg(a,DK_LOG_LEVEL_PROGRESS,errmsgs,3);
	      } else {
		char *errmsgs[3];
		errmsgs[0] = open_and_close_strings[10];
		errmsgs[1] = ff->buffer;
		errmsgs[2] = open_and_close_strings[11];
		dkapp_log_msg(a,DK_LOG_LEVEL_ERROR,errmsgs,3);
	      }
	    }
	  }
	}
        dk_delete(fnb);
      } else {
        dkapp_err_memory(a,sizeof(char),(ff->mpl));
      }
      delete_file_finder(ff);
    } else {
      dkapp_err_memory(a,sizeof(file_finder_t),1);
    }
  }
  
  return back;
}

static int
int_find_file DK_P5(dk_app_t *,a,char *,name,char *,buffer,size_t,sz,int,uud)
{
  int back = 0;
  file_finder_t *ff;
  
  if(a && name && buffer && sz) {
    ff = new_file_finder();
    if(ff) {
      ff->logenab = 1;
      ff->a = a;
      ff->buffer = buffer;
      ff->sz = sz;
      ff->name = name;
      ff->sl = NULL;
      ff->useunsecdir = uud;
      ff_run(ff);
      if((ff->index_found) > -1) {
        back = 1;
      }
      delete_file_finder(ff);
    } else {
      dkapp_err_memory(a,sizeof(file_finder_t),1);
    }
  }
  
  return back;
}

int
dkapp_find_file DK_P4(dk_app_t *,a,char *,name,char *,buffer,size_t,sz)
{
  int back = 0;
  
  back = int_find_file(a,name,buffer,sz,1);
  
  return back;
}

int
dkapp_find_cfg DK_P4(dk_app_t *,a,char *,name,char *,buffer,size_t,sz)
{
  int back = 0;
  
  back = int_find_file(a,name,buffer,sz,0);
  
  return back;
}

dk_stream_t *
dkapp_read_file DK_P2(dk_app_t *, a, char *, filename)
{
  dk_stream_t *back = NULL;
  
  back = my_read_file(a,filename,1,1);
  
  return back;
}


dk_stream_t *
dkapp_read_cfg DK_P2(dk_app_t *, a, char *, filename)
{
  dk_stream_t *back = NULL;
  
  back = my_read_file(a,filename,1,0);
  
  return back;
}


static char suffix_stt[] = { ".stt" };

static stt_entry *
find_stt_entry DK_P2(dk_app_t *, app, char *, table)
{
    stt_entry *stteptr = NULL;
    char *buffer, *cptr;
    long maxpathlen;
    dk_stream_t *strm;
    dk_stt_t *sttptr;
    int found;
    
    if(((app->loc).s) && ((app->loc).si)) {
      stteptr =
      (stt_entry *)dksto_it_find_like(((app->loc).si),((void *)table),1);
      if(!stteptr) { 
	maxpathlen = dksf_get_maxpathlen();
	if(maxpathlen < 0L) { maxpathlen = 1024L; }
	buffer = dk_new(char,maxpathlen);
	if(buffer) { 
	  if((strlen(table) + 7) < maxpathlen) {
	    strcpy(buffer, table);
	    found = 0;
	    cptr = dksf_get_file_type_dot(buffer);
	    if(cptr) {
	      if(strcmp(cptr, suffix_stt) == 0) {
		found = 1;
	      }
	    }
	    if(!found) {
	      strcat(buffer, suffix_stt);
	    } 
	    strm = my_read_file(app,buffer,0,1);
	    if(strm) { 
		sttptr = dkstt_open(strm);
		if(sttptr) { 
		  stteptr = dk_new(stt_entry,1);
		  if(stteptr) { 
		    stteptr->name = NULL;
		    stteptr->st = NULL;
		    stteptr->name = dkstr_dup(table);
		    if(stteptr->name) { 
		      stteptr->st = sttptr; 
		      if(!dksto_add(((app->loc).s),((void *)stteptr))) {
			
			stt_entry_free(stteptr);
			stteptr = NULL;
			sttptr = NULL;
		      }
		    } else {
		      stteptr->name = NULL;
		      dk_delete(stteptr);
		      stteptr = NULL;
		      dkstt_close(sttptr);
		      sttptr = NULL;
		    }
		  } else {
		    dkstt_close(sttptr);
		    sttptr = NULL;
		  }
		}
		dkstream_close(strm);
	    }
	  }
	}
	if(buffer) { dk_delete(buffer) ; }
      } 
    } 
      return stteptr;
}

static char *log_level_keywords[] = {
  NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL
};
static dk_string_finder_t log_level_string_finder[] = {
  { (char *)"/ll/n"  , (&(log_level_keywords[0])), (char *)"" },
  { (char *)"/ll/pa" , (&(log_level_keywords[1])), (char *)"PANIC:    " },
  { (char *)"/ll/f"  , (&(log_level_keywords[2])), (char *)"FATAL:    " },
  { (char *)"/ll/e"  , (&(log_level_keywords[3])), (char *)"ERROR:    " },
  { (char *)"/ll/w"  , (&(log_level_keywords[4])), (char *)"Warning:  " },
  { (char *)"/ll/i"  , (&(log_level_keywords[5])), (char *)"" },
  { (char *)"/ll/pr" , (&(log_level_keywords[6])), (char *)"" },
  { (char *)"/ll/d"  , (&(log_level_keywords[7])), (char *)"Debug:    " },
  { (char *)"/ll/ign", (&(log_level_keywords[8])), (char *)"" },
  { NULL, NULL, NULL }
} ;


/*
static char *priority_strings[] = {
  "",
  "PANIC: ",
  "FATAL: ",
  "ERROR: ",
  "Warning: ",
  "Info: ",
  "Progress: ",
  "Debug: ",
  NULL
};
*/

static char *
get_prio_string DK_P1(int, i)
{
  int x;
  char *back = NULL;
  x = i;
  if(x < 1) x = 1;
  if(x > 7) x = 7;
  back = log_level_keywords[x];
  return back;
}


static char *ide_output_words[] = {
  (char *)"line",
  (char *)"error",
  (char *)"warning",
  (char *)"C2000",
  (char *)"C4000",
  NULL
};

static char *tasm_keywords[] = {
  (char *)"",
  (char *)"*Warning*",
  (char *)"**Error**",
  (char *)"**Fatal**",
  NULL
};

static void
file_log DK_P11(FILE *,f,int,fl,char *,ts,int,min,int,p,char **,msg,int,az,unsigned char *,cp,int,ide_type,char *,name,unsigned long,lineno)
{
  int i;
  char *x, *kw;
  
  if(f && (p <= min)) {		
    if(fl & 1) {		
      if(fl & 2) {		
	fputs("# ", f);
      }				
      fputs(ts, f);
      if(fl & 2) {		
	fputc('\n', f);
      }
    }
    
    if(ide_type && name && lineno) {
      switch(ide_type) {
        case DK_APP_IDE_TASM : {
	  
	  kw = tasm_keywords[0];
	  if(p <= DK_LOG_LEVEL_WARNING) {
	    kw = tasm_keywords[1];
	    if(p <= DK_LOG_LEVEL_ERROR) {
	      kw = tasm_keywords[2];
	      if(p <= DK_LOG_LEVEL_FATAL) {
	        kw = tasm_keywords[3];
	      }
	    }
	  }
	  fprintf(f, "%s %s(%lu) ", kw, name, lineno);
	} break;
        case DK_APP_IDE_WORKSHOP : {
	  
	  if(p <= DK_LOG_LEVEL_ERROR) {
	    fprintf(f, "\"%s\", %s %lu: ",
	      name, ide_output_words[0], lineno
	    );
	  } else {
	    if(p == DK_LOG_LEVEL_WARNING) {
	      fprintf(f, "\"%s\". %s %lu: %s: ",
	        name, ide_output_words[0], lineno, ide_output_words[2]
	      );
	    }
	  }
	} break;
	case DK_APP_IDE_MSVC : {
	  
	  if(p <= DK_LOG_LEVEL_ERROR) {
	    fprintf(f, "%s(%lu) : %s %s: ",
	      name, lineno, ide_output_words[1], ide_output_words[3]
	    );
	  } else {
	    if(p == DK_LOG_LEVEL_WARNING) {
	      fprintf(f, "%s(%lu) : %s %s: ",
	        name, lineno, ide_output_words[2], ide_output_words[4]
	      );
	    }
	  }
	} break;
	default : {
	  
	  if(p <= DK_LOG_LEVEL_ERROR) {
	    fprintf(f, "%s:%lu: ", 
	      name, lineno
	    );
	  } else {
	    if(p == DK_LOG_LEVEL_WARNING) {
	      fprintf(f, "%s:%lu: %s: ",
	        name, lineno, ide_output_words[2]
	      );
	    } else {
	      fputs(name, f); fputs(": ", f);
	    }
	  }
	} break;
      }
      
    } else {
      
      if(name) {
        if(lineno) {
          fprintf(f, "%s:%lu: ",
	    name, lineno
	  );
	} else {
	  fputs(name, f); fputs(": ", f);
	}
      }
      x = get_prio_string(p);	
      if(x) { fputs(x, f); }
    }
    
#if DK_HAVE_CODEPAGES
				
    if(cp) {			
      for(i = 0; i < az; i++) {	
	if(msg[i]) {		
	  dkcp_fputs(f, cp, msg[i]);	
	}
      }
    } else {
#else
				
#endif
      for(i = 0; i < az; i++) {	
        if(msg[i]) {		
	  fputs(msg[i], f);
        }
      }
#if DK_HAVE_CODEPAGES
    }
#endif
    
    fputc('\n', f);
    fflush(f);
  } 
}


int
dkapp_log_msg DK_P4(dk_app_t *, a, int, p, char **, msg, int, az)
{
  int back = 0;
  char time_buffer[32];
#if DK_HAVE_TIME_H
  time_t timer;
#endif
#if DK_HAVE_SYSLOG
  size_t lgt;
  int i, xp;
  char *buffer, *ps;
#endif
  
  time_buffer[0] = '\0';
  if(p < ((a->l).max)) { (a->l).max = p; }
#if DK_HAVE_TIME_H
  time(&timer);
  dksf_time_convert(time_buffer,timer);
#endif
  if(a && msg && (az > 0)) {
    
    back = 1;
    
    if(((a->l).nostdio) < 1) {
      file_log(
        stdout, ((a->l).o.f), time_buffer,((a->l).o.m),
	p,msg,az,((a->l).o.c), ((a->l).o.ide_type),
	((a->l).ef.n), ((a->l).ef.lineno)
      );
    }
    
    if(((a->l).nostdio) < 2) {
      file_log(
        stderr,((a->l).e.f),time_buffer,((a->l).e.m),
        p,msg,az,((a->l).e.c), ((a->l).e.ide_type),
	((a->l).ef.n), ((a->l).ef.lineno)
      );
    }
    
    file_log(
      ((a->l).f.t),((a->l).f.f),time_buffer,((a->l).f.m),
      p,msg,az,((a->l).f.c), ((a->l).e.ide_type),
      ((a->l).ef.n), ((a->l).ef.lineno)
    );
    
#if DK_HAVE_SYSLOG
    
    if(((a->l).s.o) && (p <= ((a->l).s.m))) {
      
      ps = get_prio_string(p);
      lgt = strlen(ps);
      for(i = 0; i < az; i++) {
        if(msg[i]) { lgt += strlen(msg[i]); }
      }
      lgt++;
      buffer = dk_new(char,lgt);
      if(buffer) {
        strcpy(buffer, ps);
        for(i = 0; i < az; i++) {
	  if(msg[i]) { strcat(buffer, msg[i]); }
        }
        xp = LOG_INFO;
        switch(p) {
	  case DK_LOG_LEVEL_DEBUG: xp = LOG_DEBUG; break;
	  case DK_LOG_LEVEL_PROGRESS: xp = LOG_DEBUG; break;
	  case DK_LOG_LEVEL_INFO: xp = LOG_INFO; break;
	  case DK_LOG_LEVEL_WARNING: xp = LOG_WARNING; break;
	  case DK_LOG_LEVEL_ERROR: xp = LOG_ERR; break;
	  case DK_LOG_LEVEL_FATAL: xp = LOG_CRIT; break;
	  case DK_LOG_LEVEL_PANIC: xp = LOG_EMERG; break;
        }
        syslog(xp,"%s",buffer);
        dk_delete(buffer);
      }
    }
#endif
  } 
  return back;
}

static char *specials[] = {
  /*	0 */ (char *)"app.name",
  /*	1 */ (char *)"app.dir",
  /*	2 */ (char *)"shared.dir",
  /*	3 */ (char *)"temp.dir",
  /*	4 */ (char *)"user.name",
  /*	5 */ (char *)"user.home",
  /*	6 */ (char *)"user.uid",
  /*	7 */ (char *)"user.gid",
  /*	8 */ (char *)"user.euid",
  /*	9 */ (char *)"user.egid",
  /*	10 */ (char *)"host.name",
  /*	11 */ (char *)"host.domain",
  /*	12 */ (char *)"process.pid",
  /*	13 */ (char *)"process.ppid",
  /*	14 */ (char *)"process.pgid",
  NULL
};

static char str_pref[] = { "pref:" };
static char str_env[]  = { "env:" };
static char str_unknown[] = { "x" };

static int
transformSpecial DK_P3(dk_app_t *,app,char *,str,size_t,sz)
{
  int back = 0;
  int i;
  char *xbuffer, *ptr;
  size_t lgt1, lgt2;
  long l;
  i = dkstr_array_index(specials, str, 0);
  if(i >= 0) {
    switch(i) {
      case 0: { 
        if((app->n).a) {
	  if(strlen((app->n).a) < sz) {
	    strcpy(str, (app->n).a);
	    back = 1;
	  }
	}
      } break;
      case 1: { 
	if((app->d).a) {
	  if(strlen((app->d).a) < sz) {
	    strcpy(str, (app->d).a);
	    back = 1;
	  }
	}
      } break;
      case 2: {
	if((app->d).s) {
	  if(strlen((app->d).s) < sz) {
	    strcpy(str, (app->d).s);
	    back = 1;
	  }
	}
      } break;
      case 3: {
	if((app->d).t) {
	  if(strlen((app->d).t) < sz) {
	    strcpy(str, (app->d).t);
	    back = 1;
	  }
	}
      } break;
      case 4: {
	if((app->n).u) {
	  if(strlen((app->n).u) < sz) {
	    strcpy(str, (app->n).u);
	    back = 1;
	  }
	}
      } break;
      case 5: {
	if((app->d).h) {
	  if(strlen((app->d).h) < sz) {
	    strcpy(str, (app->d).h);
	    back = 1;
	  }
	}
      } break;
      case 6: {
	if(sz > 11) {
	  back = 1;
	  if(dksf_have_getuid()) {
	    l = dksf_getuid();
	    sprintf(str, "%lu", (unsigned long)l);
	  } else {
	    strcpy(str, str_unknown);
	  }
	}
      } break;
      case 7: {
	if(sz > 11) {
	  back = 1;
	  if(dksf_have_getgid()) {
	    l = dksf_getgid();
	    sprintf(str, "%lu", (unsigned long)l);
	  } else {
	    strcpy(str, str_unknown);
	  }
	}
      } break;
      case 8: {
	if(sz > 11) {
	  back = 1;
	  if(dksf_have_geteuid()) {
	    l = dksf_geteuid();
	    sprintf(str, "%lu", (unsigned long)l);
	  } else {
	    strcpy(str, str_unknown);
	  }
	}
      } break;
      case 9: {
	if(sz > 11) {
	  back = 1;
	  if(dksf_have_getegid()) {
	    l = dksf_getegid();
	    sprintf(str, "%lu", (unsigned long)l);
	  } else {
	    strcpy(str, str_unknown);
	  }
	}
      } break;
      case 10: {
	back = dksf_get_hostname(str,sz);
      } break;
      case 11: {
	back = dksf_get_domainname(str,sz);
      } break;
      case 12: {
	if(sz > 11) {
	  back = 1;
	  if(dksf_have_getpid()) {
	    l = dksf_getpid();
	    sprintf(str, "%lu", (unsigned long)l);
	  } else {
	    strcpy(str, str_unknown);
	  }
	}
      } break;
      case 13: {
	if(sz > 11) {
	  back = 1;
	  if(dksf_have_getppid()) {
	    l = dksf_getppid();
	    sprintf(str, "%lu", (unsigned long)l);
	  } else {
	    strcpy(str, str_unknown);
	  }
	}
      } break;
      case 14: {
	if(sz > 11) {
	  back = 1;
	  if(dksf_have_getpgid()) {
	    l = dksf_getpgid(dksf_getpid());
	    sprintf(str, "%lu", (unsigned long)l);
	  } else {
	    strcpy(str, str_unknown);
	  }
	}
      } break;
    }
  } else {
    lgt1 = strlen(str);
    lgt2 = strlen(str_env);
    if(lgt1 > lgt2) {
      if(strncmp(str,str_env,lgt2) == 0) {
	ptr = getenv(&(str[lgt2]));
	if(ptr) {
	  if(strlen(ptr) < sz) {
	    strcpy(str,ptr);
	    back = 1;
	  }
	}
      }
      if(!back) {
	lgt2 = strlen(str_pref);
	if(lgt1 > lgt2) {
	  if(strncmp(str,str_pref,lgt2) == 0) {
	    xbuffer = dk_new(char,sz);
	    if(xbuffer) {
	      strcpy(xbuffer,&(str[lgt2]));
	      back = dkapp_get_pref(app,xbuffer,str,sz,0);
	      dk_delete(xbuffer);
	    }
	  }
	}
      }
    }
  }
  return back;
}

#if DK_HAVE_FEATURE_BACKSLASH
#define FNSBS 1
#else
#define FNSBS 0
#endif

static int fnsbs = FNSBS;

int
dkapp_transform_string_ext1 DK_P5(dk_app_t *,app,char *,dest,size_t,sz,char *,src,int,fn)
{
  int back = 1;
  char tmp[256];
  char *dptr, *tptr, *sptr; size_t dsz, tsz;
  int  state;
  
  if(app && dest && src && sz) { 
    dptr = dest; tptr = tmp; sptr = src; state = 0;
    dsz = tsz = 0;
    while(*sptr) {
      switch(state) {
	case 1: { 
	  switch(*sptr) {
	    case '(': {
	      tptr = tmp; tsz = 0; state = 2;
	    } break;
	    default: {
	      if(dsz + 2 < sz) {
		*(dptr++) = '$';
		*(dptr++) = *sptr;
		dsz++; dsz++;
	      } else {
		back = 0;
	      }
	      state = 0;
	    } break;
	  }
	} break;
        case 2: { 
	  switch(*sptr) {
	    case ')': { 
	      *tptr = (char)0; 
	      if(transformSpecial(app, tmp, sizeof(tmp))) {
		
		if((dsz + strlen(tmp)) < sz) {
		  tptr = tmp;
		  while(*tptr) {
		    *(dptr++) = *(tptr++);
		  }
		} else {
		  back = 0;
		}
	      }
	      state = 0;
	    } break;
	    default: {
	      if(tsz + 1 < sizeof(tmp)) {
		*(tptr++) = *sptr;
		tsz++;
	      } else {
		back = 0;
	      }
	    }
	  }
	} break;
	case 3: { 
	  switch(*sptr) {
	    case '\\':
	    case  '$':
	    case  '/': {
	      if(dsz + 1 < sz) {
		*(dptr++) = *sptr; dsz++;
	      } else { back = 0; }
	    } break;
	    default: {
	      if(dsz + 2 < sz) {
		*(dptr++) = '\\';
		*(dptr++) = *sptr;
		dsz++; dsz++;
	      } else { back = 0; }
	    } break;
	  } state = 0;
	} break;
	default: { 
	  switch(*sptr) {
	    case '$': {
	      state = 1;
	    } break;
	    case '\\': {
	      if(fnsbs && fn) {
	        state = 0;
		if(dsz + 1 < sz) {
		  *(dptr++) = *sptr;
		  dsz++;
		} else {
		  back = 0;
		}
	      } else {
	        state = 3;
	      }
	    } break;
	    case '/' : {
	      state = 0;
	      if(dsz + 1 < sz) {
		*(dptr++) = fn_sep[0];
		dsz++;
	      } else {
		back = 0;
	      }
	    } break;
	    default: {
	      state = 0;
	      if(dsz + 1 < sz) {
		*(dptr++) = *sptr;
		dsz++;
	      } else {
		back = 0;
	      }
	    } break;
	  }
	} break;
      }
      sptr++;
    }
    *dptr = (char)0;
  } 
  return back;
}


int
dkapp_transform_string DK_P4(dk_app_t *,app,char *,dest,size_t,sz,char *,src)
{
  int back;
  back = dkapp_transform_string_ext1(app,dest,sz,src,0);
  return back;
}


static void
my_find_multi DK_P4(dk_app_t *, app, dk_string_finder_t *, f, char *, table,int,logenab)
{
  dk_string_finder_t *ptr;
  stt_entry *stteptr;
  dk_stt_t  *sttptr;
  int found;
  char *errmsgs[3];

  found = 0;
  
  if(app && f && table) {
    stteptr = find_stt_entry(app,table);
    if(stteptr) { 
      sttptr = stteptr->st;
      if(sttptr) { 
	ptr = f; found = 1;
	while((ptr->key) && (ptr->value_pointer) && (ptr->default_value)) {
	  *(ptr->value_pointer) =
	  dkstt_find(sttptr, (ptr->key), (ptr->default_value));
	  ptr++;
	}
      }
    }
  }
  if(!found) { 
    if(f) { 
      ptr = f;
      while((ptr->key) && (ptr->value_pointer) && (ptr->default_value)) {
	*(ptr->value_pointer) = (ptr->default_value);
	ptr++;
      }
    }
    /* ##### ERROR NO STRING TABLE FOUND */
    if(logenab) {
    errmsgs[0] = open_and_close_strings[12];
    errmsgs[1] = table;
    errmsgs[2] = open_and_close_strings[13];
    dkapp_log_msg(app,DK_LOG_LEVEL_WARNING,errmsgs,3);
    }
  } else {
    /* ##### DEBUG STRING TABLE FOUND */
    if(logenab) {
    errmsgs[0] = open_and_close_strings[14];
    errmsgs[1] = table;
    errmsgs[2] = open_and_close_strings[15];
    dkapp_log_msg(app,DK_LOG_LEVEL_DEBUG,errmsgs,3);
    }
  }
  
}


void
dkapp_find_multi DK_P3(dk_app_t *, app, dk_string_finder_t *, f, char *, table)
{
  my_find_multi(app,f,table,1);
}



static char *
my_find_string DK_P5(dk_app_t *,app,char *,table,char *,key,char *,def,int,logenab)
{
  char *back = NULL;
  stt_entry *stteptr;
  char *errmsgs[3];
  
  if(app && table && key) {
    stteptr = find_stt_entry(app,table);
    if(stteptr) {
      if(stteptr->st) {
	back = dkstt_find((stteptr->st),key,def);
      } else {
	/* ##### ERROR NO STRING TABLE FOUND */
	if(logenab) {
	errmsgs[0] = open_and_close_strings[12];
	errmsgs[1] = table;
	errmsgs[2] = open_and_close_strings[13];
	dkapp_log_msg(app,DK_LOG_LEVEL_WARNING,errmsgs,3);
	}
      }
    } else {
        /* ##### ERROR NO STRING TABLE FOUND */
	if(logenab) {
	errmsgs[0] = open_and_close_strings[4];
	errmsgs[1] = table;
	errmsgs[2] = open_and_close_strings[5];
	dkapp_log_msg(app,DK_LOG_LEVEL_DEBUG,errmsgs,3);
	}
    }
  }
  if(!back) {
    back = def;
  } 
  return back;
}


char *
dkapp_find_string DK_P4(dk_app_t *,app,char *,table,char *,key,char *,def)
{
  char *back;
  back = my_find_string(app,table,key,def,1);
  return back;
}


int dkapp_get_argc DK_P1(dk_app_t *,a)
{
  int back = 0;
  if(a) {
    back = (a->a).a.argc;
  }
  return back;
}

char **dkapp_get_argv DK_P1(dk_app_t *,a)
{
  char **back = NULL;
  if(a) {
    back = (a->a).a.argv;
  }
  return back;
}

void dkapp_help DK_P3(dk_app_t *,a,char *,filename,char **,str)
{
  int found;
  char buffer[128], b2[128];
  int  bytes;
  dk_stream_t *strm;
  char **ptr;
#if DK_HAVE_CODEPAGES
  int   convi;
  unsigned char *convptr;
#endif

  if(a && filename && str) {
  found = 0;
  strm = my_read_file(a,filename,1,1);
  if(strm) {
    while((bytes = dkstream_read(strm,buffer,sizeof(buffer))) > 0) {
      if((a->l).f.t) {
        DK_MEMCPY(b2,buffer,bytes) ;
#if DK_HAVE_CODEPAGES
        if((a->l).f.c) {
	  convptr = (unsigned char *)b2;
	  for(convi = 0; convi < bytes; convi++) {
	    *convptr = dkcp_convert((a->l).f.c, *convptr);
	    convptr++;
	  }
	}
#endif
        fwrite(b2,1,bytes,(a->l).f.t);
      }
#if DK_HAVE_CODEPAGES
      if((a->l).o.c) {
        convptr = (unsigned char *)buffer;
	for(convi = 0; convi < bytes; convi++) {
	  *convptr = dkcp_convert((a->l).o.c, *convptr);
	  convptr++;
	}
      }
#endif
      fwrite(buffer,1,bytes,stdout);
      found = 1;
    }
    dkstream_close(strm);
    if(found) {
      fputc('\n', stdout);
      if((a->l).f.t) {
	fputc('\n', (a->l).f.t);
      }
    }
  }
  if(!found) {
    ptr = str;
    while(*ptr) {
#if DK_HAVE_CODEPAGES
      if((a->l).o.c) {
        dkcp_fputs(stdout, (a->l).o.c, *ptr);
	fputc('\n', stdout);
      } else {
	fputs(*ptr, stdout); fputc('\n', stdout);
      }
#else
      fputs(*ptr, stdout); fputc('\n', stdout);
#endif
      if((a->l).f.t) {
#if DK_HAVE_CODEPAGES
        if((a->l).f.c) {
	  dkcp_fputs((a->l).f.t, (a->l).f.c, *ptr);
	  fputc('\n', (a->l).f.t);
	} else {
	  fputs(*ptr, (a->l).f.t); fputc('\n', (a->l).f.t);
	}
#else
	fputs(*ptr, (a->l).f.t); fputc('\n', (a->l).f.t);
#endif
      }
      ptr++;
    }
  }
  }
}

static char wb[] = { "wb" };

dk_stream_t *
dkapp_write_file DK_P2(dk_app_t *, app, char *, filename)
{
  dk_stream_t *back = NULL;
  dk_stream_suffix_t *suffix_ptr, *sptr2;
  dk_stream_open_fct_t *fct;
  char *buffer;
  size_t lgt, maxlgt;

  if(app && filename) {
    suffix_ptr = dkstream_get_write_suffixes();
    sptr2 = suffix_ptr;
    maxlgt = 0;
    while((sptr2->suffix) && (sptr2->fct)) {
      lgt = strlen(sptr2->suffix);
      if(lgt > maxlgt) maxlgt = lgt;
      sptr2++;
    }
    maxlgt += strlen(filename);
    maxlgt++;
    buffer = dk_new(char,maxlgt);
    if(buffer) {
      sptr2 = suffix_ptr;
      while((sptr2->suffix) && (sptr2->fct) && (!back)) {
	fct = sptr2->fct;
	strcpy(buffer, filename);
	strcat(buffer, sptr2->suffix);
	app->relaxed_fopen_reason = 0;
	back = ((*fct)(buffer, wb, app->relaxed_fopen_check, &(app->relaxed_fopen_reason)));
	if((!back) && (app->relaxed_fopen_reason)) {
	  dkapp_err_nowrite(app, buffer, app->relaxed_fopen_reason);
	}
	sptr2++;
      }
      dk_delete(buffer);
    }
  }
  return back;
}

static char str_w[] = { "w" };

void
dkapp_close DK_P1(dk_app_t *, a)
{
  char *cptr, **argptr; int i;
  dk_preference_t *pptr, *ppt2;
  stt_entry *stteptr;
  char *logmsgs[3];
  long maxpathlen;
  char *filename;
#if !DK_HAVE_WINREG_H
  FILE *outputfile;
  int print_it;
#endif
  
  if(a) {
    if((a->n).a) {
      logmsgs[0] = open_and_close_strings[2];
      logmsgs[1] = (a->n).a;
      logmsgs[2] = open_and_close_strings[3];
      dkapp_log_msg(a,DK_LOG_LEVEL_PROGRESS,logmsgs,3);
    }
    /*
      Free string table space
    */
    if(((a->loc).s) && ((a->loc).si)) {
      dksto_it_reset((a->loc).si);
      while((stteptr = (stt_entry *)dksto_it_next((a->loc).si)) != NULL) {
	stt_entry_free(stteptr);
      }
      dksto_close((a->loc).s);
      (a->loc).s = NULL;
      (a->loc).si = NULL;
    }
    /*
      Write preferences back to file
    */
#if !DK_HAVE_WINREG_H
    if(((a->d).h) && ((a->n).a)) {
      maxpathlen = dksf_get_maxpathlen();
      if(maxpathlen < 0L) { maxpathlen = 1024L; }
      filename = dk_new(char,maxpathlen);
      if(filename) {
	if((strlen((a->d).h) + strlen(defaults_sub) + strlen((a->n).a) + 2) < maxpathlen) {
	  strcpy(filename, (a->d).h);
          strcat(filename, fn_sep);
	  strcat(filename, defaults_sub);
	  strcat(filename, fn_sep);
	  strcat(filename, (a->n).a);
#if NO_FOPEN_CHECKING
	  outputfile = fopen(filename, str_w);
#else
	  a->relaxed_fopen_reason = 0;
	  outputfile = dksf_msfo(filename, str_w, a->relaxed_fopen_check, &(a->relaxed_fopen_reason));
	  if((!outputfile) && (a->relaxed_fopen_reason)) {
	    dkapp_err_nowrite(a, filename, a->relaxed_fopen_reason);
	  }
#endif
	  if(!((a->p).unc)) {
	  if(outputfile) {
	    if(((a->p).u) && ((a->p).ui)) {
	      dksto_it_reset((a->p).ui);
              while((pptr = (dk_preference_t *)dksto_it_next((a->p).ui)) != NULL) {
		if((pptr->n) && (pptr->v)) {
		if((pptr->p) & 2) {
		  print_it = 1;
		  if(((a->p).a) && ((a->p).ai)) {
		    ppt2 = (dk_preference_t *)dksto_it_find_like((a->p).ai, pptr->n, 1);
		    if(ppt2) {
		      print_it = 0;
		    }
		  }
		  if(print_it) {
		    fputs(pptr->n, outputfile);
		    fputs("\t=\t", outputfile);
		    fputs(pptr->v, outputfile);
		    fputc('\n', outputfile);
		  }
		}
		}
	      }
	      dksto_it_reset((a->p).ai);
	      while((pptr = (dk_preference_t *)dksto_it_next((a->p).ai)) != NULL) {
		if((pptr->n) && (pptr->v)) {
		  fputs(pptr->n, outputfile);
		  fputs("\t=\t", outputfile);
		  fputs(pptr->v, outputfile);
		  fputc('\n', outputfile);
		}
	      }
	    }
	    fclose(outputfile);
	  }
	  }
	}
        dk_delete(filename);
      }
    }
#endif
    /* 
      Close log system
    */
    if((a->l).f.t) {
      fclose((a->l).f.t);
      (a->l).f.t = NULL;
    }
    if((a->l).f.n) {
      if(((a->l).max) > ((a->l).f.k)) {
        dksf_remove_file((a->l).f.n);
      }
      cptr = (a->l).f.n; dk_delete(cptr);
      (a->l).f.n = NULL;
    }
#if DK_HAVE_SYSLOG
    if((a->l).s.o) {
      closelog();
      (a->l).s.o = 0;
    }
#endif
    if((a->l).o.c) {
      cptr = (char *)((a->l).o.c) ; dk_delete(cptr);
    } (a->l).o.c = NULL;
    if((a->l).e.c) {
      cptr = (char *)((a->l).e.c) ; dk_delete(cptr);
    } (a->l).e.c = NULL;
    if((a->l).f.c) {
      cptr = (char *)((a->l).f.c) ; dk_delete(cptr);
    } (a->l).f.c = NULL;
    /*
      Release names
    */
    if((a->loc).l) { cptr = (a->loc).l; dk_delete(cptr); } (a->loc).l = NULL;
    if((a->loc).r) { cptr = (a->loc).r; dk_delete(cptr); } (a->loc).r = NULL;
    if((a->loc).e) { cptr = (a->loc).e; dk_delete(cptr); } (a->loc).e = NULL;
    if((a->x).d) {
      cptr = (a->x).d; dk_delete(cptr);
    }
    if((a->x).f) {
      cptr = (a->x).f; dk_delete(cptr);
    }
    (a->x).d = NULL; (a->x).f = NULL;
    if((a->d).etc) {
      cptr = (a->d).etc; dk_delete(cptr);
      (a->d).etc = NULL;
    }
    if((a->d).pt) {
      (void)dksf_remove_directory((a->d).pt);
      cptr = (a->d).pt; dk_delete(cptr);
    }
    if((a->d).a) {
      cptr = (a->d).a; dk_delete(cptr);
    }
    if((a->d).s) {
      cptr = (a->d).s; dk_delete(cptr);
    }
    if((a->d).t) {
      cptr = (a->d).t; dk_delete(cptr);
    }
    if((a->d).h) {
      cptr = (a->d).h; dk_delete(cptr);
    }
    /*
    if((a->n).a) {
      cptr = (a->n).a; dk_delete(cptr);
    }
    */
    if((a->n).h) {
      cptr = (a->n).h; dk_delete(cptr);
    }
    if((a->n).u) {
      cptr = (a->n).u; dk_delete(cptr);
    }
    /* (a->n).a = NULL; */ (a->n).h = NULL; (a->n).u = NULL;
    (a->d).s = NULL; (a->d).a = NULL;
    (a->d).h = NULL; (a->d).t = NULL; (a->d).pt = NULL;
    /*
      Release preferences data
    */
    if(((a->p).a) && ((a->p).ai)) {
      dksto_it_reset((a->p).ai);
      while((pptr = (dk_preference_t *)dksto_it_next((a->p).ai)) != NULL) {
	dkpref_delete(pptr);
      }
      dksto_close((a->p).a);
      (a->p).a = NULL; (a->p).ai = NULL;
    }
    if(((a->p).c) && ((a->p).ci)) {
      dksto_it_reset((a->p).ci);
      while((pptr = (dk_preference_t *)dksto_it_next((a->p).ci)) != NULL) {
	dkpref_delete(pptr);
      }
      dksto_close((a->p).c);
      (a->p).c = NULL; (a->p).ci = NULL;
    }
#if DK_HAVE_WINREG_H
    if(((a->p).what) & 1) { RegCloseKey((a->p).hklm_all); }
    if(((a->p).what) & 2) { RegCloseKey((a->p).hklm_app); }
    if(((a->p).what) & 4) { RegCloseKey((a->p).hkcu_all); }
    if(((a->p).what) & 8) { RegCloseKey((a->p).hkcu_app); }
    if((a->p).unc) {
      maxpathlen = 1024L;
      filename = dk_new(char,maxpathlen);
      if(filename) {
        if((strlen((a->n).a) + strlen(hkcu) + 1) < maxpathlen) {
	  strcpy(filename, hkcu);
	  strcat(filename, fn_sep);
	  strcat(filename, (a->n).a);
	  RegDeleteKey(HKEY_CURRENT_USER, filename);
        }
        dk_delete(filename);
      }
    }
#else
    if(((a->p).u) && ((a->p).ui)) {
      dksto_it_reset((a->p).ui);
      while((pptr = (dk_preference_t *)dksto_it_next((a->p).ui)) != NULL) {
	dkpref_delete(pptr);
      }
      dksto_close((a->p).u);
      (a->p).u = NULL; (a->p).ui = NULL;
    }
    if(((a->p).s) && ((a->p).si)) {
      dksto_it_reset((a->p).si);
      while((pptr = (dk_preference_t *)dksto_it_next((a->p).si)) != NULL) {
	dkpref_delete(pptr);
      }
      dksto_close((a->p).s);
      (a->p).s = NULL; (a->p).si = NULL;
    }
#endif
    if((a->n).g) {
      cptr = (a->n).g ; dk_delete(cptr);
    }
    (a->n).g = NULL;
    if((a->n).a) {
      cptr = (a->n).a ; dk_delete(cptr);
    }
    (a->n).a = NULL;
    /*
      Release command line args
    */
    if( (a->a).a.argv ) {
      argptr = (a->a).a.argv ;
      for(i = 0; i < (a->a).o.argc; i++) {
	argptr[i] = NULL;
      }
      dk_delete(argptr);
    }
    if( (a->a).o.argv ) {
      argptr = (a->a).o.argv ;
      if(argptr) {
        for(i = 0; i < (a->a).o.argc; i++) {
	  cptr = argptr[i];
	  if(cptr) dk_delete(cptr);
        }
        dk_delete(argptr);
      }
    }

    /*
      Finally delete the application itself
    */
    app_init_empty(a);
    dk_delete(a);
  } 
}


#if !DK_HAVE_WINREG_H
static char local_default_configuration[] = {
#if DK_HAVE_FEATURE_BACKSLASH
  "C:\\ETC\\ALL.DEF"
#else
  "/etc/appdefaults"
#endif
};
#endif


int
dkapp_get_min_loglevel DK_P1(dk_app_t *,a)
{
  int back = DK_LOG_LEVEL_NONE;
  if(a) {
    back = (a->l).o.m;
    if(back < ((a->l).e.m)) { back = (a->l).e.m; }
    if(back < ((a->l).f.m)) { back = (a->l).f.m; }
#if DK_HAVE_SYSLOG
    if(back < ((a->l).s.m)) { back = (a->l).s.m; }
#endif
  }
  return back;
}


static char key_sec_ign_file_owner[] = { "/sec/ign/link-owner" };
static char key_sec_ign_dir_group_writable[] = { "/sec/ign/dir-group-wriable" };
static char key_sec_ign_dir_world_writable[] = { "/sec/ign/dir-world-wriable" };

static char *ide_keys[] = {
  (char *)"g$cc",
  (char *)"ms$vc",
  (char *)"w$orkshop",
  (char *)"t$asm",
  NULL
};

static char str_dkapp[] = { "dkapp" };
static char str_dkappe[] = { "dkappe" };
static char str_dkappd[] = { "dkappd" };

int
dkapp_ide_type DK_P1(char *,str)
{
  int back = 0;
  
  if(str) {
    back = 1 + dkstr_array_abbr(ide_keys,str,'$',0);
  }
  
  return back;
}

static
dk_app_t *
my_app_open DK_P6(int,argc,char **,argv,char *,grname,char *,etc,int,sil,int,nostdio)
{
  dk_app_t *back = NULL;
  int    i, j, my_set_silent;
  char *cptr, *cptr2, **argptr, **argptr2;
  long maxpathlen, mypid;
  char *buffer, *buffer2, *bptr;
  dk_stat_t st;
  char *logmsgs[3];
#if DK_HAVE_WINREG_H
  HKEY hkTemp;
  PHKEY phkSubkey;
  DWORD disp;
  LONG retval;
#endif
#if DK_HAVE_CODEPAGES
  dk_stream_t *cp_stream;
#endif

  
  maxpathlen = dksf_get_maxpathlen();
  if(maxpathlen < 0L) { maxpathlen = 1024L; }
  buffer = dk_new(char,maxpathlen);
  buffer2 = dk_new(char,maxpathlen);
  if(argc && argv) {
    my_set_silent = dkapp_silence_check(argc,argv);
    if(sil) { my_set_silent = 1; }
    back = dk_new(dk_app_t,1);
    if(back) {
      /*
	Initialize member variables
      */
      app_init_empty(back);

      if(set_silent || my_set_silent || sil) {
        (back->l).f.k = DK_LOG_LEVEL_NONE;
	(back->l).f.m = DK_LOG_LEVEL_NONE;
	(back->l).e.m = DK_LOG_LEVEL_NONE;
	(back->l).o.m = DK_LOG_LEVEL_NONE;
      }
      if(nostdio) {
        (back->l).nostdio = 1;
      }
      /*
	Copy command line arguments
      */
      if(argc > 0) {
	argptr = dk_new(CHARPTR,argc);
	if(argptr) {
	  for(i = 0; i < argc; i++) {
	    argptr[i] = NULL;
	    if(argv[i]) {
	      argptr[i] = dkstr_dup(argv[i]);
	    }
	  }
	  (back->a).o.argc = argc;
	  (back->a).o.argv = argptr;
          argptr2 = dk_new(CHARPTR,argc);
	  if(argptr2) {
	    for(i = 0; i < argc; i++) {
	      argptr2[i] = argptr[i];
	    }
	    (back->a).a.argc = argc;
	    (back->a).a.argv = argptr2;
	  }
	}
      }
      /*
	Find names and some directories
      */
      {
        if(grname) {
	  (back->n).g = dkstr_dup(grname);
	}
        if(etc) {
	  (back->d).etc = dkstr_dup(etc);
	  if((back->d).etc) {
	    dksf_correct_fnsep((back->d).etc);
	  }
	}
	if(buffer) {
	  if(dksf_get_uname(buffer,maxpathlen)) {
	    (back->n).u = dkstr_dup(buffer);
	  }
	  if(dksf_get_hostname(buffer,maxpathlen)) {
	    (back->n).h = dkstr_dup(buffer);
	  }
	  if(argc > 0) {
	  if(argv[0]) {
	    bptr = dkstr_rchr(argv[0], fn_sep[0]);
	    if(bptr) bptr++;
	    else bptr = argv[0];
	    (back->n).a = dkstr_dup(bptr);
	    if((back->n).a) {
	      bptr = dkstr_rchr((back->n).a, '.');
	      if(bptr) {
	        *bptr = (char)0;
	      }
	    }
	  }
	  }
	  if(dksf_get_home(buffer,maxpathlen)) {
	    (back->d).h = dkstr_dup(buffer);
	  }
	  if(dksf_get_tempdir(buffer,maxpathlen)) {
	    (back->d).t = dkstr_dup(buffer);
	  }
	  if(argc > 0) {
	  if(argv[0]) {
	    if(buffer2) {
	      if(dksf_getcwd(buffer2,maxpathlen)) {
	        if(dksf_get_executable(buffer,maxpathlen,buffer2,argv[0],1)) {
		  (back->x).f = dkstr_dup(buffer);
		  bptr = dkstr_rchr(buffer, fn_sep[0]);
		  if(bptr) {
		    *bptr = '\0';
		    (back->x).d = dkstr_dup(buffer);
		  }
	        }
	      }
	    }
	  }
	  }
	  if((back->x).d) {
	    strcpy(buffer, (back->x).d);
	    bptr = dkstr_rchr(buffer, fn_sep[0]);
	    if(bptr) {
	      bptr++;
	      if(strcmp(bptr, bin_dir) == 0) {
		strcpy(bptr, lib_dir);
		(back->d).s = dkstr_dup(buffer);
                if((strlen(buffer) + strlen((back->n).a) + 1) < maxpathlen) {
		  strcat(buffer, fn_sep);
		  strcat(buffer, (back->n).a);
		  (back->d).a = dkstr_dup(buffer);
		}
	      } else  {
		(back->d).a = dkstr_dup(buffer);
		(back->d).s = dkstr_dup(buffer);
	      }
	    }
	  }
	}
      }
      /*
	Initialize preferences management
      */
#if DK_HAVE_WINREG_H
	hkTemp = 0;
	phkSubkey = &hkTemp;
	retval = RegCreateKeyExA(
	  HKEY_LOCAL_MACHINE,
          hklm,
	  0,
	  NULL,
	  REG_OPTION_NON_VOLATILE,
	  KEY_QUERY_VALUE,
	  NULL,
	  phkSubkey,
	  &disp
	);
	if(retval == ERROR_SUCCESS) {
	  (back->p).hklm_all = hkTemp;
	  (back->p).what |= 1;
	}
	if(buffer && ((back->n).a)) {
	if((strlen(hklm) + strlen((back->n).a) + 1) < maxpathlen) {
	  strcpy(buffer, hklm);
	  strcat(buffer, backslash);
	  strcat(buffer, (back->n).a);
	  hkTemp = 0; phkSubkey = &hkTemp;
	  retval = RegCreateKeyExA(
	    HKEY_LOCAL_MACHINE, buffer,
	    0, NULL, REG_OPTION_NON_VOLATILE, KEY_QUERY_VALUE,
	    NULL, phkSubkey, &disp
	  );
	  if(retval == ERROR_SUCCESS) {
	    (back->p).hklm_app = hkTemp;
	    (back->p).what |= 2;
	  }
	}
	}
	hkTemp = 0;
	phkSubkey = &hkTemp;
	retval = RegCreateKeyExA(
	  HKEY_CURRENT_USER,
	  hkcu,
	  0,
	  NULL,
	  REG_OPTION_NON_VOLATILE,
	  KEY_ALL_ACCESS,
	  NULL,
	  phkSubkey,
	  &disp
	);
	if(retval == ERROR_SUCCESS) {
	  (back->p).hkcu_all = hkTemp;
	  (back->p).what |= 4;
	}
	if(buffer && ((back->n).a)) {
	if((strlen(hkcu) + strlen((back->n).a) + 1) < maxpathlen) {
	  strcpy(buffer, hkcu);
	  strcat(buffer, backslash);
	  strcat(buffer, (back->n).a);
	  hkTemp = 0; phkSubkey = &hkTemp;
	  retval = RegCreateKeyExA(
	    HKEY_CURRENT_USER, buffer, 0, NULL,
	    REG_OPTION_NON_VOLATILE, KEY_ALL_ACCESS, NULL,
	    phkSubkey, &disp
	  );
	  if(retval == ERROR_SUCCESS) {
	    (back->p).hkcu_app = hkTemp;
	    (back->p).what |= 8;
	  }
	}
	}
#else
      (back->p).s = dksto_open(DK_STO_SIZE_TINY);
      if((back->p).s) {
	(back->p).si = dksto_it_open((back->p).s);
	if((back->p).si) {
	  dksto_set_comp((back->p).s, dkapp_pref_compare, 0);
	  dksto_use_trees((back->p).s, 0);
	} else {
	  dksto_close((back->p).s);
	  (back->p).s = NULL;
	}
      }
      (back->p).u = dksto_open(DK_STO_SIZE_TINY);
      if((back->p).u) {
	(back->p).ui = dksto_it_open((back->p).u);
	if((back->p).ui) {
	  dksto_set_comp((back->p).u, dkapp_pref_compare, 0);
	  dksto_use_trees((back->p).u, 0);
	} else {
	  dksto_close((back->p).u);
	  (back->p).u = NULL;
	}
      }
#endif
      (back->p).a = dksto_open(DK_STO_SIZE_TINY);
      if((back->p).a) {
	(back->p).ai = dksto_it_open((back->p).a);
	if((back->p).ai) {
	  dksto_set_comp((back->p).a, dkapp_pref_compare, 0);
	  dksto_use_trees((back->p).a, 0);
	} else {
	  dksto_close((back->p).a);
	  (back->p).a = NULL;
	}
      }
      (back->p).c = dksto_open(DK_STO_SIZE_TINY);
      if((back->p).c) {
	(back->p).ci = dksto_it_open((back->p).c);
	if((back->p).ci) {
	  dksto_set_comp((back->p).c, dkapp_pref_compare, 0);
	  dksto_use_trees((back->p).c, 0);
	} else {
	  dksto_close((back->p).c);
	  (back->p).c = NULL;
	}
      }
      argptr = (back->a).a.argv;
      for(i = 0; i < (back->a).a.argc; i++) {
	cptr = argptr[i];
	if(cptr[0] == '-') {
	if(cptr[1] == '-') {
	if(cptr[2] == '/') {
	  for(j = i; j < ((back->a).a.argc - 1); j++) {
	    argptr[j] = argptr[j+1];
	  }
	  (back->a).a.argc = (back->a).a.argc - 1 ;
	  cptr2 = dkstr_chr(cptr,'=');
	  if(cptr2) {
	    *(cptr2++) = (char)0;
	    add_pref_to_list((back->p).c, (back->p).ci, &(cptr[2]), cptr2, 7);
	  }
	  i--;
	}
	}
	}
      }
      /*
	Read preferences from files
      */
#if !DK_HAVE_WINREG_H
      if(buffer) {
	  if((back->d).s) { 
	    if(((back->p).s) && ((back->p).si)) { 
	      if((strlen((back->d).s) + strlen(appdef) + 1) < maxpathlen) {
		
		strcpy(buffer, (back->d).s);
		strcat(buffer, fn_sep);
		strcat(buffer, appdef);
		
		read_cfg(
		  (back->p).s, (back->p).si, buffer,
		  (back->n).u, (back->n).a, (back->n).h,
		  0
		);
	      }
	      if((back->n).a) {
if((strlen((back->d).s)+strlen(appdef)+strlen((back->n).a)+2) < maxpathlen)
{
		strcpy(buffer, (back->d).s);
		strcat(buffer, fn_sep);
		strcat(buffer, appdef);
		strcat(buffer, dot);
		strcat(buffer, (back->n).a);
		
		read_cfg(
		  (back->p).s, (back->p).si, buffer,
		  (back->n).u, (back->n).a, (back->n).h,
		  2
		);
}
	      }
	    }
	  }
	  if(((back->p).s) && ((back->p).si)) {
	    
	    read_cfg(
	      (back->p).s, (back->p).si,
	      local_default_configuration,
	      (back->n).u, (back->n).a, (back->n).h,
	      0
	    );
#if DK_HAVE_FEATURE_BACKSLASH
	    if((back->n).a) {
	      if((strlen(local_default_configuration) + strlen((back->n).a) + 4) < maxpathlen) {
		strcpy(buffer, local_default_configuration);
		cptr = dkstr_chr(buffer, fn_sep[0]);
		if(cptr) {
		  cptr++;
		  strcpy(cptr, (back->n).a);
		  strcat(cptr, ".DEF");
		  
		  read_cfg(
		    (back->p).s, (back->p).si, buffer,
		    (back->n).u, (back->n).a, (back->n).h,
		    2
		  );
		}
	      }
	    }
#else
	    if((back->n).a) {
	    if((strlen((back->n).a) + strlen(local_default_configuration) + 2) < maxpathlen) {
	    }
	      strcpy(buffer, local_default_configuration);
              strcat(buffer, ".");
	      strcat(buffer, (back->n).a);
	      
	      read_cfg(
		(back->p).s, (back->p).si, buffer,
		(back->n).u, (back->n).a, (back->n).h,
		2
	      );
	    }
#endif
	  }
	  if((back->d).h) { 
	    if((strlen((back->d).h) + strlen(defaults_sub) + 1) < maxpathlen) {
	      strcpy(buffer, (back->d).h);
	      strcat(buffer, fn_sep);
	      strcat(buffer, defaults_sub);
	      dksf_mkdir(buffer, DK_PERM_CREATE_PROTECTED);
	      if((strlen(buffer) + strlen(all_files) + 1) < maxpathlen) {
		strcat(buffer, fn_sep);
		strcat(buffer, all_files);
                if(((back->p).u) && ((back->p).ui)) {
		  
		  read_cfg(
		    (back->p).u, (back->p).ui, buffer,
                    (back->n).u, (back->n).a, (back->n).h,
		    4
		  );
		}
	      }
	      strcpy(buffer, (back->d).h);
	      strcat(buffer, fn_sep);
	      strcat(buffer, defaults_sub);
	      if((back->n).a) {
		if((strlen(buffer) + strlen((back->n).a) + 1) < maxpathlen) {
		  strcat(buffer, fn_sep);
		  strcat(buffer, (back->n).a);
		  if(((back->p).u) && ((back->p).ui)) {
		    
		    read_cfg(
		      (back->p).u, (back->p).ui, buffer,
		      (back->n).u, (back->n).a, (back->n).h,
		      6
		    );
		  }
		}
	      }
	    }
	  }
      }
#endif
      /*
	Correct file and dir names
      */
      
      if(buffer && buffer2) {
        
	if(dkapp_get_pref(back,key_sec_ign_file_owner,buffer,maxpathlen,(~(DK_APP_PREF_EXCL_CMD)))) {
	  if(dkstr_is_on(buffer)) {
	    back->relaxed_fopen_check |= DK_SF_SEC_OWNER;
	  }
	}
	if(dkapp_get_pref(back,key_sec_ign_dir_world_writable,buffer,maxpathlen,(~(DK_APP_PREF_EXCL_CMD)))) {
	  if(dkstr_is_on(buffer)) {
	    back->relaxed_fopen_check |= DK_SF_SEC_WO;
	    back->relaxed_fopen_check |= DK_SF_SEC_WG;
	  }
	} else {
	  if(dkapp_get_pref(back,key_sec_ign_dir_group_writable,buffer,maxpathlen,(~(DK_APP_PREF_EXCL_CMD)))) {
	    if(dkstr_is_on(buffer)) {
	      back->relaxed_fopen_check |= DK_SF_SEC_WG;
	    }
	  }
	}
	
	if(dkapp_get_pref(back,log_file_name,buffer,maxpathlen,0)) {
	  if(dkapp_transform_string_ext1(back,buffer2,maxpathlen,buffer,1)) {
	    dksf_correct_fnsep(buffer2);
            (back->l).f.n = dkstr_dup(buffer2);
	  }
	}
	if(dkapp_get_pref(back,app_dir,buffer,maxpathlen,0)) {
	  if(dkapp_transform_string_ext1(back,buffer2,maxpathlen,buffer,1)) {
	    dksf_correct_fnsep(buffer2);
            if(dkstat_get(&st,buffer2)) {
	      if((st.filetype & (~(DK_FT_SYMLINK))) == DK_FT_DIR) {
		cptr = dkstr_dup(buffer2);
		if(cptr) {
		  if((back->d).a) {
		    cptr2 = (back->d).a;
		    dk_delete(cptr2); (back->d).a = NULL;
		  }
		  (back->d).a = cptr;
		}
	      }
	    }
	  }
	}
	if(dkapp_get_pref(back,shared_dir,buffer,maxpathlen,0)) {
	  if(dkapp_transform_string_ext1(back,buffer2,maxpathlen,buffer,1)) {
	    dksf_correct_fnsep(buffer2);
            if(dkstat_get(&st,buffer2)) {
	      if((st.filetype & (~(DK_FT_SYMLINK))) == DK_FT_DIR) {
		cptr = dkstr_dup(buffer2);
		if(cptr) {
		  if((back->d).s) {
		    cptr2 = (back->d).s;
		    dk_delete(cptr2); (back->d).s = NULL;
		  }
		  (back->d).s = cptr;
		}
	      }
	    }
	  }
	}
	if(dkapp_get_pref(back,tmp_dir,buffer,maxpathlen,0)) {
	  if(dkapp_transform_string_ext1(back,buffer2,maxpathlen,buffer,1)) {
	    dksf_correct_fnsep(buffer2);
	    if(dkstat_get(&st,buffer2)) {
	      if((st.filetype & (~(DK_FT_SYMLINK))) == DK_FT_DIR) {
		cptr = dkstr_dup(buffer2);
		if(cptr) {
		  if((back->d).t) {
		    cptr2 = (back->d).t;
		    dk_delete(cptr2); (back->d).t = NULL;
		  }
		  (back->d).t = cptr;
		}
	      }
	    } else {
	      (void)dksf_mkdir(buffer2, DK_PERM_CREATE_DIR);
	      if(dkstat_get(&st,buffer2)) {
		if((st.filetype & (~(DK_FT_SYMLINK))) == DK_FT_DIR) {
		  cptr = dkstr_dup(buffer2);
		  if(cptr) {
		    if((back->d).t) {
		      cptr2 = (back->d).t;
		      dk_delete(cptr2); (back->d).t = NULL;
		    }
		    (back->d).t = cptr;
		  }
		}
	      }
	    }
	  }
	}
	if((back->d).t) {
          if(dksf_have_getpid()) {
	    if((strlen((back->d).t) + 13) < maxpathlen) {
	      mypid = dksf_getpid();
	      sprintf(buffer2, "%08lX", mypid);
	      strcpy(buffer, (back->d).t);
	      strcat(buffer, fn_sep);
	      strcat(buffer, buffer2);
	      strcat(buffer, dot_tmp); 
	      if(!dkstat_get(&st,buffer)) {
		
		(void)dksf_mkdir(buffer, DK_PERM_CREATE_DIR);
		if(dkstat_get(&st,buffer)) {
		  
		  if(((st.filetype) & (~(DK_FT_SYMLINK))) == DK_FT_DIR) {
		    
		    cptr = dkstr_dup(buffer);
		    if(cptr) {
		      
		      (back->d).pt = cptr;
		    } else {
		      
		      dksf_remove_directory(buffer);
		    }
		  }
		}
	      }
	    }
	  } else {
	    unsigned long myul;
	    myul = 1UL;
	    while(myul != 0UL) {
	      if((strlen((back->d).t) + 13) < maxpathlen) {
		sprintf(buffer2, "%08lX", myul);
		strcpy(buffer, (back->d).t);
		strcat(buffer, fn_sep);
		strcat(buffer, buffer2);
		strcat(buffer, dot_tmp);
		if(!dkstat_get(&st,buffer)) {
		  (void)dksf_mkdir(buffer, DK_PERM_CREATE_DIR);
		  if(dkstat_get(&st,buffer)) {
		    if(((st.filetype) & (~(DK_FT_SYMLINK))) == DK_FT_DIR) {
		      cptr = dkstr_dup(buffer);
		      if(cptr) {
			(back->d).pt = cptr;
			myul = 0UL;
		      } else {
			dksf_remove_directory(buffer);
			myul = 0UL;
		      }
		    } else {
		      myul++;
		    }
		  } else {
		    myul++;
		  }
		} else {
		  myul++;
		}
	      } else {
		myul = 0UL;
	      }
	    }
	  }
	}
#if DK_HAVE_CODEPAGES
	if(dkapp_get_pref(back,log_file_codepage,buffer,maxpathlen,0)) {
	  if(strcmp(minus_sign, buffer)) {
	  cp_stream = my_read_file(back, buffer, 0, 1);
	  if(cp_stream) {
	    (back->l).f.c = dkcp_open(cp_stream);
	    dkstream_close(cp_stream);
	  }
	  cp_stream = NULL;
	  }
	}
	if(dkapp_get_pref(back,log_stdout_codepage,buffer,maxpathlen,0)) {
	  if(strcmp(minus_sign, buffer)) {
	  cp_stream = my_read_file(back, buffer, 0, 1);
	  if(cp_stream) {
	    (back->l).o.c = dkcp_open(cp_stream);
	    dkstream_close(cp_stream);
	  }
	  cp_stream = NULL;
	  }
	}
	if(dkapp_get_pref(back,log_stderr_codepage,buffer,maxpathlen,0)) {
	  if(strcmp(minus_sign, buffer)) {
	  cp_stream = my_read_file(back, buffer, 0, 1);
	  if(cp_stream) {
	    (back->l).e.c = dkcp_open(cp_stream);
	    dkstream_close(cp_stream);
	  }
	  cp_stream = NULL;
	  }
	}
#endif
      }
      /*
	Set up logging
      */
      if(buffer) {
	char *appname;
	size_t lgt;
	int i;
	if(!(set_silent || my_set_silent || sil)) {
	  if(dkapp_get_pref(back,log_file_level,buffer,maxpathlen,0)) {
	    i = get_log_level(buffer);
	    if(i >= 0) {
	      (back->l).f.m = i;
	    }
	  }
	  if(dkapp_get_pref(back,log_file_keep,buffer,maxpathlen,0)) {
	    i = get_log_level(buffer);
	    if(i >= 0) {
	      (back->l).f.k = i;
	    }
	  }
	}
	i = 0;
	if(dkapp_get_pref(back,log_file_time,buffer,maxpathlen,0)) {
	  if(dkstr_is_on(buffer)) {
	    i = 1;
	  }
	}
	if(dkapp_get_pref(back,log_file_split,buffer,maxpathlen,0)) {
	  if(dkstr_is_on(buffer)) {
	    i |= 2;
	  }
	}
	(back->l).f.f = i;
	if(dkapp_get_pref(back,log_file_ide,buffer,maxpathlen,0)) {
	  (back->l).f.ide_type = dkapp_ide_type(buffer);
	  /* 1 + dkstr_array_abbr(ide_keys,buffer,'$',0); */
	  
	}
	if(!(set_silent || my_set_silent || sil)) {
	  if(dkapp_get_pref(back,log_stdout_level,buffer,maxpathlen,0)) {
	    i = get_log_level(buffer);
	    if(i >= 0) {
	      (back->l).o.m = i;
	    }
	  }
	}
	i = 0;
	if(dkapp_get_pref(back,log_stdout_time,buffer,maxpathlen,0)) {
	  if(dkstr_is_on(buffer)) {
	    i = 1;
	  }
	}
	if(dkapp_get_pref(back,log_stdout_split,buffer,maxpathlen,0)) {
	  if(dkstr_is_on(buffer)) {
	    i |= 2;
	  }
	}
	(back->l).o.f = i;
	if(dkapp_get_pref(back,log_stdout_ide,buffer,maxpathlen,0)) {
	  (back->l).o.ide_type =
	  1 + dkstr_array_abbr(ide_keys, buffer, '$', 0);
	  
	}
	if(!(set_silent || my_set_silent || sil)) {
	  if(dkapp_get_pref(back,log_stderr_level,buffer,maxpathlen,0)) {
	    i = get_log_level(buffer);
	    if(i >= 0) {
	      (back->l).e.m = i;
	    }
	  }
	}
	i = 0;
	if(dkapp_get_pref(back,log_stderr_time,buffer,maxpathlen,0)) {
	  if(dkstr_is_on(buffer)) {
	    i = 1;
	  }
	}
	if(dkapp_get_pref(back,log_stderr_split,buffer,maxpathlen,0)) {
	  if(dkstr_is_on(buffer)) {
	    i |= 2;
	  }
	}
	(back->l).e.f = i;
	if(dkapp_get_pref(back,log_stderr_ide,buffer,maxpathlen,0)) {
	  (back->l).e.ide_type =
	  1 + dkstr_array_abbr(ide_keys, buffer, '$', 0);
	  
	}
#if DK_HAVE_SYSLOG
	i = DK_APP_PREF_EXCL_CMD;
	i |= DK_APP_PREF_EXCL_PROG;
	i |= DK_APP_PREF_EXCL_USER;
        if(dkapp_get_pref(back,log_syslog_level,buffer,maxpathlen,i)) {
	  i = get_log_level(buffer);
	  if(i >= 0) {
	    (back->l).s.m = i;
	  }
	}
#endif
	appname = (back->n).a;
	appname = (appname ? appname : app);
	lgt = strlen(appname) + strlen(suffix_log) + 1;
        if(dksf_have_getpid()) {
	  lgt += 16;
        }
	if(lgt < maxpathlen) {
	  strcpy(buffer, appname);
	  if(dksf_have_getpid()) {
	    strcat(buffer, dot);
	    appname = buffer;
	    while(*appname) appname++;
	    sprintf(appname, "%lu", (unsigned long)dksf_getpid());
	  }
	  strcat(buffer, dot);
	  strcat(buffer, suffix_log);
	}
	if(!((back->l).f.n)) {
	  (back->l).f.n = dkstr_dup(buffer);
	}
	
	if((back->l).f.n) {
#if NO_FOPEN_CHECKING
	  (back->l).f.t = fopen((back->l).f.n, str_w);
#else
	  back->relaxed_fopen_reason = 0;
	  (back->l).f.t = dksf_msfo((back->l).f.n, str_w, back->relaxed_fopen_check, &(back->relaxed_fopen_reason));
	  if((!((back->l).f.t)) && (back->relaxed_fopen_reason)) {
	    dkapp_err_nowrite(back, (back->l).f.n, back->relaxed_fopen_reason);
	  }
#endif
	}
#if DK_HAVE_SYSLOG
	appname = (back->n).a;
	appname = (appname ? appname : app);
	if((back->l).s.m > DK_LOG_LEVEL_NONE) {
	  (back->l).s.o = 1;
	  openlog(appname, (LOG_PID|LOG_CONS), LOG_USER);
	}
#endif
	i = DK_APP_PREF_EXCL_USER | DK_APP_PREF_EXCL_SYSTEM;
	if(dkapp_get_pref(back,ui_lang,buffer,maxpathlen,i)) {
	  save_lang(back,buffer);
	} else {
	  int use_env_lang;
	  use_env_lang = 1;
	  if(dkapp_get_pref(back,ui_lang_env,buffer,maxpathlen,0)) {
	    if(dkstr_is_on(buffer)) {
	      use_env_lang = 1;
	    } else {
	      use_env_lang = 0;
	    }
	  }
	  if(cptr) free(cptr);
          cptr = (char *)malloc(16);
          strcpy(cptr, "en");
	  if(cptr && use_env_lang) {
	    if(strlen(cptr) < maxpathlen) {
	      strcpy(buffer,cptr);
	      save_lang(back,buffer);
	    }
	  } else {
	    if(dkapp_get_pref(back,ui_lang,buffer,maxpathlen,0)) {
	      save_lang(back,buffer);
	    }
	  }
	}
	if(dkapp_get_pref(back,storage_trees,buffer,maxpathlen,0)) {
	  if(dkstr_is_on(buffer)) {
	    dksto_use_trees(NULL,1);
	  } else {
	    dksto_use_trees(NULL,0);
	  }
	}
      }
      cptr = (back->loc).l;
      if(cptr) {
	while(*cptr) {
	  if(isupper(*cptr)) { *cptr = tolower(*cptr); }
	  cptr++;
	}
      }
      cptr = (back->loc).r;
      if(cptr) {
	while(*cptr) {
	  if(isupper(*cptr)) { *cptr = tolower(*cptr); }
	  cptr++;
	}
      }
      cptr = (back->loc).e;
      if(cptr) {
	while(*cptr) {
	  if(isupper(*cptr)) { *cptr = tolower(*cptr); }
	  cptr++;
	}
      }

      /*
	Set up string table handling
      */
      (back->loc).s = dksto_open(DK_STO_SIZE_TINY);
      if((back->loc).s) {
	(back->loc).si = dksto_it_open((back->loc).s);
	if((back->loc).si) {
	  dksto_set_comp((back->loc).s, stt_entry_comp, 0);
	  dksto_use_trees((back->loc).s, 0);
	} else {
	  dksto_close((back->loc).s);
	  (back->loc).s = NULL;
	}
      }
      my_find_multi(back,log_level_string_finder,str_dkapp,0);
      my_find_multi(back,open_and_close_finder,str_dkapp,0);
      if(!app_err_conf) {
	my_find_multi(back,app_err_find,str_dkappe,0);
	app_err_conf = 1;
      }
      /*
	Debug output
      */
      if(grname && (!((back->n).g))) {
        dkapp_err_memory(back, sizeof(char), (1+strlen(grname)));
      }
      if(etc && (!((back->d).etc))) {
        dkapp_err_memory(back, sizeof(char), (1+strlen(etc)));
      }
      if(dkapp_get_min_loglevel(back) >= DK_LOG_LEVEL_DEBUG) {
	my_find_multi(back,debug_string_finder,str_dkappd,0);
	logmsgs[0] = debug_strings[0];
	logmsgs[1] = (((back->n).u) ? ((back->n).u) : (debug_strings[13]));
	dkapp_log_msg(back,DK_LOG_LEVEL_DEBUG,logmsgs,2);
	logmsgs[0] = debug_strings[1];
	logmsgs[1] = (((back->n).a) ? ((back->n).a) : (debug_strings[13]));
	dkapp_log_msg(back,DK_LOG_LEVEL_DEBUG,logmsgs,2);
	if((back->n).g) {
	  logmsgs[0] = debug_strings[19];
	  logmsgs[1] = (back->n).g;
	  dkapp_log_msg(back,DK_LOG_LEVEL_DEBUG,logmsgs,2);
	}
	logmsgs[0] = debug_strings[2];
	logmsgs[1] = (((back->n).h) ? ((back->n).h) : (debug_strings[13]));
	dkapp_log_msg(back,DK_LOG_LEVEL_DEBUG,logmsgs,2);
	logmsgs[0] = debug_strings[3];
	logmsgs[1] = (((back->d).h) ? ((back->d).h) : (debug_strings[13]));
	dkapp_log_msg(back,DK_LOG_LEVEL_DEBUG,logmsgs,2);
	logmsgs[0] = debug_strings[4];
	logmsgs[1] = (((back->d).t) ? ((back->d).t) : (debug_strings[13]));
	dkapp_log_msg(back,DK_LOG_LEVEL_DEBUG,logmsgs,2);
	logmsgs[0] = debug_strings[20];
	logmsgs[1] = (((back->d).etc) ? ((back->d).etc) : unix_sysconfdir);
	dkapp_log_msg(back,DK_LOG_LEVEL_DEBUG,logmsgs,2);
	logmsgs[0] = debug_strings[14];
	logmsgs[1] = (((back->d).t) ? ((back->d).t) : (debug_strings[13]));
	dkapp_log_msg(back,DK_LOG_LEVEL_DEBUG,logmsgs,2);
	logmsgs[0] = debug_strings[5];
	logmsgs[1] = (((back->d).a) ? ((back->d).a) : (debug_strings[13]));
	dkapp_log_msg(back,DK_LOG_LEVEL_DEBUG,logmsgs,2);
	logmsgs[0] = debug_strings[6];
	logmsgs[1] = (((back->d).s) ? ((back->d).s) : (debug_strings[13]));
	dkapp_log_msg(back,DK_LOG_LEVEL_DEBUG,logmsgs,2);
	logmsgs[0] = debug_strings[7];
	logmsgs[1] = (((back->x).f) ? ((back->x).f) : (debug_strings[13]));
	dkapp_log_msg(back,DK_LOG_LEVEL_DEBUG,logmsgs,2);
	logmsgs[0] = debug_strings[8];
	logmsgs[1] = (((back->x).d) ? ((back->x).d) : (debug_strings[13]));
	dkapp_log_msg(back,DK_LOG_LEVEL_DEBUG,logmsgs,2);
	logmsgs[0] = debug_strings[9];
	logmsgs[1] = (((back->l).f.n) ? ((back->l).f.n) : (debug_strings[13]));
	dkapp_log_msg(back,DK_LOG_LEVEL_DEBUG,logmsgs,2);
	logmsgs[0] = debug_strings[10];
	logmsgs[1] = (((back->loc).l) ? ((back->loc).l) : (debug_strings[13]));
	dkapp_log_msg(back,DK_LOG_LEVEL_DEBUG,logmsgs,2);
	logmsgs[0] = debug_strings[11];
	logmsgs[1] = (((back->loc).r) ? ((back->loc).r) : (debug_strings[13]));
	dkapp_log_msg(back,DK_LOG_LEVEL_DEBUG,logmsgs,2);
	logmsgs[0] = debug_strings[12];
	logmsgs[1] = (((back->loc).e) ? ((back->loc).e) : (debug_strings[13]));
	dkapp_log_msg(back,DK_LOG_LEVEL_DEBUG,logmsgs,2);
      } else {
        my_find_multi(back,debug_string_finder,NULL,0);
      }
      
      
      
      
      
      
      
      
      
      
#if DK_HAVE_WINREG_H
      
#endif
      
      
      
      
      if((back->n).a) {
        logmsgs[0] = open_and_close_strings[0];
        logmsgs[1] = (back->n).a;
        logmsgs[2] = open_and_close_strings[1];
	dkapp_log_msg(back,DK_LOG_LEVEL_PROGRESS,logmsgs,3);
      }
    }
  }
  if(buffer2) {
    dk_delete(buffer2);
  }
  if(buffer) {
    dk_delete(buffer);
  }
  
  return back;
}

dk_app_t *
dkapp_open DK_P2(int, argc, char **, argv)
{
  dk_app_t *back = NULL;
  back = my_app_open(argc, argv, NULL, NULL, 0, 0);
  return back;
}

dk_app_t *
dkapp_open_ext1 DK_P6(int,argc,char **,argv,char *,grname,char *,etc,int,sil,int,nostd)
{
  dk_app_t *back = NULL;
  back = my_app_open(argc, argv, grname, etc, sil, nostd);
  return back;
}

int dkapp_tmpnam DK_P3(dk_app_t *, a, char *, buffer, size_t, sz)
{
  int back = 0;
  char number_buffer[16];
  if(a && buffer && sz) {
    if((a->d).pt) {
      if((strlen((a->d).pt) + 13) < sz) {
	sprintf(number_buffer, "%08lX", (a->td).l);
	(a->td).l += 1UL;
	strcpy(buffer, (a->d).pt);
	strcat(buffer, fn_sep);
	strcat(buffer, number_buffer);
	strcat(buffer, dot_tmp);
	back = 1;
      }
    }
    if(!back) {
      if((a->d).t) {
	if((strlen((a->d).t) + 13) < sz) {
	  sprintf(number_buffer, "%08lX", (a->td).l);
	  (a->td).l += 1UL;
	  strcpy(buffer, (a->d).t);
	  strcat(buffer, fn_sep);
	  strcat(buffer, number_buffer);
	  strcat(buffer, dot_tmp);
	  back = 1;
	}
      }
    }
  }
  return back;
}

void
dkapp_unconfigure DK_P1(dk_app_t *,a)
{
  if(a) {
    (a->p).unc = 1;
  }
}

void dkapp_err_traverse_dir DK_P2(dk_app_t *, app, char *, dirname)
{
  char *logmsg[3];
  if(app && dirname) {
    logmsg[0] = app_err_str[0];
    logmsg[1] = dirname;
    logmsg[2] = app_err_str[1];
    dkapp_log_msg(app, DK_LOG_LEVEL_ERROR, logmsg, 3);
  }
}

void dkapp_err_stat_failed DK_P2(dk_app_t *, app, char *, name)
{
  char *logmsg[3];
  if(app && name) {
    logmsg[0] = app_err_str[2];
    logmsg[1] = name;
    logmsg[2] = app_err_str[3];
    dkapp_log_msg(app, DK_LOG_LEVEL_ERROR, logmsg, 3);
  }
}

void dkapp_err_cwd DK_P1(dk_app_t *, app)
{
  char *logmsg[3];
  if(app) {
    logmsg[0] = app_err_str[4];
    dkapp_log_msg(app, DK_LOG_LEVEL_ERROR, logmsg, 1);
  }
}

void dkapp_err_matchfile DK_P2(dk_app_t *, app, char *, name)
{
  char *logmsg[3];
  if(app && name) {
    logmsg[0] = app_err_str[8];
    logmsg[1] = name;
    logmsg[2] = app_err_str[9];
    dkapp_log_msg(app, DK_LOG_LEVEL_ERROR, logmsg, 3);
  }
}

void dkapp_err_matchdir DK_P2(dk_app_t *, app, char *, name)
{
  char *logmsg[3];
  if(app && name) {
    logmsg[0] = app_err_str[10];
    logmsg[1] = name;
    logmsg[2] = app_err_str[11];
    dkapp_log_msg(app, DK_LOG_LEVEL_ERROR, logmsg, 3);
  }
}

void dkapp_err_fopenr DK_P2(dk_app_t *, app, char *, name)
{
  char *logmsg[3];
  if(app && name) {
    logmsg[0] = app_err_str[12];
    logmsg[1] = name;
    logmsg[2] = app_err_str[13];
    dkapp_log_msg(app, DK_LOG_LEVEL_ERROR, logmsg, 3);
  }
}

void dkapp_err_fopenw DK_P2(dk_app_t *, app, char *, name)
{
  char *logmsg[3];
  if(app && name) {
    logmsg[0] = app_err_str[14];
    logmsg[1] = name;
    logmsg[2] = app_err_str[15];
    dkapp_log_msg(app, DK_LOG_LEVEL_ERROR, logmsg, 3);
  }
}

void dkapp_err_fwrite DK_P2(dk_app_t *, app, char *, name)
{
  char *logmsg[3];
  if(app && name) {
    logmsg[0] = app_err_str[16];
    logmsg[1] = name;
    logmsg[2] = app_err_str[17];
    dkapp_log_msg(app, DK_LOG_LEVEL_ERROR, logmsg, 3);
  }
}


int
dkapp_log_level_in_use DK_P2(dk_app_t *, app, int, loglevel)
{
  int back = 0;
  if(app) {
    if(((app->l).o.m) >= loglevel) {
      back = 1;
    }
    if(((app->l).e.m) >= loglevel) {
      back = 1;
    }
    if(((app->l).f.m) >= loglevel) {
      back = 1;
    }
#if DK_HAVE_SYSLOG
    if(((app->l).s.m) >= loglevel) {
      back = 1;
    }
#endif
  }
  return back;
}

void dkapp_err_fread DK_P2(dk_app_t *, app, char *, name)
{
  char *logmsg[3];
  if(app && name) {
    logmsg[0] = app_err_str[18];
    logmsg[1] = name;
    logmsg[2] = app_err_str[19];
    dkapp_log_msg(app, DK_LOG_LEVEL_ERROR, logmsg, 3);
  }
}

int
dkapp_get_relaxed_fopen_check DK_P1(dk_app_t *,app)
{
  int back = 0;
  if(app) {
    back = app->relaxed_fopen_check;
  }
  return back;
}

dk_stream_t *
dkapp_stream_openfile DK_P3(dk_app_t *,a,char *,n,char *,m)
{
  dk_stream_t *back = NULL;
  int ign, reason;
  if(n && m) {
    reason = ign = 0;
    if(a) { ign = a->relaxed_fopen_check; }
    back = dkstream_openfile(n,m,ign,&reason);
    if(a && (!back) && reason) {
      dkapp_err_nowrite(a,n,reason);
    }
  }
  return back;
}

dk_stream_t *
dkapp_stream_opengz DK_P3(dk_app_t *,a,char *,n,char *,m)
{
  dk_stream_t *back = NULL;
  int ign, reason;
  if(n && m) {
    reason = ign = 0;
    if(a) { ign = a->relaxed_fopen_check; }
    back = dkstream_opengz(n,m,ign,&reason);
    if(a && (!back) && reason) {
      dkapp_err_nowrite(a,n,reason);
    }
  }
  return back;
}

dk_stream_t *
dkapp_stream_openbz2 DK_P3(dk_app_t *,a,char *,n,char *,m)
{
  dk_stream_t *back = NULL;
  int ign, reason;
  if(n && m) {
    reason = ign = 0;
    if(a) { ign = a->relaxed_fopen_check; }
    back = dkstream_openbz2(n,m,ign,&reason);
    if(a && (!back) && reason) {
      dkapp_err_nowrite(a,n,reason);
    }
  }
  return back;
}

FILE *
dkapp_fopen DK_P3(dk_app_t *,a,char *,n,char *,m)
{
  FILE *back = NULL;
  int ign, reason;
  if(n && m) {
    ign = reason = 0;
    if(a) { ign = a->relaxed_fopen_check; }
    back = dksf_msfo(n,m,ign,&reason);
    if(a && (!back) && reason) {
      dkapp_err_nowrite(a,n,reason);
    }
  }
  return back;
}

char *
dkapp_get_appname DK_P1(dk_app_t *,a)
{
  char *back = NULL;
  if(a) {
    back = (a->n).a;
  }
  return back;
}

void
dkapp_set_silent DK_P2(dk_app_t *,a,int,v)
{
  char buffer[32];
  size_t maxpathlen;
  int i;
  maxpathlen = sizeof(buffer);
  if(a) {
    if(v) {
      (a->l).f.m = DK_LOG_LEVEL_NONE;
      (a->l).f.k = DK_LOG_LEVEL_NONE;
      (a->l).e.m = DK_LOG_LEVEL_NONE;
      (a->l).o.m = DK_LOG_LEVEL_NONE;
    } else {
      if(dkapp_get_pref(a,log_file_level,buffer,maxpathlen,0)) {
        i = get_log_level(buffer);
        if(i >= 0) {
          (a->l).f.m = i;
        }
      }
      if(dkapp_get_pref(a,log_file_keep,buffer,maxpathlen,0)) {
        i = get_log_level(buffer);
        if(i >= 0) {
          (a->l).f.k = i;
        }
      }
      if(dkapp_get_pref(a,log_stdout_level,buffer,maxpathlen,0)) {
        i = get_log_level(buffer);
        if(i >= 0) {
          (a->l).o.m = i;
        }
      }
      if(dkapp_get_pref(a,log_stderr_level,buffer,maxpathlen,0)) {
        i = get_log_level(buffer);
        if(i >= 0) {
          (a->l).e.m = i;
        }
      }
    }
  } else {
    set_silent = (v ? 1 : 0);
  }
}



EXTERN void
dkapp_set_nostdio DK_P2(dk_app_t *,a, int,v)
{
}

int
dkapp_set_groupname DK_P2(dk_app_t *,a,char *,gn)
{
  int back = 0;
  char *ptr, *cptr;
  if(a && gn) {
    ptr = dkstr_dup(gn);
    if(ptr) {
      if((a->n).g) {
        cptr = (a->n).g ; dk_delete(cptr);
	(a->n).g = NULL;
      }
      (a->n).g = ptr; back = 1;
    } else {
      dkapp_err_memory(a, sizeof(char), (1+strlen(gn)));
    }
  }
  return back;
}

void
dkapp_err_tcpip DK_P1(dk_app_t *,app)
{
  if(app) dkapp_log_msg(app, DK_LOG_LEVEL_ERROR, &(app_err_str[20]), 1);
}


void
dkapp_err_no_such_file DK_P2(dk_app_t *, app, char *,pat)
{
  char *msgptr[3];
  if(app && pat) {
    msgptr[0] = open_and_close_strings[22];
    msgptr[1] = pat;
    msgptr[2] = open_and_close_strings[23];
    dkapp_log_msg(app, DK_LOG_LEVEL_ERROR, msgptr, 3);
  }
}


void
dkapp_err_multiple_files DK_P2(dk_app_t *, app, char *,pat)
{
  char *msgptr[3];
  if(app && pat) {
    msgptr[0] = open_and_close_strings[24];
    msgptr[1] = pat;
    msgptr[2] = open_and_close_strings[25];
    dkapp_log_msg(app, DK_LOG_LEVEL_ERROR, msgptr, 3);
  }
}


void
dkapp_set_source_filename DK_P2(dk_app_t *,a,char *,n)
{
  if(a) {
    (a->l).ef.n = n;
  }
}

void
dkapp_set_source_lineno DK_P2(dk_app_t *,a,unsigned long,lineno)
{
  if(a) {
    (a->l).ef.lineno = lineno;
  }
}

unsigned char *
dkapp_get_stdout_codepage DK_P1(dk_app_t *,a)
{
  unsigned char *back = NULL;
  if(a) { back = (a->l).o.c; }
  return back;
}

char *
dkapp_fne_one DK_P3(dk_app_t *,a,dk_fne_t *,f,char *,p)
{
  char *back = NULL;
  char *ptr;
  if(a && f && p) {
    if(dkfne_next(f)) {
      ptr = dkfne_get_fullname(f);
      if(ptr) {
        back = dkstr_dup(ptr);
	if(back) {
	  if(dkfne_next(f)) {
	    dk_delete(back);
	    back = NULL;
	    dkapp_err_multiple_files(a,p);
	  }
	} else {
	  dkapp_err_memory(a,1,strlen(ptr));
	}
      } else {
        dkapp_err_no_such_file(a,p);
      }
    } else {
      dkapp_err_no_such_file(a,p);
    }
  }
  return back;
}


#ifndef LINT
static char sccs_id[] = {
  "@(#)dkapp.ctr 1.101 12/18/03 \t(krause) - application module"
};
#endif

