/*
 * dkbf - bit field handling
 * Copyright (C) 2001 - Dirk Krause
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 * In this package the copy of the GNU Lesser General Public License
 * is placed in file COPYING.
 */

#include <dk.h>
#include <dkmem.h>
#define DK_BF_C 1
#include <dkbf.h>

typedef unsigned char UC;

dk_bitfield_t *
dkbf_open DK_P1(size_t, bits)
{
  dk_bitfield_t *back = NULL;
  size_t bytes;
  if(bits) {
    bytes = (bits / 8) + 1;
    back = dk_new(dk_bitfield_t,1);
    if(back) {
      back->data = NULL;
      back->bits = bits;
      back->data = dk_new(UC,bytes);
      if(back->data) {
	DK_MEMRES((back->data),bytes) ;
      } else {
	dk_delete(back);
	back = NULL;
      }
    }
  }
  return back;
}

void
dkbf_close DK_P1(dk_bitfield_t *, bf)
{
  char *ptr;
  if(bf) {
    ptr = (char *)(bf->data);
    if(ptr) { dk_delete(ptr); }
    bf->data = NULL; bf->bits = 0;
    dk_delete(bf);
  }
}

static
unsigned char 
the_bits[] = {
  0x80, 0x40, 0x20, 0x10, 0x08, 0x04, 0x02, 0x01
};

void
dkbf_set DK_P3(dk_bitfield_t *, bf, size_t, bit, int, val)
{
  size_t i1, i2;
  if(bf) {
    if((bf->bits) > bit) {
      i1 = bit / 8;
      i2 = bit % 8;
      if(val) {
	(bf->data)[i1] |= the_bits[i2];
      } else {
	(bf->data)[i2] &= (~(the_bits[i2]));
      }
    }
  }
}

int
dkbf_get DK_P2(dk_bitfield_t *, bf, size_t, bit)
{
  int back = 0;
  size_t i1, i2;
  if(bf) {
    if((bf->bits) > bit) {
      i1 = bit / 8;
      i2 = bit % 8;
      if( ( (bf->data)[i1] ) & (the_bits[i2]) ) {
	back = 1;
      }
    }
  }
  return back;
}


#ifndef LINT
static char sccs_id[] = {
"@(#)dkbf.ctr 1.4 08/25/03\t(krause) - bitfield module"
};
#endif

