/*
 * dkcp - Dirk Krause's libraries / codepages module
 * Copyright (C) 2001 - Dirk Krause
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 * In this package the copy of the GNU Lesser General Public License
 * is placed in file COPYING.
 */

#include <dk.h>
#include <dktypes.h>
#include <dkmem.h>
#include <dkstream.h>
#include <dkstr.h>

#define DK_CP_C 1
#include <dkcp.h>

#include <stdio.h>

typedef unsigned char UCHAR;

unsigned char *
dkcp_open DK_P1(dk_stream_t *,st)
{
  unsigned char *back = NULL;
#if DK_HAVE_CODEPAGES
  unsigned index, i1, i2;
  unsigned char uc;
  int can_continue;
  char input_line[256];
  char s1[sizeof(input_line)], s2[sizeof(input_line)];
  char *cptr;

  if(st) {
    back = dk_new(UCHAR,256);
    if(back) {
      for(index = 0; index < 256; index++) {
	uc = index;
	back[index] = uc;
      }
      can_continue = 1;
      while(can_continue) {
	if(dkstream_gets(st, input_line, sizeof(input_line))) {
	  cptr = dkstr_chr(input_line, '#');
	  if(cptr) *cptr = '\0';
	  if(sscanf(input_line, "%s %s", s1, s2) == 2) {
	    if(sscanf(s1, "%u", &i1) == 1) {
	      if(i1 > 255) i1 = i1 & 0xFF;
	      if(sscanf(s2, "%u", &i2) == 1) {
		if(i2 > 255) i2 = i2 & 0xFF;
		uc = i2;
		back[i1] = uc;
	      } else {
		if(strcmp(s2, "-") == 0) {
		  back[i1] = ' ';
		}
	      }
	    }
	  }
	} else {
	  can_continue = 0;
	}
      }
    }
  }
#endif
  return back;
}

unsigned char
dkcp_convert DK_P2(unsigned char *, codepage, unsigned char, c)
{
  unsigned char back;
#if DK_HAVE_CODEPAGES
  unsigned index;
  if(codepage) {
    index = c;
    if(index > 255)  index = 255;
    back = codepage[index];
  } else {
    back = c;
  }
#else
  back = c;
#endif
  return back;
}

void
dkcp_fputs DK_P3(FILE *, out, unsigned char *, cp, char *, str)
{
  if(out && str) {
    if(cp) {
#if DK_HAVE_CODEPAGES
      char c, *cptr;
      cptr = str;
      while(*cptr) {
	c = (char)dkcp_convert(cp, (unsigned char)(*cptr));
	fputc(c, out);
	cptr++;
      }
#else
      fputs(str, out);
#endif
    } else {
      fputs(str, out);
    }
  }
}

#ifndef LINT
static char sccs_id[] = {
"@(#)dkcp.ctr 1.5 08/25/03\t(krause) - codepage support module"
};
#endif

