/* 
 * The dk libraries (libdkport + libdkc)
 * Copyright (C) 2000, 2001 - Dirk Krause
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 * In this package the copy of the GNU Library General Public License
 * is placed in file COPYING.
 */

#include <dk.h>
#include <dktypes.h>

#define DK_ENC_C 1
#include <dkenc.h>


#line 29 "dkenc.ctr"


typedef unsigned long UL;
typedef unsigned short US;
typedef unsigned char UC;

#if !DK_WORDS_BIGENDIAN

static unsigned long	swapul	DK_P1(unsigned long, x)
{
  unsigned long back;
  unsigned char *uc1, *uc2;
  int i;
  uc1 = (unsigned char *)(&x);
  uc2 = (unsigned char *)(&back);
  for(i = 0; i < sizeof(UL); i++) {
    uc2[i] = uc1[(sizeof(UL) - 1 - i)];
  }
  return back;
}

static unsigned short	swapus	DK_P1(unsigned short, x)
{
  unsigned short back;
  unsigned char *uc1, *uc2;
  int i;
  uc1 = (unsigned char *)(&x);
  uc2 = (unsigned char *)(&back);
  for(i = 0; i < sizeof(US); i++) {
    uc2[i] = uc1[(sizeof(US) - 1 - i)];
  }
  return back;
}

#endif

unsigned long dkenc_ntohl	DK_P1(unsigned long, x)
{
  unsigned long back;
#if DK_WORDS_BIGENDIAN
  back = x;
#else
  back = swapul(x);
#endif
  return back;
}

unsigned long dkenc_htonl	DK_P1(unsigned long, x)
{
  unsigned long back;
#if DK_WORDS_BIGENDIAN
  back = x;
#else
  back = swapul(x);
#endif
  return back;
}

unsigned short
dkenc_ntohs DK_P1(unsigned short, x)
{
  unsigned short back;
#if DK_WORDS_BIGENDIAN
  back = x;
#else
  back = swapus(x);
#endif
  return back;
}

unsigned short
dkenc_htons DK_P1(unsigned short, x)
{
  unsigned short back;
#if DK_WORDS_BIGENDIAN
  back = x;
#else
  back = swapus(x);
#endif
  return back;
}

int
dkenc_uc2utf8 DK_P3(dk_unicodechar, c, dk_utf8 *, u8p, int, u8l)
{
  int back = 0;
  dk_unicodechar x;
  if(u8p) {
  if(u8l > 0) {
    if(c > 0x7FUL) {
      if(c > 0x000007FFUL) {
	if(c > 0x0000FFFFUL) {
	  if(c > 0x001FFFFFUL) {
	    if(c > 0x03FFFFFFUL) {
	      if(c < 0x80000000UL) {		/* 6 Byte */
		if(u8l >= 6) {
		  back = 6;
		  x = c >> 30;
		  x &= 0x00000001UL;
		  x |= 0x000000FCUL;
		  u8p[0] = (dk_utf8)x;
		  x = c >> 24; x &= 0x0000003FUL; x |= 0x00000080UL;
		  u8p[1] = (dk_utf8)x;
		  x = c >> 18; x &= 0x0000003FUL; x |= 0x00000080UL;
		  u8p[2] = (dk_utf8)x;
		  x = c >> 12; x &= 0x0000003FUL; x |= 0x00000080UL;
		  u8p[3] = (dk_utf8)x;
		  x = c >>  6; x &= 0x0000003FUL; x |= 0x00000080UL;
		  u8p[4] = (dk_utf8)x;
		  x = c & 0x0000003FUL; x |= 0x00000080UL;
		  u8p[5] = (dk_utf8)x;
		}
	      }
	    } else {				/* 5 Byte */
	      if(u8l >= 5) {
		back = 5;
		x = c >> 24;
		x &= 0x00000003UL;
		x |= 0x000000F8UL;
		u8p[0] = (dk_utf8)x;
		x = c >> 18; x &= 0x0000003FUL; x |= 0x00000080UL;
		u8p[1] = (dk_utf8)x;
		x = c >> 12; x &= 0x0000003FUL; x |= 0x00000080UL;
		u8p[2] = (dk_utf8)x;
		x = c >>  6; x &= 0x0000003FUL; x |= 0x00000080UL;
		u8p[3] = (dk_utf8)x;
		x = c & 0x0000003FUL; x |= 0x00000080UL;
		u8p[4] = (dk_utf8)x;
	      }
	    }
	  } else {				/* 4 Byte */
	    if(u8l >= 4) {
	      back = 4;
	      x = c >> 18;
	      x &= 0x00000007UL;
	      x |= 0x000000F0UL;
	      u8p[0] = (dk_utf8)x;
	      x = c >> 12;
	      x &= 0x0000003FUL;
	      x |= 0x00000080UL;
	      u8p[1] = (dk_utf8)x;
	      x = c >>  6;
	      x &= 0x0000003FUL;
	      x |= 0x00000080UL;
	      u8p[2] = (dk_utf8)x;
	      x = c & 0x0000003FUL;
	      x |= 0x00000080UL;
	      u8p[3] = (dk_utf8)x;
	    }
	  }
	} else {				/* 3 Byte */
	  if(u8l >= 3) {
	    back = 3;
	    x = c >> 12;
	    x &= 0x0000000FUL;
	    x |= 0x000000E0UL;
	    u8p[0] = (dk_utf8)x;
	    x = c >>  6;
	    x &= 0x0000003FUL;
	    x |= 0x00000080UL;
	    u8p[1] = (dk_utf8)x;
	    x = c & 0x0000003FUL;
	    x |= 0x00000080UL;
	    u8p[2] = (dk_utf8)x;
	  }
	}
      } else {					/* 2 Byte */
	if(u8l >= 2) {
	  back = 2;
	  x = c >> 6;
	  x &= 0x0000001FUL;
	  x |= 0x000000C0UL;
	  u8p[0] = (dk_utf8)x;
	  x = c & 0x0000003FUL;
	  x |= 0x00000080UL;
	  u8p[1] = (dk_utf8)x;
	}
      }
    } else {					/* 1 Byte */
      if(u8l >= 1) {
	back = 1;
	u8p[0] = (dk_utf8)(c & 0x0000007FUL);
      }
    }
  }
  }
  return back;
}



/* ********************************************************************* */
/* *                                                                   * */
/* *   0 on error                                                      * */
/* *   1 on success, *u8u contains number of input bytes processed     * */
/* *                                                                   * */
/* ********************************************************************* */

#line 222 "dkenc.ctr"

int
dkenc_utf82uc DK_P4(dk_unicodechar *, ucp, dk_utf8 *, u8p, int , u8l, int *, u8u)
{
  int back = 0;
  dk_utf8	u1;
  dk_unicodechar res, x1, x2, x3, x4, x5, x6;
  int needed_length;
  needed_length = 0;
  if(ucp) {
  if(u8p) {
  if(u8l) {
  if(u8u) {
  if((u8l) > 0) {
    u1 = *u8p; needed_length = 1;
    if(u1 > 0x7F) {					/* > 1 Byte */
      if((u1 & 0xE0) == 0xC0) {				/* 2 Byte */
	needed_length = 2;
	if(u8l >= needed_length) {
	if((ucp[1] & 0xC0) == 0x80) {
	  x1 = u1;
	  x1 = x1 << 6;
	  x1 = x1 & 0x000007C0UL;
	  x2 = ucp[1];
	  x2 = x2 & 0x0000003FUL;
	  res = (x1 | x2); back = 1;
	}
	}
      } else {
	if((u1 & 0xF0) == 0xE0) {			/* 3 Byte */
	  needed_length = 3;
	  if(u8l >= needed_length) {
	    if((ucp[1] & 0xC0) == 0x80) {
	    if((ucp[2] & 0xC0) == 0x80) {
	      x1 = u1;
	      x1 = x1 << 12;
	      x1 = x1 & 0x0000F000UL;
	      x2 = ucp[1];
	      x2 = x2 << 6;
	      x2 = x2 & 0x00000FC0UL;
	      x3 = ucp[2];
	      x3 = x3 & 0x0000003FUL;
	      res = (x1 | x2 | x3); back = 1;
	    }
	    }
	  }
	} else {
	  if((u1 & 0xF8) == 0xF0) {			/* 4 Byte */
	    needed_length = 4;
	    if(u8l >= needed_length) {
	      if((ucp[1] & 0xC0) == 0x80) {
	      if((ucp[2] & 0xC0) == 0x80) {
	      if((ucp[3] & 0xC0) == 0x80) {
		x1 = u1;
		x2 = ucp[1];
		x3 = ucp[2];
		x4 = ucp[3];
		x1 = x1 << 18;
		x1 = x1 & 0x001C0000UL;
		x2 = x2 << 12;
		x2 = x2 & 0x0003F000UL;
		x3 = x3 << 6;
		x3 = x3 & 0x00000FC0UL;
		x4 = x4 & 0x0000003FUL;
		res = (x1 | x2 | x3 | x4); back = 1;
	      }
	      }
	      }
	    }
	  } else {
	    if((u1 & 0xFC) == 0xF8) {			/* 5 Byte */
	      needed_length = 5;
	      if(u8l >= needed_length) {
		if((ucp[1] & 0xC0) == 0x80) {
		if((ucp[2] & 0xC0) == 0x80) {
		if((ucp[3] & 0xC0) == 0x80) {
		if((ucp[4] & 0xC0) == 0x80) {
		  x1 = u1;
		  x2 = ucp[1];
		  x3 = ucp[2];
		  x4 = ucp[3];
		  x5 = ucp[4];
		  x1 = x1 << 24;
		  x1 = x1 & 0x03000000UL;
		  x2 = x2 << 18;
		  x2 = x2 & 0x00FC0000UL;
		  x3 = x3 << 12;
		  x3 = x3 & 0x0003F000UL;
		  x4 = x4 <<  6;
		  x4 = x4 & 0x00000FC0UL;
		  x5 = x5 & 0x0000003FUL;
		  res = (x1 | x2 | x3 | x4 | x5); back = 1;
		}
		}
		}
		}
	      }
	    } else {
	      if((u1 & 0xFE) == 0xFC) {			/* 6 Byte */
		needed_length = 6;
		if(u8l >= needed_length) {
		  if((ucp[1] & 0xC0) == 0x80) {
		  if((ucp[2] & 0xC0) == 0x80) {
		  if((ucp[3] & 0xC0) == 0x80) {
		  if((ucp[4] & 0xC0) == 0x80) {
		  if((ucp[5] & 0xC0) == 0x80) {
		    x1 = 0x40000000UL;
		    x2 = ucp[1]; x3 = ucp[2]; x4 = ucp[3];
		    x5 = ucp[4]; x6 = ucp[5];
		    x2 = x2 << 24;
		    x3 = x3 << 18;
		    x4 = x4 << 12;
		    x5 = x5 <<  6;
		    x2 = x2 & 0x3F000000UL;
		    x3 = x3 & 0x00FC0000UL;
		    x4 = x4 & 0x0003F000UL;
		    x5 = x5 & 0x00000FC0UL;
		    x6 = x6 & 0x0000003FUL;
		    res = (x1 | x2 | x3 | x4 | x5 | x6);
		    back = 1;
		  }
		  }
		  }
		  }
		  }
		}
	      }
	    }
	  }
	}
      }
    } else {						/* 1 Byte */
      res = u1;
      back = 1;
    }
  }
  }
  }
  }
  }
  if(back) {
    *u8u = needed_length;
    *ucp = res;
  }
  return back;
}

#if !DK_HAVE_ISDIGIT
static int
my_is_digit DK_P1(char,c)
{
  int back = 0;
  switch(c) {
    case '0': case '1': case '2': case '3': case '4':
    case '5': case '6': case '7': case '8': case '9':
    {
      back = 1;
    }
    break;
  }
  return back;
}
#endif

static unsigned long char_to_ul DK_P1(char, c)
{
  unsigned long back = 0UL;
  if(c == '1') back = 1UL;
  if(c == '2') back = 2UL;
  if(c == '3') back = 3UL;
  if(c == '4') back = 4UL;
  if(c == '5') back = 5UL;
  if(c == '6') back = 6UL;
  if(c == '7') back = 7UL;
  if(c == '8') back = 8UL;
  if(c == '9') back = 9UL;
  return back;
}

int
dkenc_ipaddr_to_ul DK_P2(char *,str,unsigned long *,ul)
{
  int back = 0;
  int state; char *ptr;
  unsigned long ul1, ul2, ul3, ulval;
  
  if(str && ul) {
    state = 0; ptr = str; back = 1;
    ul1 = ul2 = ul3 = ulval = 0UL;
    while(back && (*ptr)) {
      
      if(isdigit(*ptr)) {
        
	switch(state) {
	  case 0:
	  case 1:
	  case 2:
	  case 4:
	  case 5:
	  case 6:
	  case 8:
	  case 9:
	  case 10:
	  case 12:
	  case 13:
	  case 14: {
	    ulval = 10UL * ulval + char_to_ul(*ptr); state++;
	    
	  } break;
	  default: {
	    back = 0;
	  } break;
	}
      } else {
        if(*ptr == '.') {
	  
	  switch(state) {
	    case 0:
	    case 1:
	    case 2:
	    case 3: {
	      ul1 = ulval; ulval = 0UL; state = 4;
	    } break;
	    case 4:
	    case 5:
	    case 6:
	    case 7: {
	      ul2 = ulval; ulval = 0UL; state = 8;
	    } break;
	    case 8:
	    case 9:
	    case 10:
	    case 11: {
	      ul3 = ulval; ulval = 0UL; state = 12;
	    } break;
	  }
	} else {
	  back = 0; 
	}
      }
      ptr++;
    }
    if((state < 12) || (state > 15)) {
      back = 0;
      
    }
    if(back) {
      if(ul1 > 255UL) back = 1;
      if(ul2 > 255UL) back = 1;
      if(ul3 > 255UL) back = 1;
      if(ulval > 255UL) back = 1;
    }
    if(back) {
      ul1 = ul1 << 24;
      ul1 &= 0xFF000000UL;
      ul2 = ul2 << 16;
      ul2 &= 0x00FF0000UL;
      ul3 = ul3 <<  8;
      ul3 &= 0x0000FF00UL;
      ulval &= 0x000000FFUL;
      ulval = ulval | ul1 | ul2 | ul3;
      *ul = ulval;
    }
  }
  
  return back;
}

#ifndef LINT
static char sccs_id[] = {
"@(#)dkenc.ctr\t1.6\t08/25/03\t(krause) - Encoding module"
};
#endif

