/* 
 * The dk libraries (libdkport + libdkc)
 * Copyright (C) 2000, 2001 - Dirk Krause
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 * In this package the copy of the GNU Library General Public License
 * is placed in file COPYING.
 */


#include <dk.h>

#include <stdio.h>
#if DK_TIME_WITH_SYS_TIME
#include <sys/time.h>
#include <time.h>
#else
#if DK_HAVE_SYS_TIME_H
#include <sys/time.h>
#else
#if DK_HAVE_TIME_H
#include <time.h>
#endif
#endif
#endif
#include <dktypes.h>
#include <dkmem.h>
#include <dkss.h>
#define DK_LOG_C 1
#include <dklog.h>
#include <dklogc.h>

static char *my_strings[] = {
  (char *)"loglevels",
  (char *)"/log/level/panic",
  (char *)"PANIC",
  (char *)"/log/level/fatal",
  (char *)"FATAL",
  (char *)"/log/level/error",
  (char *)"ERROR",
  (char *)"/log/level/warning",
  (char *)"Warning",
  (char *)"/log/level/info",
  (char *)"Info",
  (char *)"/log/level/progress",
  (char *)"Progress",
  (char *)"/log/level/debug",
  (char *)"Debug",
  (char *)":",
  (char *)" ",

  NULL
};

static void log_to_stderr	DK_P4(void *, obj, int, l, char **, p, int, n)
{
  char **lfdptr;
  char *ptr;
  int i;
  int xl;
  
  if(p) {
#if DK_HAVE_TIME_H || DK_HAVE_SYS_TIME_H
    time_t timer;
    struct tm *tm;
    time(&timer);
    tm = localtime(&timer);
    if(tm) {
      fprintf(stderr,
	"# %04d/%02d/%02d %02d:%02d:%02d\n",
	(1900 + tm->tm_year),
	(1 + tm->tm_mon),
	tm->tm_mday,
	tm->tm_hour,
	tm->tm_min,
	tm->tm_sec
      );
    }
#endif
    xl = l;
    if(xl < 0) { xl = 0; }
    if(xl > DK_LOG_LEVEL_DEBUG) { xl = DK_LOG_LEVEL_DEBUG; }
    fprintf(stderr,
      "%s%s%s",
      dkss_find(
	my_strings[0],
	my_strings[1+2*xl],
	my_strings[2+2*xl]
      ),
      my_strings[15], my_strings[16]
    );
    lfdptr = p;
    for(i = 0; i < n; i++) {
      ptr = *(lfdptr++);
      if(ptr) {
	fprintf(stderr, "%s", ptr);
      }
    }
    fprintf(stderr, "\n"); 
  }
  
}

static dk_log_t default_object = {
  NULL,
  log_to_stderr,
  DK_LOG_LEVEL_PROGRESS,
  DK_LOG_LEVEL_IGNORE
};

static dk_log_t *the_object = NULL;

static int is_in_log = 0;

dk_log_t *dklog_new	DK_P0()
{
  dk_log_t *back = NULL;
  back = dk_new(dk_log_t,1);
  
  return back;
}

dk_log_t *dklog_new_for	DK_P2(void *, obj, dk_fct_type_log *, fct)
{
  dk_log_t *back = NULL;
  
  if(fct) {
    back = dk_new(dk_log_t,1);
    if(back) {
      back->obj = obj;
      back->fct = fct;
      back->level = DK_LOG_LEVEL_PROGRESS;
      back->maxlevel = DK_LOG_LEVEL_IGNORE;
    }
  } 
  return back;
}

void     dklog_delete	DK_P1(dk_log_t *, p)
{
  
  if(p) {
    p->obj = NULL; p->fct = NULL;
    dk_delete(p);
  }
}

dk_log_t *dklog_set	DK_P1(dk_log_t *, new_object)
{
  dk_log_t *back;
  
  back = the_object;
  the_object = new_object;
  return back;
}

dk_log_t *dklog_get	DK_P0()
{
 
  return the_object;
}

void     dklog_set_minlevel	DK_P2(dk_log_t *, p, int, minlevel)
{
  
  if(p) {
    p->level = minlevel;
  }
}

int      dklog_get_minlevel	DK_P1(dk_log_t *, p)
{
  int back = DK_LOG_LEVEL_NONE;
  if(p) {
    back = p->level;
  } 
  return back;
}

int      dklog_get_maxlevel	DK_P1(dk_log_t *, p)
{
  int back = DK_LOG_LEVEL_IGNORE;
  if(p) {
    back = p->maxlevel;
  } 
  return back;
}

int      dklog_msg	DK_P3(int, level, char **, strings, int, num)
{
  int back;
  dk_log_t *objptr;
  void *obj;
  dk_fct_type_log *fct;
  
  back = level;
  if(!is_in_log) { 
  is_in_log = 1;
  if(strings && (num > 0)) {
  objptr = (the_object ? the_object : &default_object);
  if(level <= objptr->level) { 
    if(level < objptr->maxlevel) {
      objptr->maxlevel = level;
    }
    obj = objptr->obj;
    fct = objptr->fct;
    (*fct)(obj, level, strings, num);
  }
  }
  is_in_log = 0; 
  } 
  return back;
}

int	 dklog_msg_ss	DK_P3(int, level, char **, strings, int, num)
{
  int back, i;
  char *buffer[16];
  size_t buffer_size = sizeof(buffer) / sizeof(dk_str);
  
  back = level;
  if(strings && (num > 0)) {
    if(!is_in_log) {
      is_in_log = 1;
      if((unsigned)num <= buffer_size) {
	char **temp; 
        temp = buffer;
	for(i = 0; i < num; i++) { temp[i] = NULL; }
        for(i = 0; i < num; i++) {
	  if(strings[3*i] && strings[3*i+1]) {
	    temp[i] = dkss_find(strings[3*i], strings[3*i+1], strings[3*i+2]);
	  } else {
	    temp[i] = strings[3*i+2];
	  }
	}
	is_in_log = 0;
	dklog_msg(level,temp,num);
	is_in_log = 1;
      } else {
	char **temp; 
	temp = dk_new(dk_str,num);
	if(temp) {
	  for(i = 0; i < num; i++) { temp[i] = NULL; }
	  for(i = 0; i < num; i++) {
	  if(strings[3*i] && strings[3*i+1]) {
	    temp[i] = dkss_find(strings[3*i], strings[3*i+1], strings[3*i+2]);
	  } else {
	    temp[i] = strings[3*i+2];
	  }
	  }
	  is_in_log = 0;
	  dklog_msg(level,temp,num);
	  is_in_log = 1;
	  dk_delete(temp);
	}
      }
      is_in_log = 0;
    }
  } 
  return back;
}


#ifndef LINT
static char sccs_id[] = {
"@(#)dklog.ctr\t1.5\t08/25/03\t(Dirk Krause) - Log module"
};
#endif

