/* 
 * The dk libraries (libdkport + libdkc)
 * Copyright (C) 2000, 2001 - Dirk Krause
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 * In this package the copy of the GNU Library General Public License
 * is placed in file COPYING.
 */


#include <dk.h>
#include <dkma.h>
#include <dkerror.h>

#if DK_HAVE_MATH_H
#include <math.h>
#endif


#line 32 "dkma.ctr"


#if DK_HAVE_LONG_LONG_INT
#define MSBULLI	0x8000000000000000ULL
#define MSBULLD 9223372036854775808.0
#define BITSULL 64
#define MAXULLI 0xFFFFFFFFFFFFFFFFULL
#else
#define MSBULLI 0x80000000UL
#define MSBULLD 2147483648.0
#define BITSULL 32
#define MAXULLI 0xFFFFFFFFUL
#endif

#if SIZEOF_LONG == 8
/* MSB for unsigned long in hex notation */
#define MSBULI 0x8000000000000000UL
/* MSB for unsigned long in double notation */
#define MSBULD 9223372036854775808.0
/* 0xFFFFFFFFFFFFFFFFUL in double notation */
#define MAXULD 18446744073709551615.0
/* MSB for long in hex notation */
#define MSBLI  0x4000000000000000L
/* MSB for long in double notation */
#define MSBLD  4611686018427387904.0
/* 0x7FFFFFFFFFFFFFFFL in double notation */
#define MAXLD  9223372036854775807.0
#define MAXLI  0x7FFFFFFFFFFFFFFFL
/* number of bits in unsigned long */
#define BITSUL 64
/* number of bits in long */
#define BITSL  63
/* negativest long number */
#define NEGLI  (-9223372036854775808L)
#define NEGLD  (-9223372036854775808.0)
#else
#define MSBULI 0x80000000UL
#define MSBULD 2147483648.0
#define MAXULD 4294967295.0
#define MSBLI  0x40000000L
#define MSBLD  1073741824.0
#define MAXLD  2147483647.0
#define MAXLI  0x7FFFFFFFL
#define BITSUL 32
#define BITSL  31
#define NEGLI  (-2147483648L)
#define NEGLD  (-2147483648.0)
#endif

static int error_code = DK_ERR_NONE;

int dkma_get_error DK_P1(int, del)
{
  int back;
  back = error_code;
  if(del) { error_code = 0; }
  
  return back;
}

static
unsigned long
st_add_ulong_ok DK_P3(unsigned long, u1, unsigned long, u2, int *, ok)
{
  unsigned long back = 0UL;
  
  if((DK_MAX_ULONG - u1) >= u2) {
    back = u1 + u2;
  } else {
    *ok = DK_ERR_MATH_OOR;
  }
  
  return back;
}

static
dk_long_long_unsigned_t
st_add_ull_ok DK_P3(dk_long_long_unsigned_t,o1,dk_long_long_unsigned_t,o2,int *,ok)
{
  dk_long_long_unsigned_t back = DK_ZERO_LONG_LONG_UNSIGNED;
  if(o1) {
    if(o2) {
      back = o1 + o2;
      if(o2 > (DK_MAX_LONG_LONG_UNSIGNED - o1)) {
        *ok = DK_ERR_MATH_OOR;
      }
    } else {
      back = o1;
    }
  } else {
    back = o2;
  }
  return back;
}

static
unsigned long
st_sub_ulong_ok DK_P3(unsigned long, u1, unsigned long, u2, int *, ok)
{
  unsigned long back = 0UL;
  
  if(u1 >= u2) {
    back = u1 - u2;
  } else {
    *ok = DK_ERR_MATH_OOR;
  }
  
  return back;
}
static
unsigned long
st_mul_ulong_ok DK_P3(unsigned long, u1, unsigned long, u2, int *, ok)
{
  unsigned long back = 0UL;
  
  if(u1 && u2) {
    if((DK_MAX_ULONG / u1) >= u2) {
      back = u1 * u2;
    } else {
      *ok = DK_ERR_MATH_OOR;
    }
  }
  
  return back;
}
static
unsigned long
st_div_ulong_ok DK_P3(unsigned long, u1, unsigned long, u2, int *, ok)
{
  unsigned long back = 0UL;
  
  if(u2) {
    back = u1 / u2;
  } else {
     *ok = DK_ERR_DIV_ZERO;
  }
  
  return back;
}
unsigned long
dkma_add_ulong_ok DK_P3(unsigned long, u1, unsigned long, u2, int *, ok)
{
  unsigned long back;
  back = st_add_ulong_ok(u1,u2,(ok ? ok : &error_code));
  return back;
}
dk_long_long_unsigned_t
dkma_add_ull_ok DK_P3(dk_long_long_unsigned_t,o1,dk_long_long_unsigned_t,o2,int *,ok)
{
  dk_long_long_unsigned_t back = DK_ZERO_LONG_LONG_UNSIGNED;
  back = st_add_ull_ok(o1,o2,(ok ? ok : &error_code));
  return back;
}
unsigned long
dkma_sub_ulong_ok DK_P3(unsigned long, u1, unsigned long, u2, int *, ok)
{
  unsigned long back;
  back = st_sub_ulong_ok(u1,u2,(ok ? ok : &error_code));
  return back;
}
unsigned long
dkma_mul_ulong_ok DK_P3(unsigned long, u1, unsigned long, u2, int *, ok)
{
  unsigned long back;
  back = st_mul_ulong_ok(u1,u2,(ok ? ok : &error_code));
  return back;
}
unsigned long
dkma_div_ulong_ok DK_P3(unsigned long, u1, unsigned long, u2, int *, ok)
{
  unsigned long back;
  back = st_div_ulong_ok(u1,u2,(ok ? ok : &error_code));
  return back;
}
unsigned long
dkma_add_ulong DK_P2(unsigned long, u1, unsigned long, u2)
{
  unsigned long back;
  back = st_add_ulong_ok(u1,u2,&error_code);
  return back;
}
unsigned long
dkma_sub_ulong DK_P2(unsigned long, u1, unsigned long, u2)
{
  unsigned long back;
  back = st_sub_ulong_ok(u1,u2,&error_code);
  return back;
}
unsigned long
dkma_mul_ulong DK_P2(unsigned long, u1, unsigned long, u2)
{
  unsigned long back;
  back = st_mul_ulong_ok(u1,u2,&error_code);
  return back;
}
unsigned long
dkma_div_ulong DK_P2(unsigned long, u1, unsigned long, u2)
{
  unsigned long back;
  back = st_div_ulong_ok(u1,u2,&error_code);
  return back;
}

static
long
st_add_long_ok DK_P3(long, l1, long, l2, int *, ok)
{
  long back = 0L;
  
  if(l1 >= 0L) {
    if(l2 >= 0L) {
      if((DK_MAX_LONG - l1) >= l2) {
	back = l1 + l2;
      } else {
	*ok = DK_ERR_MATH_OOR;
      }
    } else {
      back = l1 + l2;
    }
  } else {
    if(l2 >= 0L) {
      back = l1 + l2;
    } else { 
      
      if(((0L - DK_MAX_LONG) - l1) <= l2) {
	
	back = l1 + l2;
      } else {
	*ok = DK_ERR_MATH_OOR;
      }
    }
  } 
  return back;
}
static
long
st_sub_long_ok DK_P3(long, l1, long, l2, int *, ok)
{
  long back = 0L;
  
  if(l1 >= 0L) {
    if(l2 >= 0L) {
      back = l1 - l2;
    } else {
      if((DK_MAX_LONG + l2) >= l1) {
	back = l1 - l2;
      } else {
	*ok = DK_ERR_MATH_OOR;
      }
    }
  } else {
    if(l2 >= 0L) {
      if((DK_MAX_LONG + l1) >= l2) {
	back = l1 - l2;
      } else {
	*ok = DK_ERR_MATH_OOR;
      }
    } else {
      back = l1 - l2;
    }
  } 
  return back;
}
static
long
st_mul_long_ok DK_P3(long, l1, long, l2, int *, ok)
{
  long back = 0L;
  
  if(l1 && l2) {
    if(l1 > 0L) {
      if(l2 > 0L) {
	if((DK_MAX_LONG / l1) >= l2) {
	  back = l1 * l2;
	} else {
	  *ok = DK_ERR_MATH_OOR;
	}
      } else {
	if(l2 >= (0L - DK_MAX_LONG)) {
	  back = 0L - st_mul_long_ok(l1, (0L - l2), ok);
	} else {
	  *ok = DK_ERR_MATH_OOR;
	}
      }
    } else {
      if(l2 > 0L) {
	if(l1 >= (0L - DK_MAX_LONG)) {
	  back = 0L - st_mul_long_ok((0L - l1), l2, ok);
	} else {
	  *ok = DK_ERR_MATH_OOR;
	}
      } else {
	if(l1 >= (0L - DK_MAX_LONG)) {
	  if(l2 >= (0L - DK_MAX_LONG)) {
	    back = st_mul_long_ok((0L - l1),(0L - l2),ok);
	  } else {
	    *ok = DK_ERR_MATH_OOR;
	  }
	} else {
	  *ok = DK_ERR_MATH_OOR;
	}
      }
    }
  } 
  return back;
}
static
long
st_div_long_ok DK_P3(long, l1, long, l2, int *, ok)
{
  long back = 0L;
  
  if(l2) {
    back = l1 / l2;
  } else {
    *ok = DK_ERR_DIV_ZERO;
  } 
  return back;
}
long dkma_add_long_ok DK_P3(long, l1, long, l2, int *, ok)
{
  long back;
  back = st_add_long_ok(l1,l2,(ok ? ok : &error_code));
  return back;
}
long dkma_sub_long_ok DK_P3(long, l1, long, l2, int *, ok)
{
  long back;
  back = st_sub_long_ok(l1,l2,(ok ? ok : &error_code));
  return back;
}
long dkma_mul_long_ok DK_P3(long, l1, long, l2, int *, ok)
{
  long back;
  back = st_mul_long_ok(l1,l2,(ok ? ok : &error_code));
  return back;
}
long dkma_div_long_ok DK_P3(long, l1, long, l2, int *, ok)
{
  long back;
  back = st_div_long_ok(l1,l2,(ok ? ok : &error_code));
  return back;
}
long dkma_add_long DK_P2(long, l1, long, l2)
{
  long back;
  back = st_add_long_ok(l1,l2,&error_code);
  return back;
}
long dkma_sub_long DK_P2(long, l1, long, l2)
{
  long back;
  back = st_sub_long_ok(l1,l2,&error_code);
  return back;
}
long dkma_mul_long DK_P2(long, l1, long, l2)
{
  long back;
  back = st_mul_long_ok(l1,l2,&error_code);
  return back;
}
long dkma_div_long DK_P2(long, l1, long, l2)
{
  long back;
  back = st_div_long_ok(l1,l2,&error_code);
  return back;
}
static
double
st_add_double_ok DK_P3(double, d1, double, d2, int *, ok)
{
  double back = 0.0;
  
  if(d1 >= 0.0) {
    if(d2 >= 0.0) {
      if((DK_MAX_DOUBLE - d1) > d2) {
	back = d1 + d2;
      } else {
	*ok = DK_ERR_MATH_OOR;
      }
    } else {
      back = d1 + d2;
    }
  } else {
    if(d2 >= 0.0) {
      back = d1 + d2;
    } else {
      if((DK_MAX_DOUBLE + d1) > (0.0 - d2)) {
	back = d1 + d2;
      } else {
	*ok = DK_ERR_MATH_OOR;
      }
    }
  } 
  return back;
}
static
double
st_sub_double_ok DK_P3(double, d1, double, d2, int *, ok)
{
  double back = 0.0;
  
  if(d1 >= 0.0) {
    if(d2 >= 0.0) {
      back = d1 - d2;
    } else {
      if((DK_MAX_DOUBLE + d2) > d1) {
	back = d1 - d2;
      } else {
	*ok = DK_ERR_MATH_OOR;
      }
    }
  } else {
    if(d2 >= 0.0) {
      if((DK_MAX_DOUBLE + d1) > d2) {
	back = d1 - d2;
      } else {
	*ok = DK_ERR_MATH_OOR;
      }
    } else {
      back = d1 - d2;
    }
  } 
  return back;
}
static
double
st_mul_double_ok DK_P3(double, d1, double, d2, int *, ok)
{
  double back = 0.0;
  
  if(d1 != 0.0) {
    if(d2 != 0.0) {
      if(d1 >= 0.0) {
	if(d2 >= 0.0) {
	  if(d1 > 1.0) {
	    if((DK_MAX_DOUBLE / d1) >= d2) {
	      back = d1 * d2;
	    } else {
	      *ok = DK_ERR_MATH_OOR;
	    }
	  } else {
	    back = d1 * d2;
	  }
	} else {
	  back = 0.0 - st_mul_double_ok(d1, (0.0 - d2), ok);
	}
      } else {
	if(d2 >= 0.0) {
	  back = 0.0 - st_mul_double_ok((0.0 - d1), d2, ok);
	} else {
	  back = st_mul_double_ok((0.0 - d1), (0.0 - d2), ok);
	}
      }
    }
  } 
  return back;
}
static
double
st_div_double_ok DK_P3(double, d1, double, d2, int *, ok)
{
  double back = 0.0;
  
  if(d2 != 0.0) {
    if(d1 >= 0.0) {
      if(d2 >= 0.0) {
	if(d2 < 1.0) {
	  if((DK_MAX_DOUBLE * d2) > d1) {
	    back = d1 / d2;
	  } else {
	    *ok = DK_ERR_DIV_ZERO;
	  }
	} else {
	  back = d1 / d2;
	}
      } else {
	back = 0.0 - st_div_double_ok(d1, (0.0 - d2), ok);
      }
    } else {
      if(d2 >= 0.0) {
	back = 0.0 - st_div_double_ok((0.0 - d1), d2, ok);
      } else {
	back = st_div_double_ok((0.0 - d1), (0.0 - d2), ok);
      }
    }
  } else {
    *ok = DK_ERR_DIV_ZERO;
  } 
  return back;
}
double
dkma_add_double_ok DK_P3(double, d1, double, d2, int *, ok)
{
  double back;
  back = st_add_double_ok(d1,d2,(ok ? ok : &error_code));
  return back;
}
double
dkma_sub_double_ok DK_P3(double, d1, double, d2, int *, ok)
{
  double back;
  back = st_sub_double_ok(d1,d2,(ok ? ok : &error_code));
  return back;
}
double
dkma_mul_double_ok DK_P3(double, d1, double, d2, int *, ok)
{
  double back;
  back = st_mul_double_ok(d1,d2,(ok ? ok : &error_code));
  return back;
}
double
dkma_div_double_ok DK_P3(double, d1, double, d2, int *, ok)
{
  double back;
  back = st_div_double_ok(d1,d2,(ok ? ok : &error_code));
  return back;
}
double dkma_add_double DK_P2(double, d1, double, d2)
{
  double back;
  back = st_add_double_ok(d1,d2,&error_code);
  
  return back;
}
double dkma_sub_double DK_P2(double, d1, double, d2)
{
  double back;
  back = st_sub_double_ok(d1,d2,&error_code);
  
  return back;
}
double dkma_mul_double DK_P2(double, d1, double, d2)
{
  double back;
  back = st_mul_double_ok(d1,d2,&error_code);
  
  return back;
}
double dkma_div_double DK_P2(double, d1, double, d2)
{
  double back;
  back = st_div_double_ok(d1,d2,&error_code);
  
  return back;
}

double dkma_rint DK_P1(double, d)
{
  double back = 0.0;
#if DK_HAVE_RINT
  back = rint(d);
#else
  if(d >= 0.0) {
    if(d < (DK_MAX_DOUBLE - 0.5)) {
      back = floor(d + 0.5);
    } else {
      back = DK_MAX_DOUBLE;
    }
  } else {
    back = 0.0 - dkma_rint(0.0 - d);
  }
#endif
  
  return back;
}
static unsigned long
st_double_to_ul_ok DK_P2(double, d, int *, ok)
{
  unsigned long back = 0UL;
  double xval, testd; unsigned long testval; int trials;
  
  if(d >= 0.0) {
    xval = dkma_rint(d); 
    if(xval > MAXULD) {
      *ok = DK_ERR_MATH_OOR;
    } else {
      testd = MSBULD; testval = MSBULI; trials = BITSUL;
      while(trials--) {
	if(xval >= testd) {
	  back |= testval;
	  xval -= testd;
	}
	testd = testd / 2.0; testval = testval / 2UL;
      }
    }
  } else {
    *ok = DK_ERR_MATH_OOR;
  } 
  return back;
}
static long
st_double_to_l_ok  DK_P2(double, d, int *, ok)
{
  long back = 0L; long x;
  long testval;
  double xval, testd;
  int trials;
  
  if(d < 0.0) {
    x = st_double_to_l_ok((0.0 - d), ok);
    if(x <= DK_MAX_ULONG) {
      back = 0L - x;
    } else {
      *ok = DK_ERR_MATH_OOR;
    }
  } else {
    xval = dkma_rint(d);
    if(xval > MAXLD) {
      *ok = DK_ERR_MATH_OOR;
    } else {
      testval = MSBLI; testd = MSBLD; trials = BITSL;
      while(trials--) {
        if(xval >= testd) {
	  xval -= testd;
	  back |= testval;
        }
        testval = testval / 2L; testd = testd / 2.0;
      }
    }
  } 
  return back;
}
static double
st_ul_to_double_ok DK_P2(unsigned long, l, int *, ok)
{
  double back = 0.0;
  double testd; unsigned long testval; int trials;
  
  testd = MSBULD; testval = MSBULI; trials = BITSUL;
  while(trials--) {
    if(l & testval) { back += testd; }
    testd = testd / 2.0; testval = testval / 2UL;
  } 
  return back;
}
static double
st_ull_to_double_ok DK_P2(dk_long_long_unsigned_t,l,int *,ok)
{
  double back = 0.0;
  double testd;
  dk_long_long_unsigned_t testval; int trials;
#if DK_HAVE_LONG_LONG_INT
  
#else
  
#endif
  testd = MSBULLD;
  testval = MSBULLI;
  trials = BITSULL; 
  while(trials--) {
#if DK_HAVE_LONG_LONG_INT
    
#else
    
#endif
    
    if(l & testval) {
      back += testd;
      
    }
#if DK_HAVE_LONG_LONG_INT
    testd = testd / 2.0; testval = testval / 2ULL;
#else
    testd = testd / 2.0; testval = testval / 2UL;
#endif
  } 
  return back;
}
static double
st_l_to_double_ok  DK_P2(long, l, int *, ok)
{
  double back = 0.0;
  double testd;
  long testval;
  int trials;
  
  if(l < 0) {
    back = 0.0 - st_l_to_double_ok((0L -l), ok);
    if(l == NEGLI) { back = NEGLD; }
  } else {
    testd = MSBLD; testval = MSBLI; trials = BITSL;
    while(trials--) {
      if(l & testval) { back += testd; }
      testd = testd / 2.0; testval = testval / 2UL;
    }
  } 
  return back;
}
unsigned long
dkma_double_to_ul_ok DK_P2(double, d, int *, ok)
{
  unsigned long back;
  back = st_double_to_ul_ok(d,(ok ? ok : &error_code));
  return back;
}
long
dkma_double_to_l_ok  DK_P2(double, d, int *, ok)
{
  long back;
  back = st_double_to_l_ok(d,(ok ? ok : &error_code));
  return back;
}
double
dkma_ul_to_double_ok DK_P2(unsigned long, l, int *, ok)
{
  double back;
  back = st_ul_to_double_ok(l,(ok ? ok : &error_code));
  return back;
}
double
dkma_ull_to_double_ok DK_P2(dk_long_long_unsigned_t,l,int *,ok)
{
  double back;
  back = st_ull_to_double_ok(l,(ok ? ok : &error_code));
  return back;
}
double
dkma_l_to_double_ok  DK_P2(long, l, int *, ok)
{
  double back;
  back = st_l_to_double_ok(l, (ok ? ok : &error_code));
  return back;
}
unsigned long dkma_double_to_ul DK_P1(double, d)
{ unsigned long back; back = st_double_to_ul_ok(d, &error_code); return back; }
long dkma_double_to_l DK_P1(double, d)
{ long back; back = st_double_to_l_ok(d, &error_code); return back; }
double dkma_ul_to_double DK_P1(unsigned long, ul)
{ double back; back = st_ul_to_double_ok(ul, &error_code); return back; }
double dkma_ull_to_double DK_P1(dk_long_long_unsigned_t, ul)
{ double back; back = st_ull_to_double_ok(ul, &error_code); return back; }
double dkma_l_to_double DK_P1(long, l)
{ double back; back = st_l_to_double_ok(l, &error_code); return back; }

#ifndef LINT
static char sccs_id[] = {
"@(#)dkma.ctr\t1.13\t08/25/03\t(Dirk Krause) - math operations w/ range checks"
};
#endif

