/* 
 * The dk libraries (libdkport + libdkc)
 * Copyright (C) 2000, 2001 - Dirk Krause
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 * In this package the copy of the GNU Library General Public License
 * is placed in file COPYING.
 */


#include <dk.h>

#if DK_HAVE_STRING_H
#include <string.h>
#else
#if DK_HAVE_STRINGS_H
#include <strings.h>
#endif
#endif
#if DK_HAVE_STDLIB_H
#include <stdlib.h>
#endif
#if DK_HAVE_UNISTD_H
#include <unistd.h>
#endif
#if DK_HAVE_ALLOC_H
#include <alloc.h>
#endif

#define DK_MEM_C 1
#include <dkmem.h>



#line 47 "dkmem.ctr"



/* ********************************************************************* */
/* *                                                                   * */
/* *   Check whether the product of elsize and                         * */
/* *   nelem fits into a size_t.                                       * */
/* *   On success return the product, otherwise 0.                     * */
/* *                                                                   * */
/* ********************************************************************* */

#line 53 "dkmem.ctr"

static size_t check_size	DK_P2(size_t, elsize, size_t, nelem)
{
  size_t back = 0UL;
  unsigned long i, j, k;
  i = elsize; j = nelem;
  if(i < (0xFFFFFFFFUL / j)) {
    k = i * j;
    if(k <= ((size_t)0xFFFFFFFFUL)) {
      back = ((size_t)k);
    }
  } 
  return back;
}


/* ********************************************************************* */
/* *                                                                   * */
/* *   Reset memory.                                                   * */
/* *                                                                   * */
/* ********************************************************************* */

#line 70 "dkmem.ctr"

void  dkmem_res		DK_P2(void *, ptr, size_t, bytes)
{
  if(ptr && bytes) {
#if DK_HAVE_MEMSET
    memset(ptr,0,bytes);
#else
#if DK_HAVE_BZERO
    bzero(ptr,bytes);
#else
    char *x; size_t y;
    x = (char *)ptr; y = bytes;
    while(y--) { *(x++) = '\0'; }
#endif
#endif
  }
}


/* ********************************************************************* */
/* *                                                                   * */
/* *   Copy memory regions from s to d (n bytes).                      * */
/* *                                                                   * */
/* ********************************************************************* */

#line 90 "dkmem.ctr"

void  dkmem_cpy		DK_P3(void *, d, void *, s, size_t, n)
{
  if(d && s && n) {
#if DK_HAVE_MEMCPY
    memcpy(d,s,n);
#else
#if DK_HAVE_BCOPY
    bcopy(s,d,n);
#else
    char *x, *y; size_t z;
    z = n; x = (char *)s; y = (char *)d;
    while(z--) { *(y++) = *(x++); }
#endif
#endif
  }
}


/* ********************************************************************* */
/* *                                                                   * */
/* *   Compare memory regions.                                         * */
/* *   Returns 0 for equality, non-0 for unequal memory regions.       * */
/* *                                                                   * */
/* ********************************************************************* */

#line 111 "dkmem.ctr"

int   dkmem_cmp		DK_P3(void *, s1, void *, s2, size_t, n)
{
  int back = 0;
  if(s1 && s2 && n) {
#if DK_HAVE_MEMCMP
    back = memcmp(s1,s2,n);
#else
#if DK_HAVE_BCMP
    back = bcmp(s1,s2,n);
#else
    char *x, *y; size_t z;
    z = n; x = (char *)s1; y = (char *)s2;
    while((z--) && (!back)) {
      if((*(x++)) != (*(y++))) {
	back = 1;
      }
    }
#endif
#endif
  }
  return back;
}


/* ********************************************************************* */
/* *                                                                   * */
/* *   Allocate memory for nelem elements of size elsize.              * */
/* *   Return a valid pointer on success, NULL on                      * */
/* *   error.                                                          * */
/* *                                                                   * */
/* ********************************************************************* */

#line 139 "dkmem.ctr"

void *dkmem_alloc	DK_P2(size_t, elsize, size_t, nelem)
{
  void *back = NULL;
  size_t bytes;
  
  if(elsize && nelem) {
    bytes = check_size(elsize,nelem); 
    if(bytes) {
#if DK_HAVE_FARMALLOC
      
#if DK_HAVE_FARCORELEFT
      
      if(bytes < farcoreleft()) {
#endif
	
        back = farmalloc(bytes);
#if DK_HAVE_FARCORELEFT
      }
#endif
#else
      
#if DK_HAVE_MALLOC
#if DK_HAVE_CORELEFT
      
      if(bytes < coreleft()) {
#endif
	
        back = malloc(bytes);
#if DK_HAVE_CORELEFT
      }
#endif
#endif
#endif
    }
  }
  if(back) { 
#if DK_HAVE_MEMSET
    
    memset(back, 0, bytes);
#else
#if DK_HAVE_BZERO
    
    bzero(back,bytes);
#else
    
    dkmem_res(back,bytes);
#endif
#endif
  } 
  return back;
}


/* ********************************************************************* */
/* *                                                                   * */
/* *   Free memory allocated by dkmem_alloc.                           * */
/* *                                                                   * */
/* ********************************************************************* */

#line 194 "dkmem.ctr"

void  dkmem_free		DK_P1(void *, ptr)
{
  
  if(ptr) {
#if DK_HAVE_FARMALLOC
    farfree(ptr);
#else
#if DK_HAVE_MALLOC
    free(ptr);
#endif
#endif
  }
  
}

#ifndef LINT
static char sccs_id[] = {
"@(#)dkmem.ctr\t1.10\t08/25/03\t(Dirk Krause) - Memory handling"
};
#endif

