/* 
 * The dk libraries (libdkport + libdkc)
 * Copyright (C) 2000, 2001 - Dirk Krause
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 * In this package the copy of the GNU Library General Public License
 * is placed in file COPYING.
 */

#ifndef DK_MEM_INCLUDED
#define DK_MEM_INCLUDED 1

#include <dk.h>

#if DK_HAVE_STRING_H
#include <string.h>
#endif
#if DK_HAVE_STRINGS_H
#include <strings.h>
#endif

#ifndef DK_MEM_C
#if !DK_HAVE_PROTOTYPES
#define EXTERN extern
#else
#define EXTERN /* nix */
#endif
#else
#define EXTERN /* nix */
#endif

#ifdef __cplusplus
extern "C" {
#endif

EXTERN void  dkmem_res		DK_PR((void *ptr, size_t bytes));
EXTERN void  dkmem_cpy		DK_PR((void *d, void *s, size_t n));
EXTERN int   dkmem_cmp		DK_PR((void *s1, void *s2, size_t n));
EXTERN void *dkmem_alloc	DK_PR((size_t elsize, size_t nelem));
EXTERN void  dkmem_free		DK_PR((void *ptr));
#ifdef __cplusplus
}
#endif

#if DK_HAVE_MEMSET
#define DK_MEMRES(ptr,sz)	memset(((void *)(ptr)),0,((size_t)(sz)))
#else
#if DK_HAVE_BZERO
#define DK_MEMRES(ptr,sz)	bzero(((void *)(ptr)),((size_t)(sz)))
#else
#define DK_MEMRES(ptr,sz)	dkmem_res(((void *)(ptr)),((size_t)(sz)))
#endif
#endif

#if DK_HAVE_MEMCPY
#define DK_MEMCPY(d,s,n) memcpy(((void *)(d)),((void *)(s)),((size_t)(n)))
#else
#if DK_HAVE_BCOPY
#define DK_MEMCPY(d,s,n) bcopy(((void *)(s)),((void *)(d)),((size_t)(n)))
#else
#define DK_MEMCPY(d,s,n) dkmem_cpy(((void *)(d)),((void *)(s)),((size_t)(n)))
#endif
#endif

#if DK_HAVE_MEMCMP
#define DK_MEMCMP(a,b,n) memcmp(((void *)(a)),((void *)(b)),((size_t)(n)))
#else
#if DK_HAVE_BCMP
#define DK_MEMCMP(a,b,n) bcmp(((void *)(a)),((void *)(b)),((size_t)(n)))
#else
#define DK_MEMCMP(a,b,n) dkmem_cmp(((void *)(a)),((void *)(b)),((size_t)(n)))
#endif
#endif

#define dk_new(t,s)	(t *)dkmem_alloc(sizeof(t),((size_t)s))
#define dk_delete(p)	dkmem_free((void *)(p))

#endif
/* DK_MEM_INCLUDED */

