/* 
 * The dk libraries (libdkport + libdkc)
 * Copyright (C) 2000, 2001 - Dirk Krause
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 * In this package the copy of the GNU Library General Public License
 * is placed in file COPYING.
 */

#define DK_OF_C	1
#include <dkof.h>
#include <dkstream.h>
#include <dkmem.h>
#include <dkerror.h>


#line 29 "dkof.ctr"

#define TR_UL(x) ((unsigned long)x)

static size_t
add_to_cell DK_PR((dk_of_t *o,dk_of_cell_t *c,size_t n,char *b,size_t l));


static unsigned long a85[] = {
  (85UL * 85UL * 85UL * 85UL),
  (85UL * 85UL * 85UL),
  (85UL * 85UL),
  85UL,
  1UL
};



/* {{{ add_from_cell: write bytes downstream */
static size_t
add_from_cell DK_P4(dk_of_t *,o,size_t,n,char *,b,size_t,l)
{
  size_t back = 0;
  
  if(n) {	
    back = add_to_cell(o,&((o->cells)[(n-1)]),(n-1),b,l);
  } else {	
    if(o->target) {
      back = dkstream_write(o->target, b, l);
    } else {	
    }
  }
  
  return back;
}
/* }}} add_from_cell */



/* {{{ add_to_cell: add buffer contents to a cell */

#if DK_HAVE_ZLIB_H
static int
write_flate_not_final DK_P3(dk_of_t *,o,dk_of_cell_t *,c,size_t,n)
{
  int back = 0;
  int err, cc;
  uLong odiff;
  size_t sz;
  
  cc = 1;
  ((c->c).flate.zs)->next_in = (c->c).flate.ibuffer;
  ((c->c).flate.zs)->avail_in = (c->c).flate.used;
  back = 1;
  while(cc) {
    cc = 0;
    ((c->c).flate.zs)->next_out = (c->c).flate.obuffer;
    ((c->c).flate.zs)->avail_out = (c->c).flate.o_length;
    err = deflate((c->c).flate.zs, 0);
    switch(err) {
      case Z_OK: {
        if(((c->c).flate.zs)->avail_in) {
	  cc = 1;
	}
	odiff = (c->c).flate.o_length - ((c->c).flate.zs)->avail_out;
	sz = (size_t)odiff;
	if(add_from_cell(o,n,(char *)((c->c).flate.obuffer),sz) != sz) {
	  back = 0;
	}
      } break;
      default: {
        back = 0; (c->c).flate.stillok = -1;
      } break;
    }
  }
  
  return back;
}

static size_t
add_to_flate_cell DK_P5(dk_of_t *,o,dk_of_cell_t *,c,size_t,n,char *,b,size_t,l)
{
  size_t back = 0;
  char *ptr; size_t i;
  
  if(((c->c).flate.zs) && ((c->c).flate.ibuffer) && ((c->c).flate.obuffer)) {
    if(((c->c).flate.i_length) && ((c->c).flate.o_length)) {
      if(((c->c).flate.stillok) == 1) {
        ptr = b;
	for(i = 0; i < l; i++) {
	  ((c->c).flate.ibuffer)[(c->c).flate.used] = (Bytef)(*(ptr++));
	  (c->c).flate.used += (uLong)1UL;
	  if(((c->c).flate.used) >= ((c->c).flate.i_length)) {
	    if(write_flate_not_final(o,c,n)) {
	      back++;
	    }
	    (c->c).flate.used = (uLong)0UL;
	  } else {
	    back++;
	  }
	}
      } else {	
      }
    } else {	
    }
  } else {	
  }
  
  return back;
}
#endif

static char eol1[] = { "\r\n" };
static char eol2[] = { "\n" };

static int
write_a85_buffer DK_P3(dk_of_t *,o,dk_of_cell_t *,c,size_t,n)
{
  int back = 0;
  char buffer[6];
  unsigned long v, ul;
  unsigned char uc;
  size_t i;
  
  v = (c->c).a85.a85val;
  for(i = 0; i < 5; i++) {
    ul = 33UL + v / a85[i];
    v  = v % a85[i];
    uc = (unsigned char)((unsigned)ul);
    buffer[i] = uc;
  }
  if(add_from_cell(o,n,buffer,(1 + (c->c).a85.used)) == (1 + (c->c).a85.used)) {
    back = 1;
  }
  (c->c).a85.linepos += (1 + (c->c).a85.used);
  if(((c->c).a85.linepos) > 64) {
    if((o->flags) & 1) {
      if(add_from_cell(o,n,eol1,2) != 2) {
        back = 0;
      }
    } else {
      if(add_from_cell(o,n,eol2,1) != 1) {
        back = 0;
      }
    }
    (c->c).a85.linepos = 0;
  }
  (c->c).a85.a85val = 0UL;
  (c->c).a85.used = 0;
  
  return back;
}

static size_t
add_to_a85_cell DK_P5(dk_of_t *,o,dk_of_cell_t *,c,size_t,n,char *,b,size_t,l)
{
  size_t back = 0;
  unsigned long ul; char *ptr; size_t i;
  
  ptr = b;
  for(i = 0; i < l; i++) {
    ul = (unsigned long)((unsigned)((unsigned char)(*(ptr++))));
    (c->c).a85.a85val = ul + 256UL * (c->c).a85.a85val;
    (c->c).a85.used += 1;
    if(((c->c).a85.used) >= 4) {
      if(write_a85_buffer(o,c,n)) {
        back++;
      }
    } else {
      back++;
    }
  }
  
  return back;
}

static size_t
add_to_buf_cell DK_P5(dk_of_t *,o,dk_of_cell_t *,c,size_t,n,char *,b,size_t,l)
{
  size_t back = 0;
  char *ptr; size_t i;
  
  if(((c->c).b.buffer) && ((c->c).b.lgt)) {
    ptr = b;
    for(i = 0; i < l; i++) {
      ((c->c).b.buffer)[(c->c).b.used] = *(ptr++);
      (c->c).b.used += 1;
      if(((c->c).b.used) >= ((c->c).b.lgt)) {
        
	
#line 218 "dkof.ctr"
	
#line 219 "dkof.ctr"
	
#line 220 "dkof.ctr"
	
#line 221 "dkof.ctr"
	
        if(add_from_cell(o,n,(c->c).b.buffer,(c->c).b.used)==((c->c).b.used)) {
	  back++;
	}
	(c->c).b.used = 0;
      } else {
        back++;
      }
    }
  } else {		
  }
  
  return back;
}

static size_t
add_to_cell DK_P5(dk_of_t *,o,dk_of_cell_t *,c,size_t,n,char *,b,size_t,l)
{
  size_t back = 0;
  
  switch(c->what) {
#if DK_HAVE_ZLIB_H
    case DK_OF_TYPE_FLATE: {
      back = add_to_flate_cell(o,c,n,b,l);
    } break;
#endif
    case DK_OF_TYPE_ASCII85: {
      back = add_to_a85_cell(o,c,n,b,l);
    } break;
    case DK_OF_TYPE_BUFFERED: {
      back = add_to_buf_cell(o,c,n,b,l);
    } break;
    /* further cell types here */
    default: {		
    } break;
  } 
  return back;
}

/* }}} add_to_cell */



/* {{{ cell_init: set cell encoding type */

#if DK_HAVE_ZLIB_H
static int
cell_flate_init DK_P1(dk_of_cell_t *,c)
{
  int back = 0;
  
  (c->c).flate.zs = dk_new(z_stream,1);
  (c->c).flate.ibuffer = dk_new(Bytef,1024);
  (c->c).flate.obuffer = dk_new(Bytef,1057);
  if(((c->c).flate.zs) && ((c->c).flate.ibuffer) && ((c->c).flate.obuffer)) {
    (c->c).flate.i_length = (uLong)1024UL;
    (c->c).flate.o_length = (uLong)1057UL;
    (c->c).flate.used     = (uLong)0UL;
    (c->c).flate.stillok  = 0;
    c->what = DK_OF_TYPE_FLATE;
    back = 1;
  } else {	
    if((c->c).flate.zs) { dk_delete((c->c).flate.zs); }
    if((c->c).flate.ibuffer) { dk_delete((c->c).flate.ibuffer); }
    if((c->c).flate.obuffer) { dk_delete((c->c).flate.obuffer); }
    (c->c).flate.zs = NULL;
    (c->c).flate.ibuffer = NULL;
    (c->c).flate.obuffer = NULL;
  }
  
  return back;
}
#endif

static int
cell_a85_init DK_P1(dk_of_cell_t *,c)
{
  int back = 1;
  
  (c->c).a85.a85val = 0UL;
  (c->c).a85.linepos = 0;
  (c->c).a85.used = 0;
  c->what = DK_OF_TYPE_ASCII85;
  
  return back;
}

static int
cell_buffered_init DK_P1(dk_of_cell_t *,c)
{
  int back = 0;
  
  (c->c).b.buffer = dk_new(char,512);
  if((c->c).b.buffer) {
    (c->c).b.lgt = 512;
    (c->c).b.used = 0;
    c->what = DK_OF_TYPE_BUFFERED;
    back = 1;
  }
  
  return back;
}

static int
cell_init DK_P2(dk_of_cell_t *,c,int,w)
{
  int back = 0;
  
  if((c->what) == DK_OF_TYPE_NONE) {
    switch(w) {
#if DK_HAVE_ZLIB_H
      case DK_OF_TYPE_FLATE: {
        back = cell_flate_init(c);
      } break;
#endif
      case DK_OF_TYPE_ASCII85: {
        back = cell_a85_init(c);
      } break;
      case DK_OF_TYPE_BUFFERED: {
        back = cell_buffered_init(c);
      } break;
      /* further cell types here */
      default: {	
      } break;
    }
  } else {		
  }
  
  return back;
}

/* }}} cell_init */



/* {{{ cell_start: prepare cell for a new chunk */

#if DK_HAVE_ZLIB_H
static int
cell_flate_start DK_P3(dk_of_t *,o,dk_of_cell_t *,c,size_t,n)
{
  int back = 0;
  
  if(((c->c).flate.zs) && ((c->c).flate.ibuffer) && ((c->c).flate.obuffer)) {
    if(((c->c).flate.i_length) && ((c->c).flate.o_length)) {
      (c->c).flate.used = (uLong)0UL;
      (c->c).flate.stillok = 0;
      ((c->c).flate.zs)->zfree = (free_func)0;
      ((c->c).flate.zs)->zalloc = (alloc_func)0;
      ((c->c).flate.zs)->opaque = (voidpf)0;
      
      if(deflateInit((c->c).flate.zs, 9) == Z_OK) {
        
        back = 1; (c->c).flate.stillok = 1;
      } else {		
      }
    } else {	
    }
  } else {	
  }
  
  return back;
}
#endif

static int
cell_a85_start DK_P3(dk_of_t *,o,dk_of_cell_t *,c,size_t,n)
{
  int back = 1;
  
  (c->c).a85.a85val = 0UL;
  (c->c).a85.linepos = 0;
  (c->c).a85.used = 0;
  
  return back;
}

static int
cell_buf_start DK_P3(dk_of_t *,o,dk_of_cell_t *,c,size_t,n)
{
  int back = 0;
  
  if(((c->c).b.lgt) && ((c->c).b.buffer)) {
    (c->c).b.used = 0; back = 1;
  } else {	
  }
  
  return back;
}

static int
cell_start DK_P3(dk_of_t *,o,dk_of_cell_t *,c,size_t,n)
{
  int back = 0;
  
  switch(c->what) {
#if DK_HAVE_ZLIB_H
    case DK_OF_TYPE_FLATE: {
      back = cell_flate_start(o,c,n);
    } break;
#endif
    case DK_OF_TYPE_ASCII85: {
      back = cell_a85_start(o,c,n);
    } break;
    case DK_OF_TYPE_BUFFERED: {
      back = cell_buf_start(o,c,n);
    } break;
    default: {		
    } break;
  }
  
  return back;
}
/* }}} cell_start */



/* {{{ cell_end: finish chunk and clean up cell */

#if DK_HAVE_ZLIB_H
static int
write_flate_final DK_P3(dk_of_t *,o,dk_of_cell_t *,c,size_t,n)
{
  int back = 0;
  int cc, err;
  size_t sz;
  
  cc = 1;
  ((c->c).flate.zs)->next_in = (c->c).flate.ibuffer;
  ((c->c).flate.zs)->avail_in = (c->c).flate.used;
  back = 1;
  
  while(cc) {
    
    cc = 0;
    ((c->c).flate.zs)->next_out = (c->c).flate.obuffer;
    ((c->c).flate.zs)->avail_out = (c->c).flate.o_length;
    err = deflate((c->c).flate.zs, Z_FINISH);
    
    
    switch(err) {
      case Z_STREAM_END: {	
        sz = (size_t)(((c->c).flate.o_length) - (((c->c).flate.zs)->avail_out));
	
	if(sz) {
	  if(add_from_cell(o,n,(void *)((c->c).flate.obuffer),sz) != sz) {
	    back = 0;
	  }
	}
      } break;
      case Z_OK: {	
        cc = 1;
        sz = (size_t)(((c->c).flate.o_length) - (((c->c).flate.zs)->avail_out));
	if(sz) {
	  if(add_from_cell(o,n,(void *)((c->c).flate.obuffer),sz) != sz) {
	    back = 0;
	  }
	}
      } break;
      default: {	
        back = 0;
      } break;
    }
    
  }
  if(deflateEnd((c->c).flate.zs) != Z_OK) {
    back = 0;
  }
  (c->c).flate.used = (uLong)0UL;
  (c->c).flate.stillok = 0;
  
  
  return back;
}

static int
cell_flate_end DK_P3(dk_of_t *,o,dk_of_cell_t *,c,size_t,n)
{
  int back = 0;
  
  if(((c->c).flate.zs) && ((c->c).flate.ibuffer) && ((c->c).flate.obuffer)) {
    if(((c->c).flate.o_length) && ((c->c).flate.i_length)) {
      if((c->c).flate.stillok) {
        back = write_flate_final(o,c,n);
      } else {	
      }
    } else {	
    }
  } else {	
  }
  
  return back;
}
#endif

static char a85end[] = { "~>" };

static int
cell_a85_end DK_P3(dk_of_t *,o,dk_of_cell_t *,c,size_t,n)
{
  int back = 1;
  size_t i;
  
  if((c->c).a85.used) {
    back = 0;
    for(i = (c->c).a85.used; i < 4; i++) {
      (c->c).a85.a85val = 256UL * (c->c).a85.a85val;
    }
    if(write_a85_buffer(o,c,n)) {
      back = 1;
    }
  }
  if(((c->c).a85.linepos) > 62) {
    if((o->flags) & 1) {
      if(add_from_cell(o,n,eol1,2) != 2) {
        back = 0;
      }
    } else {
      if(add_from_cell(o,n,eol2,1) != 1) {
        back = 0;
      }
    }
  }
  if(add_from_cell(o,n,a85end,2) != 2) {
    back = 0;
  }
  if((o->flags) & 1) {
    if(add_from_cell(o,n,eol1,2) != 2) {
      back = 0;
    }
  } else {
    if(add_from_cell(o,n,eol2,1) != 1) {
      back = 0;
    }
  }
  (c->c).a85.a85val = 0UL;
  (c->c).a85.used = 0;
  (c->c).a85.linepos = 0;
  
  return back;
}

static int
cell_buf_end DK_P3(dk_of_t *,o,dk_of_cell_t *,c,size_t,n)
{
  int back = 0;
  size_t sz;
  
  if(((c->c).b.lgt) && ((c->c).b.buffer)) {
    if((c->c).b.used) {
      sz = (c->c).b.used;
      if(add_from_cell(o,n,(c->c).b.buffer,sz) == sz) {
        back = 1;
      }
    } else {
      back = 1;	
    }
  } else {	
  }
  (c->c).b.used = 0;
  
  return back;
}

static int
cell_end DK_P3(dk_of_t *,o,dk_of_cell_t *,c,size_t,n)
{
  int back = 0;
  
  switch(c->what) {
#if DK_HAVE_ZLIB_H
    case DK_OF_TYPE_FLATE: {
      back = cell_flate_end(o,c,n);
    } break;
#endif
    case DK_OF_TYPE_ASCII85: {
      back = cell_a85_end(o,c,n);
    } break;
    case DK_OF_TYPE_BUFFERED: {
      back = cell_buf_end(o,c,n);
    } break;
    /* further cell types here */
    default: {		
    } break;
  }
  
  return back;
}

/* }}} cell_end */



/* {{{ start_chunk: initialize all cells to start new chunk */
static int
start_chunk DK_P1(dk_of_t *,o)
{
  int back = 0;
  dk_of_cell_t *c; size_t n;
  
  if((o->have_data) == 0) {
    if((o->cells) && (o->n_of_cells)) {
      back = 1;
      for(n = 0; n < (o->n_of_cells); n++) {
        if(!cell_start(o,&((o->cells)[n]),n)) {
	  back = 0;
	}
      }
      if(back) {
        o->have_data = 1;
      }
    } else {		
    }
  } else {		
  }
  
  return back;
}
/* }}} start_chunk */



/* {{{ end_chunk: flush chunk data through all the cells */
static int
end_chunk DK_P1(dk_of_t *,o)
{
  int back = 0;
  dk_of_cell_t *c; size_t n;
  
  if((o->have_data) == 1) {
    if((o->cells) && (o->n_of_cells)) {
      back = 1;
      n = o->n_of_cells;
      while(n--) {
        if(!cell_end(o,&((o->cells)[n]),n)) {
	  back = 0;
	}
      }
      o->have_data = (back ? 0 : -1);
    } else {		
    }
  } else {		
  }
  
  return back;
}
/* }}} end_chunk */



/* {{{ delete_of: destroy output filter */

#if DK_HAVE_ZLIB_H
static void
delete_flate_cell DK_P1(dk_of_cell_t *,c)
{
  void *ptr;
  
  if((c->c).flate.zs) {
    ptr = (c->c).flate.zs; dk_delete(ptr);
  }
  if((c->c).flate.ibuffer) {
    ptr = (c->c).flate.ibuffer; dk_delete(ptr);
  }
  if((c->c).flate.obuffer) {
    ptr = (c->c).flate.obuffer; dk_delete(ptr);
  }
  
}
#endif

static void
delete_a85_cell DK_P1(dk_of_cell_t *,c)
{
  void *ptr;
  
  (c->c).a85.a85val = 0UL;
  (c->c).a85.linepos = 0;
  (c->c).a85.used = 0;
  
}

static void
delete_buf_cell DK_P1(dk_of_cell_t *,c)
{
  void *ptr;
  
  if((c->c).b.buffer) {
    ptr = (c->c).b.buffer; dk_delete(ptr);
  }
  
}

static void
delete_cell DK_P1(dk_of_cell_t *,c)
{
  switch(c->what) {
#if DK_HAVE_ZLIB_H
    case DK_OF_TYPE_FLATE: {
      delete_flate_cell(c);
    } break;
#endif
    case DK_OF_TYPE_ASCII85: {
      delete_a85_cell(c);
    } break;
    case DK_OF_TYPE_BUFFERED: {
      delete_buf_cell(c);
    } break;
    /* further cell types here */
    default: {	
    } break;
  }
  DK_MEMRES(c,sizeof(dk_of_cell_t));
}

static int
delete_of DK_P1(dk_of_t *,o)
{
  int back = 1;
  dk_of_cell_t *c; size_t i;
  void *ptr;
  
  if(o->have_data == 1) {
    if(!end_chunk(o)) { back = 0; }
  }
  c = o->cells;
  if(c && (o->n_of_cells)) {
    for(i = 0; i < o->n_of_cells; i++) {
      delete_cell(c++);
    }
    ptr = (void *)(o->cells);
    dk_delete(ptr);
  }
  o->target = NULL;
  o->cells = NULL;
  o->n_of_cells = 0;
  o->have_data = 0;
  dk_delete(o);
  
  return back;
}
/* }}} delete_of */


void
dkof_stream_fct DK_P1(dk_stream_api_t *, a)
{
  dk_stream_t *s;
  dk_of_t *o;
  dk_of_cell_t *c;
  if(a) {
    a->return_value = 0;
    a->error_code = DK_ERR_FUNCTION_UNSUPPORTED;
    s = (dk_stream_t *)(a->strm);
    if(s) {
      o = (dk_of_t *)(s->data);
      if(o) {
        c = o->cells;
	if(c) {
	  switch(a->cmd) {
	    case DK_STREAM_CMD_TEST: {
	      switch( (a->params).cmd ) {
	        case DK_STREAM_CMD_TEST:
		case DK_STREAM_CMD_FINISH:
		case DK_STREAM_CMD_WRBUF: {
		  a->return_value= 1;
		} break;
	      }
	    } break;
	    case DK_STREAM_CMD_FINISH: {
	      a->return_value = delete_of(o);
	      s->data = NULL; s->fct = NULL;
	    } break;
	    case DK_STREAM_CMD_WRBUF: {
	      void *b; size_t sz, res, n;
	      if(o->n_of_cells) {
	        sz = (a->params).length;
		b  = (a->params).buffer;
		n  = o->n_of_cells - 1;
		if(b && sz) {
		  res = add_to_cell(o, &(c[n]), n, b, sz);
		  if(res) {
		    a->return_value = 1;
		    (a->results).used = res;
		  } else {	
		    a->error_code = DK_ERR_IO;
		  }
		} else {	
		}
	      } else {		
	      }
	    } break;
	  }
	} else {	
	}
      } else {	
      }
    } else {	
    }
  } else {	
  }
}

dk_stream_t *
dkof_open DK_P2(dk_stream_t *,t,size_t,nof)
{
  dk_stream_t *back = NULL;
  dk_of_t *o = NULL;
  dk_of_cell_t *c = NULL, *cptr = NULL;
  size_t n;
  
  if(t && nof) {	
    o = dk_new(dk_of_t,1); c = dk_new(dk_of_cell_t,nof);
    if(o && c) {
      back = dkstream_new((void *)o, (dk_stream_fct_t *)dkof_stream_fct);
      if(back) {	
        o->flags = 0;
	o->target = t;
	o->cells = c;
	o->n_of_cells = nof;
	o->have_data = 0;
	cptr = c; n = nof;
	while(nof--) {
	  DK_MEMRES(cptr,sizeof(dk_of_cell_t));
	  cptr->what = DK_OF_TYPE_NONE;
	  cptr++;
	}
      } else {		
      }
    } else {		
    }
    if(!back) {
      if(o) dk_delete(o);
      if(c) dk_delete(c);
    }
  } else {		
  }
  
  return back;
}

void
dkof_close DK_P1(dk_stream_t *,s)
{
  
  if(s) {
    dkstream_delete(s);
  }
  
}

int
dkof_set DK_P3(dk_stream_t *,s,size_t,n,int,w)
{
  int back = 0;
  dk_of_t *o; dk_of_cell_t *c;
  
  if(s) {
    o = (dk_of_t *)(s->data);
    if(o) {
      c = o->cells;
      if(c) {
        if(n < (o->n_of_cells)) {
	  back = cell_init(&((o->cells)[n]), w);
	} else {	
	}
      } else {	
      }
    } else {	
    }
  } else {	
  }
  
  return back;
}

int
dkof_start_chunk DK_P1(dk_stream_t *,s)
{
  int back = 0;
  dk_of_t *o;
  
  if(s) {
    o = (dk_of_t *)(s->data);
    if(o) {
      back = start_chunk(o);
    } else {		
    }
  } else {		
  }
  
  return back;
}

int
dkof_end_chunk DK_P1(dk_stream_t *,s)
{
  int back = 0;
  dk_of_t *o;
  
  if(s) {
    o = (dk_of_t *)(s->data);
    if(o) {
      back = end_chunk(o);
    } else {		
    }
  } else {		
  }
  
  return back;
}

void
dkof_set_crnl DK_P2(dk_stream_t *,s,int,v)
{
  dk_of_t *o;
  
  if(s) {
    o = (dk_of_t *)(s->data);
    if(o) {
      if(v) {
        o->flags |= 1;
      } else {
        o->flags &= (~1);
      }
    } else {	
    }
  }
}


static char sccs_id[] = {
"@(#)dkof.ctr 1.4 08/25/03 (krause)\toutput filter"
};


