/* 
 * The dk libraries (libdkport + libdkc)
 * Copyright (C) 2000, 2001 - Dirk Krause
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 * In this package the copy of the GNU Library General Public License
 * is placed in file COPYING.
 */


#include <dk.h>
#include <dktypes.h>
#define DK_SIGNAL_C
#include <dksignal.h>

#if DK_HAVE_UNISTD_H
#include <unistd.h>
#endif
#if DK_HAVE_STDDEF_H
#include <stddef.h>
#endif
#if DK_HAVE_SIGNAL_H
#include <signal.h>
#endif

dk_signal_disp_t dksignal_set DK_P2(int, signo, dk_signal_disp_t, disp)
{
  dk_signal_disp_t back = NULL;
#if DK_HAVE_SIGACTION
  struct sigaction act, oact;
  back = (dk_signal_disp_t)SIG_ERR;
  act.sa_handler =
#ifdef __cplusplus
  (dk_extc_signal_disp_t)disp;
#else
  disp;
#endif
  sigemptyset(&act.sa_mask);
  act.sa_flags = 0;
#if !DK_NO_SIGNAL_RESTART_ALARM
  if(signo == SIGALRM) {
#ifdef SA_INTERRUPT
  act.sa_flags |= SA_INTERRUPT;
#else
#ifdef SA_RESTART
  act.sa_flags |= SA_RESTART;
#endif
#endif
  }
#endif
  if(sigaction(signo, &act, &oact) < 0) {
    back = (dk_signal_disp_t)SIG_ERR;
  } else {
    back = (dk_signal_disp_t)oact.sa_handler;
  }
#else
#if DK_HAVE_SIGSET
  back = (dk_signal_disp_t)sigset(signo,disp);
#else
#if DK_HAVE_SIGNAL
  back = (dk_signal_disp_t)signal(signo,disp);
#endif
#endif
#endif
  return back;
}

int		 dksignal_available DK_P0()
{
  int back = 0;
#if DK_HAVE_SIGACTION
  back = 3;
#else
#if DK_HAVE_SIGSET
  back = 2;
#else
#if DK_HAVE_SIGNAL
  back = 1;
#endif
#endif
#endif
  return back;
}

#ifndef LINT
static char sccs_id[] = {
"@(#)dksignal.ctr\t1.6\t08/25/03\t(Dirk Krause) - signal handling"
};
#endif

