/* 
 * The dk libraries (libdkport + libdkc)
 * Copyright (C) 2000, 2001 - Dirk Krause
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 * In this package the copy of the GNU Library General Public License
 * is placed in file COPYING.
 */


#include <dk.h>
#include <dktypes.h>
#include <dkmem.h>

#define DK_STRS_C 1
#include <dkss.h>


/* ********************************************************************* */
/* *                                                                   * */
/* *   The current string search object.                               * */
/* *                                                                   * */
/* ********************************************************************* */

#line 33 "dkss.ctr"

static dk_ss_t	*the_object = NULL;

dk_ss_t	*
dkss_get_object	DK_P0()
{
  
  return the_object;
}


/* ********************************************************************* */
/* *                                                                   * */
/* *   Set a new current string search object.                         * */
/* *   Returns a pointer to the formerly used object.                  * */
/* *                                                                   * */
/* ********************************************************************* */

#line 46 "dkss.ctr"

dk_ss_t *
dkss_set_object	DK_P1(dk_ss_t *,obj)
{
  dk_ss_t *back;
  
  back = the_object;
  the_object = obj;
  return back;
}


/* ********************************************************************* */
/* *                                                                   * */
/* *   Searches for a string table entry.                              * */
/* *   The table name is given in table, the entry                     * */
/* *   name in key. def is a pointer to a default                      * */
/* *   string which is returned if no matching                         * */
/* *   entry is found.                                                 * */
/* *                                                                   * */
/* ********************************************************************* */

#line 63 "dkss.ctr"

char	*
dkss_find		DK_P3(char *, table, char *, key, char *, def)
{
  char *back = NULL;
  
  if(table) {
  if(key) {
  if(the_object) {
    void *obj;
    dk_fct_type_ss *fct;
    obj = the_object->obj; fct = the_object->fct;
    if(fct) {
      back = (*fct)(obj, table, key, def);
    }
  }
  }
  }
  if(!back) { 
    back = def;
  } 
  return back;
}


/* ********************************************************************* */
/* *                                                                   * */
/* *   Create a new string search object, initialize                   * */
/* *   to 0.                                                           * */
/* *   Returns a pointer to the new object or NULL on error.           * */
/* *                                                                   * */
/* ********************************************************************* */

#line 91 "dkss.ctr"

dk_ss_t *dkss_new	DK_P0()
{
  dk_ss_t *back = NULL;
  back = dk_new(dk_ss_t,1);
  
  return back;
}


/* ********************************************************************* */
/* *                                                                   * */
/* *   Create new string search object for a given                     * */
/* *   target object and a given function.                             * */
/* *   Returns a pointer to the new object or NULL on error.           * */
/* *                                                                   * */
/* ********************************************************************* */

#line 104 "dkss.ctr"

dk_ss_t *dkss_new_for	DK_P2(void *, obj, dk_fct_type_ss *, fct)
{
  dk_ss_t *back = NULL;
  if(fct) {
  back = dkss_new();
  if(back) {
    back->obj = obj;
    back->fct = fct;
  }
  } 
  return back;
}


/* ********************************************************************* */
/* *                                                                   * */
/* *   Deallocates memory used by a dynamically allocated              * */
/* *   string search object.                                           * */
/* *                                                                   * */
/* ********************************************************************* */

#line 121 "dkss.ctr"

void    dkss_delete	DK_P1(dk_ss_t *, p)
{
  
  if(p) {
    p->obj = NULL;
    p->fct = NULL;
    dk_delete(p);
  }
}


#ifndef LINT
static char sccs_id[] = {
"@(#)dkss.ctr\t1.3\t08/25/03\t(Dirk Krause) - String search module"
};
#endif
