/* 
 * The dk libraries (libdkport + libdkc)
 * Copyright (C) 2000, 2001 - Dirk Krause
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 * In this package the copy of the GNU Library General Public License
 * is placed in file COPYING.
 */

#include <dk.h>
#include <dkmem.h>
#include <dktypes.h>
#include <dksf.h>
#include <dkenc.h>

#if DK_HAVE_ZLIB_H
#include <zlib.h>
#endif
#if DK_HAVE_BZLIB_H
#include <bzlib.h>
#endif

#define DK_STREAM_C 1
#include <dkstream.h>


#line 39 "dkstream.ctr"


#include <stdio.h>

dk_stream_t *dkstream_new DK_P2(void *, data, dk_stream_fct_t *, fct)
{
  dk_stream_t *back = NULL;
  
  if(data && fct) { 
    back = dk_new(dk_stream_t,1); 
    if(back) {
      back->data = data; back->fct = fct;
      (back->api).strm = (void *)back;
      back->bytes = 0UL;
      back->flags = 0;
    }
  } 
  return back;
}

size_t dkstream_write DK_P3(dk_stream_t *, st, char *, b, size_t, l)
{
  size_t back = 0;
  unsigned long size_to_add;
  dk_stream_fct_t *f; void *d; dk_stream_api_t *a;
  
  if(st && b && l) { 
    d = st->data; f = st->fct; a = &(st->api);
    if(d && f) { 
      a->cmd = DK_STREAM_CMD_WRBUF;
      (a->params).buffer = b;
      (a->params).length = l;
      (a->results).buffer = b;
      (a->results).length = l;
      (*f)(a); 
      if(a->return_value) { 
	back = (a->results).used;
	size_to_add = (unsigned long)back;
	st->bytes = st->bytes + size_to_add;
      }
    }
  } 
  return back;
}

size_t dkstream_read  DK_P3(dk_stream_t *, st, char *, b, size_t, l)
{
  size_t back = 0;
  dk_stream_fct_t *f; void *d; dk_stream_api_t *a;
  
  if(st && b && l) { 
    d = st->data; f = st->fct; a = &(st->api);
    if(d && f) { 
      a->cmd = DK_STREAM_CMD_RDBUF;
      (a->params).buffer = b;
      (a->params).length = l;
      (a->results).buffer = b;
      (a->results).length = l;
      (*f)(a); 
      if(a->return_value) { 
	back = (a->results).used;
      } else {
	a->cmd = DK_STREAM_CMD_TEST;
	(a->params).cmd = DK_STREAM_CMD_AT_END;
	(*f)(a);
	if(a->return_value) {
	  a->cmd = DK_STREAM_CMD_AT_END;
	  (*f)(a);
	  if(a->return_value) {
	    st->flags |= 1;
	  }
	}
      }
    }
  } 
  return back;
}

void  dkstream_delete DK_P1(dk_stream_t *, st)
{
  dk_stream_api_t *a; dk_stream_fct_t *f;
  
  if(st) {
    if((st->data) && (st->fct)) {
      a = &(st->api); f = st->fct;
      a->cmd = DK_STREAM_CMD_FINAL;
      (*f)(a);
    }
    st->data = NULL; st->fct = NULL; (st->api).strm = NULL;
    
    dk_delete(st);
  }
  
}

void file_stream_fct DK_P1(dk_stream_api_t *, a)
{
  dk_stream_t *s;
  void *d;
  FILE *f;
  char *b;
  size_t l, x;
  
  if(a) { 
    a->return_value = 0;
    s = (dk_stream_t *)(a->strm);
    if(s) { 
      d = s->data;
      f = (FILE *)d;
      if(f) { 
	switch(a->cmd) {
	  case DK_STREAM_CMD_TEST: {
	    switch( (a->params).cmd ) {
	      case DK_STREAM_CMD_TEST: a->return_value = 1; break;
	      case DK_STREAM_CMD_AT_END: a->return_value = 1; break;
	      case DK_STREAM_CMD_RDBUF: a->return_value = 1; break;
	      case DK_STREAM_CMD_WRBUF: a->return_value = 1; break;
	      case DK_STREAM_CMD_FINAL: a->return_value = 1; break;
	      case DK_STREAM_CMD_FINISH: a->return_value = 1; break;
	      case DK_STREAM_CMD_REWIND: a->return_value = 1; break;
	      case DK_STREAM_CMD_FLUSH: a->return_value = 1; break;
	      case DK_STREAM_CMD_GETS: a->return_value = 1; break;
	      case DK_STREAM_CMD_PUTS: a->return_value = 1; break;
	    }
	  } break;
	  case DK_STREAM_CMD_GETS: {
	    b = (a->params).buffer; l = (a->params).length;
	    if(b && l) { 
              if(fgets(b,l,f)) {
		a->return_value = 1;
	      }
	    }
	  } break;
	  case DK_STREAM_CMD_PUTS: {
	    int i;
	    b = (a->params).buffer;
	    if(b) { 
	      i = fputs(b,f);
	      if(i > 0) {
		a->return_value = 1; (a->results).length = i;
	      }
	    }
	  } break;
	  case DK_STREAM_CMD_AT_END: {		
	    if(feof(f)) {
	      a->return_value = 1;
	    } else {
	      a->return_value = 0;
	    }
	  } break;
	  case DK_STREAM_CMD_RDBUF: {		
	    b = (a->params).buffer; l = (a->params).length;
	    if(b && l) { 
	      x = fread((void *)b, 1, l, f); 
	      if(x) { 
		(a->results).used = x;
		a->return_value = 1;
	      }
	    }
	  } break;
	  case DK_STREAM_CMD_WRBUF: {		
	    b = (a->params).buffer; l = (a->params).length;
	    if(b && l) { 
	      x = fwrite((void *)b, 1, l, f); 
	      if(x) { 
		(a->results).used = x;
		a->return_value = 1;
	      }
	    }
	  } break;
	  case DK_STREAM_CMD_FINAL: {		
	    if(fflush(f) == 0) {
	      a->return_value = 1;
	    }
	  } break;
	  case DK_STREAM_CMD_FINISH: {		
	    if(fclose(f) == 0) {
	      a->return_value = 1;
	    }
	    s->data = NULL; /* file is closed now */
	    s->fct  = NULL; /* nothing more to do */
	  } break;
	  case DK_STREAM_CMD_REWIND: {		
	    rewind(f); a->return_value = 1;
	  } break;
	  case DK_STREAM_CMD_FLUSH: {		
	    if(fflush(f) == 0) {
	      a->return_value = 1;
	    }
	  } break;
	}
      } else {	/* NO FILE */
	
      }
    } else {	/* INVALID POINTER TO STREAM */
      
    }
    
  } else {	/* INVALID API POINTER */
    
    
  }
}

void file_stream_function DK_P1(dk_stream_api_t *, a)
{
  if(a) {
    if(a->cmd == DK_STREAM_CMD_FINISH) {
      a->cmd = DK_STREAM_CMD_FINAL;
      file_stream_fct(a);
      a->cmd = DK_STREAM_CMD_FINISH;
    } else {
      file_stream_fct(a);
    }
  }
}

void bz2_stream_fct DK_P1(dk_stream_api_t *, a)
{
#if DK_HAVE_BZLIB_H
  BZFILE *bzf;
  dk_stream_t *st;
  void *d;
  void *b;
  int   l;
  int   x;
#endif
  if(a) {
    a->return_value = 0;
#if DK_HAVE_BZLIB_H
    st = (dk_stream_t *)(a->strm);
    if(st) {
      d = st->data;
      bzf = (BZFILE *)d;
      switch(a->cmd) {
	case DK_STREAM_CMD_TEST : {
	  switch((a->params).cmd) {
	    case DK_STREAM_CMD_TEST : {
	      a->return_value = 1;
	    } break;
	    case DK_STREAM_CMD_RDBUF : {
	      a->return_value = 1;
	    } break;
	    case DK_STREAM_CMD_WRBUF : {
	      a->return_value = 1;
	    } break;
	    case DK_STREAM_CMD_FINAL : {
	      a->return_value = 1;
	    } break;
	    case DK_STREAM_CMD_FINISH : {
	      a->return_value = 1;
	    } break;
	    case DK_STREAM_CMD_REWIND : {
	    } break;
	    case DK_STREAM_CMD_FLUSH : {
	      a->return_value = 1;
	    } break;
	    case DK_STREAM_CMD_AT_END : {
	    } break;
	    case DK_STREAM_CMD_GETS : {
	    } break;
	    case DK_STREAM_CMD_PUTS : {
	    } break;
	  }
	} break;
	case DK_STREAM_CMD_RDBUF : {
	  if(!((st->flags) & 1)) {
	    b = ((void *)((a->params).buffer));
	    l = ((int)((a->params).length));
	    if(b && (l > 0)) {
	      x = BZ2_bzread(bzf,b,l);
	      if(x > 0) {
	        (a->results).used = x;
	        a->return_value = 1;
	      } else {
	        st->flags |= 1;
	      }
	    }
	  }
	} break;
	case DK_STREAM_CMD_WRBUF : {
	  b = ((void *)((a->params).buffer));
	  l = ((int)((a->params).length));
	  if(b && (l > 0)) {
	    x = BZ2_bzwrite(bzf,b,l);
	    if(x > 0) {
	      (a->results).used = x;
	      a->return_value = 1;
	    }
	  }
	} break;
	case DK_STREAM_CMD_FINAL : {
	  BZ2_bzflush(bzf);
	  a->return_value = 1;
	} break;
	case DK_STREAM_CMD_FINISH : {
	  BZ2_bzclose(bzf);
	  a->return_value = 1;
	  st->data = NULL;
	  st->fct  = NULL;
	} break;
	case DK_STREAM_CMD_REWIND : {
	} break;
	case DK_STREAM_CMD_FLUSH : {
	  BZ2_bzflush(bzf);
	  a->return_value = 1;
	} break;
	case DK_STREAM_CMD_AT_END : {
	} break;
	case DK_STREAM_CMD_GETS : {
	} break;
	case DK_STREAM_CMD_PUTS : {
	} break;
      }
    }
#endif
  }
}


void bz2_stream_function DK_P1(dk_stream_api_t *, a)
{
#if DK_HAVE_BZLIB_H
  if(a) {
    if(a->cmd == DK_STREAM_CMD_FINISH) {
      a->cmd = DK_STREAM_CMD_FINAL;
      bz2_stream_fct(a);
      a->cmd = DK_STREAM_CMD_FINISH;
    } else {
      bz2_stream_fct(a);
    }
  }
#endif
}

void gz_stream_fct DK_P1(dk_stream_api_t *, a)
{
#if DK_HAVE_ZLIB_H
  gzFile gzf;
  dk_stream_t *st;
  char        *b;
  size_t       l;
  void        *d;
  int x;
#endif
  
  if(a) {
    a->return_value = 0;
#if DK_HAVE_ZLIB_H
    st = (dk_stream_t *)(a->strm);
    if(st) {					
      d = st->data;
      if(d) {					
	gzf = (gzFile)d;
	switch(a->cmd) {
	  case DK_STREAM_CMD_TEST: {		
	    switch( (a->params).cmd) {
	      case DK_STREAM_CMD_TEST: {
		a->return_value = 1;
	      } break;
	      case DK_STREAM_CMD_RDBUF: {
		a->return_value = 1;
	      } break;
	      case DK_STREAM_CMD_WRBUF: {
		a->return_value = 1;
	      } break;
	      case DK_STREAM_CMD_FINAL: {
		a->return_value = 1;
	      } break;
	      case DK_STREAM_CMD_FINISH: {
		a->return_value = 1;
	      } break;
	      case DK_STREAM_CMD_REWIND: {
		a->return_value = 1;
	      } break;
	      case DK_STREAM_CMD_FLUSH: {
		a->return_value = 1;
	      } break;
	      case DK_STREAM_CMD_AT_END: {
		a->return_value = 1;
	      } break;
	      case DK_STREAM_CMD_GETS: {
		a->return_value = 1;
	      } break;
	      case DK_STREAM_CMD_PUTS: {
		a->return_value = 1;
	      } break;
	    }
	  } break;
	  case DK_STREAM_CMD_RDBUF: {		
	    b = (a->params).buffer; l = (a->params).length;
	    if(b && l) {			
	      x = gzread(gzf,b,l);
	      if(x > 0) {			
		(a->results).used = x;
		a->return_value = 1;
	      }
	    }
	  } break;
	  case DK_STREAM_CMD_WRBUF: {		
	    b = (a->params).buffer; l = (a->params).length;
	    if(b && l) {			
	      x = gzwrite(gzf,b,l);
	      if(x > 0) {			
		(a->results).used = x;
		a->return_value = 1;
	      }
	    }
	  } break;
	  case DK_STREAM_CMD_FINAL: {		
	    if(gzflush(gzf, Z_FULL_FLUSH) == Z_OK) {		
	      a->return_value = 1;
	    }
	  } break;
	  case DK_STREAM_CMD_FINISH: {		
	    if(gzclose(gzf) == Z_OK) {		
	      a->return_value = 1;
	    }
	    st->data = NULL;
	    st->fct  = NULL;
	  } break;
	  case DK_STREAM_CMD_REWIND: {		
	    if(gzrewind(gzf) == Z_OK) {		
	      a->return_value = 1;
	    }
	  } break;
	  case DK_STREAM_CMD_FLUSH: {		
	    if(gzflush(gzf, Z_FULL_FLUSH) == Z_OK) {		
	      a->return_value = 1;
	    }
	  } break;
	  case DK_STREAM_CMD_AT_END: {		
	    if(gzeof(gzf) == 1) {		
	      a->return_value = 1;
	    }
	  } break;
	  case DK_STREAM_CMD_GETS: {		
	    b = (a->params).buffer; l = (a->params).length;
            (a->results).buffer = b; (a->results).length = l;
            if(b && l) {
	      if(gzgets(gzf,b,l)) {		
		a->return_value = 1;
	      }
	    }
	  } break;
	  case DK_STREAM_CMD_PUTS: {
	    int i;
	    
	    b = (a->params).buffer;
	    if(b) {
	      i = gzputs(gzf,b);
	      if(i > 0) {
		a->return_value = 1; (a->results).length = i;
	      }
	    }
	  } break;
	}
      } else {				
      }
    } else {				
    }
#endif
    
  } else { 
  }
}


void gz_stream_function DK_P1(dk_stream_api_t *, a)
{
#if DK_HAVE_ZLIB_H
  if(a) {
    if(a->cmd == DK_STREAM_CMD_FINISH) {
      a->cmd = DK_STREAM_CMD_FINAL;
      gz_stream_fct(a);
      a->cmd = DK_STREAM_CMD_FINISH;
    } else {
      gz_stream_fct(a);
    }
  }
#endif
}


int dkstream_puts	DK_P2(dk_stream_t *, st, char *, str)
{
  int back = 0;
  dk_stream_api_t *a;
  dk_stream_fct_t *fct;

  if(st && str) {
    a = (dk_stream_api_t *)(&(st->api));
    a->cmd = DK_STREAM_CMD_TEST; (a->params).cmd = DK_STREAM_CMD_PUTS;
    fct = st->fct;
    (*fct)(a);
    if(a->return_value) {
      a->cmd = DK_STREAM_CMD_PUTS;
      (a->params).buffer = str;
      (*fct)(a);
      back = a->return_value;
      if(back) {
        st->bytes += strlen(str);
      }
    } else {
      size_t sz;
      sz = strlen(str);
      a->cmd = DK_STREAM_CMD_WRBUF;
      (a->params).buffer = str;
      (a->params).length = sz;
      (a->results).buffer = str;
      (a->results).length = sz;
      (*fct)(a);
      back = a->return_value;
      if(back) {
        st->bytes += (a->results).used;
      }
      if(((a->results).length) != sz) {
	back = 0;
      }
    }
  }
  return back;
}

char *dkstream_gets DK_P3(dk_stream_t *, st, char *, b, size_t, l)
{
  char *back = NULL;
  dk_stream_api_t *a;
  void *d;
  dk_stream_fct_t *fct;
  int goon;
  char c;
  size_t used;
  
  if(st && b && l) { 
    a = &(st->api);
    d = st->data; fct = st->fct;
    if(d && fct) { 
      a->cmd = DK_STREAM_CMD_TEST;
      (a->params).cmd = DK_STREAM_CMD_GETS;
      (*fct)(a);
      if(a->return_value) {	
	(a->params).buffer = b;
	(a->params).length = l;
	a->cmd = DK_STREAM_CMD_GETS;
	(*fct)(a);
	if(a->return_value) {
	  back = b;
	}
      } else {			
	goon = 1; used = 0;
	while(goon) {
	  (a->params).buffer = &c;
	  (a->params).length = 1;
	  (a->results).buffer = &c;
	  (a->results).length = 1;
	  a->cmd = DK_STREAM_CMD_RDBUF;
	  (*fct)(a);
	  if(a->return_value) {
            if(((a->results).used) == 1) {
	      switch(c) {
		case '\0': {
		  b[used++] = '\0'; goon = 0;
		} break;
		case '\r':
		case '\n': {
		  b[used++] = c; b[used] = '\0'; goon = 0;
		} break;
		default: {
		  b[used++] = c;
		  if(used >= (l - 1)) {
		    goon = 0; b[used] = '\0';
		  }
		} break;
	      }
	    } else {
	      b[used] = '\0'; goon = 0;
	    }
	  } else {
	    b[used] = '\0'; goon = 0;
	  }
	}
	if(used) {
	  back = b;
	}
      }
    }
  }
  
  return back;
}

dk_stream_t *dkstream_openfile DK_P4(char *, name, char *, mode, int, ign, int *, reason)
{
  FILE *fipo;
  dk_stream_t *back = NULL;
  if(name && mode) {
#if NO_FOPEN_CHECKING
    fipo = fopen(name, mode);
#else
    fipo = dksf_msfo(name,mode,ign,reason);
#endif
    if(fipo) {
      back = dkstream_new((void *)fipo, file_stream_fct);
      if(!back) {
	fclose(fipo);
      }
    }
  }
  return back;
}

dk_stream_t *dkstream_for_file DK_P1(FILE *, fipo)
{
  dk_stream_t *back = NULL;
  if(fipo) {
    back = dkstream_new((void *)fipo, file_stream_function);
  }
  return back;
}

#if DK_HAVE_ZLIB_H
static void correct_gz_mode(char *m, size_t l)
{
  int want_to_write;
  int have_compression_level;
  char *ptr;
  size_t lgt;
  
  want_to_write = 0; have_compression_level = 0;
  lgt = strlen(m);
  ptr = m;
  while(*ptr) {
    switch(*ptr) {
      case 'w': {
	want_to_write = 1;
      } break;
      case '0':
      case '1':
      case '2':
      case '3':
      case '4':
      case '5':
      case '6':
      case '7':
      case '8':
      case '9':
      {
	have_compression_level = 1;
      } break;
    }
    ptr++;
  }
  if(want_to_write && (!have_compression_level)) {
    if(lgt < (l - 1)) {
      m[lgt] = '9'; m[lgt+1] = '\0';
    }
  } 
}
#endif

static int
is_open_mode_write DK_P1(char *,mode)
{
  int back = 0;
  char *ptr;
  if(mode) {
    ptr = mode;
    while(*ptr) {
      switch(*ptr) {
        case 'w':
	case 'a':
	case '+':
	{
	  back = 1;
	} break;
      }
      ptr++;
    }
  }
  
  return back;
}

dk_stream_t *dkstream_opengz DK_P4(char *, name, char *, mode, int, ign, int *, reason)
{
  dk_stream_t *back = NULL;
#if DK_HAVE_ZLIB_H
  size_t my_mode_used;
  char my_mode[8];
  gzFile gzf;
  if(name && mode) {
    my_mode_used = strlen(mode);
    if(my_mode_used < sizeof(my_mode)) {
      strcpy(my_mode, mode);
      correct_gz_mode(my_mode, sizeof(my_mode));
#if NO_FOPEN_CHECKING
      gzf = gzopen(name, my_mode);
#else
      gzf = NULL;
      if(is_open_mode_write(my_mode)) {
      if(dksf_allowed_to_write(name,ign,reason)) {
        gzf = gzopen(name, my_mode);
      }
      } else {
        gzf = gzopen(name, my_mode);
      }
#endif
      if(gzf) {
	back = dkstream_new((void *)gzf, gz_stream_fct);
	if(!back) {
	  gzclose(gzf);
	}
      }
    }
  }
#endif
  return back;
}

#if DK_HAVE_ZLIB_H
dk_stream_t *dkstream_for_gz DK_P1(gzFile, gzf)
{
  dk_stream_t *back = NULL;
  if(gzf) {
    back = dkstream_new((void *)gzf, gz_stream_function);
  }
  return back;
}
#endif

dk_stream_t *dkstream_openbz2 DK_P4(char *, name, char *, mode, int, ign, int *, reason)
{
  dk_stream_t *back = NULL;
#if DK_HAVE_BZLIB_H
  BZFILE *bzf;
  if(name && mode) {
#if NO_FOPEN_CHECKING
    bzf = BZ2_bzopen(name,mode);
#else
    bzf = NULL;
    if(is_open_mode_write(mode)) {
    if(dksf_allowed_to_write(name,ign,reason)) {
      bzf = BZ2_bzopen(name,mode);
    }
    } else {
      bzf = BZ2_bzopen(name,mode);
    }
#endif
    if(bzf) {
      back = dkstream_new((void *)bzf, bz2_stream_fct);
      if(!back) {
	BZ2_bzclose(bzf);
      }
    }
  }
#endif
  return back;
}

#if DK_HAVE_BZLIB_H
dk_stream_t *dkstream_for_bz2 DK_P1(BZFILE *, bzf)
{
  dk_stream_t *back = NULL;
  if(bzf) {
    back = dkstream_new((void *)bzf, bz2_stream_function);
  }
  return back;
}
#endif

void dkstream_close DK_P1(dk_stream_t *, st)
{
  dk_stream_api_t *a;
  dk_stream_fct_t *f;
  void *d;
  if(st) {
    f = st->fct; d = st->data;
    if(f && d) {
      a = &(st->api);
      a->cmd = DK_STREAM_CMD_FINISH;
      (*f)(a);
      /* ##### 2001/07/04 - D. Krause */
      /* prevent FINAL from dkstream_delte() after FINISH */
      st->fct = NULL;
      st->data = NULL;
      /* ##### */
    }
    dkstream_delete(st);
  }
}


/* ********************************************************************* */
/* *                                                                   * */
/* *         The following functions do not depend on low              * */
/* *         level functionality directly.                             * */
/* *                                                                   * */
/* ********************************************************************* */

#line 834 "dkstream.ctr"


int dkstream_wb_word DK_P2(dk_stream_t *, st, dk_word, w)
{
  int back = 0;
  dk_word x;
  if(st) {
    x = dkenc_htons(w);
    if(dkstream_write(st, ((char *)(&x)), sizeof(dk_word)) == sizeof(dk_word)) {
      back = 1;
    }
  }
  return back;
}

int dkstream_wb_uword DK_P2(dk_stream_t *, st, dk_uword, w)
{
  int back = 0;
  dk_uword x;
  if(st) {
    x = dkenc_htons(w);
if(dkstream_write(st, ((char *)(&x)), sizeof(dk_uword)) == sizeof(dk_uword)) {
      back = 1;
}
  }
  return back;
}

int dkstream_wb_dword DK_P2(dk_stream_t *, st, dk_dword, w)
{
  int back = 0;
  dk_dword x;
  if(st) {
    x = dkenc_htonl(w);
if(dkstream_write(st, ((char *)(&x)), sizeof(dk_dword)) == sizeof(dk_dword)) {
      back = 1;
}
  }
  return back;
}

int dkstream_wb_udword DK_P2(dk_stream_t *, st, dk_udword, w)
{
  int back = 0;
  dk_udword x;
  if(st) {
    x = dkenc_htonl(w);
if(dkstream_write(st, ((char *)(&x)), sizeof(dk_udword)) == sizeof(dk_udword)) {
    back = 1;
}
  }
  return back;
}

int dkstream_rb_word DK_P2(dk_stream_t *, st, dk_word *, w)
{
  int back = 0;
  dk_word x;
  if(st && w) {
if(dkstream_read(st, ((char *)(&x)), sizeof(dk_word)) == sizeof(dk_word)) {
    *w = dkenc_ntohs(x);
    back = 1;
}
  }
  return back;
}


int dkstream_rb_uword DK_P2(dk_stream_t *, st, dk_uword *, w)
{
  int back = 0;
  dk_uword x;
  if(st && w) {
if(dkstream_read(st, ((char *)(&x)), sizeof(dk_uword)) == sizeof(dk_uword)) {
    *w = dkenc_ntohs(x);
    back = 1;
}
  }
  return back;
}

int dkstream_rb_dword DK_P2(dk_stream_t *, st, dk_dword *, w)
{
  int back = 0;
  dk_dword x;
  if(st && w) {
if(dkstream_read(st, ((char *)(&x)), sizeof(dk_dword)) == sizeof(dk_dword)) {
    *w = dkenc_ntohl(x);
    back = 1;
}
  }
  return back;
}

int dkstream_rb_udword DK_P2(dk_stream_t *, st, dk_udword *, w)
{
  int back = 0;
  dk_udword x;
  if(st && w) {
if(dkstream_read(st, ((char *)(&x)), sizeof(dk_udword)) == sizeof(dk_udword)) {
    *w = dkenc_ntohl(x);
    back = 1;
}
  }
  return back;
}

int dkstream_wb_string DK_P2(dk_stream_t *, st, char *, str)
{
  int back = 0;
  dk_uword lgt;
  lgt = 0UL;
  if(st) {
    if(str) {
      lgt = strlen(str);
      lgt++;
      if(dkstream_wb_uword(st,lgt)) {
	if(lgt == dkstream_write(st, str, lgt)) {
	  back = 1;
	}
      }
    } else {
      dkstream_wb_uword(st,0UL);
    }
  }
  return back;
}

char *dkstream_rb_string DK_P1(dk_stream_t *, st)
{
  char *back = NULL;
  dk_uword lgt;
  if(st) {
    if(dkstream_rb_uword(st,&lgt)) {
      back = dk_new(char,lgt);
      if(back) {
	if(dkstream_read(st, back, lgt) != lgt) {
	  dk_delete(back); back = NULL;
	}
      }
    }
  }
  return back;
}

static char suffix_none[] = { "" };

#if DK_HAVE_ZLIB_H
static char suffix_gz[] = { ".gz" };
#endif

#if DK_HAVE_BZLIB_H
static char suffix_bz2[] = { ".bz2" };
#endif

/*
	Table for file types to read
*/
dk_stream_suffix_t dkstream_read_suffix_list[] = {
  { suffix_none, (dk_stream_open_fct_t *)dkstream_openfile },
#if DK_HAVE_ZLIB_H
  { suffix_gz,   (dk_stream_open_fct_t *)dkstream_opengz },
#endif
#if DK_HAVE_BZLIB_H
  { suffix_bz2,  (dk_stream_open_fct_t *)dkstream_openbz2 },
#endif
  { NULL, (dk_stream_open_fct_t *)NULL }
};

/*
	Table for file types to write
*/
dk_stream_suffix_t dkstream_write_suffix_list[] = {
#if DK_HAVE_BZLIB_H
  { suffix_bz2,  (dk_stream_open_fct_t *)dkstream_openbz2 },
#endif
#if DK_HAVE_ZLIB_H
  { suffix_gz,   (dk_stream_open_fct_t *)dkstream_opengz },
#endif
  { suffix_none, (dk_stream_open_fct_t *)dkstream_openfile },
  { NULL, (dk_stream_open_fct_t *)NULL }
};

dk_stream_suffix_t *dkstream_get_read_suffixes DK_P0()
{
  return dkstream_read_suffix_list;
}

dk_stream_suffix_t *dkstream_get_write_suffixes DK_P0()
{
  return dkstream_write_suffix_list;
}

unsigned long
dkstream_get_bytes_written DK_P1(dk_stream_t *,st)
{
  unsigned long back = 0UL;
  if(st) { back = st->bytes; }
  return back;
}


#ifndef LINT
static char sccs_id[] = {
  "@(#)dkstream.ctr\t1.31\t12/17/03\t(krause) streams module"
};
#endif

