/* 
 * The dk libraries (libdkport + libdkc)
 * Copyright (C) 2000, 2001 - Dirk Krause
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 * In this package the copy of the GNU Library General Public License
 * is placed in file COPYING.
 */

#ifndef DK_STREAM_INC
#define DK_STREAM_INC 1

#include <dk.h>
#include <dktypes.h>

#if DK_HAVE_ZLIB_H
#include <zlib.h>
#endif
#if DK_HAVE_BZLIB_H
#include <bzlib.h>
#endif

#ifndef DK_STREAM_C
#if !DK_HAVE_PROTOTYPES
#define EXTERN extern
#else
#define EXTERN /* nix */
#endif
#else
#define EXTERN /* nix */
#endif

#ifdef __cplusplus
extern "C" {
#endif

EXTERN
dk_stream_t *dkstream_new DK_PR((void *data, dk_stream_fct_t *fct));

EXTERN
size_t dkstream_write DK_PR((dk_stream_t *s, char *b, size_t l));

EXTERN
size_t dkstream_read  DK_PR((dk_stream_t *s, char *b, size_t l));

EXTERN
void  dkstream_delete DK_PR((dk_stream_t *st));

EXTERN
dk_stream_t *dkstream_openfile DK_PR((char *name, char *mode, int ign, int *reason));

EXTERN
dk_stream_t *dkstream_opengz DK_PR((char *name, char *mode, int ign, int *reason));

EXTERN
dk_stream_t *dkstream_openbz2 DK_PR((char *name, char *mode, int ign, int *reason));

EXTERN
char *dkstream_gets DK_PR((dk_stream_t *st, char *buffer, size_t length));

EXTERN
int dkstream_puts DK_PR((dk_stream_t *st, char *buffer));

EXTERN
void dkstream_close DK_PR((dk_stream_t *st));

EXTERN
int dkstream_wb_word DK_PR((dk_stream_t *st, dk_word w));
EXTERN
int dkstream_wb_uword DK_PR((dk_stream_t *st, dk_uword w));
EXTERN
int dkstream_wb_dword DK_PR((dk_stream_t *st, dk_dword w));
EXTERN
int dkstream_wb_udword DK_PR((dk_stream_t *st, dk_udword w));
EXTERN
int dkstream_wb_string DK_PR((dk_stream_t *st, char *str));
EXTERN
int dkstream_rb_word DK_PR((dk_stream_t *st, dk_word *w));
EXTERN
int dkstream_rb_uword DK_PR((dk_stream_t *st, dk_uword *w));
EXTERN
int dkstream_rb_dword DK_PR((dk_stream_t *st, dk_dword *w));
EXTERN
int dkstream_rb_udword DK_PR((dk_stream_t *st, dk_udword *w));
EXTERN
char *dkstream_rb_string DK_PR((dk_stream_t *st));
EXTERN
dk_stream_suffix_t *dkstream_get_read_suffixes DK_PR((void));
EXTERN
dk_stream_suffix_t *dkstream_get_write_suffixes DK_PR((void));
EXTERN
unsigned long dkstream_get_bytes_written DK_PR((dk_stream_t *st));
EXTERN
dk_stream_t *dkstream_for_file DK_PR((FILE *f));
#if DK_HAVE_ZLIB_H
EXTERN
dk_stream_t *dkstream_for_gz DK_PR((gzFile gzf));
#endif
#if DK_HAVE_BZLIB_H
EXTERN
dk_stream_t *dkstream_for_bz2 DK_PR((BZFILE *bzf));
#endif
#ifdef __cplusplus
}
#endif

#define DK_STREAM_CMD_TEST	1
#define DK_STREAM_CMD_RDBUF	2
#define DK_STREAM_CMD_WRBUF	3
#define DK_STREAM_CMD_FINAL	4
#define DK_STREAM_CMD_FINISH	5
#define DK_STREAM_CMD_REWIND	6
#define DK_STREAM_CMD_FLUSH	7
#define DK_STREAM_CMD_AT_END	8
#define DK_STREAM_CMD_GETS	9
#define DK_STREAM_CMD_PUTS	10

#endif

