/* 
 * The dk libraries (libdkport + libdkc)
 * Copyright (C) 2000, 2001 - Dirk Krause
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 * In this package the copy of the GNU Library General Public License
 * is placed in file COPYING.
 */


#ifndef DKTYPES_INC
#define DKTYPES_INC 1

#ifndef DKCONFIG_INC
#include <dkconfig.h>
#endif

#include <stdio.h>

#if DK_HAVE_LIMITS_H
#include <limits.h>
#endif
#if DK_HAVE_SYS_PARAM_H
#include <sys/param.h>
#endif

#if DK_HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif

#if DK_HAVE_SYS_STAT_H
#include <sys/stat.h>
#endif

#if DK_HAVE_SYS_SOCKET_H
#include <sys/socket.h>
#else
#if DK_HAVE_WINSOCK2_H
#ifndef WINSOCK2_INCLUDED
#include <winsock2.h>
#define WINSOCK2_INCLUDED 1
#endif
#endif
#endif

#if DK_TIME_WITH_SYS_TIME
#include <sys/time.h>
#include <time.h>
#else
#if DK_HAVE_SYS_TIME_H
#include <sys/time.h>
#else
#if DK_HAVE_TIME_H
#include <time.h>
#endif
#endif
#endif
#if DK_HAVE_STDDEF_H
#include <stddef.h>
#endif
#if DK_HAVE_UNISTD_H
#include <unistd.h>
#endif
#if DK_HAVE_DIRECT_H
#include <direct.h>
#endif
#if DK_HAVE_DIRENT_H
#include <dirent.h>
#endif
#if DK_HAVE_DIR_H
#include <dir.h>
#endif
#if DK_HAVE_IO_H
#include <io.h>
#endif
#if DK_HAVE_NETINET_IN_H
#include <netinet/in.h>
#endif
#if DK_HAVE_FCNTL_H
#include <fcntl.h>
#endif
#if DK_HAVE_WCHAR_H
#include <wchar.h>
#endif
#if DK_HAVE_ZLIB_H
#include <zlib.h>
#endif

#include <dkwin.h>

#ifdef PATH_MAX
#define DK_MAX_PATH (PATH_MAX + 1)
#else
#ifdef  MAXPATHLEN
#define DK_MAX_PATH (MAXPATHLEN + 1)
#else
#define DK_MAX_PATH 1025
#endif
#endif

#define DK_MAX_DOUBLE 1.7e308

#ifdef __BORLANDC__
typedef int mode_t;
#endif

#if !DK_HAVE_WCHAR_H
#if !WCHAR_T_DEFINED
#define WCHAR_T_DEFINED 1
typedef unsigned short wchar_t;
#endif
#endif

/*
	Simple data types, may be instantiated statically.
*/
typedef char		*dk_str;
typedef	dk_str		*dk_str_ptr;

typedef char		dk_byte;
typedef unsigned char	dk_ubyte;
typedef short		dk_word;
typedef unsigned short	dk_uword;
typedef long		dk_dword;
typedef unsigned long	dk_udword;

/* change the following void to int if necessary */
typedef void		dk_signal_ret_t;
typedef dk_signal_ret_t dk_signal_fct_t DK_PR((int));
typedef dk_signal_fct_t *dk_signal_disp_t;
#ifdef __cplusplus
extern "C" {
#endif
typedef dk_signal_ret_t dk_extc_signal_fct_t DK_PR((int));
#ifdef __cplusplus
}
#endif
typedef dk_extc_signal_fct_t *dk_extc_signal_disp_t;

typedef	char * dk_fct_type_ss DK_PR((void *obj, char *table, char *key, char *def));

typedef void dk_fct_type_log DK_PR((void *obj, int level, char **ptr, int num));


typedef	dk_ubyte	dk_utf8;
typedef	dk_uword	dk_utf16;
typedef	dk_udword	dk_unicodechar;

#if DK_HAVE_LONG_LONG_INT
typedef long long		dk_long_long_int_t;
typedef long long unsigned	dk_long_long_unsigned_t;
#else
typedef long		dk_long_long_int_t;
typedef long unsigned	dk_long_long_unsigned_t;
#endif

#if DK_HAVE_SOCKLEN_T
#define dk_socklen_t	socklen_t
#else
#if DK_HAVE_SIZE_T
#define dk_socklen_t	size_t
#else
#define dk_socklen_t	unsigned int
#endif
#endif

/*
 * Combined data types start here.
 * Do not instantiate directly.
 */

typedef struct {
  char  *key;			/* key to look up */
  char **value_pointer;		/* pointer to a pointer storing the result */
  char  *default_value;		/* default value */
} dk_string_finder_t;

/*
	More complex data types.
	Do not instatiate them statically.
	Use the ...open/new-functions to create them
	dynamically.
*/

typedef struct {
  unsigned char *data;
  size_t        bits;
} dk_bitfield_t ;


typedef struct {
#if DK_HAVE_SYS_STAT_H
#if DK_HAVE_STAT
  struct stat st;
#else
#if DK_HAVE__STAT
  struct _stat st;
#endif
#endif
#if DK_HAVE_LSTAT
  struct stat lst;
#endif
#endif
  int permissions;
  int filetype;
  unsigned long inode_number;
  unsigned long device_number;
  unsigned long rdevice_number;
  unsigned long number_of_links;
  /* unsigned long size; */
  dk_long_long_unsigned_t   size;
  int      size_math_error;
  long     uid;
  long     gid;
  char ctime[24];
  char atime[24];
  char mtime[24];
  short	is_far_link;	/* lst->st_device != st->st_device */
} dk_stat_t;

typedef struct {
  char *dirname;
  char *shortname;
  char *fullname;
  dk_stat_t stbuf;
  int  state;
  long maxpathlen;
#if DK_HAVE_DIRENT_H
  DIR *dir;
#else
#if DK_HAVE_FINDFIRST32
  long hFile;
  struct _finddata_t fdt;
#else
#if DK_HAVE_FINDFIRST
  struct ffblk dir;
#else
#endif
#endif
#endif
  int error_code;
} dk_dir_t;

typedef dk_dir_t dk_fne_t ;

typedef struct {
	void	*obj;
	dk_fct_type_log *fct;
	int	level;
	int	maxlevel;
} dk_log_t;

typedef struct {
	void	*obj;
	dk_fct_type_ss	*fct;
} dk_ss_t;

typedef struct {
  int cmd;
  struct {
    size_t length;
    char   *buffer;
    int cmd;
  } params;
  struct {
    size_t length;
    size_t used;
    char   *buffer;
  } results;
  int return_value;
  int error_code;
  void *strm;
} dk_stream_api_t;

typedef void dk_stream_fct_t DK_PR((dk_stream_api_t *api));

typedef struct {
  dk_stream_api_t api;
  void            *data;
  dk_stream_fct_t *fct;
  unsigned long bytes;
  int flags;
} dk_stream_t;

/* encoder on top of stream (closing the encoder does not close the stream) */
typedef struct {
  dk_stream_t *target;
  char *rlbuffer;
  char *flate_inbuffer;
  char *flate_outbuffer;
  int  flags;
  size_t rlb_size;
  size_t fib_size;
  size_t fob_size;
} dk_output_encoder_t;

typedef dk_stream_t *dk_stream_open_fct_t DK_PR((char *name, char *mode, int ign, int *reason));

typedef struct {
  char *suffix;
  dk_stream_open_fct_t *fct;
} dk_stream_suffix_t;

struct _dk_storage_node {
  struct _dk_storage_node *p;	/* parent node in tree */
  struct _dk_storage_node *l;	/* left sun or preceeding node */
  struct _dk_storage_node *r;	/* right sun or following node */
  short                    b;	/* balance flags in trees */
  short                    w;	/* direction to walk */
  void                    *o;	/* object to store */
  union {			/* object evaluation */
    char c;
    unsigned char uc;
    short s;
    unsigned short us;
    int i;
    unsigned int ui;
    long l;
    unsigned long ul;
    float f;
    double d;
  } v;
};

typedef struct _dk_storage_node dk_storage_node_t;

typedef dk_storage_node_t *dk_storage_node_p;

typedef char dk_fct_eval_c_t(void *obj, int crit);
typedef unsigned char dk_fct_eval_uc_t(void *obj, int crit);
typedef short dk_fct_eval_s_t(void *obj, int crit);
typedef unsigned short dk_fct_eval_us_t(void *obj, int crit);
typedef int dk_fct_eval_i_t(void *obj, int crit);
typedef unsigned int dk_fct_eval_ui_t(void *obj, int crit);
typedef long dk_fct_eval_l_t(void *obj, int crit);
typedef unsigned long dk_fct_eval_ul_t(void *obj, int crit);
typedef float dk_fct_eval_f_t(void *obj, int crit);
typedef double dk_fct_eval_d_t(void *obj, int crit);
typedef int dk_fct_comp_t(void *o1, void *o2, int crit);

typedef struct {
  dk_storage_node_p *d;
  short              l;
  dk_storage_node_t *r; 	/* the trees root element, lists left */
  int 		     h;		/* comparison algorithm */
  int                c;		/* comparison criteria */
  int		     t;		/* is tree or sorted list */
  union {
    dk_fct_eval_c_t *c;
    dk_fct_eval_uc_t *uc;
    dk_fct_eval_s_t *s;
    dk_fct_eval_us_t *us;
    dk_fct_eval_i_t *i;
    dk_fct_eval_ui_t *ui;
    dk_fct_eval_l_t *l;
    dk_fct_eval_ul_t*ul;
    dk_fct_eval_f_t *f;
    dk_fct_eval_d_t *d;
    dk_fct_comp_t *comp;
  } e;
  void              *i;		/* double linked list of iterators */
} dk_storage_t;

struct _dk_storage_iterator {
  struct _dk_storage_iterator *l;	/* preceeding iterator */
  struct _dk_storage_iterator *r;	/* following iterator */
  dk_storage_t                *s;	/* storage the iterator belongs to */
  dk_storage_node_t           *c;	/* current storage node */
};

typedef struct _dk_storage_iterator dk_storage_iterator_t;

typedef struct {
  dk_uword	  vers_major;	/* major version number */
  dk_uword	  vers_minor;	/* minor version number */
  dk_uword	  elems;	/* number of elements */
  char		**keys;		/* array of keys */
  char		**values;	/* array of values */
} dk_stt_t;

typedef struct {
  char *n;		/* name */
  char *v;		/* value */
  int   p;		/* priority user=4, app=2, host=1 or combined */
} dk_preference_t;

typedef struct {
  struct {           /* copy of the command line arguments */
    struct {            /* copy of original cmd line args */
      int    argc;
      char **argv;
    } o;
    struct {            /* preferences removed from the list */
      int    argc;
      char **argv;
    } a;
  } a;
  struct {           /* names often needed */
    char *u;            /* user login name */
    char *a;            /* application name */
    char *h;            /* host name */
    char *g;		/* program group or package */
  } n;
  struct {           /* directories often needed */
    char *h;            /* users home directory */
    char *t;            /* temporary directory */
    char *a;            /* application directory */
    char *s;            /* shared directory */
    char *pt;		/* processes temporary directory */
    char *etc;		/* etc directory */
  } d;
  struct {           /* preferences management */
    dk_storage_t *c;    /* command line preferences */
    dk_storage_iterator_t *ci;
    dk_storage_t *a;    /* application set prefs */
    dk_storage_iterator_t *ai;
#if DK_HAVE_WINREG_H
    int what;
    HKEY hklm_all;
    HKEY hklm_app;
    HKEY hkcu_all;
    HKEY hkcu_app;
#else
    dk_storage_t *s;	/* system set preferences */
    dk_storage_iterator_t *si;
    dk_storage_t *u;	/* user set preferences */
    dk_storage_iterator_t *ui;
#endif
    int unc;		/* unconfigure application */
  } p;
  struct {
    char *f;		/* executable file name */
    char *d;		/* executable directory */
  } x;
  struct {		/* loggin */
    int max;
    int nostdio;	/* no logging to stdout/stderr */
    struct {		/*   stdout */
      int m;		/*     minimum priority */
      int f;		/*     flags: 1=date+time, 2=split */
      unsigned char *c;	/*     codepage */
      int ide_type;	/* IDE type */
    } o;
    struct {		/*   stderr */
      int m;
      int f;
      unsigned char *c;
      int ide_type;	/* IDE type */
    } e;
    struct {		/*   file */
      int m;
      int k;		/*     level required to keep log file */
      int f;
      FILE *t;		/*     target */
      char *n;		/*     target file name */
      unsigned char *c;
      int ide_type;	/* IDE type */
    } f;
#if DK_HAVE_SYSLOG
    struct {		/*     syslog */
      int m;
      char *i;		/*     ident arg to openlog */
      int f;		/*     syslog facility */
      int o;		/*     1=syslog opened, 0=not opened */
    } s;
#endif
    struct {		/* error file and line */
      char *n;
      unsigned long lineno;
    } ef;
  } l;
  struct {
    char *l;		/* language */
    char *r;		/* region */
    char *e;		/* encoding */
    dk_storage_t *s;	/* string tables */
    dk_storage_iterator_t *si;
  } loc;
  struct {	      /* data for temporary filenames */
    unsigned long l;
  } td;
  int relaxed_fopen_check;
  int relaxed_fopen_reason;
} dk_app_t;


#if DK_HAVE_PROTOTYPES
typedef int dk_fct_tokenizer(void *data, void *tok, char *s, int *err);
#else
typedef int dk_fct_tokenizer();
#endif

typedef struct {
  char *token_buffer;
  size_t buffer_size;
  size_t buffer_used;
  char **quotes;		/* pairs of opening and closing quotes */
  char  *sct;			/* single character tokens */
  char  *whsp;			/* whitespaces */
  char  *nl;			/* newlines */
  char   comment;		/* comment introducer */
  char  *endquote;		/* this ends the current quote */
  int    error_code;		/* we can continue while this is untouched */
  dk_fct_tokenizer *fct;	/* function to process a token */
  void  *user_data;		/* additional data for function */
  int    state;			/* current state of the tokenizer */
  unsigned long lineno;		/* line number in input file */
} dk_tokenizer_t ;

typedef struct {
  int what;
  union {
    struct {		/* for buffered output */
      char *buffer;
      size_t lgt;
      size_t used;
    } b;
    struct {		/* ASCII85 encoding */
      unsigned long a85val;
      size_t used;
      size_t linepos;
    } a85;
#if DK_HAVE_ZLIB_H
    struct {		/* flate compression */
      z_stream *zs;
      Bytef *ibuffer;
      uLong i_length;
      Bytef *obuffer;
      uLong o_length;
      uLong used;
      int   stillok;
    } flate;
#endif
  } c;
} dk_of_cell_t;

typedef struct {
  int flags;
  dk_stream_t	*target;
  dk_of_cell_t	*cells;
  int		 n_of_cells;
  int		 have_data;
} dk_of_t;

#if DK_HAVE_SYS_SOCKET_H || DK_HAVE_WINSOCK2_H

typedef struct {
  int what;
  struct {
    struct sockaddr_in sin;
    unsigned short portmin;
    unsigned short portmax;
  } ip4;
} dk_ip_addr_t;

typedef struct {
  int state;
  int flags;
#if DK_HAVE_WINSOCK2_H
  SOCKET s;
#else
  int s;
#endif
  struct {
    struct {		/* local addresses */
      dk_ip_addr_t w;	/*    wished */
      dk_ip_addr_t f;	/*    found  */
    } l;
    struct {		/* remote addresses */
      dk_ip_addr_t w;	/*    wished */
      dk_ip_addr_t f;	/*    found  */
    } r;
  } a;
  struct {
    time_t seconds;
    long   usecs;
  } to;
  int error_code;
} dk_tcpip_t;

#endif

#if DK_HAVE_LONG_LONG_INT
#define DK_MAX_LONG_LONG_UNSIGNED 0xFFFFFFFFFFFFFFFFULL
#define DK_MAX_LONG_LONG_INT	  0x7FFFFFFFFFFFFFFFLL
#define DK_ZERO_LONG_LONG_UNSIGNED 0ULL
#define DK_ZERO_LONG_LONG_INT      0LL
#define DK_ONE_LONG_LONG_UNSIGNED  1ULL
#define DK_ONE_LONG_LONG_INT       1LL
#else
#define DK_MAX_LONG_LONG_UNSIGNED 0xFFFFFFFFUL
#define DK_MAX_LONG_LONG_INT	  0x7FFFFFFFL
#define DK_ZERO_LONG_LONG_UNSIGNED 0UL
#define DK_ZERO_LONG_LONG_INT      0L
#define DK_ONE_LONG_LONG_UNSIGNED  1UL
#define DK_ONE_LONG_LONG_INT       1L
#endif

#if SIZEOF_LONG == 8
#define DK_MAX_ULONG	(0xFFFFFFFFFFFFFFFFUL)
#define DK_MAX_LONG	(0x7FFFFFFFFFFFFFFFL)
#else
#define DK_MAX_ULONG	(0xFFFFFFFFUL)
#define DK_MAX_LONG	(0x7FFFFFFFL)
#endif

#if SIZEOF_INT == 4
#define DK_MAX_UNSIGNED	(0xFFFFFFFFU)
#define DK_MAX_INT	(0x7FFFFFFF)
#else
#define DK_MAX_UNSIGNED	(0xFFFFU)
#define DK_MAX_INT	(0x7FFF)
#endif

#if SIZEOF_SHORT == 4
#define DK_MAX_USHORT	(0xFFFFFFFFU)
#define DK_MAX_SHORT	(0x7FFFFFFF)
#else
#define DK_MAX_USHORT	(0xFFFFU)
#define DK_MAX_SHORT	(0x7FFF)
#endif

#if SIZEOF_SIZE_T == 2
#define DK_MAX_SIZE_T	0xFFFFU
#else
#if SIZEOF_SIZE_T == 4
#define DK_MAX_SIZE_T	0xFFFFFFFFUL
#else
#if SIZEOF_SIZE_T == 8
#define DK_MAX_SIZE_T	0xFFFFFFFFFFFFFFFFULL
#endif
#endif
#endif

#endif

