/*
 * traceco - tracecc output module
 * Copyright (C) 2001 - Dirk Krause
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 * In this package the copy of the GNU General Public License
 * is placed in file COPYING.
 */

#define TRACECO_C 1
#include <tracecc.h>

#define TRACE_CONTROL 512



/* typedefs {{{1 */


/* pointer to char {{{2 */

/* ********************************************************************* */
/* *                                                                   * */
/* *   This type is needed as argument to the sizeof                   * */
/* *   operator in memory allocations.                                 * */
/* *                                                                   * */
/* ********************************************************************* */

typedef char *p_char_t;
/* }}}2 */


/* message_contents_t {{{2 */
typedef struct {
  char *key;
  char *def;
} message_contents_t;
/* }}}2 */


/* suffix_t {{{2 */

/* ********************************************************************* */
/* *                                                                   * */
/* *   The relationship between the file name suffixes                 * */
/* *   in source and destination file names.                           * */
/* *                                                                   * */
/* ********************************************************************* */

typedef struct {
  char *old;
  char *new;
} suffix_t;
/* }}}2 */


/* file_t {{{2 */

/* ********************************************************************* */
/* *                                                                   * */
/* *   The relationship between destination file name                  * */
/* *   suffix and the file type.                                       * */
/* *   2=Java, 1=C                                                     * */
/* *                                                                   * */
/* ********************************************************************* */

typedef struct {
  char *suffix;
  int   ft;
} file_t;
/* }}}2 */


/* IFCT {{{2 */

/* ********************************************************************* */
/* *                                                                   * */
/* *   The function type invoked when interpreting                     * */
/* *   $(...)                                                          * */
/* *   or                                                              * */
/* *   $!...                                                           * */
/* *                                                                   * */
/* ********************************************************************* */

#if DK_HAVE_PROTOTYPES
typedef int IFCT(tracecc_t *tc, char *args);
#else
typedef int IFCT();
#endif
typedef IFCT *i_fct_t;
/* }}}2 */


/* function_info_t {{{2 */

/* ********************************************************************* */
/* *                                                                   * */
/* *   Relationshipo between interpreter function name                 * */
/* *   and function pointer. The need_lineno part indicates            * */
/* *   whether or not we need to print a line number statement         * */
/* *   after executing the function.                                   * */
/* *                                                                   * */
/* ********************************************************************* */

typedef struct {
  char  *name;
  IFCT	*fct;
  int    need_lineno;
} function_info_t ;
/* }}}2 */


/* }}}1 */

/* local variables {{{1 */

static char str_tracecc[] = { "tracecc" };

/* message contents {{{2 */

/* ********************************************************************* */
/* *                                                                   * */
/* *   The messages printed by this application.                       * */
/* *   Here we have the keys and the default values.                   * */
/* *                                                                   * */
/* ********************************************************************* */

static message_contents_t mc[] = {
  { "/m/00",
  "Small buffer in use."},

  { "/m/01",
  "Comman not executed!"},

  { "/m/02",
  "No popen() function!"},

  { "/m/03",
  "Arguments needed!"},

  { "/m/04",
  "Missing arguments!"},

  { "/m/05",
  "No such function!"},

  { "/m/06",
  "Empty function name!"},

  { "/m/07",
  "Failed to write file!"},

  { "/m/08",
  "Failed to read file!"},

  { "/m/09",
  "Multiple files for pattern!"},

  { "/m/10",
  "No file for pattern!"},

  { "/m/11",
  "Multiple files for pattern!"},

  { "/m/12",
  "No file for pattern!"},

  { "/m/13",
  "Destination file name too long!"},

  { "/m/14",
  "Failed to open directory!"},

  { "/m/15",
  "No directory for pattern!"},

  { "/m/16",
  "Processing file \""},
  { "/m/17",
  "\"."},
  { "/m/18",
  "Finished processing file \""},
  { "/m/19",
  "\"."},
  { "/m/20",
  "Processing directory \""},
  { "/m/21",
  "\"."},
  { "/m/22",
  "Finished processing directory \""},
  { "/m/23",
  "\"."},
  { "/m/24",
  "Skipping source file \"" },
  { "/m/25",
  "\". Destination file is newer." },
  { "/m/26",
  "Input line too long!" },

  { NULL, NULL }
};
/* }}}2 */

/* number of messages {{{2 */
static size_t number_of_messages =
sizeof(mc)/sizeof(message_contents_t);
/* }}}2 */

/* IDE styles {{{2 */

/* ********************************************************************* */
/* *                                                                   * */
/* *   The destination program will behave like one of the             * */
/* *   following programs (in terms of formatting the                  * */
/* *   trace messages).                                                * */
/* *                                                                   * */
/* ********************************************************************* */

static char *ide_styles[] = {
  (char *)"none",
  (char *)"gcc",
  (char *)"msvc",
  (char *)"workshop",
  (char *)"tasm",
  NULL
};
/* }}}2 */

/* preferences keywords {{{2 */

/* ********************************************************************* */
/* *                                                                   * */
/* *   These keywords are used to retrieve and save                    * */
/* *   preferences.                                                    * */
/* *                                                                   * */
/* ********************************************************************* */

static char pk_debug_enable[] = { "/debug/enable" };
static char pk_debug_stdout[] = { "/debug/stdout" };
static char pk_debug_timestamp[] = { "/debug/timestamp" };
static char pk_debug_ide[] = { "/debug/ide" };
static char pk_debug_keyword[] = { "/debug/keyword" };
static char pk_linenumbers[] = { "/linenumbers" };
static char pk_make[] = { "/make" };
static char pk_cpp_comments[] = { "/cpp-comments" };
static char pk_boxwidth[] = { "/box-width" };
static char pk_ide_style[] = { "/debug/ide-style" };
static char pk_input_length[] = { "/input/length" };
static char pk_command_length[] = { "/command/length" };
/* }}}2 */

/* keywords to turn bools on or off {{{2 */

/* ********************************************************************* */
/* *                                                                   * */
/* *   on or off are used to save bool preferences.                    * */
/* *                                                                   * */
/* ********************************************************************* */

static char str_on[] = { "on" };
static char str_off[] = { "off" };
static char str_spc[] = { " " };
static char str_bropen[] = { "(" };
static char str_brclose[] = { ")" };
/* }}}2 */

/* function names {{{2 */

/* ********************************************************************* */
/* *                                                                   * */
/* *   The names of the functions used by C programs.                  * */
/* *                                                                   * */
/* ********************************************************************* */

static char str_c_fct_get_file[] = { "dktrace_file" };
static char str_c_fct_init[] = { "dktrace_init" };
static char str_c_fct_end[] = { "dktrace_end" };
static char str_c_fct_time[] = { "dktrace_time" };
static char str_c_fct_stdout_time[] = { "dktrace_stdout_time" };
/* }}}2 */

/* list of suffixes {{{2 */

/* ********************************************************************* */
/* *                                                                   * */
/* *   The list of source/destination file name suffix                 * */
/* *   pairs.                                                          * */
/* *                                                                   * */
/* ********************************************************************* */

static suffix_t suffixlist[] = {
  { ".ctr", ".c"    },
  { ".cpt", ".cpp"  },
  { ".mtr", ".m"    },
  { ".jtr", ".java" },
  { NULL, NULL }
};
/* }}}2 */

/* list of file types {{{2 */

/* ********************************************************************* */
/* *                                                                   * */
/* *   The list of file name suffix / file type pairs.                 * */
/* *                                                                   * */
/* ********************************************************************* */

static file_t ftlist[] = {
  { ".c",    1 },
  { ".cpp",  1 },
  { ".m",    1 },
  { ".java", 2 },
  { NULL, 0 }
};
/* }}}2 */

/* file name separator {{{2 */

/* ********************************************************************* */
/* *                                                                   * */
/* *   The separator for file name components.                         * */
/* *                                                                   * */
/* ********************************************************************* */

static char fnsep[] = {
#if DK_HAVE_FEATURE_BACKSLASH
"\\"
#else
"/"
#endif
};
/* }}}2 */

/* hex digits {{{ */
static char hexdigits[] = { "0123456789ABCDEF0000" };
/* }}} */

/* }}}1 */

/* Local (static) functions {{{1 */


/* change a mode bit {{{2 */

/* ********************************************************************* */
/* *                                                                   * */
/* *   Change a bit in bitset (mode) for a preference (key).           * */
/* *                                                                   * */
/* ********************************************************************* */

static
void
set_mode_bit_bool DK_P4(int *,mode,dk_app_t *,app,char *,key,int,bit)
{
  int val;
  char inputline[256], *ptr;
  val = *mode;
  if(dkapp_get_pref(app,key,inputline,sizeof(inputline),0)) {
    ptr = dkstr_start(inputline, NULL);
    if(ptr) {
      dkstr_chomp(ptr, NULL);
      if(dkstr_is_on(ptr)) {
        val |= bit;
      } else {
        val &= (~bit);
      }
    }
  }
  *mode = val;
}
/* }}}2 */

/* set a bool preference {{{2 */

/* ********************************************************************* */
/* *                                                                   * */
/* *   Save a boolean preference.                                      * */
/* *                                                                   * */
/* ********************************************************************* */

static
void
set_bool DK_P3(dk_app_t *,app,char *,key,int,v)
{
  dkapp_set_pref(app, key, (v ? str_on : str_off));
}
/* }}}2 */


/* set an integer preference {{{2 */

/* ********************************************************************* */
/* *                                                                   * */
/* *   Set an integer value from a preference.                         * */
/* *                                                                   * */
/* ********************************************************************* */

static
void
set_int_from_pref DK_P3(int *,iptr,dk_app_t *,app,char *,key)
{
  char inputline[32];
  int i;
  if(dkapp_get_pref(app,key,inputline,sizeof(inputline),0)) {
    if(sscanf(inputline, "%d", &i) == 1) {
      *iptr = i;
    }
  }
}
/* }}}2 */


/* reset message strings {{{2 */

/* ********************************************************************* */
/* *                                                                   * */
/* *   Mark all strings in the message array as                        * */
/* *   non-existant.                                                   * */
/* *                                                                   * */
/* ********************************************************************* */

static
void
reset_strings_for_messages DK_P1(char **,msgarray)
{
  char **lfdptr; size_t i;
  
  lfdptr = msgarray;
  for(i = 0; i < number_of_messages; i++) {
    *(lfdptr++) = NULL;
  }
  
}
/* }}}2 */


/* set message strings {{{2 */

/* ********************************************************************* */
/* *                                                                   * */
/* *   Set message array to the default values as specified            * */
/* *   in mc->def.                                                     * */
/* *                                                                   * */
/* ********************************************************************* */

static
void
set_strings_for_messages DK_P1(char **,msgarray)
{
  char **lfdptr;
  size_t i;
  message_contents_t *mcptr;
  
  mcptr = mc; lfdptr = msgarray;
  for(i = 0; i < number_of_messages; i++) {
    *(lfdptr++) = (mcptr++)->def;
  }
  
}
/* }}}2 */


/* set strings using the app {{{2*/

/* ********************************************************************* */
/* *                                                                   * */
/* *   Try to set the message array using the given                    * */
/* *   application.                                                    * */
/* *                                                                   * */
/* ********************************************************************* */

static
void
set_strings_by_app DK_P2(char **,msgarray,dk_app_t *,app)
{
  char **lfdptr;
  dk_string_finder_t *fptr, *fptr2;
  message_contents_t *mcptr;
  size_t i;
  
  fptr = dk_new(dk_string_finder_t,number_of_messages);
  if(fptr) {
    
    lfdptr = msgarray; fptr2 = fptr; mcptr = mc;
    for(i = 0; i < number_of_messages; i++) {
      fptr2->key = mcptr->key;
      fptr2->value_pointer = lfdptr;
      fptr2->default_value = mcptr->def;
      lfdptr++; fptr2++; mcptr++;
    }
    dkapp_find_multi(app, fptr, "tracecc");
    dk_delete(fptr);
  } else {
    
    set_strings_for_messages(msgarray);
    dkapp_err_memory(app,sizeof(dk_string_finder_t),number_of_messages);
  }
  
}
/* }}}2 */


/* fputul {{{2 */

/* ********************************************************************* */
/* *                                                                   * */
/* *   Put unsigned long to file (as text).                            * */
/* *                                                                   * */
/* ********************************************************************* */

static
void
fputul DK_P2(unsigned long,ul,FILE *,f)
{
  char buffer[32];
  sprintf(buffer, "%lu", ul);
  fputs(buffer,f);
}
/* }}}2 */


/* is_trace_control {{{2 */

/* ********************************************************************* */
/* *                                                                   * */
/* *   Check whether or not a character is the trace                   * */
/* *   control character for a given file type.                        * */
/* *                                                                   * */
/* ********************************************************************* */

static
int
is_trace_control DK_P2(int,ft,char,c)
{
  int back = 0;
  
  switch(ft) {
    /* Java */
    case 2:  { if(c == '$') back = 1; } break;
    /* C, C++, Objective-C */
    case 1:  { if(c == '$') back = 1; } break;
    /* all others */
    default: { if(c == '$') back = 1; } break;
  } 
  return back;
}
/* }}}2 */


/* automata_reaction {{{2 */

/* ********************************************************************* */
/* *                                                                   * */
/* *   The flow control unit.                                          * */
/* *                                                                   * */
/* ********************************************************************* */

static
int
automata_reaction DK_P2(int *, ptr, int, input)
{
  int back = 0;
  int newstate = 0;
  if(ptr) {
    
    newstate = back = 0;
    switch(*ptr) {
      case 0: { /* initial state */
        switch(input) {
          case '\\': { newstate = back = 1; } break;
	  case '"': { newstate = 2; back = 1; } break;
	  case '\'': { newstate = 4; back = 1; } break;
	  case '/': { newstate = 6; back = 0; } break;
	  case TRACE_CONTROL: { newstate = 10; back = 0; } break;
	  default: { newstate = 0; back = 1; } break;
	}
      } break;
      case 1: { /* after backslash */
        newstate = 0; back = 1;
      } break;
      case 2: { /* in a string */
	back = 1;
        switch(input) {
          case '"': { newstate = 0; } break;
	  case '\\': { newstate = 3; } break;
	  default: { newstate = 2; } break;
	}
      } break;
      case 3: { /* after backslash in string */
        newstate = 2; back = 1;
      } break;
      case 4: { /* single character constant */
	back = 1;
        switch(input) {
          case '\'': { newstate = 0; } break;
	  case '\\': { newstate = 5; } break;
	  default: { newstate = 4; } break;
	}
      } break;
      case 5: { /* after backslash in single character constant */
        newstate = 4; back = 1;
      } break;
      case 6: { /* after slash */
        switch(input) {
          case '/': { newstate = 7; back = 16; } break;
	  case '*': { newstate = 8; back = 18; } break;
	  default: { newstate = 0; back = 18; } break;
	}
      } break;
      case 7: { /* in C++ style comment */
	newstate = 7; back = 1;
        if(input == '\n') {
          newstate = 0; back = 17;
	}
      } break;
      case 8: { /* in normal comment */
        newstate = 8; back = 1;
	if(input == '*') newstate = 9;
      } break;
      case 9: { /* after asterisk in normal C comment */
	back = 1;
        switch(input) {
          case '*': { newstate = 9; } break;
	  case '/': { newstate = 0; } break;
	  default: { newstate = 8; } break;
	}
      } break;
      case 10: { /* after TRACE_CONTROL */
        switch(input) {
          case '(': { newstate = 15; back = 0; } break;
	  case '!': { newstate = 12; back = 0; } break;
	  case '?': { newstate = 11; back = 3; } break;
	  case '*': { newstate = 17; back = 11; } break;
	  default: { newstate = 0; back = 2; } break;
	}
      } break;
      case 11: { /* after TRACE_CONTROL, question mark */
        switch(input) {
          case '\n': { newstate = 0; back = 10; } break;
	  case ' ' : { newstate = 11; back = 0; } break;
	  case '\t': { newstate = 11; back = 0; } break;
	  default: { newstate = 13; back = 4; } break;
	}
      } break;
      case 12: { /* after TRACE_CONTROL, exclam */
        switch(input) {
          case '\n': { newstate = 0; back = 1; } break;
	  case ' ':
	  case '\t': { newstate = 12; back = 0; } break;
	  default: { newstate = 14; back = 6; } break;
	}
      } break;
      case 13: { /* in debug output */
	newstate = 13; back = 9;
        if(input == '\n') {
          newstate = 0; back = 5;
	}
      } break;
      case 14: { /* in interprete context */
	newstate = 14; back = 7;
        if(input == '\n') {
          newstate = 0; back = 19;
	}
      } break;
      case 15: { /* in brackets in interprete context*/
        switch(input) {
          case '\n': { newstate = 0; back = 1; } break;
	  case ')': { newstate = 1; back = 0; } break;
	  case ' ': 
	  case '\t': { newstate = 15; back = 0; } break;
	  default: { newstate = 16; back = 6; } break;
	}
      } break;
      case 16: {
	newstate = 16; back = 7;
        if(input == ')') {
          newstate = 0; back = 8;
	}
      } break;
      case 17: { /* in comment box */
        switch(input) {
          case '\n': { newstate = 17; back = 12; } break;
	  case TRACE_CONTROL: { newstate = 18; back = 0; } break;
	  default: { newstate = 17; back = 13; } break;
	}
      } break;
      case 18: { /* in comment box, found TRACE_CONTROL */
	newstate = 17; back = 14;
        if(input == '*') {
          newstate = 0; back = 15;
	}
      } break;
    }
    *ptr = newstate;
    
  }
  return back;
}
/* }}}2 */


/* get_new_file_type {{{2 */

/* ********************************************************************* */
/* *                                                                   * */
/* *   Find the file type by checking the destination                  * */
/* *   file name suffix.                                               * */
/* *                                                                   * */
/* ********************************************************************* */

static
int
get_new_file_type DK_P1(char *,suffix)
{
  int back = 0; file_t *ptr;
  
  ptr = ftlist;
  while((ptr->suffix) && (!back)) {
    if(strcmp(ptr->suffix, suffix) == 0) {
      back = ptr->ft;
    } else {
      ptr++;
    }
  }
  
  return back;
}
/* }}}2 */


/* get_new_suffix {{{2 */

/* ********************************************************************* */
/* *                                                                   * */
/* *   Try to find a destination file name suffix for a                * */
/* *   given source file name suffix.                                  * */
/* *   If we fail to find one the source file is skipped.              * */
/* *                                                                   * */
/* ********************************************************************* */

static
char *
get_new_suffix DK_P1(char *,suffix)
{
  char *back = NULL;
  suffix_t *sptr;
  
  sptr = suffixlist;
  while((sptr->old) && (!back)) {
    if(strcmp(sptr->old, suffix) == 0) {
      back = sptr->new;
    }
    sptr++;
  }
  
  return back;
}
/* }}}2 */


/* indent_this_line {{{2 */

/* ********************************************************************* */
/* *                                                                   * */
/* *   Print spaces to go to the current level                         * */
/* *   of indent.                                                      * */
/* *                                                                   * */
/* ********************************************************************* */

static
void
indent_this_line DK_P1(tracecc_t *,tc)
{
  int i;
  
  for(i = 0; i < tc->indent; i++) {
    fputc(' ', tc->outfile);
  }
  
}
/* }}}2 */


/* is_show_timestamp {{{2 */

/* ********************************************************************* */
/* *                                                                   * */
/* *   Check whether or not we want time stamps                        * */
/* *   in the trace output.                                            * */
/* *                                                                   * */
/* ********************************************************************* */

static
int
is_show_timestamp DK_P1(tracecc_t *,tc)
{
  int back;
  
  back = (((tc->f).o) & TRACECC_OPT_DATETIME) ? 1 : 0;
  
  return back;
}
/* }}}2 */


/* print_current_char {{{2 */

/* ********************************************************************* */
/* *                                                                   * */
/* *   Print current character to destination file.                    * */
/* *                                                                   * */
/* ********************************************************************* */

static
void
print_current_char DK_P1(tracecc_t *,tc)
{
  
  fputc(tc->currentchar, tc->outfile);
  if(tc->currentchar == '\n') fflush(tc->outfile);
}
/* }}}2 */


/* tc_and_current_char {{{2 */

/* ********************************************************************* */
/* *                                                                   * */
/* *   Print tracecc control character and current character           * */
/* *   to destination file.                                            * */
/* *                                                                   * */
/* ********************************************************************* */

static
void
tc_and_current_char DK_P1(tracecc_t *,tc)
{
  char c;
  
  switch(tc->newft) {
    case 2:
    case 1:
    default: {
      c = '$';
    } break;
  }
  fputc(c, tc->outfile);
  print_current_char(tc);
  
}
/* }}}2 */


/* have_cpp_comment {{{2 */

/* ********************************************************************* */
/* *                                                                   * */
/* *   Check whether or not we can use C++ comments                    * */
/* *                                                                   * */
/* ********************************************************************* */

static
int
have_cpp_comment DK_P1(tracecc_t *,tc)
{
  int back;
  back = (((tc->o).o) & ((tc->f).o) & TRACECC_OPT_CPPCOMMENT) ? 1 : 0;
  
  return back;
}
/* }}}2 */


/* is_debug {{{2 */

/* ********************************************************************* */
/* *                                                                   * */
/* *   Check whether or not we run in debug mode.                      * */
/* *                                                                   * */
/* ********************************************************************* */

static
int
is_debug DK_P1(tracecc_t *,tc)
{
  int back;
  back = (((tc->o).o) & ((tc->f).o) & TRACECC_OPT_DEBUG) ? 1 : 0;
  
  return back;
}
/* }}}2 */


/* print_debug_char {{{2 */

/* ********************************************************************* */
/* *                                                                   * */
/* *   Print a character only if we are in debugging mode.             * */
/* *                                                                   * */
/* ********************************************************************* */

static
void
print_debug_char DK_P1(tracecc_t *,tc)
{
  
  if(is_debug(tc)) {
    print_current_char(tc);
  }
  
}
/* }}}2 */


/* is_debst {{{2 */

/* ********************************************************************* */
/* *                                                                   * */
/* *   Check whether or not debug messages go to stdout.               * */
/* *                                                                   * */
/* ********************************************************************* */

static
int
is_debst DK_P1(tracecc_t *,tc)
{
  int back = 0;
  if(is_debug(tc)) {
    back = (((tc->o).o) & ((tc->f).o) & TRACECC_OPT_DEBST) ? 1 : 0;
  }
  
  return back;
}
/* }}}2 */


/* nl {{{2 */

/* ********************************************************************* */
/* *                                                                   * */
/* *   Print newline to destination file.                              * */
/* *                                                                   * */
/* ********************************************************************* */

static
void
nl DK_P1(tracecc_t *,tc)
{
  fputc('\n', tc->outfile);
  
}
/* }}}2 */


/* need_source_position {{{2 */
static
int
need_source_position DK_P1(tracecc_t *,tc)
{
  int back = 0;
  switch(tc->newft) {
    case 1:
    default: {
      if(((tc->o).o) & ((tc->f).o) & TRACECC_OPT_LINENO) back = 1;
    } break;
  }
  return back;
}
/* }}}2 */


/* print_source_position {{{2 */
static
void
print_source_position DK_P2(tracecc_t *,tc,int,cor)
{
  FILE *f;
  unsigned long ul;

  f = tc->outfile; ul = tc->lineno;
  switch(cor) {
    case -1: ul--; break;
    case  1: ul++; break;
  }
  fputs("#line ", f),
  fputul(ul, f);
  fputc(' ', f);
  fputc('"', f);
  fputs(tc->in_short, f);
  fputc('"', f);
  nl(tc);
}
/* }}}2 */


/* show_filename_and_lineno {{{2 */

/* ********************************************************************* */
/* *                                                                   * */
/* *   Print file name and line number depending on the IDE            * */
/* *   style we have.                                                  * */
/* *                                                                   * */
/* ********************************************************************* */

static
void
show_filename_and_lineno DK_P1(tracecc_t *,tc)
{
  FILE *f;
  
  f = tc->outfile;
  fputs("\"", f);
  switch((tc->f).i) {
    case DK_APP_IDE_TASM: {	
      fputs("*Warning* ", f);
      fputs(tc->in_short, f);
      fputc('(', f);
      fputul(tc->lineno, f);
      fputc(')', f);
    } break;
    case DK_APP_IDE_WORKSHOP: {	
      fputc('"', f);
      fputs(tc->in_short, f);
      fputc('"', f);
      fputs(", line ", f);
      fputul(tc->lineno, f);
      fputc(':', f);
    } break;
    case DK_APP_IDE_MSVC: {	
      fputs(tc->in_short, f);
      fputc('(', f);
      fputul(tc->lineno, f);
      fputs(") :", f);
    } break;
    case DK_APP_IDE_GCC:
    default: {		
      fputs(tc->in_short, f);
      fputc(':', f);
      fputul(tc->lineno, f);
      fputc(':', f);
    } break;
  }
  if(((tc->f).o) & TRACECC_OPT_KEYWORD) {
    fputs(" trace", f);
  }
  fputs(" \"", f);
  
}
/* }}}2 */


/* start_c_line {{{2 */

/* ********************************************************************* */
/* *                                                                   * */
/* *   Begin a trace output line for C files.                          * */
/* *                                                                   * */
/* ********************************************************************* */

static
void
start_c_line DK_P1(tracecc_t *,tc)
{
  FILE *f;
  
  f = tc->outfile;
  if(is_debst(tc)) {
    fputc( '{', f);
    nl(tc);
    indent_this_line(tc);
    fputs( "printf(", f);
  } else {
    fputs("{ if(", f);
    fputs(str_c_fct_get_file, f);
    fputs("()) {", f);
    nl(tc);
    indent_this_line(tc);
    fputs("fprintf(", f);
    fputs(str_c_fct_get_file, f);
    fputs("(), ", f);
  }
  
}
/* }}}2 */


/* end_c_line {{{2 */

/* ********************************************************************* */
/* *                                                                   * */
/* *   End trace output line in C files.                               * */
/* *                                                                   * */
/* ********************************************************************* */

static
void
end_c_line DK_P1(tracecc_t *,tc)
{
  FILE *f;
  
  f = tc->outfile;
  fputs(");", f);
  if(is_debst(tc)) {
    nl(tc);
    indent_this_line(tc);
  } else {
    nl(tc);
    indent_this_line(tc);
    fputs("} ", f);
  }
  fputs("}\n", f);
  
}
/* }}}2 */


/* begin_java_trace {{{2 */

/* ********************************************************************* */
/* *                                                                   * */
/* *   Begin trace output block in Java files.                         * */
/* *                                                                   * */
/* ********************************************************************* */

static
void
begin_java_trace DK_P1(tracecc_t *,tc)
{
  FILE *f;
  f = tc->outfile;
  if(is_debug(tc)) {
    if(is_show_timestamp(tc)) {
    }
    fputs("KrTrace.printMessage(", f);
    show_filename_and_lineno(tc);
    fputs(" + ", f);
  }
}
/* }}}2 */


/* begin_c_trace {{{2 */

/* ********************************************************************* */
/* *                                                                   * */
/* *   Begin trace output block in C files.                            * */
/* *                                                                   * */
/* ********************************************************************* */

static
void
begin_c_trace DK_P1(tracecc_t *,tc)
{
  FILE *f;
  
  f = tc->outfile;
  fputs("/* TRACE BEGIN */\n", f);
  indent_this_line(tc);
  if(is_show_timestamp(tc)) {
    if(is_debst(tc)) {
      fputs(str_c_fct_stdout_time, f);
    } else {
      fputs(str_c_fct_time, f);
    }
    fputs("();\n", f);
    indent_this_line(tc);
  }
  /* start_c_line(tc); */
  /* show_filename_and_lineno(tc); */
  /* fputc(' ', f); */
  /* end_c_line(tc); */
  if(!is_debst(tc)) {
    fputs("{ if(", f);
    fputs(str_c_fct_get_file, f);
    fputs("()) ", f);
  }
  fputs("{ fputs(", f);
  show_filename_and_lineno(tc);
  fputs(", ", f);
  if(is_debst(tc)) {
    fputs("stdout", f);
  } else {
    fputs(str_c_fct_get_file, f);
    fputs("()", f);
  }
  fputs("); }", f);
  if(!is_debst(tc)) {
    fputs(" }", f);
  }
  nl(tc);
  
}
/* }}}2 */


/* begin_trace {{{2 */

/* ********************************************************************* */
/* *                                                                   * */
/* *   Begin a trace output block.                                     * */
/* *                                                                   * */
/* ********************************************************************* */

static
void
begin_trace_block DK_P1(tracecc_t *,tc)
{
  
  if(is_debug(tc)) {
    switch(tc->newft) {
      case 2: {
        begin_java_trace(tc);
      } break;
      default: {
        begin_c_trace(tc);
      } break;
    }
  }
  
}
/* }}}2 */


/* begin_c_trace_contents {{{2 */

/* ********************************************************************* */
/* *                                                                   * */
/* *   We begin to transfer the trace message contents                 * */
/* *   from the source to the destination.                             * */
/* *                                                                   * */
/* ********************************************************************* */

static
void
begin_c_trace_contents DK_P1(tracecc_t *,tc)
{
  
  indent_this_line(tc);
  start_c_line(tc);
  print_current_char(tc);
  
}
/* }}}2 */


/* begin_java_trace_contents {{{2 */
static
void
begin_java_trace_contents DK_P1(tracecc_t *,tc)
{
  print_current_char(tc);
}
/* }}}2 */


/* begin_trace_contents {{{2 */
static
void
begin_trace_contents DK_P1(tracecc_t *,tc)
{
  
  if(is_debug(tc)) {
    switch(tc->newft) {
      case 2: {
        begin_java_trace_contents(tc);
      } break;
      default: {
        begin_c_trace_contents(tc);
      } break;
    }
  }
  
}
/* }}}2 */


/* end_c_trace_contents {{{2 */

/* ********************************************************************* */
/* *                                                                   * */
/* *   Trace message finished.                                         * */
/* *                                                                   * */
/* ********************************************************************* */

static
void
end_c_trace_contents DK_P1(tracecc_t *,tc)
{
  FILE *f;
  
  f = tc->outfile;
  end_c_line(tc);
  indent_this_line(tc);

  /*
  start_c_line(tc);
  fputs("\"\\n\"", f);
  end_c_line(tc);
  */

  if(!is_debst(tc)) {
    fputs("{ if(", f);
    fputs(str_c_fct_get_file,f);
    fputs("()) ", f);
  }
  fputs("{ fputc('\\n', ", f);
  if(is_debst(tc)) {
    fputs("stdout", f);
  } else {
    fputs(str_c_fct_get_file,f);
    fputs("()", f);
  }
  fputs("); }", f);
  if(!is_debst(tc)) {
    fputs(" }", f);
  }
  fputc('\n', f);
  indent_this_line(tc);
  if(is_debst(tc)) {
    fputs("fflush(stdout);\n", f);
  } else {
    fputs(
      "{ if(dktrace_file()) { fflush(dktrace_file()); } }\n",
      f
    );
  }
  indent_this_line(tc);
  fputs("/* TRACE END */", f);
  if(need_source_position(tc)) {
    nl(tc);
    print_source_position(tc, 0);
  }
  
}
/* }}}2 */


/* end_java_trace_contents {{{2 */
static
void
end_java_trace_contents DK_P1(tracecc_t *,tc)
{
  if(is_debug(tc)) {
    fputs(");", tc->outfile);
  }
}
/* }}}2 */


/* end_trace_contents {{{2 */

/* ********************************************************************* */
/* *                                                                   * */
/* *   Trace message finished.                                         * */
/* *                                                                   * */
/* ********************************************************************* */

static
void
end_trace_contents DK_P1(tracecc_t *,tc)
{
  
  if(is_debug(tc)) {
    switch(tc->newft) {
      case 2: {
        end_java_trace_contents(tc);
      } break;
      default: {
        end_c_trace_contents(tc);
      } break;
    }
  }
  
  fputc('\n', tc->outfile);
}
/* }}}2 */


/* add_to_interpreter_command {{{2 */

/* ********************************************************************* */
/* *                                                                   * */
/* *   Add one character to the command line for                       * */
/* *   the interpreter.                                                * */
/* *                                                                   * */
/* ********************************************************************* */

static
int
add_to_interpreter_command DK_P1(tracecc_t *,tc)
{
  int back = 0;
  
  if((tc->icmdu) < ((tc->icmdsz) - 1)) {
    back = 1;
    (tc->icmdbuf)[tc->icmdu] = tc->currentchar;
    tc->icmdu += 1;
  } else {
    if(tc->app && (!(tc->line_too_long_printed))) {
      char *xxxptr[3];
      dkapp_set_source_filename(tc->app, tc->in_short);
      dkapp_set_source_lineno(tc->app, tc->lineno);
      xxxptr[0] = (tc->messages)[26];
      dkapp_log_msg(tc->app, DK_LOG_LEVEL_ERROR, xxxptr, 1);
      dkapp_set_source_filename(tc->app, NULL);
      dkapp_set_source_lineno(tc->app, 0UL);
      tc->line_too_long_printed = 1;
    }
  }
  
  return back;
}
/* }}}2 */


/* begin_interpreter_command {{{2 */

/* ********************************************************************* */
/* *                                                                   * */
/* *   Start to save a new interpreter command in                      * */
/* *   the command line buffer.                                        * */
/* *                                                                   * */
/* ********************************************************************* */

static
void
begin_interpreter_command DK_P1(tracecc_t *,tc)
{
  
  tc->icmdu = 0;
  (void)add_to_interpreter_command(tc);
  tc->line_too_long_printed = 0;
}
/* }}}2 */


/* fct_timestamp_off {{{2 */
static
int
fct_timestamp_off DK_P2(tracecc_t *,tc,char *,args)
{
  int back = 1;
  (tc->f).o &= (~(TRACECC_OPT_DATETIME));
  return back;
}
/* }}}2 */


/* fct_timestamp_on {{{2 */
static
int
fct_timestamp_on DK_P2(tracecc_t *,tc,char *,args)
{
  int back = 1;
  (tc->f).o |= TRACECC_OPT_DATETIME;
  return back;
}
/* }}}2 */


/* fct_lineno_off {{{2 */
static
int
fct_lineno_off DK_P2(tracecc_t *,tc,char *,args)
{
  int back = 1;
  (tc->f).o &= (~(TRACECC_OPT_LINENO));
  return back;
}
/* }}}2 */


/* fct_lineno_on {{{2 */
static
int
fct_lineno_on DK_P2(tracecc_t *,tc,char *,args)
{
  int back = 1;
  if(((tc->o).o) & TRACECC_OPT_LINENO) {
    (tc->f).o |= TRACECC_OPT_LINENO;
  }
  return back;
}
/* }}}2 */


/* fct_trace_off {{{2 */
static
int
fct_trace_off DK_P2(tracecc_t *,tc,char *,args)
{
  int back = 1;
  (tc->f).o &= (~(TRACECC_OPT_DEBUG));
  return back;
}
/* }}}2 */


/* fct_trace_on {{{2 */
static
int
fct_trace_on DK_P2(tracecc_t *,tc,char *,args)
{
  int back = 1;
  if(((tc->o).o) & TRACECC_OPT_DEBUG) {
    (tc->f).o |= TRACECC_OPT_DEBUG;
  }
  return back;
}
/* }}}2 */


/* fct_trace_pipe {{{2 */

/* ********************************************************************* */
/* *                                                                   * */
/* *   Receive input from a pipe.                                      * */
/* *                                                                   * */
/* ********************************************************************* */

static
int
fct_trace_pipe DK_P2(tracecc_t *,tc,char *,args)
{
  int back = 0;
  FILE *f;
  char smallbuffer[128], *largebuffer, *tb;
  size_t bs, u;
  int cc;
  f = NULL;
  if(args) {
#if DK_HAVE_POPEN || DK_HAVE__POPEN
#if DK_HAVE_POPEN
    f = popen(args, "r");
#else
    f = _popen(args, "r");
#endif
    if(f) {
      back = 1;
      largebuffer = dk_new(char,4096);
      if(largebuffer) {
        tb = largebuffer; bs = 4096;
      } else {
        tb = smallbuffer; bs = sizeof(smallbuffer);
	/* XXXXX WARNING: Only small buffer */
        if(tc->app) {
          char *xxxptr[8];
	  dkapp_set_source_filename(tc->app, tc->in_short);
	  dkapp_set_source_lineno(tc->app, tc->lineno);
          xxxptr[0] = (tc->messages)[0];
          dkapp_log_msg(tc->app, DK_LOG_LEVEL_ERROR, xxxptr, 1);
	  dkapp_set_source_filename(tc->app, NULL);
	  dkapp_set_source_lineno(tc->app, 0UL);
        }
      }
      cc = 1;
      while(cc) {
        u = fread(tb,1,bs,f);
	if(u > 0) {
	  fwrite(tb,1,u,tc->outfile);
	} else {
	  cc = 0;
	}
      }
      if(largebuffer) {
        dk_delete(largebuffer);
      }
#if DK_HAVE_POPEN
      pclose(f);
#else
      _pclose(f);
#endif
    } else {
      /* XXXXX ERROR: Failed to execute command */
      if(tc->app) {
        char *xxxptr[8];
        xxxptr[0] = (tc->messages)[1];
        xxxptr[1] = str_spc;
        xxxptr[2] = str_bropen;
        xxxptr[3] = args;
        xxxptr[4] = str_brclose;
	dkapp_set_source_filename(tc->app, tc->in_short);
	dkapp_set_source_lineno(tc->app, tc->lineno);
        dkapp_log_msg(tc->app, DK_LOG_LEVEL_ERROR, xxxptr, 5);
	dkapp_set_source_filename(tc->app, NULL);
	dkapp_set_source_lineno(tc->app, 0UL);
      }
    }
#else
    /* XXXXX ERROR: No popen function on this system! */
    if(tc->app) {
      char *xxxptr[8];
      xxxptr[0] = (tc->messages)[2];
      dkapp_set_source_filename(tc->app, tc->in_short);
      dkapp_set_source_lineno(tc->app, tc->lineno);
      dkapp_log_msg(tc->app, DK_LOG_LEVEL_ERROR, xxxptr, 1);
      dkapp_set_source_filename(tc->app, NULL);
      dkapp_set_source_lineno(tc->app, 0UL);
    }
#endif
  } else {
    /* XXXXXX ERROR: Missing arguments */
    if(tc->app) {
      char *xxxptr[8];
      xxxptr[0] = (tc->messages)[3];
      dkapp_set_source_filename(tc->app, tc->in_short);
      dkapp_set_source_lineno(tc->app, tc->lineno);
      dkapp_log_msg(tc->app, DK_LOG_LEVEL_ERROR, xxxptr, 1);
      dkapp_set_source_filename(tc->app, NULL);
      dkapp_set_source_lineno(tc->app, 0UL);
    }
  }
  return back;
}
/* }}}2 */


/* fct_trace_code {{{2 */
static
int
fct_trace_code DK_P2(tracecc_t *,tc,char *,args)
{
  int back = 1;
  if(is_debug(tc)) {
    fputs(args, tc->outfile);
  }
  return back;
}
/* }}}2 */


/* encoded_fputc {{{ */
static
void
encoded_fputc DK_P3(char,c,FILE *,f,int,ft)
{
  switch(c) {
    case '\n': {
      switch(ft) {
        case 2: {
	  fputs("\",\n\"", f);
	} break;
	case 1:
	default: {
	  fputs("\",\n(char *)\"", f);
	} break;
      }
    } break;
    case '\t': {
      fputs("\\t", f);
    } break;
    case '"' : {
      fputs("\\\"", f);
    } break;
    default: {
      fputc(c,f);
    } break;
  }
}
/* }}} */


/* hex_fputc {{{ */
static
void
hex_fputc DK_P2(char,c,FILE *,f)
{
  unsigned char uc, uc1;
  unsigned u;
  uc = c;
  fputs("0x", f);
  uc1 = ((uc >> 4) & 0x0F); u = uc1;
  fputc(hexdigits[u], f);
  uc1 = (uc & 0x0F); u = uc1;
  fputc(hexdigits[u], f);
}
/* }}} */


/* fct_byte_file {{{ */
static
int
fct_byte_file DK_P2(tracecc_t *,tc,char *,args)
{
  int back = 0;
  int r, cc;
  char smallbuffer[128], *largebuffer, *buffer, *ptr1, *ptr2;
  size_t bs, used, i;
  FILE *inputfile, *f;
  unsigned long bytesused;

  r = 0;
  f = tc->outfile; bytesused = 0UL;
  ptr1 = dkstr_start(args, NULL);
  if(ptr1) {
    dkstr_chomp(ptr1, NULL);
    if(tc->app) {
      inputfile = dkapp_fopen(tc->app, ptr1, "r");
    } else {
      inputfile = dksf_msfo(ptr1, "r", 0, &r);
    }
    if(inputfile) {
      buffer = smallbuffer; bs = sizeof(smallbuffer);
      largebuffer = dk_new(char, 4096);
      if(largebuffer) {
        buffer = largebuffer; bs = 4096;
      } else {
        /* XXXXX Warning for small buffer */
        if(tc->app) {
          char *xxxptr[8];
	  dkapp_set_source_filename(tc->app, tc->in_short);
	  dkapp_set_source_lineno(tc->app, tc->lineno);
          xxxptr[0] = (tc->messages)[0];
          dkapp_log_msg(tc->app, DK_LOG_LEVEL_ERROR, xxxptr, 1);
	  dkapp_set_source_filename(tc->app, NULL);
	  dkapp_set_source_lineno(tc->app, 0UL);
        }
      }
      switch(tc->newft) {
        case 2:
	case 1:
	default: {
	  /* transfer file contents */
	  cc = 1;
	  while(cc) {
	    used = fread(buffer,1,bs,inputfile);
	    if(used > 0) {
	      for(i = 0; i < used; i++) {
	        if(bytesused) {
		  fputc(',', f);
		  if(!(bytesused % 12)) {
		    fputc('\n', f);
		  }
		}
	        hex_fputc(buffer[i], f);
		bytesused++;
	      }
	    } else {
	      cc = 0;
	    }
	  }
	  if(bytesused) {
	    fputc('\n', f);
	  }
	} break;
      }
      if(largebuffer) {
        dk_delete(largebuffer);
	largebuffer = NULL;
      }
      fclose(inputfile);
      back = 1;
    } else {
      /* ERROR: Failed to read file */
    }
  }
  return back;
}
/* }}} */



/* fct_string_file {{{ */
static
int
fct_string_file DK_P2(tracecc_t *,tc,char *,args)
{
  int back = 0;
  int r, cc;
  char smallbuffer[128], *largebuffer, *buffer, *ptr1, *ptr2;
  size_t bs, used, i;
  FILE *inputfile, *f;

  r = 0;
  f = tc->outfile;
  ptr1 = dkstr_start(args, NULL);
  if(ptr1) {
    dkstr_chomp(ptr1, NULL);
    if(tc->app) {
      inputfile = dkapp_fopen(tc->app, ptr1, "r");
    } else {
      inputfile = dksf_msfo(ptr1, "r", 0, &r);
    }
    if(inputfile) {
      buffer = smallbuffer; bs = sizeof(smallbuffer);
      largebuffer = dk_new(char, 4096);
      if(largebuffer) {
        buffer = largebuffer; bs = 4096;
      } else {
        /* XXXXX Warning for small buffer */
        if(tc->app) {
          char *xxxptr[8];
	  dkapp_set_source_filename(tc->app, tc->in_short);
	  dkapp_set_source_lineno(tc->app, tc->lineno);
          xxxptr[0] = (tc->messages)[0];
          dkapp_log_msg(tc->app, DK_LOG_LEVEL_ERROR, xxxptr, 1);
	  dkapp_set_source_filename(tc->app, NULL);
	  dkapp_set_source_lineno(tc->app, 0UL);
        }
      }
      switch(tc->newft) {
        case 2: {
	} break;
	case 1:
	default: {
	  /* show file position if necessary */
	  if(need_source_position(tc)) {
	    nl(tc);
	    fputs("#line 1 \"", f);
	    ptr2 = ptr1;
	    while(*ptr2) {
	      if(*ptr2 == '\\') {
	        fputc('/', f);
	      } else {
	        fputc(*ptr2, f);
	      }
	      ptr2++;
	    }
	    fputs("\"\n", f);
	  }
	  /* transfer file contents */
	  cc = 1;
	  fputs("(char *)\"", f);
	  while(cc) {
	    used = fread(buffer,1,bs,inputfile);
	    if(used > 0) {
	      for(i = 0; i < used; i++) {
	        encoded_fputc(buffer[i], f, tc->newft);
	      }
	    } else {
	      cc = 0;
	    }
	  }
	  fputs("\",\nNULL\n", f);
	} break;
      }
      if(largebuffer) {
        dk_delete(largebuffer);
	largebuffer = NULL;
      }
      fclose(inputfile);
      back = 1;
    } else {
      /* ERROR: Failed to read file */
    }
  }
  return back;
}
/* }}} */


/* fct_trace_init {{{2 */
static
int
fct_trace_init DK_P2(tracecc_t *,tc,char *,args)
{
  int back = 1;
  char *ptr;
  FILE *f;
  
  f = tc->outfile;
  switch(tc->newft) {
    case 2: {
      if(is_debug(tc)) {
        if(is_debst(tc)) {
	  fputs("KrTrace.start();", tc->outfile);
	} else {
	  ptr = dkstr_start(args, NULL);
	  if(ptr) {
	    (void)dkstr_next(ptr,NULL);
	    fputs("KrTrace.start(\"", tc->outfile);
	    fputs(ptr, tc->outfile);
	    fputs("\");", tc->outfile);
	  } else {
            if(tc->app) {
              char *xxxptr[8];
              xxxptr[0] = (tc->messages)[3];
              dkapp_set_source_filename(tc->app, tc->in_short);
              dkapp_set_source_lineno(tc->app, tc->lineno);
              dkapp_log_msg(tc->app, DK_LOG_LEVEL_ERROR, xxxptr, 1);
              dkapp_set_source_filename(tc->app, NULL);
              dkapp_set_source_lineno(tc->app, 0UL);
            }
	  }
	}
      }
    } break;
    case 1:
    default: {
      if(is_debug(tc) && (!is_debst(tc))) {
        if(args) {
          ptr = dkstr_start(args, NULL);
          if(ptr) {
            (void)dkstr_next(ptr, NULL);
            back = 1;
            fputs(str_c_fct_init, f);
            fputs("(\"", f);
            fputs(ptr, f);
            fputs("\");", f);
          }
        } else {
          /* XXXXX ERROR: Need argument */
          if(tc->app) {
            char *xxxptr[8];
            xxxptr[0] = (tc->messages)[3];
            dkapp_set_source_filename(tc->app, tc->in_short);
            dkapp_set_source_lineno(tc->app, tc->lineno);
            dkapp_log_msg(tc->app, DK_LOG_LEVEL_ERROR, xxxptr, 1);
            dkapp_set_source_filename(tc->app, NULL);
            dkapp_set_source_lineno(tc->app, 0UL);
          }
        }
      } else {
        back = 1;
      }
    } break;
  }
  
  return back;
}
/* }}}2 */


/* fct_trace_end {{{2 */
static
int
fct_trace_end DK_P2(tracecc_t *,tc,char *,args)
{
  FILE *f;
  f = tc->outfile;
  switch(tc->newft) {
    case 2: {
      if(is_debug(tc)) {
        fputs("KrTrace.stop();", tc->outfile);
      }
    } break;
    case 1:
    default: {
      if(is_debug(tc) && (!is_debst(tc))) {
        fputs(str_c_fct_end, f);
        fputs("();", f);
      }
    } break;
  }
  return 1;
}
/* }}}2 */


/* fct_trace_include {{{2 */
static
int
fct_trace_include DK_P2(tracecc_t *,tc,char *,args)
{
  if(is_debug(tc)) {
    switch(tc->newft) {
      case 2: {
        fputs("import KrTrace;\n", tc->outfile);
      } break;
      case 1:
      default: {
        fputs("#include <dktrace.h>\n", tc->outfile);
      } break;
    } 
  }
  return 1;
}
/* }}}2 */


/* fctlist {{{2 */

/* ********************************************************************* */
/* *                                                                   * */
/* *   List of functions for interpreter.                              * */
/* *                                                                   * */
/* ********************************************************************* */

static function_info_t fctlist[] = {
  { (char *)"trace-init",    fct_trace_init,    0 },
  { (char *)"trace-end",     fct_trace_end,     0 },
  { (char *)"trace-include", fct_trace_include, 0 },
  { (char *)"trace-code",    fct_trace_code,    0 },
  { (char *)"!",             fct_trace_code,    0 },
  { (char *)"trace-on",      fct_trace_on,      0 },
  { (char *)"trace-off",     fct_trace_off,     0 },
  { (char *)"debug-on",      fct_trace_on,      0 },
  { (char *)"debug-off",     fct_trace_off,     0 },
  { (char *)"timestamp-on",  fct_timestamp_on,  0 },
  { (char *)"timestamp-off", fct_timestamp_off, 0 },
  { (char *)"lineno-on",     fct_lineno_on,     0 },
  { (char *)"lineno-off",    fct_lineno_off,    0 },
  { (char *)"pipe",          fct_trace_pipe,    1 },
  { (char *)"|",             fct_trace_pipe,    1 },
  { (char *)"string-file",   fct_string_file,   1 },
  { (char *)"byte-file",     fct_byte_file,     1 },
  { NULL, NULL, 0 }
};
/* }}}2 */


/* i_run_interpreter_command {{{2 */

/* ********************************************************************* */
/* *                                                                   * */
/* *   Run the interpreter command as saved in the                     * */
/* *   command line buffer.                                            * */
/* *                                                                   * */
/* ********************************************************************* */

static
int
i_run_interpreter_command DK_P1(tracecc_t *,tc)
{
  int back = 0;
  char *ptr, *args;
  function_info_t *fctptr, *fctlfd;
  i_fct_t fct;

  
  if((tc->icmdu) < (tc->icmdsz)) {
    (tc->icmdbuf)[tc->icmdu] = '\0';
    ptr = dkstr_start(tc->icmdbuf, NULL);
    if(ptr) {
      dkstr_chomp(ptr, NULL);
      args = dkstr_next(ptr, NULL);
      fctptr = NULL; fctlfd = fctlist;
      while((fctlfd->name) && (!fctptr)) {
        if(strcmp(fctlfd->name, ptr) == 0) {
	  fctptr = fctlfd;
	} else {
	  fctlfd++;
	}
      }
      if(fctptr) {
        fct = fctptr->fct;
	back = (*fct)(tc, args);
      } else {
        /* XXXXX ERROR: Function not found */
        if(tc->app) {
          char *xxxptr[8];
          xxxptr[0] = (tc->messages)[5];
          xxxptr[1] = str_spc;
          xxxptr[2] = str_bropen;
          xxxptr[3] = ptr;
          xxxptr[4] = str_brclose;
	  dkapp_set_source_filename(tc->app, tc->in_short);
	  dkapp_set_source_lineno(tc->app, tc->lineno);
          dkapp_log_msg(tc->app, DK_LOG_LEVEL_ERROR, xxxptr, 5);
	  dkapp_set_source_filename(tc->app, NULL);
	  dkapp_set_source_lineno(tc->app, 0UL);
        }
      }
    } else {
      /* XXXXX ERROR: Empty function name */
      if(tc->app) {
        char *xxxptr[8];
        xxxptr[0] = (tc->messages)[6];
	dkapp_set_source_filename(tc->app, tc->in_short);
	dkapp_set_source_lineno(tc->app, tc->lineno);
        dkapp_log_msg(tc->app, DK_LOG_LEVEL_ERROR, xxxptr, 1);
	dkapp_set_source_filename(tc->app, NULL);
	dkapp_set_source_lineno(tc->app, 0UL);
      }
    }
  }
  
  return back;
}
/* }}}2 */


/* run_interpreter_command {{{2 */

/* ********************************************************************* */
/* *                                                                   * */
/* *   Simply run the interpreter.                                     * */
/* *                                                                   * */
/* ********************************************************************* */

static
int
run_interpreter_command DK_P1(tracecc_t *,tc)
{
  int back;
  back = i_run_interpreter_command(tc);
  if(need_source_position(tc)) {
    nl(tc);
    print_source_position(tc, 0);
  }
  return back;
}
/* }}}2 */


/* run_interpreter_command_nl {{{2 */

/* ********************************************************************* */
/* *                                                                   * */
/* *   Run the interpreter and print a newline to the                  * */
/* *   source file.                                                    * */
/* *   The $!... - variant of interpretation consumes                  * */
/* *   a newline, so we need to add one explicitly.                    * */
/* *                                                                   * */
/* ********************************************************************* */

static
int
run_interpreter_command_nl DK_P1(tracecc_t *,tc)
{
  int back;
  back = i_run_interpreter_command(tc);
  fputc('\n', tc->outfile);
  if(need_source_position(tc)) {
    print_source_position(tc, 1);
  }
  return back;
}
/* }}}2 */


/* open_comment_box_file {{{2 */

/* ********************************************************************* */
/* *                                                                   * */
/* *   Begin to save data for the next comment box.                    * */
/* *                                                                   * */
/* ********************************************************************* */

static
int
open_comment_box_file DK_P1(tracecc_t *,tc)
{
  int back = 0;
  
  if(tc->tmpfile) {
    fclose(tc->tmpfile);
  }
  tc->tmpfile = fopen(tc->tmp_name_buf, "w+");
  if(tc->tmpfile) {
    back = 1;
  }
  tc->cmbll = tc->cmbmax = 0;
  
  return back;
}
/* }}}2 */


/* add_newline_to_comment_box {{{2 */

/* ********************************************************************* */
/* *                                                                   * */
/* *   Add a newline to the comment box data file.                     * */
/* *                                                                   * */
/* ********************************************************************* */

static
void
add_newline_to_comment_box DK_P1(tracecc_t *,tc)
{
  
  if(tc->tmpfile) {
    fputc('\n', tc->tmpfile);
    tc->cmbll = 0;
  }
  
}
/* }}}2 */


/* add_char_to_comment_box {{{2 */

/* ********************************************************************* */
/* *                                                                   * */
/* *   Add a character to the comment box data file.                   * */
/* *                                                                   * */
/* ********************************************************************* */

static
void
add_char_to_comment_box DK_P1(tracecc_t *,tc)
{
  
  if(tc->tmpfile) {
    fputc(tc->currentchar, tc->tmpfile);
  }
  tc->cmbll += 1;
  if((tc->cmbll) > (tc->cmbmax)) {
    tc->cmbmax = tc->cmbll;
  }
  
}
/* }}}2 */


/* add_control_and_char_to_comment_box {{{2 */

/* ********************************************************************* */
/* *                                                                   * */
/* *   Add the control character and the current character             * */
/* *   to the comment box data file.                                   * */
/* *                                                                   * */
/* ********************************************************************* */

static
void
add_control_and_char_to_comment_box DK_P1(tracecc_t *,tc)
{
  
  if(tc->tmpfile) {
    switch(tc->newft) {
      case 2:
      case 1:
      default: {
	fputc('$', tc->tmpfile);
      } break;
    }
    fputc(tc->currentchar, tc->tmpfile);
  }
  tc->cmbll += 2;
  if((tc->cmbll) > (tc->cmbmax)) { tc->cmbmax = tc->cmbll; }
  
}
/* }}}2 */


/* start_line {{{2 */
static
void
start_line DK_P1(FILE *,f)
{
  fputs("/* * ", f);
}
/* }}}2 */


/* end_line {{{2 */
static
void
end_line DK_P1(FILE *,f)
{
  fputs(" * */", f);
  fputc('\n', f);
}
/* }}}2 */


/* box_line {{{2 */
static
void
box_line DK_P2(FILE *,f,int,mll)
{
  int i, mx;
  mx = mll + 2;
  fputs("/* *", f);
  for(i = 0; i < mx; i++) { fputc('*', f); }
  fputs("* */", f);
  fputc('\n', f);
}
/* }}}2 */


/* print_comment_box {{{2 */

/* ********************************************************************* */
/* *                                                                   * */
/* *   Transfer comment box data file to the                           * */
/* *   destination file.                                               * */
/* *                                                                   * */
/* ********************************************************************* */

static
void
print_comment_box DK_P1(tracecc_t *,tc)
{
#if FIRST_ATTEMPT_FOR_COMMENT_BOX
  int maxval, cc;
  char buffer[128]; size_t max, u;
  FILE *f;
  int ll;
  f = tc->outfile;
  if(tc->tmpfile) {
    fflush(tc->tmpfile);
    rewind(tc->tmpfile);
    clearerr(tc->tmpfile);
    fputs("/*\n", f);
    cc = 1;
    while(cc) {
      max = fread(buffer,1,sizeof(buffer),tc->tmpfile);
      if(max > 0) {
        for(u = 0; u < max; u++) {
	  fputc(buffer[u], f);
	}
      } else {
        cc = 0;
      }
    }
    fputs("*/\n", f);
    fclose(tc->tmpfile); tc->tmpfile = NULL;
    dksf_remove_file(tc->tmp_name_buf);
  }
#else
  char smallbuffer[128], *largebuffer, *buffer;
  size_t bufsize, max, u;
  int mll, curpos, cc;
  FILE *fi, *fo;
  
  fi = tc->tmpfile;
  fo = tc->outfile;
  if(tc->tmpfile) {
    largebuffer = dk_new(char,4096);
    if(largebuffer) {
      buffer = largebuffer; bufsize = 4096;
    } else {
      buffer = smallbuffer; bufsize = sizeof(smallbuffer);
    }
    mll = tc->cmbmax;
    if((tc->cmbmax) < ((tc->f).b - 10)) {
      mll = (tc->f).b - 10;
    }
    curpos = 0;
    fflush(fi);
    rewind(fi);
    clearerr(fi);
    fputc('\n', fo);
    /* starting box line */
    box_line(fo, mll);
    start_line(fo);
    cc = 1;
    while(cc) {
      max = fread(buffer, 1, bufsize, fi);
      if(max > 0) {
        for(u = 0; u < max; u++) {
	  /* buffer[u] verarbeiten */
	  switch(buffer[u]) {
	    case '\n' : {
	      while(curpos++ < mll) { fputc(' ', fo); }
	      end_line(fo);
	      start_line(fo);
	      curpos = 0;
	    } break;
	    case '\t' : {
	      fputc(' ', fo); curpos++;
	      while(curpos % 8) {
	        fputc(' ', fo); curpos++;
	      }
	    } break;
	    default : {
	      fputc(buffer[u], fo);
	      curpos++;
	    } break;
	  }
	}
      } else {
        cc = 0;
      }
    }
    while(curpos++ < mll) { fputc(' ', fo); }
    end_line(fo);
    /* finishing box line */
    box_line(fo, mll);
    if(largebuffer) {
      dk_delete(largebuffer);
    }
  }
  if(need_source_position(tc)) {
    nl(tc);
    print_source_position(tc, 0);
  }
  
#endif
}
/* }}}2 */


/* start_line_comment {{{2 */

/* ********************************************************************* */
/* *                                                                   * */
/* *   A C++ style comment was recognized in the source file.          * */
/* *                                                                   * */
/* ********************************************************************* */

static
void
start_line_comment DK_P1(tracecc_t *,tc)
{
  FILE *f;
  
  f = tc->outfile;
  switch(tc->newft) {
    case 2: {
      fputs("//", f);
    } break;
    default: {
      if(have_cpp_comment(tc)) {
	fputs("//", f);
      } else {
	fputs("/*", f);
      }
    } break;
  }
  
}
/* }}}2 */


/* end_line_comment {{{2 */

/* ********************************************************************* */
/* *                                                                   * */
/* *   The end of a C++ comment was recognized in the source.          * */
/* *                                                                   * */
/* ********************************************************************* */

static
void
end_line_comment DK_P1(tracecc_t *,tc)
{
  FILE *f;
  
  f = tc->outfile;
  switch(tc->newft) {
    case 2: {
      fputc('\n', f);
    } break;
    default: {
      if(have_cpp_comment(tc)) {
	fputc('\n', f);
      } else {
	fputs(" */\n", f);
      }
    } break;
  }
  
}
/* }}}2 */


/* start_normal_comment {{{2 */

/* ********************************************************************* */
/* *                                                                   * */
/* *   A normal comment begins.                                        * */
/* *                                                                   * */
/* ********************************************************************* */

static
void
start_normal_comment DK_P1(tracecc_t *,tc)
{
  
  fputc('/', tc->outfile);
  print_current_char(tc);
  
}
/* }}}2 */


/* perform_action {{{2 */

/* ********************************************************************* */
/* *                                                                   * */
/* *   Perform action for character.                                   * */
/* *                                                                   * */
/* ********************************************************************* */

static
int
perform_action DK_P1(tracecc_t *,tc)
{
  int back = 0;
  
  switch(tc->reaction) {
    case  0: { back = 1; /* important to set back */ } break;
    case  1: { print_current_char(tc); back = 1; } break;
    case  2: { tc_and_current_char(tc); back = 1; } break;
    case  3: { begin_trace_block(tc); back = 1; } break;
    case  4: { begin_trace_contents(tc); back = 1; } break;
    case  5: { end_trace_contents(tc); back = 1; } break;
    case  6: { begin_interpreter_command(tc); back = 1; } break;
    case  7: { back = add_to_interpreter_command(tc); } break;
    case  8: { back = run_interpreter_command(tc); } break;
    case  9: { print_debug_char(tc); back = 1; } break;
    case 10: { end_trace_contents(tc); back = 1; } break;
    case 11: { back = open_comment_box_file(tc); } break;
    case 12: { add_newline_to_comment_box(tc); back = 1; } break;
    case 13: { add_char_to_comment_box(tc); back = 1; } break;
    case 14: { add_control_and_char_to_comment_box(tc); back = 1; } break;
    case 15: { print_comment_box(tc); back = 1; } break;
    case 16: { start_line_comment(tc); back = 1; } break;
    case 17: { end_line_comment(tc); back = 1; }
    case 18: { start_normal_comment(tc); back = 1; } break;
    case 19: { back = run_interpreter_command_nl(tc); } break;
  }
  
  return back;
}
/* }}}2 */


/* add_char {{{2 */

/* ********************************************************************* */
/* *                                                                   * */
/* *   Add one character.                                              * */
/* *                                                                   * */
/* ********************************************************************* */

static
int
add_char DK_P1(tracecc_t *,tc)
{
  int back = 0;
  int auin;
  
  auin =
  (is_trace_control(tc->newft,tc->currentchar)?TRACE_CONTROL:(tc->currentchar));
  tc->reaction = automata_reaction(&(tc->state), auin);
  back = perform_action(tc);
  
  return back;
}
/* }}}2 */


/* process_char {{{2 */

/* ********************************************************************* */
/* *                                                                   * */
/* *   Process one character.                                          * */
/* *                                                                   * */
/* ********************************************************************* */

static
int
process_char DK_P1(tracecc_t *,tc)
{
  int back = 0;
  
  if(!((tc->textflags) & 1)) {
    if((tc->currentchar == ' ') || (tc->currentchar == '\t')) {
      tc->indent += 1;
      if(tc->currentchar == '\t') {
        while((tc->indent) % 8) {
	  tc->indent += 1;
	}
      }
    } else {
      tc->textflags |= 1;
    }
  }
  back = add_char(tc);
  if((tc->currentchar) == '\n') {
    tc->lineno += 1UL;
    tc->textflags &= (~1);
    tc->indent = 0;
  }
  
  return back;
}
/* }}}2 */


/* run_for_opened_files {{{2 */

/* ********************************************************************* */
/* *                                                                   * */
/* *   tc->infile and tc->outfile are open, process files.             * */
/* *                                                                   * */
/* ********************************************************************* */

static
int
run_for_opened_files DK_P1(tracecc_t *,tc)
{
  int back = 1; int cc;
  
  cc = 1;
  tc->lineno = 1UL;
  tc->indent = 0;
  tc->textflags = 0;
  while(cc) {
    tc->ibu = fread((void *)(tc->inbuffer), 1, tc->ibl, tc->infile);
    if(tc->ibu) {
      tc->ibc = 0;
      while((tc->ibc) < (tc->ibu)) {
        tc->currentchar = (tc->inbuffer)[tc->ibc];
        if(!process_char(tc)) {
	  back = 0;
	}
	tc->ibc += 1;
      }
    } else {
      cc = 0;
    }
  }
  
  return back;
}
/* }}}2 */


/* must_run_for_these_files {{{2 */
static
int
must_run_for_these_files DK_P1(tracecc_t *,tc)
{
  int back = 1;
  
  if(((tc->o).o) & TRACECC_OPT_MAKE) {
#if DK_HAVE_STAT || DK_HAVE__STAT
#if DK_HAVE_STAT
  /* stat available {{{3 */
  struct stat st1, st2;
  if(stat(tc->out_real, &st2) >= 0) {
    if(stat(tc->in_real, &st1) >= 0) {
      if((st2.st_mtime) > (st1.st_mtime)) {
        back = 0;
      }
    }
  }
  /* }}}3 */
#else
  /* _stat instead {{{3 */
  struct _stat st1, st2;
  if(_stat(tc->out_real, &st2) >= 0) {
    if(_stat(tc->in_real, &st1) >= 0) {
      if((st2.st_mtime) > (st1.st_mtime)) {
        back = 0;
      }
    }
  }
  /* }}}3 */
#endif
#endif
  }
  
  return back;
}
/* }}}2 */


/* run_for_real_in_and_real_out {{{2 */

/* ********************************************************************* */
/* *                                                                   * */
/* *   We have names for input and output file,                        * */
/* *   open the files and run.                                         * */
/* *                                                                   * */
/* ********************************************************************* */

static
int
run_for_real_in_and_real_out DK_P1(tracecc_t *,tc)
{
  int back = 0;
  int reason;
  char *ptr;
  
  tc->newft = 0;
  ptr = dksf_get_file_type_dot(tc->out_real);
  if(ptr) { tc->newft = get_new_file_type(ptr); }
  tc->in_short = NULL;
  ptr = tc->in_real;
  while(*ptr) {
    if(*ptr == fnsep[0]) tc->in_short = ptr;
    ptr++;
  }
  if(tc->in_short) {
    ptr = tc->in_short; ptr++; tc->in_short = ptr;
  } else {
    tc->in_short = tc->in_real;
  }
  if(must_run_for_these_files(tc)) {
    if(tc->app) {
      tc->infile = dkapp_fopen(tc->app, tc->in_real, "r");
      if(tc->infile) {
        tc->outfile = dkapp_fopen(tc->app, tc->out_real, "w");
        if(tc->outfile) {
          if(tc->app) {
            char *xxxptr[8];
            xxxptr[0] = (tc->messages)[16];
            xxxptr[1] = tc->in_short;
            xxxptr[2] = (tc->messages)[17];
            dkapp_log_msg(tc->app, DK_LOG_LEVEL_PROGRESS, xxxptr, 3);
          }
          back = run_for_opened_files(tc);
          if(tc->app) {
            char *xxxptr[8];
            xxxptr[0] = (tc->messages)[18];
            xxxptr[1] = tc->in_short;
            xxxptr[2] = (tc->messages)[19];
            dkapp_log_msg(tc->app, DK_LOG_LEVEL_PROGRESS, xxxptr, 3);
          }
          fclose(tc->outfile); tc->outfile = NULL;
        } else {
          /* XXXXX ERROR: Failed to open output file */
          if(tc->app) {
            char *xxxptr[8];
            xxxptr[0] = (tc->messages)[7];
            xxxptr[1] = str_spc;
            xxxptr[2] = str_bropen;
            xxxptr[3] = tc->out_real;
            xxxptr[4] = str_brclose;
            dkapp_log_msg(tc->app, DK_LOG_LEVEL_ERROR, xxxptr, 5);
          }
        }
        fclose(tc->infile); tc->infile = NULL;
      } else {
        /* XXXXX ERROR: Failed to open input file */
        if(tc->app) {
          char *xxxptr[8];
          xxxptr[0] = (tc->messages)[8];
          xxxptr[1] = str_spc;
          xxxptr[2] = str_bropen;
          xxxptr[3] = tc->in_real;
          xxxptr[4] = str_brclose;
          dkapp_log_msg(tc->app, DK_LOG_LEVEL_ERROR, xxxptr, 5);
        }
      }
    } else {
      tc->infile = dksf_msfo(tc->in_real, "r", 0, &reason);
      if(tc->infile) {
        tc->outfile = dksf_msfo(tc->out_real, "w", 0, &reason);
        if(tc->outfile) {
          back = run_for_opened_files(tc);
          fclose(tc->outfile); tc->outfile = NULL;
        } else {
          /* XXXXX ERROR: Failed to open output file */
          if(tc->app) {
            char *xxxptr[8];
            xxxptr[0] = (tc->messages)[7];
            xxxptr[1] = str_spc;
            xxxptr[2] = str_bropen;
            xxxptr[3] = tc->in_real;
            xxxptr[4] = str_brclose;
            dkapp_log_msg(tc->app, DK_LOG_LEVEL_ERROR, xxxptr, 5);
          }
        }
        fclose(tc->infile); tc->infile = NULL;
      } else {
        /* XXXXX ERROR: Failed to open input file */
        if(tc->app) {
          char *xxxptr[8];
          xxxptr[0] = (tc->messages)[8];
          xxxptr[1] = str_spc;
          xxxptr[2] = str_bropen;
          xxxptr[3] = tc->in_real;
          xxxptr[4] = str_brclose;
          dkapp_log_msg(tc->app, DK_LOG_LEVEL_ERROR, xxxptr, 5);
        }
      }
    }
  } else {
    if(tc->app) {
      char *xxxptr[8];
      xxxptr[0] = (tc->messages)[24];
      xxxptr[1] = tc->in_short,
      xxxptr[2] = (tc->messages)[25];
      dkapp_log_msg(tc->app, DK_LOG_LEVEL_PROGRESS, xxxptr, 3);
    }
  }
  tc->in_short = NULL;
  
  return back;
}
/* }}}2 */


/* run_for_in_and_real_out {{{2 */

/* ********************************************************************* */
/* *                                                                   * */
/* *   We have the exact output file name but we must                  * */
/* *   check whether or not to expand wildcards                        * */
/* *   in the input file name.                                         * */
/* *                                                                   * */
/* ********************************************************************* */

static
int
run_for_in_and_real_out DK_P1(tracecc_t *,tc)
{
  int back = 0;
  
  if(dksf_must_expand_filename(tc->in_ori)) {
    tc->fne_in = dkfne_open(tc->in_ori, 1, 0);
    if(tc->fne_in) {
      if(dkfne_next(tc->fne_in)) {
        tc->in_real = dkfne_get_fullname(tc->fne_in);
	back = run_for_real_in_and_real_out(tc);
	tc->in_real = NULL;
	if(dkfne_next(tc->fne_in)) {
	  /* XXXXX WARNING: Multiple input files matching pattern */
          if(tc->app) {
            char *xxxptr[8];
            xxxptr[0] = (tc->messages)[9];
            xxxptr[1] = str_spc;
            xxxptr[2] = str_bropen;
            xxxptr[3] = tc->in_ori;
            xxxptr[4] = str_brclose;
            dkapp_log_msg(tc->app, DK_LOG_LEVEL_ERROR, xxxptr, 5);
          }
	}
      } else {
        /* XXXXX ERROR: No matching input file */
        if(tc->app) {
          char *xxxptr[10];
          xxxptr[0] = (tc->messages)[13];
          xxxptr[1] = str_spc;
          xxxptr[2] = str_bropen;
          xxxptr[3] = tc->in_ori;
          xxxptr[4] = str_brclose;
          dkapp_log_msg(tc->app, DK_LOG_LEVEL_ERROR, xxxptr, 5);
        }
      }
      dkfne_close(tc->fne_in); tc->fne_in = NULL;
    } else {
      tc->error_code = DK_ERR_NOMEM;
      if(tc->app) dkapp_err_memory(tc->app,sizeof(dk_fne_t),1);
    }
  } else {
    tc->in_real = tc->in_ori;
    back = run_for_real_in_and_real_out(tc);
    tc->in_real = NULL;
  }
  
  return back;
}
/* }}}2 */


/* run_for_in_and_out_file_name {{{2 */

/* ********************************************************************* */
/* *                                                                   * */
/* *   We must check both input and output file name                   * */
/* *   whether or not we need to expand wildcards in                   * */
/* *   the name.                                                       * */
/* *                                                                   * */
/* ********************************************************************* */

static
int
run_for_in_and_out_file_name DK_P1(tracecc_t *,tc)
{
  int back = 0;
  
  if(dksf_must_expand_filename(tc->out_ori)) {
    tc->fne_out = dkfne_open(tc->out_ori,1,0);
    if(tc->fne_out) {
      if(dkfne_next(tc->fne_out)) {
        tc->out_real = dkfne_get_fullname(tc->fne_out);
	back = run_for_in_and_real_out(tc);
	tc->out_real = NULL;
	if(dkfne_next(tc->fne_out)) {
	  /* XXXXX WARNING: Multiple file names for pattern */
          if(tc->app) {
            char *xxxptr[8];
            xxxptr[0] = (tc->messages)[11];
            xxxptr[1] = str_spc;
            xxxptr[2] = str_bropen;
            xxxptr[3] = tc->out_ori;
            xxxptr[4] = str_brclose;
            dkapp_log_msg(tc->app, DK_LOG_LEVEL_ERROR, xxxptr, 5);
          }
	}
      } else {
        /* XXXXX ERROR: No matching output file name */
        if(tc->app) {
          char *xxxptr[8];
          xxxptr[0] = (tc->messages)[12];
          xxxptr[1] = str_spc;
          xxxptr[2] = str_bropen;
          xxxptr[3] = tc->in_real;
          xxxptr[4] = str_brclose;
          dkapp_log_msg(tc->app, DK_LOG_LEVEL_ERROR, xxxptr, 5);
        }
      }
      dkfne_close(tc->fne_out); tc->fne_out = NULL;
    } else {
      tc->error_code = DK_ERR_NOMEM;
      if(tc->app) dkapp_err_memory(tc->app,sizeof(dk_fne_t),1);
    }
  } else {
    tc->out_real = tc->out_ori;
    back = run_for_in_and_real_out(tc);
    tc->out_real = NULL;
  } 
  return back;
}
/* }}}2 */


/* is_dir {{{2 */

/* ********************************************************************* */
/* *                                                                   * */
/* *   Check whether or not the given file name points                 * */
/* *   to a directory.                                                 * */
/* *                                                                   * */
/* ********************************************************************* */

static
int
is_dir DK_P1(char *,dn)
{
  int back = 0;
  dk_stat_t *stptr;
  int ft;
  
  stptr = dkstat_open(dn);
  if(stptr) {
    ft = dkstat_filetype(stptr);
    if((ft & (~DK_FT_SYMLINK)) == DK_FT_DIR) {
      back = 1;
    }
    dkstat_close(stptr);
  } else {
    
  }
  
  return back;
}
/* }}}2 */


/* run_for_real_dir {{{2 */

/* ********************************************************************* */
/* *                                                                   * */
/* *   The exact file name is known, traverse                          * */
/* *   direcotry.                                                      * */
/* *                                                                   * */
/* ********************************************************************* */

static
int
run_for_real_dir DK_P1(tracecc_t *,tc)
{
  int back = 0;
  int ft;
  char *suffixptr, *newsuffix;
  
  if(is_dir(tc->dir_cur)) {
  tc->dir_tra = dkdir_open(tc->dir_cur);
  if(tc->dir_tra) {	
    back = 1;
    while(dkdir_next(tc->dir_tra)) {
      
      ft = dkdir_filetype(tc->dir_tra);
      if((ft & (~DK_FT_SYMLINK)) == DK_FT_REG) {
        
	if(strcmp(dkdir_get_shortname(tc->dir_tra), ".")) {
	if(strcmp(dkdir_get_shortname(tc->dir_tra), "..")) {
	  tc->in_real = dkdir_get_fullname(tc->dir_tra);
	  tc->out_real = tc->out_name_buf;
	  suffixptr = dksf_get_file_type_dot(tc->in_real);
	  if(suffixptr) {
	    newsuffix = get_new_suffix(suffixptr);
	    if(newsuffix) {
if(tc->mpl > (strlen(tc->in_real) - strlen(suffixptr) + strlen(newsuffix))) {
  strcpy(tc->out_real, tc->in_real);
  suffixptr = dksf_get_file_type_dot(tc->out_real);
  if(suffixptr) {
    strcpy(suffixptr, newsuffix);
    if(!run_for_real_in_and_real_out(tc)) {
      back = 0;
    }
  }
} else {
  back = 0;
  if(tc->app) {
    char *xxxptr[8];
    xxxptr[0] = (tc->messages)[13];
    xxxptr[1] = str_spc;
    xxxptr[2] = str_bropen;
    xxxptr[3] = tc->in_real;
    xxxptr[4] = str_brclose;
    dkapp_log_msg(tc->app, DK_LOG_LEVEL_ERROR, xxxptr, 5);
  }
}
	    }
	  }
	  tc->out_real = NULL;
	  tc->in_real = NULL;
	}
	}
      }
    }
    dkdir_close(tc->dir_tra); tc->dir_tra = NULL;
  } else {
    /* XXXXX ERROR: Failed to traverse directory */
    if(tc->app) {
      char *xxxptr[8];
      xxxptr[0] = (tc->messages)[14];
      xxxptr[1] = str_spc;
      xxxptr[2] = str_bropen;
      xxxptr[3] = tc->dir_cur;
      xxxptr[4] = str_brclose;
      dkapp_log_msg(tc->app, DK_LOG_LEVEL_ERROR, xxxptr, 5);
    }
  }
  } else {	
  }
  
  return back;
}
/* }}}2 */


/* set_ide_style_from_pref {{{2 */

/* ********************************************************************* */
/* *                                                                   * */
/* *   Retrieve IDE style from preferences.                            * */
/* *                                                                   * */
/* ********************************************************************* */

static
void
set_ide_style_from_pref DK_P2(int *,i,dk_app_t *,app)
{
  char buffer[32], *ptr;
  
  if(dkapp_get_pref(app,pk_ide_style, buffer, sizeof(buffer), 0)) {
    ptr = dkstr_start(buffer, NULL);
    if(ptr) {
      dkstr_chomp(ptr, NULL);
      *i = dkapp_ide_type(ptr);
    }
  }
  
}
/* }}}2 */


/* tracecc_retrieve_preferences {{{2 */

/* ********************************************************************* */
/* *                                                                   * */
/* *   Retrieve all preferences.                                       * */
/* *                                                                   * */
/* ********************************************************************* */

static
void
tracecc_retrieve_preferences DK_P2(dk_app_t *,app,tracecc_opt_t *,tc)
{
  int val;
  char buffer[32], *ptr;
  if(app && tc) {
    val = tc->o;
    set_mode_bit_bool(&val, app, pk_debug_enable, TRACECC_OPT_DEBUG);
    set_mode_bit_bool(&val, app, pk_debug_stdout, TRACECC_OPT_DEBST);
    set_mode_bit_bool(&val, app, pk_debug_timestamp, TRACECC_OPT_DATETIME);
    set_mode_bit_bool(&val, app, pk_debug_keyword, TRACECC_OPT_KEYWORD);
    /* set_mode_bit_bool(&val, app, pk_debug_ide, TRACECC_OPT_DEB_IDE); */
    if(dkapp_get_pref(app, pk_ide_style,buffer,sizeof(buffer),0)) {
      ptr = dkstr_start(buffer, NULL);
      if(ptr) {
        dkstr_chomp(ptr, NULL);
	tc->i = dkapp_ide_type(ptr);
      }
    }
    set_mode_bit_bool(&val, app, pk_linenumbers, TRACECC_OPT_LINENO);
    set_mode_bit_bool(&val, app, pk_make, TRACECC_OPT_MAKE);
    set_mode_bit_bool(&val, app, pk_cpp_comments, TRACECC_OPT_CPPCOMMENT);
    tc->o = val;
    set_int_from_pref(&(tc->b), app, pk_boxwidth);
    set_ide_style_from_pref(&(tc->i), app);
  }
}
/* }}}2 */


/* tracecc_save_preferences {{{2 */

/* ********************************************************************* */
/* *                                                                   * */
/* *   Save preferences.                                               * */
/* *                                                                   * */
/* ********************************************************************* */

void
tracecc_save_preferences DK_P4(dk_app_t *,app,int,mode,int,bw,int,is)
{
  char buffer[32];
  int i;
  if(app) {
    set_bool(app, pk_debug_enable, (mode & TRACECC_OPT_DEBUG));
    set_bool(app, pk_debug_stdout, (mode & TRACECC_OPT_DEBST));
    set_bool(app, pk_debug_timestamp, (mode & TRACECC_OPT_DATETIME));
    /* set_bool(app, pk_debug_ide, (mode & TRACECC_OPT_DEB_IDE)); */
    set_bool(app, pk_debug_keyword, (mode & TRACECC_OPT_KEYWORD));
    i = is;
    if(i < 0) i = 0; if(i > 4) i = 4;
    dkapp_set_pref(app, pk_ide_style, ide_styles[i]);
    set_bool(app, pk_linenumbers, (mode & TRACECC_OPT_LINENO));
    set_bool(app, pk_make, (mode & TRACECC_OPT_MAKE));
    set_bool(app, pk_cpp_comments, (mode & TRACECC_OPT_CPPCOMMENT));
    sprintf(buffer, "%d", bw);
    dkapp_set_pref(app, pk_boxwidth, buffer);
  }
}
/* }}}2 */


/* tracecc_cleanup_before_run {{{2 */

/* ********************************************************************* */
/* *                                                                   * */
/* *   cleanup before we start running _and_ after                     * */
/* *   finishing the run.                                              * */
/* *                                                                   * */
/* ********************************************************************* */

static
void
tracecc_cleanup_before_run DK_P1(tracecc_t *,tc)
{
  char *ptr;
  
  if(tc) {
    (tc->f).o = (tc->o).o; (tc->f).b = (tc->o).b; (tc->f).i = (tc->o).i;
    tc->cmbll = tc->cmbmax = 0;
    tc->icmdu = 0;
    tc->state = tc->reaction = tc->indent = tc->textflags = 0;
    tc->currentchar = '\0';
    tc->ibu = tc->ibc = 0;
    if(tc->tmpfile) {
      
      fclose(tc->tmpfile);
      
    }
    tc->tmpfile = NULL;
    
    (void)dksf_remove_file(tc->tmp_name_buf);
    
    if(tc->outfile) {
      
      fclose(tc->outfile);
      
    }
    tc->outfile = NULL;
    if(tc->infile) {
      
      fclose(tc->infile);
      
    }
    tc->infile = NULL;
    tc->in_short = NULL;
    tc->lineno = 0UL;
    tc->what = 0;
    if(tc->dir_ori) {
      
      ptr = tc->dir_ori; dk_delete(ptr);
      
    }
    tc->dir_ori = NULL;
    if(tc->fne_ori) {
      
      dkfne_close(tc->fne_ori);
      
    }
    tc->fne_ori = NULL;
    tc->dir_cur = NULL;
    if(tc->dir_tra) {
      
      dkdir_close(tc->dir_tra);
      
    }
    tc->dir_tra = NULL;
    if(tc->in_ori) {
      
      ptr = tc->in_ori; dk_delete(ptr);
      
    }
    tc->in_ori = NULL;
    if(tc->out_ori) {
      
      ptr = tc->out_ori; dk_delete(ptr);
      
    }
    tc->out_ori = NULL;
    if(tc->fne_in) {
      
      dkfne_close(tc->fne_in);
      
    }
    tc->fne_in = NULL;
    if(tc->fne_out) {
      
      dkfne_close(tc->fne_out);
      
    }
    tc->fne_out = NULL;
    tc->in_real = NULL;
    tc->out_real = NULL;
    
  }
  
}
/* }}}2 */


/* tracecc_initialize_at_open {{{2 */

/* ********************************************************************* */
/* *                                                                   * */
/* *   Initialize a tracecc_t when opening it.                         * */
/* *   Reset all variables, get message texts from                     * */
/* *   string table.                                                   * */
/* *                                                                   * */
/* ********************************************************************* */

static
int
tracecc_initialize_at_open DK_P2(tracecc_t *,tc,dk_app_t *,app)
{
  int back = 0;
  dk_string_finder_t *fptr;
  message_contents_t *mcptr;
  long mpl; unsigned long ul; size_t sz;
  char buffer[32];
  
  if(tc) {
    (tc->o).o = 0; (tc->o).b = 75; (tc->o).i = DK_APP_IDE_NONE;
    tc->error_code = 0;
    tc->messages = NULL;
    tc->app = app;
    mpl = tc->mpl = dksf_get_maxpathlen();
    sz = (size_t)4096;
    /* tc->icmdsz = 128; */
    tc->icmdsz = 512;
    if(app) {
      if(dkapp_get_pref(app,pk_input_length,buffer,sizeof(buffer), 0)) {
        if(sscanf(buffer, "%lu", &ul) == 1) {
	  sz = (size_t)ul;
	  if(sz > 16384) sz = 16384;
	  if(sz <   128) sz =   128;
	}
      }
      if(dkapp_get_pref(app,pk_command_length,buffer,sizeof(buffer), 0)) {
        if(sscanf(buffer, "%lu", &ul) == 1) {
	  tc->icmdsz = (size_t)ul;
	  if(tc->icmdsz > 1024) tc->icmdsz = 1024;
	  if(tc->icmdsz < 64)   tc->icmdsz = 64;
	}
      }
    }
    tc->ibl = sz;
    tc->inbuffer = dk_new(char,sz);
    tc->messages = dk_new(p_char_t,number_of_messages);
    tc->out_name_buf = dk_new(char,mpl);
    tc->tmp_name_buf = dk_new(char,mpl);
    tc->icmdbuf = dk_new(char,(tc->icmdsz));
    if((tc->messages) && (tc->out_name_buf) && (tc->inbuffer) && (tc->tmp_name_buf) && (tc->icmdbuf)) {
      (tc->tmp_name_buf)[0] = '\0';
      reset_strings_for_messages(tc->messages);
      if(app) {
        set_strings_by_app(tc->messages, app);
	back = dkapp_tmpnam(app, tc->tmp_name_buf, mpl);
      } else {
        set_strings_for_messages(tc->messages);
	if(dksf_get_tempdir(tc->tmp_name_buf, mpl)) {
	  char xbuffer[32];
	  sprintf(xbuffer, "%ld", (long)dksf_getpid());
	  if(mpl > (strlen(str_tracecc) + strlen(xbuffer) + 3)) {
	    size_t lgt;
	    back = 1;
	    lgt = strlen(tc->tmp_name_buf);
	    sprintf(&((tc->tmp_name_buf)[lgt]), "/%s.%s", str_tracecc, xbuffer);
	  }
	}
      }
      
    } else {
      tc->error_code = DK_ERR_NOMEM;
      if(app) {
        dkapp_err_memory(app, sizeof(p_char_t),number_of_messages);
      }
    }
    if(app) {
      tracecc_retrieve_preferences(
        app, &(tc->o)
      );
    }
    tc->what = 0;
    tc->dir_ori = NULL;
    tc->fne_ori = NULL;
    tc->dir_cur = NULL;
    tc->dir_tra = NULL;
    tc->in_ori = NULL;
    tc->out_ori = NULL;
    tc->fne_in = NULL;
    tc->fne_out = NULL;
    tc->in_real = NULL;
    tc->out_real = NULL;
    tc->newft = 0;
    tc->in_short = NULL;
    tc->infile = NULL;
    tc->outfile = NULL;
    tc->tmpfile = NULL;
    tc->lineno = 0UL;
    tc->ibu = 0;
    tc->ibc = 0;
    tc->currentchar = '\0';
    tc->textflags = 0;
    tc->indent = 0;
  } 
  return back;
}
/* }}}2 */

/* }}}1 */

/* Exported functions {{{1 */

/* tracecc_close {{{2 */

/* ********************************************************************* */
/* *                                                                   * */
/* *   Close a tracecc_t dataset, release all the resources            * */
/* *   associated to it.                                               * */
/* *                                                                   * */
/* ********************************************************************* */

void
tracecc_close DK_P1(tracecc_t *,ptr)
{
  char *p;
  
  if(ptr) {
    tracecc_cleanup_before_run(ptr);
    if(ptr->icmdbuf) {
      p = ptr->icmdbuf;
      dk_delete(p);
    }
    ptr->icmdbuf = NULL;
    if(ptr->inbuffer) {
      p = ptr->inbuffer;
      dk_delete(p);
    }
    ptr->inbuffer = NULL;
    if(ptr->tmp_name_buf) {
      p = ptr->tmp_name_buf;
      dk_delete(p);
    }
    ptr->tmp_name_buf = NULL;
    if(ptr->out_name_buf) {
      p = ptr->out_name_buf;
      dk_delete(p);
    }
    ptr->out_name_buf = NULL;
    if(ptr->messages) {
      char **cptr;
      
      cptr = ptr->messages;
      dk_delete(cptr);
    }
    DK_MEMRES(ptr,sizeof(tracecc_t));
    dk_delete(ptr);
  }
  
}
/* }}}2 */


/* tracecc_open {{{2 */

/* ********************************************************************* */
/* *                                                                   * */
/* *   Create a new tracecc_t dataset and acquire further              * */
/* *   resources needed to use it.                                     * */
/* *                                                                   * */
/* ********************************************************************* */

tracecc_t *
tracecc_open DK_P1(dk_app_t *,app)
{
  tracecc_t *back = NULL;
  
  back = dk_new(tracecc_t,1);
  if(back) {
    DK_MEMRES(back,sizeof(tracecc_t));
    if(!tracecc_initialize_at_open(back,app)) {
      dk_delete(back); back = NULL;
    }
  } 
  return back;
}
/* }}}2 */


/* get operating mode {{{2 */

/* ********************************************************************* */
/* *                                                                   * */
/* *   Return operating mode.                                          * */
/* *                                                                   * */
/* ********************************************************************* */

int
tracecc_get_mode DK_P1(tracecc_t *,tc)
{
  int back = 0;
  
  if(tc) { back = (tc->o).o; }
  
  return back;
}
/* }}}2 */


/* get box width {{{2 */

/* ********************************************************************* */
/* *                                                                   * */
/* *   Get box width for comment boxes.                                * */
/* *                                                                   * */
/* ********************************************************************* */

int
tracecc_get_boxwidth DK_P1(tracecc_t *,tc)
{
  int back = 0;
  
  if(tc) { back = (tc->o).b; }
  
  return back;
}
/* }}}2 */


/* get IDE style {{{2 */
int
tracecc_get_idestyle DK_P1(tracecc_t *,tc)
{
  int back = 0;
  
  if(tc) { back = (tc->o).i; }
  
  return back;
}
/* }}}2*/


/* set operating mode {{{2 */

/* ********************************************************************* */
/* *                                                                   * */
/* *   Set operating mode.                                             * */
/* *                                                                   * */
/* ********************************************************************* */

void
tracecc_set_mode DK_P2(tracecc_t *,tc,int,mode)
{
  
  if(tc) { (tc->o).o = mode; }
  
}
/* }}}2 */


/* set box width {{{2 */

/* ********************************************************************* */
/* *                                                                   * */
/* *   Set box width for comment boxes.                                * */
/* *                                                                   * */
/* ********************************************************************* */

void
tracecc_set_boxwidth DK_P2(tracecc_t *,tc,int,bw)
{
  
  if(tc) { (tc->o).b = bw; }
  
}
/* }}}2 */


/* set IDE style {{{2 */
void
tracecc_set_idestyle DK_P2(tracecc_t *,tc,int,bw)
{
  
  if(tc) { (tc->o).i = bw; }
  
}
/* }}}2 */


/* tracecc_dir {{{2 */

/* ********************************************************************* */
/* *                                                                   * */
/* *   Run for a directory.                                            * */
/* *                                                                   * */
/* ********************************************************************* */

int
tracecc_dir DK_P2(tracecc_t *,tc,char *,d)
{
  int back = 0; char *ptr;
  
  if(tc && d) {
    tracecc_cleanup_before_run(tc);
    tc->what = 1;
    tc->dir_ori = dkstr_dup(d);
    if(tc->dir_ori) {
      dksf_correct_fnsep(tc->dir_ori);
      if(dksf_must_expand_filename(tc->dir_ori)) {
        tc->fne_ori = dkfne_open(tc->dir_ori,0,1);
	if(tc->fne_ori) {
	  if(dkfne_next(tc->fne_ori)) {
	    back = 1;
	    do {
	      tc->dir_cur = dkfne_get_fullname(tc->fne_ori);
              if(tc->app) {
                char *xxxptr[8];
                xxxptr[0] = (tc->messages)[20];
                xxxptr[1] = tc->dir_cur;
                xxxptr[2] = (tc->messages)[21];
                dkapp_log_msg(tc->app, DK_LOG_LEVEL_PROGRESS, xxxptr, 3);
              }
	      if(!run_for_real_dir(tc)) {
	        back = 0;
	      }
              if(tc->app) {
                char *xxxptr[8];
                xxxptr[0] = (tc->messages)[23];
                xxxptr[1] = tc->dir_cur;
                xxxptr[2] = (tc->messages)[24];
                dkapp_log_msg(tc->app, DK_LOG_LEVEL_PROGRESS, xxxptr, 3);
              }
	      tc->dir_cur = NULL;
	    } while(dkfne_next(tc->fne_ori));
	  } else {
	    /* XXXX ERROR: No directory matching pattern */
            if(tc->app) {
              char *xxxptr[8];
              xxxptr[0] = (tc->messages)[15];
              xxxptr[1] = str_spc;
              xxxptr[2] = str_bropen;
              xxxptr[3] = tc->dir_ori;
              xxxptr[4] = str_brclose;
              dkapp_log_msg(tc->app, DK_LOG_LEVEL_ERROR, xxxptr, 5);
            }
	  }
	  dkfne_close(tc->fne_ori); tc->fne_ori = NULL;
	} else {
	  tc->error_code = DK_ERR_NOMEM;
	  if(tc->app) dkapp_err_memory(tc->app,sizeof(dk_fne_t),1);
	}
      } else {
        tc->dir_cur = tc->dir_ori;
        if(tc->app) {
          char *xxxptr[8];
          xxxptr[0] = (tc->messages)[20];
          xxxptr[1] = tc->dir_cur;
          xxxptr[2] = (tc->messages)[21];
          dkapp_log_msg(tc->app, DK_LOG_LEVEL_PROGRESS, xxxptr, 3);
        }
	back = run_for_real_dir(tc);
        if(tc->app) {
          char *xxxptr[8];
          xxxptr[0] = (tc->messages)[22];
          xxxptr[1] = tc->dir_cur;
          xxxptr[2] = (tc->messages)[23];
          dkapp_log_msg(tc->app, DK_LOG_LEVEL_PROGRESS, xxxptr, 3);
        }
	tc->dir_cur = NULL;
      }
      ptr = tc->dir_ori; dk_delete(ptr); tc->dir_ori = NULL;
    } else {
      tc->error_code = DK_ERR_NOMEM;
      if(tc->app) dkapp_err_memory(tc->app,1,strlen(d));
    }
    tracecc_cleanup_before_run(tc);
  }
  
  return back;
}
/* }}}2 */


/* tracecc_files {{{2 */

/* ********************************************************************* */
/* *                                                                   * */
/* *   Run for a given pair of input file and output file name.        * */
/* *                                                                   * */
/* ********************************************************************* */

int
tracecc_files DK_P3(tracecc_t *,tc,char *,i,char *,o)
{
  int back = 0; char *ptr;
  
  if(tc && i && o) {
    tracecc_cleanup_before_run(tc);
    tc->what = 2;
    tc->in_ori = dkstr_dup(i);
    if(tc->in_ori) {
      tc->out_ori = dkstr_dup(o);
      if(tc->out_ori) {
        
        dksf_correct_fnsep(tc->in_ori);
	dksf_correct_fnsep(tc->out_ori);
	back = run_for_in_and_out_file_name(tc);
	ptr = tc->out_ori; dk_delete(ptr); tc->out_ori = NULL;
      } else {
        tc->error_code = DK_ERR_NOMEM;
        if(tc->app) dkapp_err_memory(tc->app,1,strlen(o));
      }
      ptr = tc->in_ori; dk_delete(ptr); tc->in_ori = NULL;
    } else {
      tc->error_code = DK_ERR_NOMEM;
      if(tc->app) dkapp_err_memory(tc->app,1,strlen(i));
    }
    tracecc_cleanup_before_run(tc);
  }
  
  return back;
}
/* }}}2 */

/* }}}1 */


#ifndef LINT
static char sccs_id[] =
{ "@(#)traceco.c 1.42 08/25/03\t(krause) - tracecc working module." };
#endif

