/*
 * The dk libraries (libdkport + libdkc)
 * Copyright (C) 2001 - Dirk Krause
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 * In this package the copy of the GNU Lesser General Public License
 * is placed in file COPYING.
 */

#include <dk.h>
#include <dkstr.h>
#include <dkapp.h>
#include <dklic.h>
#include <dklogc.h>
#include <dkmem.h>
#include <dksf.h>

#include <stdio.h>

#if DK_HAVE_STRING_H
#include <string.h>
#endif
#if DK_HAVE_STRINGS_H
#include <strings.h>
#endif
#if DK_HAVE_STDLIB_H
#include <stdlib.h>
#endif
#if DK_HAVE_UNISTD_H
#include <unistd.h>
#endif
#if DK_HAVE_PROCESS_H
#include <process.h>
#endif


#line 49 "trana.ctr"


/* Configuration directories {{{1 */
#ifndef SYSCONFDIR
#define SYSCONFDIR "/etc"
#endif
static char sysconfdir[] = { SYSCONFDIR };
static char packagename[] = { "tracecc" };
static char programname[] = { "trana" };
#ifndef VERSNUMB
#define VERSNUMB "1.3.8"
#endif
static char the_version_number[] = { VERSNUMB } ;
/* }}}1 */

/* Options for mode {{{1 */
#define OPTION_MARKER	1
#define OPTION_POSITION	2
#define OPTION_TRACEONLY 4
#define OPTION_HELP	8
#define OPTION_VERSION	16
#define OPTION_CONFIGURE	32
#define OPTION_UNCONF		64
#define OPTION_NUMERIC		128
#define OPTION_ERROR		1024
#define OPTION_SHOWCONF		2048
/* }}}1 */

/* Static variables {{{1 */
static dk_app_t *app = NULL;
static char inputline[1024];
static char copyline[1024];
/* }}}1 */

/* Keywords for comparisons and preferences {{{1 */
static char kw_trace[] = { "trace" };
static char pk_marker[] = { "/marker" };
static char pk_position[] = { "/position" };
static char pk_traceonly[] = { "/trace-only" };
static char pk_numeric[] = { "/numeric" };
/* }}}1 */

/* message {{{1 */
static char *messages[10];
static dk_string_finder_t trana_msg[] = {
{ (char *)"/e/001", &(messages[0]), (char *)"Unknown option!" },
{ (char *)"/e/002", &(messages[1]), (char *)"Too many filenames!" },
{ (char *)"/m/001", &(messages[2]), (char *)"Current configuration:" },
{ (char *)"/m/002", &(messages[3]), (char *)"Markers for vim folding" },
{ (char *)"/m/003", &(messages[4]), (char *)"Position in source" },
{ (char *)"/m/004", &(messages[5]), (char *)"\"trace\" lines only" },
{ (char *)"/m/005", &(messages[6]), (char *)"Show indent numbers" },
{ (char *)"/m/006", &(messages[7]), (char *)" enabled" },
{ (char *)"/m/007", &(messages[8]), (char *)"disabled" },
{ NULL, NULL, NULL }
};
static char str_bropen[] = { "(" };
static char str_brclose[] = { ")" };
static char str_spc[] = { " " };
/* }}}1 */

/* run for input file and output file {{{1 */
static int
run_for_files DK_P3(FILE *,in,FILE *,out,int,mode)
{
  int cc;
  int indent, direction, istrace;
  char *p1, *p2, *p3;
  cc = 1; indent = 0;
  while(cc) {
    if(fgets(inputline,sizeof(inputline),in)) {
      strcpy(copyline, inputline);
      dkstr_chomp(copyline, NULL);
      p1 = p2 = p3 = NULL; direction  = 0; istrace = 0;
      p1 = dkstr_start(inputline, NULL);
      if(p1) {
        dkstr_chomp(p1, NULL);
        p2 = dkstr_next(p1, NULL);
	if(p2) {
	  p3 = dkstr_next(p2, NULL);
	}
      }
      if(p2) {
        if(*p2 == '+') { direction = 1; }
	if(*p2 == '-') { direction = -1; }
        if(strcmp(p2, kw_trace) == 0) {
	  istrace = 1;
	  if(p3 && (direction == 0)) {
	    if(*p3 == '+') { direction = 1; }
	    if(*p3 == '-') { direction = -1; }
	  }
	}
      }
      if(direction == 1) { indent++; }
      if(mode & OPTION_NUMERIC) {
        fprintf(out,"%-5d ", indent);
      }
      { int i; for(i = 0; i < indent; i++) { fputs("  ", out); } }
      if(istrace || (!(mode & OPTION_TRACEONLY))) {
        if(istrace) {
	  if(mode & OPTION_POSITION) {
	    fputs(p3, out); fputc(' ', out); fputs(p1, out);
	  } else {
	    fputs(p3, out);
	  }
        } else {
	  if(mode & OPTION_POSITION) {
	    fputs(copyline, out);
	  } else {
	    if(p2) {
	      fputs(p2, out);
	      if(p3) {
	        fputc(' ', out); fputs(p3, out);
	      }
	    }
	  }
        }
      }
      if(mode & OPTION_MARKER) {
        switch(direction) {
          case 1: {
	    fprintf(out, " {{{%d", indent);
	  } break;
	  case -1: {
	    fprintf(out, " }}}%d", indent);
	  } break;
        }
      }
      if(direction == -1) { indent--; if(indent < 0) indent = 0; }
      fputc('\n', out);
    } else {
      cc = 0;
    }
  }
  return 1;
}
/* }}}1 */

/* run for one inputfile, write to stdout {{{1 */
static int
run_for_input DK_P3(char *,inname,int,mode,FILE *,outputfile)
{
  int back = 0; char *copy_of_inname; dk_fne_t *fne;
  FILE *in;
  copy_of_inname = dkstr_dup(inname);
  if(copy_of_inname) {
    dksf_correct_fnsep(copy_of_inname);
    if(dksf_must_expand_filename(copy_of_inname)) {
      fne = dkfne_open(copy_of_inname, 1, 0);
      if(fne) {
        if(dkfne_next(fne)) {
	  in = dkapp_fopen(app, dkfne_get_fullname(fne), "r");
	  if(in) {
	    back = run_for_files(in,outputfile,mode);
	    fclose(in);
	  } else {
	    dkapp_err_fopenr(app, dkfne_get_fullname(fne));
	  }
	} else {
	  dkapp_err_matchfile(app, copy_of_inname);
	}
        dkfne_close(fne);
      } else {
        dkapp_err_matchfile(app, copy_of_inname);
      }
    } else {
      in = dkapp_fopen(app, inname, "r");
      if(in) {
        back = run_for_files(in,outputfile,mode);
        fclose(in);
      } else {
        dkapp_err_fopenr(app, inname);
      }
    }
    dk_delete(copy_of_inname);
  } else {
    dkapp_err_memory(app,1,strlen(inname));
  }
  return back;
  /*
  in = dkapp_fopen(app, inname, "r");
  if(in) {
    back = run_for_files(in,outputfile,mode);
    fclose(in);
  } else {
    dkapp_err_fopenr(app, inname);
  }
  return back;
  */
}
/* }}}1 */

/* run for a pair of filename {{{1 */
static int
run_for_names DK_P3(char *,inname,char *,outname,int,mode)
{
  int back = 0;
  FILE *in, *out; char *copy_of_outputname; dk_fne_t *fne;

  copy_of_outputname = dkstr_dup(outname);
  if(copy_of_outputname) {
    dksf_correct_fnsep(copy_of_outputname);
    if(dksf_must_expand_filename(copy_of_outputname)) {
      fne = dkfne_open(copy_of_outputname, 1, 0);
      if(fne) {
        if(dkfne_next(fne)) {
	  out = dkapp_fopen(app, dkfne_get_fullname(fne), "w");
	  if(out) {
	    back = run_for_input(inname, mode, out);
	    fclose(out);
	  } else {
	    dkapp_err_fopenw(app, dkfne_get_fullname(fne));
	  }
	} else {
	  dkapp_err_matchfile(app, copy_of_outputname);
	}
        dkfne_close(fne);
      } else {
        dkapp_err_matchfile(app, copy_of_outputname);
      }
    } else {
      out = dkapp_fopen(app, copy_of_outputname, "w");
      if(out) {
        back = run_for_input(inname, mode, out);
        fclose(out);
      } else {
        dkapp_err_fopenw(app, outname);
      }
    }
    dk_delete(copy_of_outputname);
  } else {
    dkapp_err_memory(app,1,strlen(outname));
  }
  return back;
  /*
  in = dkapp_fopen(app, inname, "r");
  if(in) {
    out = dkapp_fopen(app, outname, "w");
    if(out) {
      back = run_for_files(in,out,mode);
      fclose(out);
    } else {
      dkapp_err_fopenw(app, outname);
    }
    fclose(in);
  } else {
    dkapp_err_fopenr(app, inname);
  }
  return back;
  */
}
/* }}}1 */

/* check for a bool preference turned on {{{1 */
static int
check_prf_on DK_P3(int,mode,char *,pk,int,bit)
{
  int back;
  char line[32];
  back = mode;
  if(dkapp_get_pref(app, pk, line, sizeof(line), 0)) {
    if(dkstr_is_on(line)) {
      back |= bit;
    } else {
      back &= (~bit);
    }
  }
  return back;
}
/* }}}1 */

/* Retrieve default configuration {{{1 */
static int
get_default_mode DK_P0()
{
  int back = OPTION_POSITION;
  back = check_prf_on(back, pk_marker, OPTION_MARKER);
  back = check_prf_on(back, pk_position, OPTION_POSITION);
  back = check_prf_on(back, pk_traceonly, OPTION_TRACEONLY);
  back = check_prf_on(back, pk_numeric, OPTION_NUMERIC);
  return back;
}
/* }}}1 */

/* option keywords {{{1 */
static char *kw_options[] = {
  (char *)"m$arker",
  (char *)"p$osition",
  (char *)"t$race-only",
  (char *)"v$ersion",
  (char *)"h$elp",
  (char *)"c$onfigure",
  (char *)"u$nconfigure",
  (char *)"n$umeric",
  (char *)"sh$owconf",
  (char *)"r$eset",
  NULL
};
/* }}}1 */

/* strings to set configuration {{{1 */
static char str_yes[] = { "yes" };
static char str_no[] = { "no" };
/* }}}1 */

/* set bit in mode {{{1 */
static int
set_bit DK_P3(int,mode,int,bit,int,val)
{
  int back;
  back = mode;
  if(val) {
    back |= bit;
  } else {
    back &= (~bit);
  }
  return back;
}
/* }}}1 */

/* set bit in mode from option {{{1 */
static int
set_bit_from_opt DK_P3(int,mode,int,bit,char *,p)
{
  int back;
  back = mode;
  if(*p) {
    if(*p == '-') {
      back &= (~bit);
    } else {
      if(*p == '+') {
        back |= bit;
      } else {
        if(dkstr_is_bool(p)) {
	  if(dkstr_is_on(p)) {
	    back |= bit;
	  } else {
	    back &= (~bit);
	  }
	}
      }
    }
  } else {
    back |= bit;
  }
  return back;
}
/* }}}1 */

/* process command line arguments {{{1 */
static void
process_arguments DK_P5(int,c,char **,v,int *,m,char **,i,char **,o)
{
  int mode, lfd; char *p1, *p2, *p3, *p4, **lfdptr;
  char *msgptr[5];
  
  lfdptr = v; lfdptr++; lfd = 1;
  mode = *m;
  while(lfd < c) {
    p1 = *lfdptr; 
    if(*p1 == '-') { 
      p1++;
      switch(*p1) {
        case '-': {
	  p1++;
	  p2 = strchr(p1, '=');
	  if(p2) {
	    *(p2++) = '\0';
	  }
	  switch(dkstr_array_abbr(kw_options,p1,'$',1)) {
	    case 0: {
	      if(p2) {
	        if(dkstr_is_on(p2)) {
		  mode |= OPTION_MARKER;
		} else {
		  mode &= (~OPTION_MARKER);
		}
	      } else {
	        mode |= OPTION_MARKER;
	      }
	    } break;
	    case 1: {
	      if(p2) {
	        if(dkstr_is_on(p2)) {
		  mode |= OPTION_POSITION;
		} else {
		  mode &= (~OPTION_POSITION);
		}
	      } else {
	        mode |= OPTION_POSITION;
	      }
	    } break;
	    case 2: {
	      if(p2) {
	        if(dkstr_is_on(p2)) {
		  mode |= OPTION_TRACEONLY;
		} else {
		  mode &= (~OPTION_TRACEONLY);
		}
	      } else {
	        mode |= OPTION_TRACEONLY;
	      }
	    } break;
	    case 3: {
	      mode |= OPTION_VERSION;
	    } break;
	    case 4: {
	      mode |= OPTION_HELP;
	    } break;
	    case 5: {
	      mode |= OPTION_CONFIGURE;
	    } break;
	    case 6: {
	      mode |= OPTION_UNCONF;
	    } break;
	    case 7: {
	      if(p2) {
	        if(dkstr_is_on(p2)) {
		  mode |= OPTION_NUMERIC;
		} else {
		  mode &= (~OPTION_NUMERIC);
		}
	      } else {
	        mode |= OPTION_NUMERIC;
	      }
	    } break;
	    case 8: {
	      mode |= OPTION_SHOWCONF;
	    } break;
	    case 9: {
mode &= ~(OPTION_TRACEONLY|OPTION_POSITION|OPTION_MARKER|OPTION_NUMERIC);
	    } break;
	    default: {
	      p1--; p1--;
	      msgptr[0] = messages[0];
	      msgptr[1] = str_spc;
	      msgptr[2] = str_bropen;
	      msgptr[3] = p1;
	      msgptr[4] = str_brclose;
	      dkapp_log_msg(app, DK_LOG_LEVEL_ERROR, msgptr, 5);
	      mode |= OPTION_ERROR;
	    } break;
	  }
	} break;
	case 'm': {
	  /* mode |= OPTION_MARKER; */
	  mode = set_bit_from_opt(mode, OPTION_MARKER, ++p1);
	} break;
	case 'p': {
	  /* mode |= OPTION_POSITION; */
	  mode = set_bit_from_opt(mode, OPTION_POSITION, ++p1);
	} break;
	case 't': {
	  /* mode |= OPTION_TRACEONLY; */
	  mode = set_bit_from_opt(mode, OPTION_TRACEONLY, ++p1);
	} break;
	case 'v': {
	  mode |= OPTION_VERSION;
	} break;
	case 'h': {
	  mode |= OPTION_HELP;
	} break;
	case 'c': {
	  mode |= OPTION_CONFIGURE;
	} break;
	case 'u': {
	  mode |= OPTION_UNCONF;
	} break;
	case 'r': {
mode &= ~(OPTION_TRACEONLY|OPTION_MARKER|OPTION_NUMERIC);
mode |= OPTION_POSITION;
	} break;
	case 'n': {
	  /* mode |= OPTION_NUMERIC; */
	  mode = set_bit_from_opt(mode, OPTION_NUMERIC, ++p1);
	} break;
	case 'C': {
	  mode |= OPTION_SHOWCONF;
	} break;
	default: {
	  p1--;
	  msgptr[0] = messages[0];
	  msgptr[1] = str_spc;
	  msgptr[2] = str_bropen;
	  msgptr[3] = p1;
	  msgptr[4] = str_brclose;
	  dkapp_log_msg(app, DK_LOG_LEVEL_ERROR, msgptr, 5);
	  mode |= OPTION_ERROR;
	} break;
      }
    } else {
      if(*i) {
        
        if(*o) {
	  
	  msgptr[0] = messages[1];
	  msgptr[1] = str_spc;
	  msgptr[2] = str_bropen;
	  msgptr[3] = p1;
	  msgptr[4] = str_brclose;
	  dkapp_log_msg(app, DK_LOG_LEVEL_ERROR, msgptr, 5);
	  mode |= OPTION_ERROR;
	} else {
	  
	  *o = p1;
	}
      } else {
        
        *i = p1;
      }
    }
    lfd++; lfdptr++;
    
  }
  
  
  
  
  
  
  
  
  
  
  
  *m = mode; 
}
/* }}}1 */

/* set boolean preference {{{1 */
static void
set_bool DK_P2(char *,pk,int,v)
{
  
  dkapp_set_pref(app,pk,(v ? (str_yes) : (str_no)));
}
/* }}}1 */

/* help text {{{1 */
static char *help_text[] = {
(char *)"The trana program can be used to analyze trace files created by",
(char *)"tracecc.",
(char *)"",
(char *)"Usage:",
(char *)"------",
(char *)"trana [-m] [-p] [-t] [-n] [ <inputfile> [ <outputfile> ] ]",
(char *)"trana -c [-m] [-p] [-t] [-n]",
(char *)"trana -u",
(char *)"",
(char *)"Inputfile specifies the name of the input file to read, outputfile",
(char *)"is the name of the output file to create. Standard input and output",
(char *)"are used for unspecified filenames.",
(char *)"-m --marker=<bool>       create markers for folding in vim",
(char *)"-p --position=<bool>     transfer source position to output file",
(char *)"-t --trace-only=<bool>   transfer only lines containing the trace keyword",
(char *)"-n --numeric=<bool>      Prepend all lines by indent level",
(char *)"-c --configure           keep configuration as default configuration",
(char *)"                   NOTE: Command line options and default configu-",
(char *)"                         ration are merged. Use -u to reset all",
(char *)"                         options.",
(char *)"-u --unconfigure         remove default configuration",
NULL
};
/* }}}1 */

/* The real main function {{{1 */
static int 
run_main DK_P2(int,argc,char **,argv)
{
  int back = 0;
  int mode = OPTION_POSITION;
  char *inputfilename, *outputfilename, **ptr;
  
  /* retrieve string tables					{{{2 */
  dkapp_find_multi(app, trana_msg, "trana" );
  /* }}}2 */
  /* process arguments						{{{2 */
  inputfilename = outputfilename = NULL;
  mode = OPTION_POSITION;
  mode = get_default_mode();
  process_arguments(argc, argv, &mode, &inputfilename, &outputfilename);
  /* }}}2 */
  if(mode & OPTION_ERROR) {
    mode |= OPTION_HELP;
  }
  if(mode & (OPTION_HELP | OPTION_VERSION)) {
    if(!(mode & OPTION_ERROR)) { back = 1; }
    /* show version text					{{{2 */
    printf("%s %s\n", programname, the_version_number);
    printf("Copyright (C) 2002 Dipl.-Ing. D. Krause\n");
    printf("Licensed under the GNU Library General Public License, version 2.\n");
    printf("This is free software with ABSOLUTELY NO WARRANTY.\n");
    printf("\nLibraries used:\n\n");
    ptr = dklic_get();
    while(*ptr) {
      printf("%s\n", *(ptr++));
    }
    printf("\n");
    /* }}}2 */
    if(mode & OPTION_HELP) {
      /* show help text						{{{2 */
      ptr = help_text;
      while(*ptr) {
        printf("%s\n", *(ptr++));
      }
      /* }}}2 */
    }
  } else {
    if(mode & (OPTION_SHOWCONF | OPTION_CONFIGURE | OPTION_UNCONF)) {
      back = 1;
      if(mode & OPTION_UNCONF) {
        /* remove configuration data				{{{2 */
	dkapp_unconfigure(app);
	/* }}}2 */
      } else {
        if(mode & OPTION_SHOWCONF) {
	  /* show configuration					{{{2 */
	  printf("%s\n", messages[2]);
          printf("-m   %s   %s\n",
	    ((mode & OPTION_MARKER) ? messages[7] : messages[8]), messages[3]
	  );
          printf("-p   %s   %s\n",
	    ((mode & OPTION_POSITION) ? messages[7] : messages[8]), messages[4]
	  );
          printf("-t   %s   %s\n",
	    ((mode & OPTION_TRACEONLY) ? messages[7] : messages[8]), messages[5]
	  );
          printf("-n   %s   %s\n",
	    ((mode & OPTION_NUMERIC) ? messages[7] : messages[8]), messages[6]
	  );
	  /* }}}2 */
	} else {
          /* save configuration data				{{{2 */
	  set_bool(pk_marker, (mode & OPTION_MARKER));
	  set_bool(pk_position, (mode & OPTION_POSITION));
	  set_bool(pk_traceonly, (mode & OPTION_TRACEONLY));
	  set_bool(pk_numeric, (mode & OPTION_NUMERIC));
	  /* }}}2 */
	}
      }
    } else {
      
      /* really run						{{{2 */
      if(inputfilename) {
        if(outputfilename) {
	  back = run_for_names(inputfilename, outputfilename, mode);
	} else {
	  back = run_for_input(inputfilename,mode,stdout);
	}
      } else {
        run_for_files(stdin,stdout,mode);
	back = 1;
      }
      /* }}}2 */
    }
  } 
  return back;
}
/* }}}1 */

/* Check whether we have to run silently {{{1 */

/* ********************************************************************* */
/* *                                                                   * */
/* *   We set the variables where rs and rf point to if we             * */
/* *   need to run completely silently or as a filter.                 * */
/* *                                                                   * */
/* ********************************************************************* */

#line 713 "trana.ctr"

static void
silence_check DK_P4(int,argc,char **,argv,int *,rs,int *,rf)
{
  int i, filenames;
  char *ptr, **lfdptr;
  
  filenames = 0; lfdptr = argv;
  lfdptr++; i = 1;
  while(i < argc) {
    ptr = *lfdptr;
    if(*ptr == '-') {
      ptr++;
      if(*ptr == 's') {
        *rs = 1;
      }
    } else {
      filenames++;
    }
    i++; lfdptr++;
  }
  if(filenames < 2) {
    *rf = 1;
  }
  
}


/* The main function {{{1 */

/* ********************************************************************* */
/* *                                                                   * */
/* *   We only create a dk_app_t variable here and leave               * */
/* *   all things up to run_main.                                      * */
/* *                                                                   * */
/* ********************************************************************* */

#line 745 "trana.ctr"

#if DK_HAVE_PROTOTYPES
int main(int argc, char *argv[])
#else
int main(argc, argv) int argc; char *argv[];
#endif
{
  int exval;
  int rs, rf;
  
#line 754 "trana.ctr"

  
  exval = rs = rf = 0;
  silence_check(argc, argv, &rs, &rf);
  app = dkapp_open_ext1(argc, argv, packagename, sysconfdir, rs, rf);
  if(app) {
    
    exval = run_main(dkapp_get_argc(app), dkapp_get_argv(app));
    dkapp_close(app);
  } else {
    
    if(!rs) {
      fputs("ERROR: Not enough memory!\n", stderr);
      fflush(stdout);
    }
  }
  exval = (exval ? 0 : 1);
  
  
#line 772 "trana.ctr"

  exit(exval); return exval;
}
/* }}}1 */


#ifndef LINT
static char sccs_id[] = {
"@(#)trana.ctr 1.10 11/25/03 (krause)\tTrace output analyzer"
};
#endif

