/* 
 * The dk libraries (libdkport + libdkc)
 * Copyright (C) 2000, 2001 - Dirk Krause
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 * In this package the copy of the GNU Library General Public License
 * is placed in file COPYING.
 */

#include <dk.h>
#include <dkstr.h>
#include <dkmem.h>
#include <dksf.h>

#include <stdio.h>
#include <string.h>
#include <winbase.h>
#include <winreg.h>

struct _key_val_ {
  char *key;
  char *value;
} ;
typedef struct _key_val_ key_val;

static key_val keys_to_set[] = {
  { "all/all:/ui/lang",    NULL },
  { "all/all:/dir/app",    NULL },
  { "all/all:/dir/shared", NULL },
  { "all/all:/dir/tmp",    NULL },
  { "all/all:/log/file/level", "warning" },
  { "all/all:/log/file/keep",  "warning" },
  { "all/all:/log/file/name",  NULL },
  { "all/all:/log/stderr/level", "error" },
  { "all/all:/log/stdout/level", "none" },
  { "all/all:/log/stdout/codepage", NULL },
  { "all/all:/log/stderr/codepage", NULL },
  { NULL, NULL }
};

static char shared_suffix[] = { "/lib" };
static char app_suffix[] = { "/lib/$(app.name)" };
static char cp850[] = { "cp.850" };
static char deflf[] = { "/$(app.name).$(process.pid).log" };

static char *appdir = NULL;
static char *shareddir = NULL;
static char *tmpdir = NULL;
static char *buffer = NULL;
static char *logfile = NULL;


#line 65 "winsetup.ctr"


static long maxpathlen;

static void correct_string(char *str)
{
  char *ptr;
  ptr = str;
  while(*ptr) {
    if(*ptr == '\\') *ptr = '/';
    ptr++;
  }
}

static void save_the_keys(void)
{
  HKEY hkTemp;
  PHKEY phkSubkey;
  DWORD disp, dwType, sz;
  LONG retval;
  key_val *kvptr;
  int ok;
  hkTemp = 0;
  phkSubkey = &hkTemp;
  retval = RegCreateKeyExA(
    HKEY_LOCAL_MACHINE, "Software\\Dkapp",
    0, NULL, REG_OPTION_NON_VOLATILE,
    KEY_ALL_ACCESS,
    NULL,
    phkSubkey,
    &disp
  );
  if(retval == ERROR_SUCCESS) { 
    kvptr = keys_to_set;
    while(kvptr->key) { 
      if(kvptr->value) { 
	ok = 0; retval = ERROR_SUCCESS - 1L;
	sz = maxpathlen;
	phkSubkey = &hkTemp;
        retval = RegQueryValueExA(
	  hkTemp,
	  kvptr->key,
	  NULL,
	  &dwType,
	  buffer,
	  &sz
	);
	if(retval == ERROR_SUCCESS) { 
	  if(dwType == REG_SZ) { 
	    ok = 1;
	  }
	}
	if(!ok) { 
	  retval = RegSetValueExA(
	    hkTemp,
	    kvptr->key,
	    0,
	    REG_SZ,
	    kvptr->value,
	    (strlen(kvptr->value) + 1)
	  );
	  
	}
      }
      kvptr++;
    }
    RegCloseKey(hkTemp);
  }
}

int main(int argc, char *argv[])
{
  
#line 137 "winsetup.ctr"

  if(argc == 3) { 
    keys_to_set[0].value = argv[2];
    maxpathlen = dksf_get_maxpathlen();
    appdir = dk_new(char,maxpathlen);
    shareddir = dk_new(char,maxpathlen);
    tmpdir = dk_new(char,maxpathlen);
    buffer = dk_new(char,maxpathlen);
    logfile = dk_new(char,maxpathlen);
    if(appdir && shareddir && tmpdir && buffer && logfile) {
      if((strlen(argv[1]) + strlen(app_suffix)) < maxpathlen) {
        strcpy(appdir, argv[1]);
        strcat(appdir, app_suffix);
        keys_to_set[1].value = appdir;
	
      }
      if((strlen(argv[1]) + strlen(shared_suffix)) < maxpathlen) {
        strcpy(shareddir, argv[1]);
        strcat(shareddir, shared_suffix);
        keys_to_set[2].value = shareddir;
	
      }
      if(dksf_get_tempdir(tmpdir,maxpathlen)) {
	keys_to_set[3].value = tmpdir;
	if((strlen(tmpdir) + strlen(deflf)) < maxpathlen) {
	  strcpy(logfile, tmpdir);
	  strcat(logfile, deflf);
	  keys_to_set[6].value = logfile;
	  correct_string(logfile);
	}
	
      }
      if(strcmp(argv[2], "de") == 0) {
	keys_to_set[9].value = cp850;
	keys_to_set[10].value = cp850;
      }
      if(keys_to_set[1].value) correct_string(keys_to_set[1].value);
      if(keys_to_set[2].value) correct_string(keys_to_set[2].value);
      if(keys_to_set[3].value) correct_string(keys_to_set[3].value);
      save_the_keys();
    }
    if(appdir) { dk_delete(appdir); }
    appdir = NULL;
    if(shareddir) { dk_delete(shareddir); }
    shareddir = NULL;
    if(tmpdir) { dk_delete(tmpdir); }
    tmpdir = NULL;
    if(buffer) { dk_delete(buffer); }
    buffer = NULL;
    if(logfile) { dk_delete(logfile); }
    logfile = NULL;
  }
  
#line 189 "winsetup.ctr"

  exit(0); return 0;
}


#ifndef LINT
static char sccs_id[] = {
"@(#)winsetup.ctr 1.5 08/25/03\t(krause) - Initial registry setup for windows"
};
#endif

