/*
 *  KANJI Code conversion routines.
 */

#ifndef PTEXENC_PTEXENC_H
#define PTEXENC_PTEXENC_H

#include <kpathsea/types.h>

#include <ptexenc/ptxconfig.h>

extern const char *ptexenc_version_string;

#if defined(WIN32)
extern int sjisterminal;
extern FILE *Poptr;
extern int infile_enc_auto;
#endif

#define KANJI_OPTS "{jis|euc|sjis|utf8}"

extern void enable_UPTEX (int enable);
extern void set_prior_file_enc(void);
extern const_string get_enc_string (void);
extern int    set_enc_string (const_string file, const_string inter);
#define getencstring  get_enc_string
#define setencstring  set_enc_string

/* decide if internal Kanji encode is SJIS/UPTEX or not */
extern int  is_internalEUC(void);
extern int  is_internalSJIS (void);
extern int  is_internalUPTEX (void);
#define isinternalEUC   is_internalEUC
#define isinternalSJIS  is_internalSJIS
#define isinternalUPTEX is_internalUPTEX

/* check char range */
extern int ismultichr(int length, int nth, int c);
extern int iskanji1(int c);
extern int iskanji2(int c);

/* internal (EUC/SJIS/UPTEX) from/to buffer (EUC/SJIS/UTF-8) code conversion */
extern int multistrlen(unsigned char * s, int len, int pos);
extern int multibytelen(int first_byte);
extern long fromBUFF(unsigned char * s, int len, int pos);
extern long toBUFF(long inter);

/* internal (EUC/SJIS/UPTEX) from/to DVI (JIS/UCS) code conversion */
extern long toDVI (long kcode);
extern long fromDVI (long kcode);

/* JIS/EUC/SJIS/KUTN/UCS to internal (EUC/SJIS/UPTEX) code conversion */
/* (only for \euc primitive, etc.) */
extern long toJIS(long kcode);
extern long fromJIS(long jis);
extern long fromEUC(long euc);
extern long fromSJIS(long sjis);
extern long fromKUTEN(long kuten);
extern long fromUCS(long ucs);
extern long toUCS(long kcode);

/* fputs/putc with encoding conversion */
extern int putc2(int c, FILE *fp);
extern int fputs2(const char *s, FILE *fp);

extern long input_line2(FILE *fp, unsigned char * buff, long pos,
                                    const long buffsize, int *lastchar);

#ifdef WIN32
extern void clear_infile_enc(FILE *fp);
#else
/* open/close through nkf */
extern void nkf_disable(void);
extern FILE *nkf_open(const char *path, const char *mode);
extern int nkf_close(FILE *fp);
#endif

#endif /* PTEXENC_PTEXENC_H */
