/*
 * make vf which simulates vert-ASCII-dvi by NTT-dvi
 *   with -tdm option, tdmjsyNN is assumed to be a font with rotated glyph
 * mkp2njvf dm 10 tmin10 /.../dmjka10.tfm /.../tmin10.tfm
 *          vchs [-tdm] vchr [VFDIR]
 *	tmin10.vf
 */
#include	<stdio.h>
#include	"defs.h"
#include	"vfcodes.h"
#include	"bifont.h"

#define	UNIT	1048576		/* 1 fixpoint real = 2^20 */

/* hard coded data */

struct specchar {
    int ct, gl;
} scdata[] = {
    {3, 112608},
    {6, 504478},
    {7, 252239},
    {0, 0},
};

#define	DEPTHRATIO	165565
int rdisp = UNIT-DEPTHRATIO;
int ddisp = UNIT/2;
int vdisp = (UNIT/2)-DEPTHRATIO;

/* end of hard coded data */

#define	LEN	256

#define	VCHS	24
#define	VCHR	29
struct vcharshift {
    int cc;
    int r, d;
} vcharshift[VCHS+1];
struct vcharrot {
    int cc;
} vcharrot[VCHR+1];

int an_ratio;		/* (ASC/NTT)*2^20 */
int nttcharwid, asccharwid;
char nttfont[64], *ntt2;
char *ascfont;
int vfds;
BOOLEAN userotfont;
char *rotfont;
int vff;
char *vfdir;
FILE *vf;
int lastku;

/* set by readtfm */
extern char param[];
/* set by readjfm */
extern char header[];
extern char width[];
extern int nt, lh, bc, ec, nw, ds;
extern struct jfmfntinfo *jfmfi;

main(ac, av)
int ac;
char *av[];
{
    FILE *fp;
    int vac;
    char name[64];

    lastku = 84;
    strcpy(nttfont, av[1]);
    strcat(nttfont, ntt2 = av[2]);
    vfds = atoi(ntt2)<<20;
    ascfont = av[3];
    if ((fp = fopen(av[4], "rb")) == NULL) {
	fprintf(stderr, "cannot read %s\n", av[4]);
	exit(1);
    }
    readtfm(fp);
    nttcharwid = makeuint(param+4*4, 4);
    if ((fp = fopen(av[5], "rb")) == NULL) {
	fprintf(stderr, "cannot read %s\n", av[5]);
	exit(1);
    }
    if (!readjfm(fp)) {
	fprintf(stderr, "%s is not jfm\n", av[5]);
	exit(1);
    }
    asccharwid = jfmfi->ch[0].tfmw;
    an_ratio = divsp(asccharwid, nttcharwid);

    readvchshift(av[6]);
    if (av[7][0] == '-') {
	userotfont = TRUE;
	rotfont = av[7]+1;
	vac = 8;
    } else {
	userotfont = FALSE;
	vac = 7;
    }
    readvchrot(av[vac]);

    vfdir = (ac > vac+1) ? av[vac+1] : "";
    sprintf(name, "%s%s.vf", vfdir, ascfont);
    if ((vf = fopen(name, "wb")) == NULL) {
	fprintf(stderr, "cannot write %s\n", name);
	exit(1);
    }
    ascvf();
    fclose(vf);

    exit(0);
}

#define	ESC	'\033'

getvch(b, e)
char *b, **e;
{
    int cc = 0;

    if (*b&0x80) {
	cc = ((*b&0x7f)<<8)|(*(b+1)&0x7f);
	b += 2;
    } else if (*b == ESC) {
	if (*(b+1) == '$') {
	    b += ('@' <= *(b+2) && *(b+2) <= 'B') ? 3 : 4;
	    cc = (*b<<8)|*(b+1);
	    b += 2;
	}
    }
    if (e != NULL) {
	for (; *b != ' '; b++)
	    ;
	*e = b;
    }
    return cc;
}

readvchshift(fn)
char *fn;
{
    FILE *f;
    char buf[LEN], *p;
    int i;

    if ((f = fopen(fn, "r")) == NULL) {
	fprintf(stderr, "cannot open %s\n", fn);
	exit(1);
    }
    for (i = 0; fgets(buf, LEN, f) != NULL && i < VCHS; i++) {
	(vcharshift+i)->cc = getvch(buf, &p);
	(vcharshift+i)->d = -(strtol(p, &p, 10)*UNIT)/1000;
	(vcharshift+i)->r = -(strtol(p, &p, 10)*UNIT)/1000;
    }
    (vcharshift+i)->cc = 0;
    fclose(f);
}

readvchrot(fn)
char *fn;
{
    FILE *f;
    char buf[LEN], *p;
    int i;

    if ((f = fopen(fn, "r")) == NULL) {
	fprintf(stderr, "cannot open %s\n", fn);
	exit(1);
    }
    for (i = 0; fgets(buf, LEN, f) != NULL && i < VCHR; i++) {
	(vcharrot+i)->cc = getvch(buf, NULL);
    }
    (vcharrot+i)->cc = 0;
    fclose(f);
}

#define	makejis(k,t)	((k+0x20)*256+t+0x20)

ascvf()
{
    int f, c, ku, ten;
    char jsubname[64];
    long len;
    int i;

    putuint(vf, VF_PRE, 1);
    putuint(vf, VF_ID, 1);
    putuint(vf, 0, 1);	/* no comment */
    putbytes(vf, header, 2*4);	/* check sum & design size */

    fontdef(1, nttfont);
    if (userotfont)
	fontdef(vff = 2, rotfont);

    vfchar0();
    for (ku = 1; ku <= lastku; ku++)
	for (ten = 1; ten <= 94; ten++)
	    vfchar(makejis(ku, ten));

    len = ftell(vf);
    for (i = 0; i < 4 - len%4; i++)
	putuint(vf, VF_POST, 1);
}

fontdef(f, fn)
int f;
char *fn;
{
    putuint(vf, FNT_DEF1, 1);
    putuint(vf, f, 1);
    putuint(vf, 0, 4);
    putuint(vf, an_ratio, 4);
    putuint(vf, vfds, 4);
    putuint(vf, 0, 1);
    putuint(vf, strlen(fn), 1);
    putbytes(vf, fn, strlen(fn));
}

vfchar0()
{
    int pl;
    int ct;
    int rn, dn;
    byte rb[sizeof(int)], db[sizeof(int)];

    pl = 5;	/* DIR 3, SET2 jis */
    rn = inttob(rb, rdisp, TRUE);
    pl += 1+rn;
    dn = inttob(db, ddisp, TRUE);
    pl += 1+dn;

    putuint(vf, VF_LONG_CHAR, 1);
    putuint(vf, pl, 4);
    putuint(vf, 0, 4);	/* char 0 */
    putuint(vf, jfmfi->ch[0].tfmw, 4);
    putuint(vf, RIGHT1+rn-1, 1);
    putbytes(vf, rb, rn);
    putuint(vf, DOWN1+dn-1, 1);
    putbytes(vf, db, dn);
    putuint(vf, DIR, 1);
    putuint(vf, 3, 1);
    putuint(vf, SET2, 1);
    putuint(vf, 0x2121, 2);
}

vfchar(jis)
int jis;
{
    int f = 1;
    int pl;
    int ct, n;
    int rd, dd;
    int rn, dn;
    byte rb[sizeof(int)], db[sizeof(int)];
    struct specchar *sd;
    struct vcharshift *vs;
    struct vcharrot *vr;

    ct = getctype(jis, jfmfi);
    pl = 3;	/* SET2 jis */
    for (sd = scdata; sd->ct != 0; sd++)
	if (sd->ct == ct)
	    break;
    for (vs = vcharshift; vs->cc != 0; vs++)
	if (vs->cc == jis)
	    break;
    for (vr = vcharrot; vr->cc != 0; vr++)
	if (vr->cc == jis)
	    break;
    if (vr->cc == 0 && vs->cc == 0)
	return;
    if (vr->cc == 0) {
	pl += 2;	/* DIR 3 */
	rd = rdisp - sd->gl + vs->r;
    } else if (userotfont) {
	pl += 3;	/* FONT_00+vff, DIR 3 */
	f = vff;
	rd = - sd->gl;
    } else
	rd = - sd->gl + vs->r;
    if (rd != 0) {
	rn = inttob(rb, rd, TRUE);
	pl += 1+rn;
    }
    if (vr->cc == 0)
	dd = ddisp + vs->d;
    else if (userotfont)
	dd = vdisp;
    else
	dd = vdisp + vs->d;
    if (dd != 0) {
	dn = inttob(db, dd, TRUE);
	pl += 1+dn;
    }
    
    putuint(vf, VF_LONG_CHAR, 1);
    putuint(vf, pl, 4);
    putuint(vf, jis, 4);
    putuint(vf, jfmfi->ch[ct].tfmw, 4);
    if (rd != 0) {
	putuint(vf, RIGHT1+rn-1, 1);
	putbytes(vf, rb, rn);
    }
    if (dd != 0) {
	putuint(vf, DOWN1+dn-1, 1);
	putbytes(vf, db, dn);
    }
    if (f != 1)
	putuint(vf, FONT_00+f, 1);
    if (vr->cc == 0 || userotfont) {
	putuint(vf, DIR, 1);
	putuint(vf, 3, 1);
    }
    putuint(vf, SET2, 1);
    putuint(vf, jis, 2);
}
