/* -*- Mode: C++; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
/* ***** BEGIN LICENSE BLOCK *****
 * Version: NPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Netscape Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * http://www.mozilla.org/NPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is sila.mozdev.org code.
 *
 * The Initial Developer of the Original Code is 
 * Keith Stribley.
 * Portions created by the Initial Developer are Copyright (C) 2005
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the NPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the NPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

#ifndef GR_IPAINTER_FACTORY_H
#define GR_IPAINTER_FACTORY_H

#include "nsColor.h"
#include <graphite/SegmentPainter.h>


namespace gr {
  class SegmentPainter;
  class Segment;
}

/**
* An interface to create platform specific segment painter objects
* This interface allows more of the application's painting code to be 
* platform independent.
*/
class nsISegPainterFactory
{
public:
  /**
  * Creates a painter object using the appropriate platform specific 
  * implementation. The caller is responsible for deleting the painter.
  * @param aSeg a graphite segment that will be passed into the Painter
  * @return pointer to a SegmentPainter object
  */
  virtual gr::SegmentPainter * getPainter(gr::Segment * aSeg) = 0;
  virtual ~nsISegPainterFactory() {}
};



#endif
