/*
 * Copyright (C) 2007 Keith Stribley
 * 
 * Distributable under the terms of either the Common Public License or the
 * GNU Lesser General Public License, as specified in the LICENSING.txt file.
 *
 */
package org.sil.graphite;

/**
 * Provides access to the Graphite GlyphInfo structure
 * It uses arrays of each type for efficiency over the JNI.
 */
public class GlyphInfo
{
    
    boolean [] bools = new boolean [GraphiteFont.GI_BOOL_MAX];
    int [] ints = new int[GraphiteFont.GI_INT_MAX];
    float [] floats = new float[GraphiteFont.GI_FLOAT_MAX];
    
    protected GlyphInfo() {}
    
    public boolean isSpace() {return bools[GraphiteFont.GI_BOOL_IS_SPACE]; }
    public boolean insertBefore() {return bools[GraphiteFont.GI_BOOL_INSERT_BEFORE]; }
    public boolean isAttached() {return bools[GraphiteFont.GI_BOOL_IS_ATTACHED]; }
    public boolean erroneous() {return bools[GraphiteFont.GI_BOOL_ERRONEOUS]; }
    
    public int glyphID() {return ints[GraphiteFont.GI_INT_GLYPH_ID]; }
    public int pseudoGlyphID() {return ints[GraphiteFont.GI_INT_PSEUDO_GLYPH_ID]; }
    public int logicalIndex() {return ints[GraphiteFont.GI_INT_LOGICAL_INDEX]; }
    public int firstChar() {return ints[GraphiteFont.GI_INT_FIRST_CHAR]; }
    public int lastChar() {return ints[GraphiteFont.GI_INT_LAST_CHAR]; }
    public int directionality() {return ints[GraphiteFont.GI_INT_DIRECTIONALITY]; }
    public int directionLevel() {return ints[GraphiteFont.GI_INT_DIRECTION_LEVEL]; }
    public int breakWeight() {return ints[GraphiteFont.GI_INT_BREAK_WEIGHT]; }
    
    public float origin() {return floats[GraphiteFont.GI_FLOAT_ORIGIN]; }
    public float advanceWidth() {return floats[GraphiteFont.GI_FLOAT_ADVANCE_WIDTH]; }
    public float advanceHeight() {return floats[GraphiteFont.GI_FLOAT_ADVANCE_HEIGHT]; }
    public float yOffset() {return floats[GraphiteFont.GI_FLOAT_Y_OFFSET]; }
    public float bbLeft() {return floats[GraphiteFont.GI_FLOAT_BB_LEFT]; }
    public float bbRight() {return floats[GraphiteFont.GI_FLOAT_BB_RIGHT]; }
    public float bbTop() {return floats[GraphiteFont.GI_FLOAT_BB_TOP]; }
    public float bbBottom() {return floats[GraphiteFont.GI_FLOAT_BB_BOTTOM]; }
}
