/*
 * Copyright (C) 2005,2007 Keith Stribley
 * 
 * Distributable under the terms of either the Common Public License or the
 * GNU Lesser General Public License, as specified in the LICENSING.txt file.
 *
 */

package org.sil.graphite.awt;
import java.text.CharacterIterator;
import java.text.DecimalFormat;
import java.awt.Font;
import java.awt.Shape;
import java.awt.font.*;
import java.awt.geom.AffineTransform;
import java.util.Map;
import java.text.AttributedCharacterIterator;
import sun.font.TrueTypeFont;
import sun.font.Font2D;
import sun.font.FontManager;
import java.awt.font.GlyphVector;
import java.awt.font.FontRenderContext;

import org.sil.graphite.GraphiteFont;

import sun.font.StandardGlyphVector;

//import sun.font.GraphiteTTF;


/**
 *
 * @author keith
 */
public class GraphiteAwtFont extends java.awt.Font
{
  /**
     * 
     */
    private static final long serialVersionUID = -368561389268636886L;
    
    // Java2D always defaults to 72 dpi
    public final static int JAVA_AWT_DPI = 72;
  
  
  private static final String FONT_NAME_PREFIX = "fileName=";
  
  
  private boolean checkedGraphite = false;
  private boolean isGraphite = false;
  private GraphiteFont jniFont = null;
  private boolean isRtl = false;
  
  /** Creates a new instance of GraphiteFont */
  public GraphiteAwtFont(Font ttf)
  {
    super(ttf.getName(), ttf.getStyle(), ttf.getSize());
  }
  public boolean isGraphiteFont()
  {
    if (!checkedGraphite)
    {
      
      isGraphite = false;
      Font2D font2D = FontManager.findFont2D(this.getName(), this.getStyle(), FontManager.NO_FALLBACK);
      TrueTypeFont ttf = null;
      if (font2D instanceof TrueTypeFont) ttf = (TrueTypeFont)font2D;
      if (ttf != null)
      {      
        //assertPrint("Font is truetype:" + ttf.toString());
        String fontDesc = ttf.toString();
        int fontNameIndex = fontDesc.indexOf(FONT_NAME_PREFIX) + 
            FONT_NAME_PREFIX.length();
        String fontFileName = fontDesc.substring(fontNameIndex);
        jniFont = new GraphiteFont(fontFileName, this.getSize(), JAVA_AWT_DPI);
        jniFont.createGraphiteFont(fontFileName, this.getSize(), JAVA_AWT_DPI, 400, 0);
        if (jniFont != null)
        {
          isGraphite = jniFont.isGraphiteFont();
          checkedGraphite = true;
          if (isGraphite)
          {
            //assertPrint(fontFileName + " is a graphite font");
            GraphiteFontManager.addFont(this);
          }
        }
        //assertPrint("table size " + GraphiteTTF.getTTFTableSize(ttf, ttf.GSUBTag));
      }
      else
      {
        checkedGraphite = true;
      }
    }
    return isGraphite;
  }
  
  
  public LineMetrics 	getLineMetrics(char[] chars, int beginIndex, int limit, FontRenderContext frc)
  {
    //assertPrint("getLineMetrics(char[] chars, int beginIndex, int limit, FontRenderContext frc)");
    return super.getLineMetrics(chars, beginIndex, limit, frc);
  }
  public LineMetrics 	getLineMetrics(CharacterIterator ci, int beginIndex, int limit, FontRenderContext frc)
  {
    //assertPrint("CharacterIterator ci, int beginIndex, int limit, FontRenderContext frc)");
    return super.getLineMetrics(ci, beginIndex, limit, frc);
  }
  public LineMetrics 	getLineMetrics(String str, FontRenderContext frc)
  {
    //assertPrint("getLineMetrics(String str, FontRenderContext frc)");
    return super.getLineMetrics(str, frc);
  }
  public LineMetrics 	getLineMetrics(String str, int beginIndex, int limit, FontRenderContext frc) 
  {
    //assertPrint("getLineMetrics(String str, int beginIndex, int limit, FontRenderContext frc");
    return super.getLineMetrics(str, beginIndex, limit, frc);
  }
  
  
  public synchronized GlyphVector layoutGlyphVector(FontRenderContext frc, char[] text, int start, int limit, int flags) 
  {
   GlyphVector gv = null;//super.layoutGlyphVector(frc, text, start, limit, flags);
   if (isGraphiteFont())
   {
       int [] gid = null;
       int [] charIndices = null;
       float [] pos = null;
       String subText = new String(text, start, limit - start);
     
       //StandardGlyphVector sgv = (StandardGlyphVector) gv;   
       int glyphCount = jniFont.countGlyphs(subText.toCharArray(), 0, 
                       subText.length(), isRtl);
       //assertPrint("layoutGlyphVector" + glyphCount);
       gid = new int[glyphCount];
       charIndices = new int[glyphCount + 1];
       charIndices[glyphCount] = subText.length();
       pos = new float[2 * glyphCount + 2];
       if (jniFont.getGlyphArray(text, start, limit, gid, pos, charIndices, isRtl))
       {
           //sgv.initGlyphVector(this, frc, gid, pos, charIndices, 
           gv = new StandardGlyphVector(this, frc, gid, pos, charIndices, 
                          GlyphVector.FLAG_COMPLEX_GLYPHS |
                          GlyphVector.FLAG_HAS_POSITION_ADJUSTMENTS
                        );
           
         DecimalFormat format = new DecimalFormat("####.#");
         StringBuffer positions = new StringBuffer();

         double prevXAfter = 0;
         for (int i = 0; i < pos.length; i+= 2)
         {
//             int minIndex = charIndices[i/2];
//             int j = i/2;
//             // combine reordered charIndices into one cluster
//             while (++j < charIndices.length && charIndices[j] < minIndex)
//             {
//                 minIndex = charIndices[j];
//             }
//             charIndices[i/2] = minIndex;
             positions.append(format.format(pos[i]));
             positions.append(',');
             positions.append(format.format(pos[i + 1]));
             positions.append('[');
             positions.append(charIndices[i/2]);
             positions.append(']');
             int j = i/2;
             if (j < glyphCount)
             {
                 positions.append(format.format(
                         gv.getGlyphVisualBounds(j).getBounds2D().getMinX()));
                 Shape sglb = gv.getGlyphLogicalBounds(j);
                 Shape sgbb = gv.getGlyphVisualBounds(j);
                 GlyphMetrics gm = gv.getGlyphMetrics(j);
                 AffineTransform at = gv.getGlyphTransform(j);
                 if (prevXAfter > gv.getGlyphPosition(j).getX())
                 {
                     assertPrint("Glyph " + j + " not after advance" + 
                             prevXAfter + "/" + gv.getGlyphPosition(j).getX());
                 }                 
                 if (at != null)
                 {
                     positions.append("at:"); positions.append(at.getTranslateX());
                 }
                 GlyphJustificationInfo gji = gv.getGlyphJustificationInfo(j);
                 if (gji != null)
                 {
                     positions.append("just:");
                     positions.append(gji.growLeftLimit);
                 }
                 prevXAfter = gv.getGlyphPosition(j).getX() + gm.getAdvanceX();
                 
             }
             positions.append('\t');
         }
         //assertPrint("Reinit glyph vector with graphite data");
         //int param = 0; // GlyphVector.FLAG_COMPLEX_GLYPHS
         //flags |= GlyphVector.FLAG_HAS_POSITION_ADJUSTMENTS;
         //flags |= GlyphVector.FLAG_COMPLEX_GLYPHS;
         assertPrint(positions.toString());
         positions.delete(0, positions.length());
         
       }
       else
       {
         assertPrint("failed to get int array");
       }
       
     
   }
   return gv;
  }
  
  
  public GlyphVector 	createGlyphVector(FontRenderContext frc, char[] chars)
  {
    assertPrint("createGlyphVector(String str, FontRenderContext frc)");
    return super.createGlyphVector(frc, chars);
  }
  public GlyphVector 	createGlyphVector(FontRenderContext frc, CharacterIterator ci)
  {
    assertPrint("createGlyphVector(String str, FontRenderContext frc)");
    return super.createGlyphVector(frc, ci);
  }
  public GlyphVector 	createGlyphVector(FontRenderContext frc, int[] glyphCodes)
  {
    assertPrint("createGlyphVector(String str, FontRenderContext frc)");
    return super.createGlyphVector(frc, glyphCodes);
  }
  public GlyphVector 	createGlyphVector(FontRenderContext frc, String str)
  {
    assertPrint("createGlyphVector(String str, FontRenderContext frc)");
    
    return super.createGlyphVector(frc, str);
  }
  
//  private Font2D getFont2D()
//  {
//    assertPrint("getFont2D");
//    return super.getFont2D();
//  }
  public Font deriveFont(int style, float size)  
  {
    return super.deriveFont(style,size);
  }
  public Font deriveFont(int style, AffineTransform trans)
  {
    return super.deriveFont(style,trans);
  }
  public Font deriveFont(float size){
    return super.deriveFont(size);
  }
  public Font deriveFont(AffineTransform trans){
    return super.deriveFont(trans);
  }
    public Font deriveFont(int style){
    return super.deriveFont(style);
  }
  public Font deriveFont(Map<? extends  AttributedCharacterIterator.Attribute, ?> attributes){
    return super.deriveFont(attributes);
  }
  
  public int getNumGlyphs()
  {
    assertPrint("gr");
    int ng = super.getNumGlyphs();
    return ng;
  }
  
  public boolean canDisplay(char code)
  {
    assertPrint("gr");
    return super.canDisplay(code);    
  }
  public boolean canDisplay(int code)
  {
    assertPrint("gr");    
    return super.canDisplay(code);      
  }
  /**
   * Delete any native objects that may be allocated.
   */
  public void destroy()
  {
    if (jniFont != null)
    {
      jniFont.dispose();
      checkedGraphite = false;
      jniFont = null;
    }
  }
  
  
  public float getAscent() 
  {
    assert(jniFont != null);
    return jniFont.getAscent();
  }
  
  public float getDescent() 
  {
    assert(jniFont != null);
    return jniFont.getDescent();
  }
  
  public float getHeight()
  {
      assert(jniFont != null);
      return jniFont.getHeight();
  }
  /**
   * Delete any native objects that may be allocated.
   */
  public void dispose()
  {
    destroy();
    assertPrint("GraphiteFont " + getName() + " disposed");
  }
  /**
   * Delete any native objects that may be allocated.
   */
  protected void finalize() throws Throwable 
  {
     dispose();
     super.finalize();
  }
  /**
   * Sets whether segments should be created right to left
   * @param rtl - false for left to right, true for right to left
   */
  public void setRightToLeft(boolean rtl)
  {
      this.isRtl = rtl;
  }
  // for debugging
  protected void assertPrint(String s) 
  {
      assert(debugPrint(s));
  }
  
  protected boolean debugPrint(String s)
  {
      System.out.println(s);
      return true;
  }
}
