/*
 * Copyright (C) 2005,2007 Keith Stribley
 * 
 * Distributable under the terms of either the Common Public License or the
 * GNU Lesser General Public License, as specified in the LICENSING.txt file.
 *
 */

package org.sil.graphite.awt;

import java.awt.Font;
import java.util.Iterator;
import java.util.HashMap;

import org.sil.graphite.GraphiteFont;

/**
 * This manager maintains a 3 level set of hashes of GraphiteFont objects 
 * Fonts are cached first by face name, then by style (bold/italic) and finally
 * by size.
 * @author keith
 */
public class GraphiteFontManager
{
  private HashMap<String,HashMap<Integer,
      HashMap<Float,GraphiteAwtFont>>> fontNameMap = null;
  private static GraphiteFontManager instance = null;
  /** Creates a new instance of GraphiteFontManager */
  public GraphiteFontManager()
  {
    fontNameMap = new HashMap<String, HashMap<Integer,
                                              HashMap<Float,GraphiteAwtFont>>>();
  }
  protected static void getInstance()
  {
    if (instance == null) instance = new GraphiteFontManager();
  }
  public static boolean contains(Font font)
  {
    boolean hasFont = false;
    getInstance();
    if (instance != null)
    {
      if (instance.fontNameMap.containsKey(font.getFontName()))
      {
        HashMap<Integer,HashMap<Float,GraphiteAwtFont>> fontStyleMap = 
          instance.fontNameMap.get(font.getFontName());
        if (fontStyleMap.containsKey(font.getStyle()))
        {
            HashMap<Float,GraphiteAwtFont> fontSizeMap = 
                fontStyleMap.get(font.getStyle());
            hasFont = fontSizeMap.containsKey(font.getSize2D());
        }
      }
    }
    return hasFont;
  }
  public static boolean isGraphite(Font font)
  {
    boolean hasFont = false;
    boolean isGraphite = false;
    getInstance();
    if (instance != null)
    {
      if (instance.fontNameMap.containsKey(font.getFontName()))
      {
        HashMap<Integer,HashMap<Float,GraphiteAwtFont>> fontStyleMap = 
          instance.fontNameMap.get(font.getFontName());
        hasFont = fontStyleMap.containsKey(font.getStyle());
        if (fontStyleMap.get(font.getStyle()) != null)
          isGraphite = true;
      }
      if (!hasFont)
      {
        GraphiteAwtFont test = new GraphiteAwtFont(font);
        isGraphite = test.isGraphiteFont(); // calls addFont internally
      }
    }
    return isGraphite;
  }
  public static GraphiteAwtFont getFont(Font font)
  {
    GraphiteAwtFont theFont = null;
    if (instance != null && instance.fontNameMap.containsKey(font.getFontName()))
    {
      
      HashMap<Integer,HashMap<Float,GraphiteAwtFont>> fontStyleMap = 
        instance.fontNameMap.get(font.getFontName());
      if (fontStyleMap.containsKey(font.getStyle()))
      {
          HashMap<Float,GraphiteAwtFont> sizeMap = fontStyleMap.get(font.getStyle());
          if (sizeMap.containsKey(font.getSize2D()))
          {
            Font gFont = sizeMap.get(font.getSize2D());
            if (gFont instanceof GraphiteAwtFont)
            {
              theFont = (GraphiteAwtFont)gFont;
            }
          }
      }      
    }
    if (theFont == null)
    {
        GraphiteAwtFont test = new GraphiteAwtFont(font);
        if (test.isGraphiteFont())
        {
            addFont(test);
            theFont = test;
        }
    }
    return theFont;
  }
  protected static void addFont(GraphiteAwtFont font)
  {
    getInstance();
    if (instance == null) return;
    GraphiteAwtFont realGrFont = null;
    if (!contains(font))
    {
      if (font.isGraphiteFont()) realGrFont = font;
      HashMap<Integer,HashMap<Float,GraphiteAwtFont>> fontStyleMap = null;
      if (instance.fontNameMap.containsKey(font.getFontName()))
      {
        fontStyleMap = instance.fontNameMap.get(font.getFontName());
        HashMap<Float,GraphiteAwtFont> sizeMap = null;
        if (fontStyleMap.containsKey(font.getStyle()))
        {
            sizeMap = fontStyleMap.get(font.getStyle());
        }
        else
        {
            sizeMap = new HashMap<Float,GraphiteAwtFont>();
            fontStyleMap.put(font.getStyle(), sizeMap);
        }
        sizeMap.put(font.getSize2D(), realGrFont);
      }
      else
      {
        fontStyleMap = new HashMap<Integer, HashMap<Float, GraphiteAwtFont>>();
        HashMap<Float,GraphiteAwtFont> sizeMap = new HashMap<Float,GraphiteAwtFont>();
        fontStyleMap.put(font.getStyle(), sizeMap);
        sizeMap.put(font.getSize2D(), realGrFont);
        instance.fontNameMap.put(font.getFontName(),fontStyleMap);
      }
    }
  }
  public static void dispose()
  {
    if (instance != null && instance.fontNameMap != null)
    {
      Iterator <String>i = instance.fontNameMap.keySet().iterator();
      while (i.hasNext())
      {
        HashMap<Integer,HashMap<Float,GraphiteAwtFont>> fontStyleMap = 
            instance.fontNameMap.get(i.next());
        Iterator <HashMap<Float,GraphiteAwtFont>> j = fontStyleMap.values().iterator();
        while (j.hasNext())
        {
            HashMap<Float,GraphiteAwtFont> sizeMap = j.next();
            Iterator <GraphiteAwtFont> k = sizeMap.values().iterator();
            while (k.hasNext())
            {
                k.next().dispose();
            }
        }
        fontStyleMap.clear();
      }
      instance.fontNameMap.clear();
    }
    instance = null;
  }
  
  protected void finalize() throws Throwable 
  {
      dispose();
  }
}
