/*
 * Copyright (C) 2007 Keith Stribley
 * 
 * Distributable under the terms of either the Common Public License or the
 * GNU Lesser General Public License, as specified in the LICENSING.txt file.
 *
 */

package org.sil.graphite.swing;

import java.awt.Component;
import java.awt.Font;
import java.awt.Graphics;
import java.awt.Graphics2D;

import javax.swing.JComponent;
import javax.swing.plaf.ComponentUI;
import javax.swing.plaf.TextUI;
import javax.swing.text.JTextComponent;
import javax.swing.text.View;
import javax.swing.text.PlainView;
import javax.swing.text.GlyphView;

import org.sil.graphite.GraphiteFont;
import org.sil.graphite.awt.GraphiteAwtFont;
import org.sil.graphite.awt.GraphiteFontManager;
import org.sil.graphite.swing.GraphiteGlyphPainter;
/**
 *
 * @author keith
 */
public class PaintHelper
{
    /** Creates a new instance of PaintHelper */
    static public boolean preparePaint(JComponent component, Graphics g)
    {
        if (component instanceof JTextComponent &&
            g instanceof Graphics2D)
        {
            JTextComponent textComponent = (JTextComponent)component;
            return preparePaint(textComponent, (Graphics2D)g, textComponent.getUI());
        }
        return false;
    }
  /** Creates a new instance of PaintHelper */
  static public boolean preparePaint(JTextComponent textComponent, Graphics2D g2d, ComponentUI ui)
  {
    if (ui instanceof TextUI)
    {
      TextUI btUI = (TextUI)ui;
      View view = btUI.getRootView(textComponent);
      if (view.getViewCount() > 0)
        return processView(view, g2d);
      else
      {
        return processLeafView(view, g2d);      
      }
    }  
    return false;
  }
  static public boolean preparePaint(View view, Graphics2D g2d)
  {
      return processView(view, g2d);
  }
  
    static protected boolean processLeafView(View child, Graphics2D g2d)
    {
        boolean changed = false;
        if (child instanceof GlyphView)
        {
            GlyphView gv = (GlyphView)child;
            Font testFont = gv.getFont();
            if (testFont == null)
            {
              Component host = gv.getContainer();
              testFont = host.getFont();
            }
            if ( (gv.getGlyphPainter() == null || 
                !(gv.getGlyphPainter() instanceof GraphiteGlyphPainter)) &&
              GraphiteFontManager.isGraphite(testFont))
            {
                //System.out.println(gv);
                GraphiteAwtFont gFont = GraphiteFontManager.getFont(gv.getFont());
                GraphiteGlyphPainter ggp = new GraphiteGlyphPainter(gFont, gv, g2d,
                        gv.getStartOffset(), gv.getEndOffset());
                ggp.setOriginalPainter(gv.getGlyphPainter());
                gv.setGlyphPainter(ggp);
                changed = true;
            }
        }
        return changed;
    }
  
    static protected boolean processView(View view,  Graphics2D g2d)
    {
        boolean changed = false;
        if (view.getViewCount() > 0)
        {
          for (int i = 0; i < view.getViewCount(); i++)
          {
            View child = view.getView(i);
            if (child.getViewCount() > 0)
            {
              changed |= processView(child, g2d);
            }
            else 
            {
              if (child instanceof PlainView)
              {
                  // view.replace doesn't work for BasicTextUI
                GlyphView [] gViews = { new GlyphView(child.getElement()) };
                gViews[0].setParent(child.getParent());
                view.replace(i, 1, gViews);
                child = gViews[0];
              }
              changed |= processLeafView(child, g2d);
            }
          }
        }
        else
        {
          changed |= processLeafView(view, g2d);
        }
        return changed;
    }
}
