/*
 * Copyright (C) 2010 Keith Stribley
 * 
 * Distributable under the terms of either the Common Public License or the
 * GNU Lesser General Public License, as specified in the LICENSING.txt file.
 *
 */
package org.sil.graphite.test;

import org.sil.graphite.GraphiteFont;
import java.text.MessageFormat;
import java.text.DecimalFormat;
import java.io.File;
import java.io.IOException;
import java.io.FileNotFoundException;
import java.io.FileInputStream;
import java.io.BufferedReader;
import java.io.InputStreamReader;

public class GraphiteConsoleTest
{
	private GraphiteFont mFont = null;
	public GraphiteConsoleTest(String fontname, float fontSize, int dpi)
	{
		mFont = new GraphiteFont(fontname, fontSize, dpi);
	}
	
	private String formatInteger(int value, int width)
	{
		StringBuilder intString = new StringBuilder();
		intString.append(value);
		while (intString.length() < width) intString.insert(0, ' ');
		return intString.toString();
	}
	
	private String formatFloat(float value, int decimals, int width)
	{
		StringBuilder floatString = new StringBuilder();
		StringBuilder format = new StringBuilder("0.");
		for (int i = 0; i < decimals; i++) format.append("0");
		DecimalFormat df = new DecimalFormat(format.toString());
		floatString.append(df.format(value));
		while (floatString.length() < width) floatString.insert(0, ' ');
		return floatString.toString();
	}
	
	public void renderText(String text, boolean rtl)
	{
		int numGlyphs = mFont.countGlyphs(text.toCharArray(), 0, text.length(), rtl);
		int [] gid = new int[numGlyphs];
		float [] pos = new float[numGlyphs*2+2];
		int [] chars = new int[numGlyphs];
		mFont.getGlyphArray(text.toCharArray(), 0, text.length(), gid, pos, chars, rtl);
		MessageFormat mf = new MessageFormat("{0} {1}, {2} {3}\n");
		StringBuffer sb = new StringBuffer();
		System.out.println(" GID    X         Y     Char");
		for (int i = 0; i < numGlyphs; i++)
		{
			mf.format(new Object[] {formatInteger(gid[i], 6), formatFloat(pos[2*i], 3, 8), formatFloat(pos[2*i+1], 3, 8), formatInteger(chars[i],4)}, sb, null);
		}
		sb.append("Width=");
		sb.append(formatFloat(pos[numGlyphs*2], 3, 8));
		System.out.println(sb.toString());
	}
	
	public static void usage()
	{
		System.err.println("Usage: GraphiteConsoleTest [-dDPI] [-sSIZE] [-r] fontFileName.ttf test");
	}
	/**
   * @param args the command line arguments
   */
    public static void main(String args[])
    {
		String [] primary = new String[2];
		int argPos = 0;
		int dpi = 96;
		float size = 12;
		boolean rtl = false;
		for (int i = 0; i < args.length; i++)
		{
			if (args[i].startsWith("-"))
			{
				if (args[i].startsWith("-d"))
				{
					dpi = Integer.parseInt(args[i].substring(2));
				}
				else if (args[i].startsWith("-s"))
				{
					size = Float.parseFloat(args[i].substring(2));
				}
				else if (args[i].startsWith("-r"))
				{
					rtl = true;
				}
				else
				{
					usage();
					System.exit(1);
				}
				continue;
			}
			if (argPos == primary.length)
			{
				usage();
				System.exit(1);
			}
			primary[argPos] = args[i];
			++argPos;
		}
		if (primary[0] == null || primary[1] == null)
		{
			usage();
			System.exit(1);
		}
		GraphiteConsoleTest gct = new GraphiteConsoleTest(primary[0], size, dpi);
		File f = new File(primary[1]);
		if (f.exists())
		{
			try
			{
				BufferedReader br = new BufferedReader(new InputStreamReader(new FileInputStream(f), "UTF-8"));
				String line = br.readLine();
				while (line != null)
				{
					gct.renderText(line, rtl);
					line = br.readLine();
				}
				br.close();
			}
			catch (IOException e)
			{
				System.err.println("Failed to read file." + e.getMessage());
			}
		}
		else
			gct.renderText(primary[1], rtl);
    }
}
